#include "feature.hh"



/**
 * @brief It constructs a new feature from its point, list of crossings and
 * proximities, and base segments relations.
 *
 * @param point The point this feature will represents.
 * @param crossingList List of crossings.
 * @param proximitList List of proximities.
 * @param baseSegmentCross Crossing with base segment.
 * @param baseSegmentProxt Proximity with base segment.
 * @param baseSegmentError Error with base segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
Feature::Feature(Point const * point,
		 list<unsigned> const & crossingList,
		 list<unsigned> const & proximitList,
		 bool baseSegmentCross,
		 bool baseSegmentProxt,
		 bool baseSegmentError)
  : point(point),

    nCrossings(crossingList.size()),
    nProximits(proximitList.size()),

    frstCrossIdx(0), lastCrossIdx(nCrossings),
    frstProxtIdx(0), lastProxtIdx(nProximits),

    baseSegmentCross(baseSegmentCross),
    baseSegmentProxt(baseSegmentProxt),
    baseSegmentError(baseSegmentError)
{
  crossings = new unsigned[nCrossings+1];
  copy(crossingList.begin(), crossingList.end(), crossings);
  crossings[nCrossings] = 1;

  proximits = new unsigned[nProximits];
  copy(proximitList.begin(), proximitList.end(), proximits);
}



/**
 * @brief It find the index of the crossing segment based on @ splitIdx.
 *
 * @param splitIdx The index of the splitting vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
unsigned Feature::findCrossIdx(unsigned splitIdx) const
{
  unsigned strtIdx = frstCrossIdx;
  unsigned pendIdx = lastCrossIdx;

  while (pendIdx > strtIdx)
    {
      unsigned midIdx = (strtIdx+pendIdx)/2;

      if      (splitIdx > crossings[midIdx]) strtIdx = midIdx+1;
      else if (splitIdx < crossings[midIdx]) pendIdx = midIdx;
      else return midIdx+1;
    }

  return strtIdx;
}



/**
 * @brief It find the index of the near segment based on @ splitIdx.
 *
 * @param splitIdx The index of the splitting vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
unsigned Feature::findProxtIdx(unsigned splitIdx) const
{
  unsigned strtIdx = frstProxtIdx;
  unsigned pendIdx = lastProxtIdx;

  while (pendIdx > strtIdx)
    {
      unsigned midIdx = (strtIdx+pendIdx)/2;

      if      (splitIdx > proximits[midIdx]) strtIdx = midIdx+1;
      else if (splitIdx < proximits[midIdx]) pendIdx = midIdx;
      else return midIdx+1;
    }

  return strtIdx;
}



/**
 * @brief It writes @a feature in the stream @a ostr.
 *
 * It writes the point or vertex coordinates this feature represent.  It
 * writes the error flag of this feature with respect to the base segment.
 * It writes the proximity relations of this feature.  It writes the
 * crossing relations of this feature.  It writes the consistency state of
 * this feature.
 *
 * @param ostr Output stream.
 * @param feature The feature to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Feature const & feature)
{
  ostr << feature.getPoint() << " -> ("
       << feature.baseSegmentError << ") ("
       << feature.baseSegmentProxt << ", "
       << feature.lastProxtIdx-feature.frstProxtIdx << ") ("
       << feature.baseSegmentCross << ", "
       << feature.lastCrossIdx-feature.frstCrossIdx << ") -> "
       << (feature.getInconsistent() ? "IN" : "  ") << "CONSISTENT";


  return ostr;
}
