#ifndef __FEATURE_HH_
#define __FEATURE_HH_


#include "point.hh"

#include <iostream>
#include <list>


using namespace std;



/**
 * @brief This class a point feature or a vertex linear or areal feature.
 * 
 * This class is used by in correction step of the consistent simplification
 * to store the feature consistency state with respect to part of a original
 * polyline and its correspondent simplifying segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
class Feature {

public:

  // Constructors
  Feature(Point const * point,
	  list<unsigned> const & crossingList,
	  list<unsigned> const & proximitList,
	  bool baseSegmentCross, bool baseSegmentProxt, bool baseSegmentError);

  Feature(Feature const & feature);


  // Destructor
  ~Feature();


  // Get functions
  Point const & getPoint() const;

  unsigned getFrstCrossIdx() const; 
  unsigned getLastCrossIdx() const;

  unsigned getFrstProxtIdx() const; 
  unsigned getLastProxtIdx() const;

  bool getInconsistent() const;


  // Set function
  void setFrstCrossIdx(unsigned frstCrossIdx); 
  void setLastCrossIdx(unsigned lastCrossIdx);

  void setFrstProxtIdx(unsigned frstProxtIdx); 
  void setLastProxtIdx(unsigned lastProxtIdx);

  void setBaseSegmentCross(bool baseSegmentCross); 
  void setBaseSegmentProxt(bool baseSegmentProxt); 
  void setBaseSegmentError(bool baseSegmentError); 

  // Find functions
  unsigned findCrossIdx(unsigned splitIdx) const;
  unsigned findProxtIdx(unsigned splitIdx) const;


  // Operators
  bool operator==(Feature const & feature) const;
  bool operator< (Feature const & feature) const;


  // Extraction operator
  friend ostream & operator<<(ostream & ostr, Feature const & feature);


private:

  Point const * point; /// Pointer to point

  // List of crossings
  unsigned  nCrossings;
  unsigned * crossings;

  // List of proximities
  unsigned  nProximits;
  unsigned * proximits;

  // Indices
  unsigned frstCrossIdx;
  unsigned lastCrossIdx;

  unsigned frstProxtIdx;
  unsigned lastProxtIdx;

  // Relationships with base segment
  bool baseSegmentCross;
  bool baseSegmentProxt;
  bool baseSegmentError;
};



/**
 * @brief The copy constructor.
 *
 * It creates a new feature from @a feature.
 *
 * @param feature The feature to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Feature::Feature(Feature const & feature)
  : point(feature.point),

    nCrossings(feature.nCrossings), crossings(feature.crossings),
    nProximits(feature.nProximits), proximits(feature.proximits),
 
    frstCrossIdx(feature.frstCrossIdx), lastCrossIdx(feature.lastCrossIdx),
    frstProxtIdx(feature.frstProxtIdx), lastProxtIdx(feature.lastProxtIdx),

    baseSegmentCross(feature.baseSegmentCross),
    baseSegmentProxt(feature.baseSegmentProxt),
    baseSegmentError(feature.baseSegmentError)
{
  ++crossings[nCrossings];
}


/**
 * @brief The destructor.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Feature::~Feature()
{
  if (--crossings[nCrossings] == 0)
    {
      delete [] crossings;
      delete [] proximits;
    }
}


/**
 * @brief It returns the point or vertex this feature represents.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point const & Feature::getPoint() const { return *point; }


/**
 * @brief It returns the index of the first crossing segment in the vector
 * of crossings of this feature.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Feature::getFrstCrossIdx() const { return frstCrossIdx; }


/**
 * @brief It returns the index of the last crossing segment in the vector of
 * crossings of this feature.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Feature::getLastCrossIdx() const { return lastCrossIdx; }


/**
 * @brief It returns the index of the first near segment in the vector of
 * proximities of this feature.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Feature::getFrstProxtIdx() const { return frstProxtIdx; }


/**
 * @brief It returns the index of the last near segment in the vector of
 * proximities of this feature.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Feature::getLastProxtIdx() const { return lastProxtIdx; }


/**
 * @brief It returns true if this feature is inconsistent with respect to
 * the CRU where is stored.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Feature::getInconsistent() const
{
  return ((baseSegmentError) ||
	  (baseSegmentProxt && lastProxtIdx-frstProxtIdx == 0) ||
	  (baseSegmentCross+lastCrossIdx-frstCrossIdx)%2 == 1) ? true : false;
}


/**
 * @brief It sets the index of the first crossing segment in the vector of
 * crossings of this feature from @a frstCrossIdx.
 *
 * @param frstCrossIdx The index of the first crossing segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setFrstCrossIdx(unsigned frstCrossIdx)
{ this->frstCrossIdx = frstCrossIdx; }


/**
 * @brief It sets the index of the last crossing segment in the vector of
 * crossings of this feature from @a lastCrossIdx.
 *
 * @param lastCrossIdx The index of the last crossing segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setLastCrossIdx(unsigned lastCrossIdx)
{ this->lastCrossIdx = lastCrossIdx; }


/**
 * @brief It sets the index of the first near segment in the vector of
 * proximities of this feature from @a frstProxIdx.
 *
 * @param frstProxtIdx The index of the first near segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setFrstProxtIdx(unsigned frstProxtIdx)
{ this->frstProxtIdx = frstProxtIdx; }


/**
 * @brief It sets the index of the last near segment in the vector of
 * proximities of this feature from @a lastProxIdx.
 *
 * @param lastProxtIdx The index of the last near segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setLastProxtIdx(unsigned lastProxtIdx)
{ this->lastProxtIdx = lastProxtIdx; }


/**
 * @brief It sets the crossing with the base segment of the CRU where this
 * is stored from @a baseSegmentCross.
 *
 * @param baseSegmentCross The crossing with the the base segment of the CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setBaseSegmentCross(bool baseSegmentCross)
{ this->baseSegmentCross = baseSegmentCross; }


/**
 * @brief It sets the proximity with the base segment of the CRU where this
 * is stored from @a baseSegmentProxt.
 *
 * @param baseSegmentProxt The proximity with the the base segment of the
 * CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setBaseSegmentProxt(bool baseSegmentProxt)
{ this->baseSegmentProxt = baseSegmentProxt; }


/**
 * @brief It sets the proximity error with the base segment of the CRU where
 * this is stored from @a baseSegmentError.
 *
 * @param baseSegmentError The proximity error with the the base segment of
 * the CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Feature::setBaseSegmentError(bool baseSegmentError)
{ this->baseSegmentError = baseSegmentError; }


/**
 * @brief It returns @c true if this feature is equal to @a feature, and @c
 * false otherwise.
 *
 * Two features are considered to be true if its point coordinates are the
 * same.
 *
 * @param feature The feature to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Feature::operator==(Feature const & feature) const
{ return *point == *feature.point; }


/**
 * @brief It returns @c true if this feature come befor @a feature in a
 * cartesian plane order, and @c false otherwise.
 *
 * The feature with the greater y coordinate comes before.  If the y
 * coordinates are the same, the feature with least x coordinate comes
 * before.
 *
 * @param feature The feature to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Feature::operator<(Feature const & feature) const
{
  return point->getY() == feature.point->getY()
    ?    point->getX() <  feature.point->getX()
    :    point->getY() >  feature.point->getY();
}



#endif // __FEATURE_HH_
