#ifndef __POINT_HH_
#define __POINT_HH_


#include "rectangle.hh"

#include <cmath>
#include <iostream>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;



/**
 * @brief This class implements a 2D point.
 * 
 * It stores the cartesian coordinates x (the absissa) and y (the ordinate)
 * of the point.  It is a subclass of the @c Rectangle class.  This class
 * also provides some geometric functions.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Point
  : public Rectangle {

public:

  // Constructors
  Point(Point const & point);             // Copy
  Point(double x = 0.0, double y = 0.0);  // Cartesian coordinates

  // Destructtor
  virtual ~Point();


  // Get rectangle functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;

  // Get coordinates functions
  double getX() const;
  double getY() const;

  // Set functions
  void setX(double x);
  void setY(double y);


  // Operators of equality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Distance and squared distance between points
  double    distance(Point const & point) const;
  double sqrDistance(Point const & point) const;

  // Distance and squared distance between point and segment
  double    distance(Point const & p, Point const & q) const;
  double sqrDistance(Point const & p, Point const & q) const;


  // The cross product between 'pq' and 'p(*this)'
  double crossProduct(Point const & p, Point const & q) const;

  // Test for sidedness
  bool atLeft (Point const & p, Point const & q) const;
  bool atRight(Point const & p, Point const & q) const;
  bool onLine (Point const & p, Point const & q) const;

  bool atLeftOrOnLine(Point const & p, Point const & q) const;
  bool atRghtOrOnLine(Point const & p, Point const & q) const;


  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;

  // Friend operators for input and output handle
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);


protected:

  // Plane cartesian coordinates
  double x; /// x cartesian coordinate (the absissa)
  double y; /// y cartesian coordiante (the ordinate)
};



/**
 * @brief The copy constructor.
 *
 * It creates a new point from @a point.
 *
 * @param point The point to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point::Point(Point const & point) : x(point.x), y(point.y) {}


/**
 * @brief The xy-coordinates constructor.
 *
 * It creates a new point from the coordinates @a x and @a y.
 *
 * @param x The absissa.
 * @param y The ordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point::Point(double x, double y) : x(x), y(y) {}


/**
 * @brief The destructor.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point::~Point() {}


/**
 * @brief It returns the absissa x of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getXMin() const { return x; }


/**
 * @brief It returns the absissa x of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getXMax() const { return x; }


/**
 * @brief It returns the ordinate y of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getYMin() const { return y; }


/**
 * @brief It returns the ordinate y of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getYMax() const { return y; }


/**
 * @brief It returns the absissa x of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getX() const { return x; }


/**
 * @brief It returns the ordinate y of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getY() const { return y; }


/**
 * @brief It sets the absissa x of this point from @a x.
 *
 * @param x The new absissa.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Point::setX(double x) { this->x = x; }


/**
 * @brief It sets the ordinate y of this point from @a y.
 *
 * @param y The new ordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Point::setY(double y) { this->y = y; }


/**
 * @brief It returns true if this point is different from @a point.
 *
 * It returns @c true if the at least one of the coordinates x and y of this
 * point is different from its correspondent coordinate of @a point, and @c
 * false, otherwise.
 *
 * @param point The point to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::operator!=(Point const & point) const
{ return !operator==(point); }


/**
 * @brief It returns true if this point is equal to @a point.
 *
 * It returns @c true if the the coordinates x and y of this point are equal
 * to their correspondent coordinates of @a point, and @c false, otherwise.
 *
 * @param point The point to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::operator==(Point const & point) const
{ return ((x == point.x) && (y == point.y)); }


/**
 * @brief It computes the distance between this point and @a point.
 *
 * @param point The point to compute the distance from.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::distance(Point const & point) const
{ return sqrt((point.x-x)*(point.x-x)+(point.y-y)*(point.y-y)); }


/**
 * @brief It computes the squared distance between this point and @a point.
 *
 * @param point The point to compute the distance from.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::sqrDistance(Point const & point) const
{ return (point.x-x)*(point.x-x)+(point.y-y)*(point.y-y); }


/**
 * @brief It computes the cross product of two vectors.
 *
 * It computes the cross product of between the vectors @a pq and @a
 * p(*this).
 *
 * @param p Base point of vector.
 * @param q End point of first vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::crossProduct(Point const & p, Point const & q) const
{ return (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y); }


/**
 * @brief It returns true if this point is at left from the straight line
 * that passes over the points @a p and @a q considering the direction of
 * the vector @a pq, and false otherwise.
 *
 * @param p First point of vector.
 * @param q Second point of vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::atLeft(Point const & p, Point const & q) const
{ return crossProduct(p, q) > 0; }


/**
 * @brief It returns true if this point is at right from the straight line
 * that passes over the points @a p and @a q considering the direction of
 * the vector @a pq, and false otherwise.
 *
 * @param p First point of vector.
 * @param q Second point of vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::atRight(Point const & p, Point const & q) const
{ return crossProduct(p, q) < 0; }


/**
 * @brief It returns true if this point is over the straight line that
 * passes over points @a p and @a q, and false otherwise.
 *
 * @param p First point of vector.
 * @param q Second point of vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::onLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) == 0; }


/**
 * @brief It returns true if this point is at left from or over the straight
 * line that passes over the points @a p and @a q considering the direction
 * of the vector @a pq, and false otherwise.
 *
 * @param p First point of vector.
 * @param q Second point of vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::atLeftOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) >= 0; }


/**
 * @brief It returns true if this point is at right from or over the
 * straight line that passes over the points @a p and @a q considering the
 * direction of the vector @a pq, and false otherwise.
 *
 * @param p First point of vector.
 * @param q Second point of vector.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::atRghtOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) <= 0; }



#endif // __POINT_HH_
