#include "polyline.hh"

#include <iomanip>




/**
 * @brief Computes farthest vertex between two vertices
 *
 * Computes farthest vertex in the range (@a frstIdx, @a lastIdx) from base
 * segment formed by first and last vertices.
 *
 * @param frstIdx The index of the first vertex in the range.
 * @param latsIdx The index of the last  vertex in the range.
 *****************************************************************************/
unsigned Polyline::computeSplitVertex(unsigned frstIdx, unsigned lastIdx)
{
  // Searching farthest vertex
  double   farthestSqrDistance = -1.0;
  unsigned farthestIdx         = frstIdx;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	(*this)[vertexIdx].sqrDistance((*this)[frstIdx],
				       (*this)[lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }

  // Setting metric of farthest vertex
  (*this)[farthestIdx].setMetric(farthestSqrDistance);


  return farthestIdx;
}



/**
 * @brief Applies the RDP algorithm to part this polyline.
 *
 * The simplification is carried out in the range [@a frstIdx, @a lastIdx]
 * with a maximum distance of @a sqrTolerance, which is given squared.
 *
 * @param sqrTolerance The squared maximum distance..
 * @param frstIdx The index of the first vertex in the range.
 * @param latsIdx The index of the last  vertex in the range.
 *****************************************************************************/
void Polyline::simplifyRDP(double sqrTolerance,
			   unsigned frstIdx, unsigned lastIdx)
{
  // Searching farthest vertex
  double   farthestSqrDistance = sqrTolerance;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	(*this)[vertexIdx].sqrDistance((*this)[frstIdx],
				       (*this)[lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  // Testing whether distance is greater than tolerance
  if (farthestSqrDistance > sqrTolerance)
    {
      simplifyRDP(sqrTolerance, frstIdx, farthestIdx);
      simplifyRDP(sqrTolerance, farthestIdx, lastIdx);
    }

  else
    {
      (*this)[frstIdx].setNextIdx(lastIdx);
      (*this)[lastIdx].setPrevIdx(frstIdx);
    }
}



/**
 * @brief Prints subpolyline of this polyline in @c ostr.
 *
 * First, prints the number of vertices of the subpolyline and the range.
 * Then, prints vertices of the subpolyline one per line, prefixed with its
 * index.
 *
 * @param ostr Output stream in where the polyline is printed.
 * @param frstIdx First index of subpolyline's range.
 * @param lastIdx Last  index of subpolyline's range.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setfill('0') << setw(3) << vertexIdx << ") " << setfill(' ');
      (*this)[vertexIdx].print(ostr);
    }
}



/**
 * @brief Writes @c polyline in @c ostr.
 *
 * First, writes the number of vertices.  Then, writes the vertices one per
 * line.  Is a redefinition of the insertion operator in the context of
 * streams.
 *
 * @param ostr Output stream in where the polyline is writed.
 * @param polyline Polyline to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;


  return ostr;
}



/**
 * @brief Reads @c polyline from @c ostr.
 *
 * First, reads number of vertices.  Then, reads vertices in the specified
 * order.  Is a redefinition of the extraction operator in the context of
 * streams.
 *
 * @param ostr Input stream from where the polyline is printed.
 * @param polyline Polyline to print.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    istr >> polyline[vertexIdx];


  return istr;
}
