#include "rips.hh"

#include <string>
#include <iomanip>
#include <algorithm>




/**
 * @brief It solves the Rectangle Intersection Problem.
 *
 * It uses the algorithm of Edelsbrunner to solve the RIP problem for the
 * rectangles given in the list @a rectangles.  The pairs of intersecting
 * rectangles are store in a internal list.
 *
 * @param rectangles List of rectangles to compute the intersections.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::solve(list<Rectangle const *> const & rectangles)
{
  // List of event points
  list<eventPoint_t> eventPoints;

  // Vector of yValues
  vector<double> yValues;
  yValues.reserve(2*rectangles.size());


  // Iniatialing lists of event points and y values
  for (constRectIter_t rIter = rectangles.begin();
       rIter != rectangles.end(); ++rIter)
    {
      eventPoints.push_back(eventPoint_t((*rIter)->getXMin(), true,  *rIter));
      eventPoints.push_back(eventPoint_t((*rIter)->getXMax(), false, *rIter));

      yValues.push_back((*rIter)->getYMin());
      yValues.push_back((*rIter)->getYMax());
   }


  // Sorting event points and y values
  eventPoints.sort();

  // Sorting and removing repeated element of y values
  sort(yValues.begin(), yValues.end());
  yValues.resize(unique(yValues.begin(), yValues.end())-yValues.begin());


  // Creating interval tree
  node_t * treeRoot;
  createTree(treeRoot, yValues, 0, yValues.size()-1);


  // Sweeping the plane horizontally
  for (eventPointIter_t pointIter = eventPoints.begin();
       pointIter != eventPoints.end(); ++pointIter)

    if (pointIter->minimum) insertRectangle(treeRoot, pointIter->rect);
    else                    removeRectangle(treeRoot, pointIter->rect);


  // Deleting allocated inverval tree
  deleteTree(treeRoot);
}



/**
 * @brief It creates the interval tree.
 *
 * It is a recursive procedure that creates the interval tree from the
 * vector of ordinates @a yValues.
 *
 * @param node Pointer to current node to be created.
 * @param yValues Lis of ordinates.
 * @param bgnIdx First index of the current interval.
 * @param bgnIdx Last index of the current interval.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::createTree(node_t * & node, vector<double> const & yValues,
		      unsigned bgnIdx, unsigned endIdx)
{
  unsigned midIdx = (bgnIdx+endIdx)/2;

  node = new node_t(yValues[midIdx]);

  if (midIdx > bgnIdx) createTree(node->leftSon, yValues, bgnIdx, midIdx-1);
  if (midIdx < endIdx) createTree(node->rghtSon, yValues, midIdx+1, endIdx);
}



/**
 * @brief It deletes the interval tree.
 *
 * @param node Root of the interval tree.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::deleteTree(node_t * node)
{
  if (!node) return;


  deleteTree(node->leftSon);
  deleteTree(node->rghtSon);

  delete node;
}



/**
 * @brief It inserts a rectangle in the interval tree.
 *
 * @param node Root of the interval tree.
 * @param rect Rectangle to be inserted.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::insertRectangle(node_t * node, Rectangle const * rect)
{
  node->empty = false;


  if (rect->getYMin() <= node->y &&
      rect->getYMax() >= node->y)
    {
      for (rectIter_t rectIter = node->leftRectangles.begin();
	   rectIter != node->leftRectangles.end(); ++rectIter)
	rectPairs.push_back(RectPair_t(rect, *rectIter));

      searchRectangles(node->leftSon, rect);
      searchRectangles(node->rghtSon, rect);


      rectIter_t leftRectIter = node->leftRectangles.begin();

      while (leftRectIter != node->leftRectangles.end() &&
	     rect->getYMin() > (*leftRectIter)->getYMin()) ++leftRectIter;

      node->leftRectangles.insert(leftRectIter, rect);


      rectIter_t rghtRectIter = node->rghtRectangles.begin();

      while (rghtRectIter != node->rghtRectangles.end() &&
	     rect->getYMax() < (*rghtRectIter)->getYMax()) ++rghtRectIter;

      node->rghtRectangles.insert(rghtRectIter, rect);
    }

  else if (rect->getYMax() < node->y)
    {
      for (rectIter_t rectIter = node->leftRectangles.begin();
	   rectIter != node->leftRectangles.end(); ++rectIter)
	{
	  if (rect->getYMax() < (*rectIter)->getYMin()) break;
	  rectPairs.push_back(RectPair_t(rect, *rectIter));
	}

      insertRectangle(node->leftSon, rect);
    }

  else if (rect->getYMin() > node->y)
    {
      for (rectIter_t rectIter = node->rghtRectangles.begin();
	   rectIter != node->rghtRectangles.end(); ++rectIter)
	{
	  if (rect->getYMin() > (*rectIter)->getYMax()) break;
	  rectPairs.push_back(RectPair_t(rect, *rectIter));
	}

      insertRectangle(node->rghtSon, rect);
    }
}



/**
 * @brief It removes a rectangle from the interval tree.
 *
 * @param node Root of the interval tree.
 * @param rect Rectangle to be removed.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::removeRectangle(node_t * node, Rectangle const * rect)
{
  if (!node) return;


  if (rect->getYMin() <= node->y &&
      rect->getYMax() >= node->y)
    {
      for (rectIter_t rectIter = node->leftRectangles.begin();
	   rectIter != node->leftRectangles.end(); ++rectIter)
	if (*rectIter == rect)
	  {
	    node->leftRectangles.erase(rectIter);
	    break;
	  }

      for (rectIter_t rectIter = node->rghtRectangles.begin();
	   rectIter != node->rghtRectangles.end(); ++rectIter)
	if (*rectIter == rect)
	  {
	    node->rghtRectangles.erase(rectIter);
	    break;
	  }
    }

  else if (rect->getYMax() < node->y) removeRectangle(node->leftSon, rect);
  else if (rect->getYMin() > node->y) removeRectangle(node->rghtSon, rect);


  if (node->leftRectangles.empty() &&
      (!node->leftSon || node->leftSon->empty) &&
      (!node->rghtSon || node->rghtSon->empty))
    node->empty = true;
}



/**
 * @brief It searchs in the interval tree for rectangles that intersects @a
 * rect.
 *
 * @param node Root of the interval tree.
 * @param rect Rectangle to search the intersecting rectangles.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
void RIPS::searchRectangles(node_t * node, Rectangle const * rect)
{
  if (!node || node->empty) return;


  if (rect->getYMin() <= node->y &&
      rect->getYMax() >= node->y)
    {
      for (rectIter_t rectIter = node->leftRectangles.begin();
	   rectIter != node->leftRectangles.end(); ++rectIter)
	rectPairs.push_back(RectPair_t(rect, *rectIter));

      searchRectangles(node->leftSon, rect);
      searchRectangles(node->rghtSon, rect);
    }
      
  else if (rect->getYMax() < node->y)
    {
      for (rectIter_t rectIter = node->leftRectangles.begin();
	   rectIter != node->leftRectangles.end(); ++rectIter)
	{
	  if (rect->getYMax() < (*rectIter)->getYMin()) break;
	  rectPairs.push_back(RectPair_t(rect, *rectIter));
	}

      searchRectangles(node->leftSon, rect);
    }

  else if (rect->getYMin() > node->y)
    {
      for (rectIter_t rectIter = node->rghtRectangles.begin();
	   rectIter != node->rghtRectangles.end(); ++rectIter)
	{
	  if (rect->getYMin() > (*rectIter)->getYMax()) break;
	  rectPairs.push_back(RectPair_t(rect, *rectIter));
	}

      searchRectangles(node->rghtSon, rect);
    }
}



/**
 * @brief It prints the interval tree in the stream @a ostr.
 *
 * @param ostr Output stream.
 * @param node Root of the interval tree.
 * @param nSpaces Number of spaces for indentation.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void RIPS::print(ostream & ostr, node_t const * node, unsigned nSpaces) const
{
  if (!node) return;

  string spaces(nSpaces, ' ');
  ostr << spaces << '(' << setw(14) << node->y << ')' << endl;


  ostr << spaces << "- left " << setw(2) << node->leftRectangles.size() << ':';
  for (constRectIter_t rectIter = node->leftRectangles.begin();
       rectIter != node->leftRectangles.end(); ++rectIter)
    ostr << " [" << setw(14) << (*rectIter)->getYMin()
	 << "; " << setw(14) << (*rectIter)->getYMax() << ']';
  ostr << endl;


  ostr << spaces << "- rght " << setw(2) << node->leftRectangles.size() << ':';
  for (constRectIter_t rectIter = node->rghtRectangles.begin();
       rectIter != node->rghtRectangles.end(); ++rectIter)
    ostr << " [" << setw(14) << (*rectIter)->getYMin()
	 << "; " << setw(14) << (*rectIter)->getYMax() << ']';
  ostr << endl;


  // Recursions
  print(ostr, node->leftSon, nSpaces+2);
  print(ostr, node->rghtSon, nSpaces+2);
}
