#ifndef __VERTEX_HH_
#define __VERTEX_HH_


#include <iostream>

#include "point.hh"


using namespace std;


/**
 * @brief This class implements a 2D vertex of a polyline.
 * 
 * This class is a subclass of the class @c Point.  Besides its coordinates,
 * it stores a flag that indicates if this vertex has been erased from the
 * polyline it belongs to.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Vertex
  : public Point {

public:

  // Constructors
  Vertex(Vertex const & vertex);           // Copy
  Vertex(Point  const & point);            // Point
  Vertex(double x = 0.0, double y = 0.0);  // Cartesien coordinates


  // Indices access functions
  unsigned getPrevIdx() const;
  unsigned getNextIdx() const;

  // Get metric function
  double getMetric() const;


  // Indices modification functions
  void setPrevIdx(unsigned prevIdx);
  void setNextIdx(unsigned nextIdx);

  // Metric modification function
  void setMetric(double metric);


  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Indices of previous and next vertices in polyline
  unsigned prevIdx; /// Index of previous not erased vertex
  unsigned nextIdx; /// Index of next not erased vertex

  // Metric for the algorithm
  double metric; /// Distance from base line segment
};



/**
 * @brief The copy constructor.
 *
 * It creates a new vertex from @a vertex.
 *
 * @param vertex The vertex to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Vertex const & vertex)
  : Point(vertex), prevIdx(vertex.prevIdx), nextIdx(vertex.nextIdx),
    metric(vertex.metric) {}


/**
 * @brief The point constructor.
 *
 * It creates a new vertex from @a point.
 *
 * @param point The point to copy coordinates.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Point const & point)
  : Point(point),  prevIdx(0), nextIdx(0), metric(0.0) {}


/**
 * @brief The xy-coordinate constructor.
 *
 * It creates a new vertex with given coordinates.
 *
 * @param x The @a x coordinate.
 * @param y The @a y coordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(double x, double y)
  : Point(x, y), prevIdx(0), nextIdx(0), metric(0.0) {}


/**
 * @brief It returns the index of the next not erased vertex in the polyline
 * this vertex belongs to.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Vertex::getPrevIdx() const { return prevIdx; }


/**
 * @brief It returns the index of the next not erased vertex in the polyline
 * this vertex belongs to.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Vertex::getNextIdx() const { return nextIdx; }


/**
 * @brief It returns the metric of this vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Vertex::getMetric() const { return metric; }


/**
 * @brief It sets the index of the previous not erased vertex of the
 * polyline this vertex belongs to from @prevIdx.
 *
 * @param prevIdx The index of the previous not erased vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Vertex::setPrevIdx(unsigned prevIdx) { this->prevIdx = prevIdx; }


/**
 * @brief It sets the index of the next not erased vertex of the polyline
 * this vertex belongs to from @nextIdx.
 *
 * @param prevIdx The index of the previous not erased vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Vertex::setNextIdx(unsigned nextIdx) { this->nextIdx = nextIdx; }


/**
 * @brief It sets the metric of this vertex.
 *
 * @param metric The new metric.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Vertex::setMetric(double metric) { this->metric = metric; }



#endif // __VERTEX_HH_
