#ifndef __CRU_HH_
#define __CRU_HH_


#include "rectangle.hh"
#include "polyline.hh"
#include "feature.hh"

#include <deque>
#include <list>


using namespace std;




/**
 * External types
 *****************************************************************************/
class CRU;
typedef list<CRU *>::iterator            CRUIter_t;
typedef list<CRU *>::const_iterator constCRUIter_t;


/**
 * Internal types
 *****************************************************************************/
typedef list<Feature>::iterator            featureIter_t;
typedef list<Feature>::const_iterator constFeatureIter_t;



struct neighbour_t
{
  neighbour_t(CRU * cru) : cru(cru) {}

  CRU * cru;
  list<neighbour_t>::iterator myIter;
};



/**
 * 'CRU' is the class for the Consistency Recovery Unit.
 *****************************************************************************/
class CRU
  : public Rectangle {

public:

  // Constructors
  CRU(Polyline const & polyline, double nearTolerance);
  CRU(Polyline const & polyline, double nearTolerance,
      unsigned frstIdx, unsigned lastIdx);

  // Virtual destructor
  virtual ~CRU();


  // Determining bounding box of 'polyline' between 'frstIdx' and 'lastIdx'
  void set(unsigned frstIdx, unsigned lastIdx);


  // Spliting convex hull at 'splitIdx'
  void splitBoth(unsigned splitIdx);
  void splitFrst(unsigned splitIdx);
  void splitScnd(unsigned splitIdx);
  void splitZero(unsigned splitIdx);


  // Determines if 'point' is inside its bounding box
  bool isInside(Point const & point);


  // Pushes a new feature into the features' list
  void pushFeature(Point   const & point);
  void pushFeature(Feature const & feature, bool firstSubPol,
		   unsigned findCrossIdx,
		   unsigned findProxtIdx);


  // Get functions
  Polyline const & getPolyline() const;

  double getNearTolerance() const;

  unsigned getFrstIdx() const;
  unsigned getLastIdx() const;
  unsigned getFarthestIdx() const;

  Point const & getFrstPoint() const;
  Point const & getLastPoint() const;

  bool getInconsistent() const;
  bool getBiggerDistance(double tolerance) const;

  unsigned numberOfFeatures() const;


  // Set function
  void setNearTolerance(double nearTolerance);


  // Test if cru overlaps this cru
  bool overlaps(CRU const & cru) const;


  // Bouding box get functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;


  // Neighbouring function
  friend void setNeighbouring(CRU & cru1, CRU & cru2);


  // Friendship relation of insertion operator
  friend ostream & operator<<(ostream & ostr, CRU const & cru);


  // Printing function
  void print(ostream & ostr) const;


  // Static list of CRUs
  static list<CRU *> cleanCRUs;
  static list<CRU *> dirtyCRUs;


private:

  // Referenced polyline
  Polyline const & polyline;

  // Limit indices of convex hull
  unsigned frstIdx;
  unsigned lastIdx;
  unsigned farthestIdx;

  // Tolerance for minimum distance
  double    nearTolerance;
  double sqrNearTolerance;

  // List of features
  list<Feature> features;

  // Vector of neighbours
  list<neighbour_t> neighbours;

  // Sidedness test
  bool inconsistent;

  // CRU iterator to list of CRUs (dirty or clean)
  CRUIter_t cruIter;

  // Bounding box coordinates
  double xMin, xMax;
  double yMin, yMax;

  // Bounding diamond coordinates
  double pMin, pMax;
  double qMin, qMax;
};



inline CRU::CRU(Polyline const & polyline, double nearTolerance)
  : polyline(polyline), nearTolerance(nearTolerance),
    sqrNearTolerance(nearTolerance*nearTolerance), inconsistent(false) {}

inline CRU::CRU(Polyline const & polyline, double nearTolerance,
		unsigned frstIdx, unsigned lastIdx)
  : polyline(polyline), nearTolerance(nearTolerance),
    sqrNearTolerance(nearTolerance*nearTolerance), inconsistent(false)
{
  set(frstIdx, lastIdx);
}

inline CRU::~CRU() {}


inline Polyline const & CRU::getPolyline() const { return polyline; }

inline double CRU::getNearTolerance() const { return nearTolerance; }

inline unsigned CRU::getFrstIdx() const     { return frstIdx; }
inline unsigned CRU::getLastIdx() const     { return lastIdx; }
inline unsigned CRU::getFarthestIdx() const { return farthestIdx; }

inline Point const & CRU::getFrstPoint() const { return polyline[frstIdx]; }
inline Point const & CRU::getLastPoint() const { return polyline[lastIdx]; }

inline bool CRU::getInconsistent() const  { return inconsistent; }
inline bool CRU::getBiggerDistance(double tolerance) const
{ return polyline[farthestIdx].getDistance() >= tolerance; }

inline unsigned CRU::numberOfFeatures() const { return features.size(); }


inline void CRU::setNearTolerance(double nearTolerance)
{
  this->   nearTolerance = nearTolerance;
  this->sqrNearTolerance = nearTolerance*nearTolerance;
}


inline bool CRU::overlaps(CRU const & cru) const
{
  // Testing bounding box
  if (cru.xMax < xMin || cru.xMin > xMax ||
      cru.yMax < yMin || cru.yMin > yMax)
    return false;

  // Testing bounding diamond
  if (cru.pMax < pMin || cru.xMin > xMax ||
      cru.pMax < pMin || cru.yMin > yMax)
    return false;

  return true;
}


inline bool CRU::isInside(Point const & point)
{
  // Testing bounding box
  if (point.getX() < xMin || point.getX() > xMax ||
      point.getY() < yMin || point.getY() > yMax)
    return false;

  // Testing bounding diamond
  double p = point.getX()+point.getY(); if (p < pMin || p > pMax) return false;
  double q = point.getX()-point.getY(); if (q < qMin || q > qMax) return false;

  return true;
}


inline double CRU::getXMin() const { return xMin; }
inline double CRU::getXMax() const { return xMax; }
inline double CRU::getYMin() const { return yMin; }
inline double CRU::getYMax() const { return yMax; }



#endif // __CRU_HH_
