#include "mapmrtcrdp.hh"
#include "rips.hh"

#include <iomanip>
#include <algorithm>
#include <sys/time.h>




/**
 * @brief Simplifies polylines of 'mapMRTCRDP' using Silva's algorithm for
 * tolerance 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void MapMRTCRDP::simplifyMRTCRDP(double nearTolerance)
{
  // Getting initial time
  timeval bgnTime;
  timeval endTime;
  double diffTime;

  gettimeofday(&bgnTime, NULL);

  cerr << "---------------------" << endl;


  // Setting RDP distances
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    polylines[polIdx].simplifyMRRDP();


  gettimeofday(&endTime, NULL);
  diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
	      (endTime.tv_usec-bgnTime.tv_usec)/1E6);
  cerr << "Time RDP:  " << setw(14) << diffTime << endl;


  // Computing initial tolerances
  computeTolerances();


  gettimeofday(&bgnTime, NULL);
  diffTime = ((bgnTime.tv_sec -endTime.tv_sec)+
	      (bgnTime.tv_usec-endTime.tv_usec)/1E6);
  cerr << "Time Tol:  " << setw(14) << diffTime << endl;


  // Set initial CRUs and symplifying for tolerance infinity
  initCRUs(nearTolerance);

  gettimeofday(&endTime, NULL);
  diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
	      (endTime.tv_usec-bgnTime.tv_usec)/1E6);
  cerr << "Time Init: " << setw(14) << diffTime << endl;


  simplifyMRTCRDP(0);


  gettimeofday(&bgnTime, NULL);
  diffTime = ((bgnTime.tv_sec -endTime.tv_sec)+
	      (bgnTime.tv_usec-endTime.tv_usec)/1E6);
  cerr << "Time Smp0: " << setw(14) << diffTime << endl;


  // Computing rest of tolerances
  for (unsigned tolIdx = 1; tolerances[tolIdx].getValue() !=
	 tolerances.back().getValue(); ++tolIdx)
    {
      if (tolerances[tolIdx  ].getValid()) addCRUs(tolIdx, crus);
      if (tolerances[tolIdx  ].getValue() !=
	  tolerances[tolIdx+1].getValue()) simplifyMRTCRDP(tolIdx, crus);
    }


  gettimeofday(&endTime, NULL);
  diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
	      (endTime.tv_usec-bgnTime.tv_usec)/1E6);
  cerr << "Time Simp: " << setw(14) << diffTime << endl;


  // Deleting allocated memory for CRUs
  for (CRUIter_t cruIter = cleanCRUs.begin();
       cruIter != cleanCRUs.end(); ++cruIter)
    delete *cruIter;
}



/**
 *
 *****************************************************************************/
void MapMRTCRDP::computeTolerances()
{
  // Calculating number of tolerances and allocating memory
  unsigned nTolerances = 1;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    nTolerances += polylines[polIdx].numberOfVertices()-2;

  tolerances.reserve(nTolerances);


  // Computing tolerances and storing in vector
  tolerances.push_back(Tolerance(MAXDOUBLE, 0, 0, 0));

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    for (unsigned vertexIdx = 1; vertexIdx <
	   polylines[polIdx].numberOfVertices()-1; ++vertexIdx)
      tolerances.
	push_back(Tolerance(polylines[polIdx][vertexIdx].getTolerance(),
			    polylines[polIdx][vertexIdx].getLevel(),
			    polIdx, vertexIdx));

  // Sorting vector of tolerances and setting indices of tolerances in vertices
  sort(tolerances.begin(), tolerances.end());

  for (unsigned tolIdx = 0; tolIdx < nTolerances; ++tolIdx)
    polylines[tolerances[tolIdx].getPolIdx()]
      [tolerances[tolIdx].getVertexIdx()].setTolIdx(tolIdx);
}



/**
 *
 *****************************************************************************/
void MapMRTCRDP::initCRUs(double nearTolerance)
{
  // List of rectangles
  list<Rectangle const *> rectangles;


  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)

    if (polylines[polIdx].numberOfVertices() == 2)
      {
	// Pushing extreme vertices in list of rectangles
	rectangles.push_back(&polylines[polIdx].front());
	rectangles.push_back(&polylines[polIdx].back());
      }

    else
      {
	// Pushing CRU in list of CRUs and rectangles
	CRU * cru = new CRU(polylines[polIdx], nearTolerance, 0,
			    polylines[polIdx].numberOfVertices()-1);

	crus.      push_back(cru);
	rectangles.push_back(cru);

	polylines[polIdx][cru->getFarthestIdx()].setCRUIter(--crus.end());
      }


  // Adding point features in list of rectangles
  for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
    rectangles.push_back(&points[pointIdx]);


  // Identifying neighbours
  RIPS rips;
  rips.solve(rectangles);


  // Adding points and setting neighbours
  RectPairList const & rectPairs = rips.getRectPairs();

  for (constRectPairIter_t rectPairIter = rectPairs.begin();
       rectPairIter != rectPairs.end(); ++rectPairIter)
    {
      Rectangle * frstRect = const_cast<Rectangle *>(rectPairIter->first);
      Rectangle * scndRect = const_cast<Rectangle *>(rectPairIter->second);

      if (typeid(*frstRect) == typeid(CRU) &&
	  typeid(*scndRect) == typeid(CRU))
	{
	  CRU & cru1 = *reinterpret_cast<CRU *>(frstRect);
	  CRU & cru2 = *reinterpret_cast<CRU *>(scndRect);

	  setNeighbouring(cru1, cru2);

	  cru1.pushFeature(cru2.getFrstPoint());
	  cru1.pushFeature(cru2.getLastPoint());

	  cru2.pushFeature(cru1.getFrstPoint());
	  cru2.pushFeature(cru1.getLastPoint());
	}

      else if (typeid(*frstRect) == typeid(CRU))
	reinterpret_cast<CRU *>(frstRect)->
	  pushFeature(*reinterpret_cast<Point *>(scndRect));

      else if (typeid(*scndRect) == typeid(CRU))
	reinterpret_cast<CRU *>(scndRect)->
	  pushFeature(*reinterpret_cast<Point *>(frstRect));
    }
}



/**
 *
 *****************************************************************************/
void MapMRTCRDP::addCRUs(unsigned tolIdx, list<CRU *> & crus)
{
  unsigned polIdx    = tolerances[tolIdx].getPolIdx();
  unsigned vertexIdx = tolerances[tolIdx].getVertexIdx();


  // Setting vertex tolerance
  polylines[polIdx][vertexIdx].setTolerance(tolerances[tolIdx].getValue());
  CRUIter_t cruIter = polylines[polIdx][vertexIdx].getCRUIter();


  /* Correcting current CRU */
  correct(tolIdx, cruIter, crus);
}



/**
 * Simplifies map 'mapmrtcrdp' with Silva's version of RDP algorithm for
 * squared tolerance 'sqrTolerance', filling up the convex hulls in 'crus'.
 */
void MapMRTCRDP::simplifyMRTCRDP(unsigned tolIdx, list<CRU *> & crus)
{
  // Correcting sidedness errors
  while (!CRU::dirtyCRUs.empty())
    correct(tolIdx, cru);
}



/**
 * Simplifies polyline 'polylines[polIdx]' of 'tcdpmap' using RDP algorithm for
 * squared tolerance 'sqrTolerance' with convex hull 'cHullIter' in convex
 * hulls list 'crus'.  New vertices are added to 'newVertices' list.
 *****************************************************************************/
void MapMRTCRDP::correct(unsigned tolIdx, CRU * cru)
{
  // Testing for inconsistency or vertex farther than tolerance
  if (cru->getInconsistent() ||
      cru->getBiggerDistance(tolerances[tolIdx].getValue()))
    {
      // Setting polyine, first, last and farthest indices from convex hull
      Polyline & polyline = const_cast<Polyline &>(cru->getPolyline());
 
      unsigned frstIdx     = cru->getFrstIdx();
      unsigned lastIdx     = cru->getLastIdx();
      unsigned farthestIdx = cru->getFarthestIdx();


      /* Setting tolerance of farthest vertex */
      polyline[farthestIdx].setTolerance(tolerances[tolIdx].getValue());
      tolerances[polyline[farthestIdx].getTolIdx()].setValid(false);


      // Setting first and second CRUs
      if      (farthestIdx > frstIdx+1 &&
	       farthestIdx < lastIdx-1) cru->splitBoth(farthestIdx);
      else if (farthestIdx > frstIdx+1) cru->splitFrst(farthestIdx);
      else if (farthestIdx < lastIdx-1) cru->splitScnd(farthestIdx);
      else                              cru->splitZero(farthestIdx);


      // Erasing remaining CRU
      delete cru;
    }
}
