#ifndef __POINT_HH_
#define __POINT_HH_


#include "rectangle.hh"

#include <cmath>
#include <iostream>
#include <list>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;

class CRU;
typedef list<CRU *>::iterator            CRUIter_t;
typedef list<CRU *>::const_iterator constCRUIter_t;



/**
 * @brief This class contains cartesian coordinates of a point in 2D space. It
 * presents geometric transformations and util functions usually applied to
 * points.
 *****************************************************************************/
class Point
  : public Rectangle{

public:

  // Constructors
  Point();                    // Default
  Point(Point const & point); // Copy
  Point(double x, double y);  // Cartesian coordinates


  // Get functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;

  double getX() const;
  double getY() const;

  // Functions 'get' for RDP's distance and tolerance
  double getDistance()  const;
  double getTolerance() const;

  // Functions 'get' for the right and left children' indices
  unsigned getLeftIdx() const;
  unsigned getRghtIdx() const;

  // Functions 'get' for level of detail and index of tolerance
  unsigned getLevel() const;
  unsigned getTolIdx() const;

  // Function 'get' for cru iterator
  CRUIter_t getCRUIter() const;


  // Set functions
  void setX(double x);
  void setY(double y);

  // Functions 'set' for DP distance and Saalfeld's tolerance
  void setDistance (double distance);
  void setTolerance(double tolerance);

  // Functions 'get' for the right and left sons' indices
  void setLeftIdx(unsigned leftIdx);
  void setRghtIdx(unsigned rghtIdx);

  // Functions 'set' for level of detail and index of tolerance
  void setLevel (unsigned level);
  void setTolIdx(unsigned tolIdx);

  // Function 'set' for cru iterator
  void setCRUIter(CRUIter_t cruIter);


  // Squared distances beween points and between a point and a segment
  double sqrDistance(Point const & p) const;
  double sqrDistance(Point const & p, Point const & q) const;


  // The cross product between 'pq' and 'p(*this)'
  double crossProduct(Point const & p, Point const & q) const;

  // Test for sidedness
  bool atLeft (Point const & p, Point const & q) const;
  bool atRight(Point const & p, Point const & q) const;
  bool onLine (Point const & p, Point const & q) const;

  bool atLeftOrOnLine(Point const & p, Point const & q) const;
  bool atRghtOrOnLine(Point const & p, Point const & q) const;


  // Operators of equality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Friend operators for input and output handle
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);

  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Cartesian coordinates
  double x;
  double y;

  // RDP's distance and tolerance
  double distance;
  double tolerance;

  // Indices of children in RDP's tree
  unsigned leftIdx;
  unsigned rghtIdx;

  // Level of detail and index of tolerance
  unsigned level;
  unsigned tolIdx;

  // Iterator to CRU
  CRUIter_t cruIter;
};


inline Point::Point()
  : x(0.0),
    y(0.0),

    distance (0.0),
    tolerance(0.0),

    leftIdx(0),
    rghtIdx(0),

    level (0),
    tolIdx(0) {}

inline Point::Point(Point const & point)
  : x(0.0),
    y(0.0),

    distance (point.distance),
    tolerance(point.tolerance),

    leftIdx(point.leftIdx),
    rghtIdx(point.rghtIdx),

    level (point.level),
    tolIdx(point.tolIdx) {}

inline Point::Point(double x, double y)
  : x(x),
    y(y),

    distance (0.0),
    tolerance(0.0),

    leftIdx(0),
    rghtIdx(0),

    level (0),
    tolIdx(0) {}


inline double Point::getX() const { return x; }
inline double Point::getY() const { return y; }

inline double Point::getDistance() const  { return distance; }
inline double Point::getTolerance() const { return tolerance; }

inline unsigned Point::getLeftIdx() const { return leftIdx; }
inline unsigned Point::getRghtIdx() const { return rghtIdx; }

inline unsigned Point::getLevel() const  { return level; }
inline unsigned Point::getTolIdx() const { return tolIdx; }

inline CRUIter_t Point::getCRUIter() const { return cruIter; }


inline void Point::setX(double x) { this->x = x; }
inline void Point::setY(double y) { this->y = y; }

inline void Point::setDistance (double dist)  { this->distance  = dist; }
inline void Point::setTolerance(double toler) { this->tolerance = toler; }

inline void Point::setLeftIdx(unsigned index) { this->leftIdx = index; }
inline void Point::setRghtIdx(unsigned index) { this->rghtIdx = index; }

inline void Point::setLevel (unsigned level)  { this->level  = level; }
inline void Point::setTolIdx(unsigned tolIdx) { this->tolIdx = tolIdx; }

inline void Point::setCRUIter(CRUIter_t cruIter) { this->cruIter = cruIter; }


inline bool Point::operator!=(Point const & p) const { return !operator==(p); }
inline bool Point::operator==(Point const & p) const { return ((x == p.x) &&
							       (y == p.y)); }
inline double Point::getXMin() const { return x; }
inline double Point::getXMax() const { return x; }
inline double Point::getYMin() const { return y; }
inline double Point::getYMax() const { return y; }


inline double Point::sqrDistance(Point const & p) const
{ return (p.x-x)*(p.x-x)+(p.y-y)*(p.y-y); }


inline double Point::crossProduct(Point const & p, Point const & q) const
{ return (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y); }

inline bool Point::atLeft(Point const & p, Point const & q) const
{ return crossProduct(p, q) > 0; }

inline bool Point::atRight(Point const & p, Point const & q) const
{ return crossProduct(p, q) < 0; }

inline bool Point::onLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) == 0; }

inline bool Point::atLeftOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) >= 0; }

inline bool Point::atRghtOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) <= 0; }



#endif // __POINT_HH_
