#include "polyline.hh"

#include <iomanip>




/**
 * @brief Compute and store the distance between each vertex and its base
 * segment inside the range ('frstIdx', 'lastIdx').
 *****************************************************************************/
void Polyline::simplifyMRRDP(unsigned frstIdx, unsigned lastIdx,
			     unsigned level, double parentTolerance)
{
  /* Searching for farthest vertex between 'frstIdx' and 'lastIdx' */
  double   farthestSqrDistance = -1.0;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	(*this)[vertexIdx].sqrDistance((*this)[frstIdx], (*this)[lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  // Setting distance and tolerance of farthest vertex
  double distance  = sqrt(farthestSqrDistance);
  double tolerance = (distance < parentTolerance) ? distance : parentTolerance;

  (*this)[farthestIdx].setDistance (distance);
  (*this)[farthestIdx].setTolerance(tolerance);


  /* Iterating */
  if (frstIdx < farthestIdx-1)
    simplifyMRRDP(frstIdx, farthestIdx, level+1, tolerance);

  if (lastIdx > farthestIdx+1)
    simplifyMRRDP(farthestIdx, lastIdx, level+1, tolerance);


  /* Setting level  of farthest vertex */
  (*this)[farthestIdx].setLevel(level);

  /* Setting children of first and last index */
  (*this)[frstIdx].setRghtIdx(farthestIdx);
  (*this)[lastIdx].setLeftIdx(farthestIdx);
}



/**
 * Prints the list of points sequentially ordered, one per line.
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setfill('0') << setw(3) << vertexIdx << ") " << setfill(' ');
      (*this)[vertexIdx].print(ostr);
    }
}



/**
 * Prints out the number of points, followed by the list of points sequentially
 * ordered, one per line. Is a redefinition of the insertion operator '<<' in
 * stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;


  return ostr;
}



/**
 * First reads the number of points, and then the list of points. Is a
 * redefinition of the extraction operator '>>' in stream context. Accepts the
 * points in the given order by considering they are ordered correctely.
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    istr >> polyline[vertexIdx];


  return istr;
}
