#include "cru.hh"

#include <iomanip>

#define ERROR_DISTANCE 1E-20


// Initialization lits of crus;
list<CRU *> CRU::cleanCRUs;
list<CRU *> CRU::dirtyCRUs;



/**
 * @brief Determines the bounding box and diamond of the subpolyline beginning
 * at the vertex 'frstIdx' and ending at the vertex 'lastIdx'.
 *****************************************************************************/
CRU::CRU(Polyline & polyline, unsigned frstIdx, unsigned lastIdx)
  : polyline(polyline), nearTolerance(polyline.getWidth()),
    sqrNearTolerance(nearTolerance*nearTolerance), state(CLEAN)
{
  set(frstIdx, lastIdx);


  // Inserting new CRU in ordered list of 
  cruIter = cleanCRUs.begin();
  while (cruIter != cleanCRUs.end() &&
	 this->getMetric() < (*cruIter)->getMetric()) ++cruIter;

  cruIter = cleanCRUs.insert(cruIter, this);
}



/**
 * @brief Determines the bounding box and diamond of the subpolyline beginning
 * at the vertex 'frstIdx' and ending at the vertex 'lastIdx'.
 *****************************************************************************/
void CRU::set(unsigned frstIdx, unsigned lastIdx)
{
  // Setting indices of first and last vertices
  this->frstIdx = frstIdx;
  this->lastIdx = lastIdx;
  this->spltIdx = polyline.computeSplitVertex(frstIdx, lastIdx);


  // Computing bounding box and diamond
  xMin = xMax = polyline[frstIdx].getX();
  yMin = yMax = polyline[frstIdx].getY();

  pMin = pMax = polyline[frstIdx].getX()+polyline[frstIdx].getY();
  qMin = qMax = polyline[frstIdx].getX()-polyline[frstIdx].getY();

  // Handling rest of points
  for (unsigned vIdx = frstIdx+1; vIdx <= lastIdx; ++vIdx)
    {
      if      (polyline[vIdx].getX() < xMin) xMin = polyline[vIdx].getX();
      else if (polyline[vIdx].getX() > xMax) xMax = polyline[vIdx].getX();
      if      (polyline[vIdx].getY() < yMin) yMin = polyline[vIdx].getY();
      else if (polyline[vIdx].getY() > yMax) yMax = polyline[vIdx].getY();

      double pCurr = polyline[vIdx].getX()+polyline[vIdx].getY();
      double qCurr = polyline[vIdx].getX()-polyline[vIdx].getY();

      if      (pCurr < pMin) pMin = pCurr;
      else if (pCurr > pMax) pMax = pCurr;
      if      (qCurr < qMin) qMin = qCurr;
      else if (qCurr > qMax) qMax = qCurr;
    }

  xMin -= nearTolerance;  xMax += nearTolerance;
  yMin -= nearTolerance;  yMax += nearTolerance;

  pMin -= M_SQRT2*nearTolerance;  pMax += M_SQRT2*nearTolerance;
  qMin -= M_SQRT2*nearTolerance;  qMax += M_SQRT2*nearTolerance;
}



/**
 * Split this CRU into two CRUs at the vertex 'splitIdx'.  The two resulting
 * CRUs are placed in 'frstCru' and 'scndCru'.
 *****************************************************************************/
void CRU::correct(double tolerance)
{
  // Defining first and second CRUs
  CRU * frstCRU = NULL;
  CRU * scndCRU = NULL;


  // Allocating memory for usable CRUs
  if (spltIdx > frstIdx+1)
    {
      frstCRU = new CRU(polyline, frstIdx, spltIdx);
      frstCRU->pushFeature(polyline[lastIdx]);

      if (state == CLEAN &&
	  polyline[frstCRU->spltIdx].getMetric() >= tolerance)
	{
	  cleanCRUs.erase(frstCRU->cruIter);
	  frstCRU->cruIter = dirtyCRUs.insert(dirtyCRUs.end(), frstCRU);
	  frstCRU->state   = DIRTY;
	}
    }


  if (spltIdx < lastIdx-1)
    {
      scndCRU = new CRU(polyline, spltIdx, lastIdx);
      scndCRU->pushFeature(polyline[frstIdx]);

      if (state == CLEAN &&
	  polyline[scndCRU->spltIdx].getMetric() >= tolerance)
	{
	  cleanCRUs.erase(scndCRU->cruIter);
	  scndCRU->cruIter = dirtyCRUs.insert(dirtyCRUs.end(), scndCRU);
	  scndCRU->state   = DIRTY;
	}
    }


  // Inserting features in new CRUs
  for (featureIter_t featureIter = features.begin();
       featureIter != features.end(); ++featureIter)
    {
      unsigned findCrossIdx = featureIter->findCrossIdx(spltIdx);
      unsigned findProxtIdx = featureIter->findProxtIdx(spltIdx);

      if (spltIdx > frstIdx+1)
	frstCRU->pushFeature(*featureIter, true,  findCrossIdx, findProxtIdx);

      if (spltIdx < lastIdx-1)
	scndCRU->pushFeature(*featureIter, false, findCrossIdx, findProxtIdx);
    }


  // Setting neighbouring
  if (spltIdx > frstIdx+1 &&
      spltIdx < lastIdx-1)
    setNeighbouring(*frstCRU, *scndCRU);


  // Setting neighbourhood relationship
  for (list<neighbour_t>::iterator neighbourIter = neighbours.begin();
       neighbourIter != neighbours.end(); ++neighbourIter)
    {
      CRU & cru = *neighbourIter->cru;

      cru.neighbours.erase(neighbourIter->myIter);
      cru.pushFeature(polyline[spltIdx]);

      if (spltIdx > frstIdx+1) setNeighbouring(*frstCRU, cru);
      if (spltIdx < lastIdx-1) setNeighbouring(*scndCRU, cru);
    }


  // Correcting links of neighbouring vertices
  polyline[frstIdx].setNextIdx(spltIdx);
  polyline[lastIdx].setPrevIdx(spltIdx);

  // Correcting its links
  polyline[spltIdx].setPrevIdx(frstIdx);
  polyline[spltIdx].setNextIdx(lastIdx);

  // Setting metric of split vertex
  polyline[spltIdx].setMetric(tolerance);


  // Removing itself from one of list of dirty CRUs
  dirtyCRUs.erase(cruIter);


  // Deallocating memory of this CRU -- bye bye CRU
  delete this;
}



/**
 * @brieft Push a new feature point 'point' into the features list and
 * compute its sidedness with respect to the subpolyline of this CRU.
 *****************************************************************************/
void CRU::pushFeature(Point const & point)
{
  // Discarding limits
  if (!isInside(point) ||
      point == polyline[frstIdx] ||
      point == polyline[lastIdx]) return;


  // Storing crossings and proximities
  list<unsigned> crossingList;
  list<unsigned> proximitList;

  bool atLeft = polyline[frstIdx].getX() <= point.getX();

  for (unsigned prevIdx = frstIdx, currIdx = frstIdx+1;
       currIdx <= lastIdx; prevIdx = currIdx++)
    {
      // Crossings
      if (polyline[currIdx].getX() <= point.getX())
	{
	  if (!atLeft && point.atLeft(polyline[prevIdx], polyline[currIdx]))
	    crossingList.push_back(currIdx);

	  atLeft = true;
	}
      else if (polyline[currIdx].getX() > point.getX())
	{
	  if (atLeft && point.atRight(polyline[prevIdx], polyline[currIdx]))
	    crossingList.push_back(currIdx);

	  atLeft = false;
	}

      // Proximities
      if (point.sqrDistance(polyline[prevIdx],
			    polyline[currIdx]) <= sqrNearTolerance)
	proximitList.push_back(currIdx);
    }


  // Determining whether base segment cross vertical ray from feature
  bool lLeft = polyline[lastIdx].getX() <= point.getX();
  bool fLeft = polyline[frstIdx].getX() <= point.getX();
  bool baseSegmentCross =
    ( lLeft && !fLeft && point.atRight(polyline[lastIdx],polyline[frstIdx])) ||
    (!lLeft &&  fLeft && point.atLeft (polyline[lastIdx],polyline[frstIdx]));


  // Determining whether base segment is near to feature
  double baseSegmentSqrDistance = point.sqrDistance(polyline[frstIdx],
						    polyline[lastIdx]);

  bool baseSegmentProxt = baseSegmentSqrDistance <= sqrNearTolerance;
  bool baseSegmentError = baseSegmentSqrDistance <= ERROR_DISTANCE;


  // Pushing feature in list
  features.push_back(Feature(&point, crossingList, proximitList,
			     baseSegmentCross, baseSegmentProxt,
			     baseSegmentError));


  // Updating convexhull's consistency
  if (state == CLEAN && features.back().getInconsistent())
    {
      cleanCRUs.erase(cruIter);
      cruIter = dirtyCRUs.insert(dirtyCRUs.end(), this);
      state   = DIRTY;
    }
}



/**
 * @brief Push a new feature point 'feature' into the features list and
 * compute its sidedness with respect to the subpolyline of this CRU.
 *****************************************************************************/
void CRU::pushFeature(Feature const & feature, bool firstCRU,
		      unsigned findCrossIdx, unsigned findProxtIdx)
{
  if (!isInside(feature.getPoint())) return;


  // Pushing feature in list
  features.push_back(feature);
  Point const & point = features.back().getPoint();


  // Updating indices
  if (firstCRU)
    {
      features.back().setLastCrossIdx(findCrossIdx);
      features.back().setLastProxtIdx(findProxtIdx);
    }
  else
    {
      features.back().setFrstCrossIdx(findCrossIdx);
      features.back().setFrstProxtIdx(findProxtIdx);
    }


  // Determining whether base segment cross vertical upward ray
  bool lLeft = polyline[lastIdx].getX() <= point.getX();
  bool fLeft = polyline[frstIdx].getX() <= point.getX();
  bool baseSegmentCross =
    ( lLeft && !fLeft && point.atRight(polyline[lastIdx],polyline[frstIdx])) ||
    (!lLeft &&  fLeft && point.atLeft (polyline[lastIdx],polyline[frstIdx]));


  // Determining whether base segment is near to feature
  double baseSegmentSqrDistance = point.sqrDistance(polyline[frstIdx],
						    polyline[lastIdx]);

  bool baseSegmentProxt = baseSegmentSqrDistance <= sqrNearTolerance;
  bool baseSegmentError = baseSegmentSqrDistance <= ERROR_DISTANCE;


  // Setting relationships with base segment
  features.back().setBaseSegmentCross(baseSegmentCross);
  features.back().setBaseSegmentProxt(baseSegmentProxt);
  features.back().setBaseSegmentError(baseSegmentError);


  // Updating convexhull's consistency
  if (state == CLEAN && features.back().getInconsistent())
    {
      cleanCRUs.erase(cruIter);
      cruIter = dirtyCRUs.insert(dirtyCRUs.end(), this);
      state   = DIRTY;
    }
}



/**
 * Prints the list of point indeces sequentially ordered, one per line.
 *****************************************************************************/
void CRU::print(ostream & ostr) const
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    ostr << setw(3) << polyline[vertexIdx] << endl;
}



/**
 * @brief Set neighbouring relationship between 'cru1' and 'cru2'.
 *****************************************************************************/
void setNeighbouring(CRU & cru1, CRU & cru2)
{
  if (!cru1.overlaps(cru2)) return;


  // Pushing neighbours
  cru1.neighbours.push_back(neighbour_t(&cru2));
  cru2.neighbours.push_back(neighbour_t(&cru1));


  // Setting iterators
  cru1.neighbours.back().myIter = --cru2.neighbours.end();
  cru2.neighbours.back().myIter = --cru1.neighbours.end();
}



/**
 * Prints out the number of indeces, followed by the list of point indeces
 * sequentially ordered, one per line.  It is a redefinition of the
 * insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, CRU const & cru)
{
  ostr << "----------------------------------" << endl;


  // Range
  ostr << "CRU (" << setw(2) << cru.lastIdx-cru.frstIdx+1 << ") "
       << setw(10) << &cru
       << " from " << setw(2) << cru.frstIdx << " ("
       << cru.polyline[cru.frstIdx] << ')'<< endl
       << "                      to " << setw(2) << cru.lastIdx << " ("
       << cru.polyline[cru.lastIdx] << ')'<< endl;


  // Bounding box
  ostr << "BB x = ("
       << setw(14) << cru.xMin << ", "
       << setw(14) << cru.xMax << ')' << endl
       << "   y = ("
       << setw(14) << cru.yMin << ", "
       << setw(14) << cru.yMax << ')' << endl;


  // Bounding diamond
  ostr << "BD p = ("
       << setw(14) << cru.pMin << ", "
       << setw(14) << cru.pMax << ')' << endl
       << "   q = ("
       << setw(14) << cru.qMin << ", "
       << setw(14) << cru.qMax << ')' << endl;


  // Neighbours
  ostr << "NEIGHBOURS: " << cru.neighbours.size() << endl;

  unsigned neighbourIdx = 0;
  for (list<neighbour_t>::const_iterator neighbourIter =
	 cru.neighbours.begin(); neighbourIter != cru.neighbours.end();
       ++neighbourIter)
    ostr << setw(2) << neighbourIdx++
	 << ")  CRU (" << setw(2)
	 << neighbourIter->cru->lastIdx-neighbourIter->cru->frstIdx+1 << ") "
	 << setw(10) << neighbourIter->cru
	 << " from "  << setw(2) << neighbourIter->cru->frstIdx
	 << " to "    << setw(2) << neighbourIter->cru->lastIdx << endl;


  // Features
  ostr << "FEATURES: " << cru.features.size() << endl;

  for (constFeatureIter_t featIter = cru.features.begin();
       featIter != cru.features.end(); ++featIter)
    ostr << "  " << *featIter << endl;

  ostr << "----------------------------------" << endl;


  return ostr;
}
