#ifndef __CRU_HH_
#define __CRU_HH_


#include "rectangle.hh"
#include "polyline.hh"
#include "feature.hh"

#include <functional>
#include <list>


using namespace std;




/**
 * External types
 *****************************************************************************/
class CRU;
typedef list<CRU *>::iterator            CRUIter_t;
typedef list<CRU *>::const_iterator constCRUIter_t;


/**
 * Internal types
 *****************************************************************************/
typedef list<Feature>::iterator            featureIter_t;
typedef list<Feature>::const_iterator constFeatureIter_t;

enum state_t { CLEAN = 0, DIRTY = 1 };


struct neighbour_t
{
  neighbour_t(CRU * cru) : cru(cru) {}

  CRU * cru;
  list<neighbour_t>::iterator myIter;
};



/**
 * 'CRU' is the class for the Consistency Recovery Unit.
 *****************************************************************************/
class CRU
  : public Rectangle {

public:

  // Constructors
  CRU(Polyline & polyline, unsigned frstIdx, unsigned lastIdx);

  // Virtual destructor
  virtual ~CRU();


  // Determining bounding box of 'polyline' between 'frstIdx' and 'lastIdx'
  void set(unsigned frstIdx, unsigned lastIdx);


  // Spliting convex hull at 'splitIdx'
  void correct(double tolerance);


  // Determines if 'point' is inside its bounding box
  bool isInside(Point const & point);


  // Pushes a new feature into the features' list
  void pushFeature(Point   const & point);
  void pushFeature(Feature const & feature, bool firstSubPol,
		   unsigned findCrossIdx, unsigned findProxtIdx);


  // Get functions
  Polyline const & getPolyline() const;

  double getNearTolerance() const;
  double getMetric() const;

  unsigned getFrstIdx() const;
  unsigned getLastIdx() const;
  unsigned getSpltIdx() const;

  Point const & getFrstPoint() const;
  Point const & getLastPoint() const;
  Point const & getSpltPoint() const;

  unsigned numberOfFeatures() const;


  // Set function
  void setNearTolerance(double nearTolerance);


  // Test if cru overlaps this cru
  bool overlaps(CRU const & cru) const;


  // Bouding box get functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;


  // Neighbouring function
  friend void setNeighbouring(CRU & cru1, CRU & cru2);


  // Friendship relation of insertion operator
  friend ostream & operator<<(ostream & ostr, CRU const & cru);


  // Printing function
  void print(ostream & ostr) const;


  // Static list of CRUs
  static list<CRU *> cleanCRUs;
  static list<CRU *> dirtyCRUs;


private:

  // Referenced polyline
  Polyline & polyline;

  // Limit indices of convex hull
  unsigned frstIdx;
  unsigned lastIdx;
  unsigned spltIdx;

  // Tolerance for minimum distance
  double    nearTolerance;
  double sqrNearTolerance;

  // List of features
  list<Feature> features;

  // Vector of neighbours
  list<neighbour_t> neighbours;

  // Sidedness test
  state_t state;

  // CRU iterator to list of CRUs (dirty or clean)
  CRUIter_t cruIter;

  // Bounding box coordinates
  double xMin, xMax;
  double yMin, yMax;

  // Bounding diamond coordinates
  double pMin, pMax;
  double qMin, qMax;
};



inline CRU::~CRU() {}



inline Polyline const & CRU::getPolyline() const { return polyline; }

inline double CRU::getNearTolerance() const { return nearTolerance; }
inline double CRU::getMetric() const { return polyline[spltIdx].getMetric(); }

inline unsigned CRU::getFrstIdx() const { return frstIdx; }
inline unsigned CRU::getLastIdx() const { return lastIdx; }
inline unsigned CRU::getSpltIdx() const { return spltIdx; }

inline Point const & CRU::getFrstPoint() const { return polyline[frstIdx]; }
inline Point const & CRU::getLastPoint() const { return polyline[lastIdx]; }
inline Point const & CRU::getSpltPoint() const { return polyline[spltIdx]; }

inline unsigned CRU::numberOfFeatures() const { return features.size(); }


inline void CRU::setNearTolerance(double nearTolerance)
{
  this->   nearTolerance = nearTolerance;
  this->sqrNearTolerance = nearTolerance*nearTolerance;
}


inline bool CRU::overlaps(CRU const & cru) const
{
  // Testing bounding box
  if (cru.xMax < xMin || cru.xMin > xMax ||
      cru.yMax < yMin || cru.yMin > yMax)
    return false;

  // Testing bounding diamond
  if (cru.pMax < pMin || cru.xMin > xMax ||
      cru.pMax < pMin || cru.yMin > yMax)
    return false;

  return true;
}


inline bool CRU::isInside(Point const & point)
{
  // Testing bounding box
  if (point.getX() < xMin || point.getX() > xMax ||
      point.getY() < yMin || point.getY() > yMax)
    return false;

  // Testing bounding diamond
  double p = point.getX()+point.getY(); if (p < pMin || p > pMax) return false;
  double q = point.getX()-point.getY(); if (q < qMin || q > qMax) return false;

  return true;
}


inline double CRU::getXMin() const { return xMin; }
inline double CRU::getXMax() const { return xMax; }
inline double CRU::getYMin() const { return yMin; }
inline double CRU::getYMax() const { return yMax; }



#endif // __CRU_HH_
