#ifndef __FEATURE_HH_
#define __FEATURE_HH_


#include "point.hh"

#include <iostream>
#include <list>


using namespace std;



/**
 * This class implements the features.
 */
class Feature {

public:

  // Constructors
  Feature(Point const * point,
	  list<unsigned> const & crossingList,
	  list<unsigned> const & proximitList,
	  bool baseSegmentCross, bool baseSegmentProxt, bool baseSegmentError);

  Feature(Feature const & feature);


  // Destructor
  ~Feature();


  // Get functions
  Point const & getPoint() const;

  unsigned getFrstCrossIdx() const; 
  unsigned getLastCrossIdx() const;

  unsigned getFrstProxtIdx() const; 
  unsigned getLastProxtIdx() const;

  bool getInconsistent() const;


  // Set function
  void setFrstCrossIdx(unsigned frstCrossIdx); 
  void setLastCrossIdx(unsigned lastCrossIdx);

  void setFrstProxtIdx(unsigned frstProxtIdx); 
  void setLastProxtIdx(unsigned lastProxtIdx);

  void setBaseSegmentCross(bool baseSegmentCross); 
  void setBaseSegmentProxt(bool baseSegmentProxt); 
  void setBaseSegmentError(bool baseSegmentError); 

  // Find functions
  unsigned findCrossIdx(unsigned splitIdx) const;
  unsigned findProxtIdx(unsigned splitIdx) const;


  // Operators
  bool operator==(Feature const & feature) const;
  bool operator< (Feature const & feature) const;


  // Extraction operator
  friend ostream & operator<<(ostream & ostr, Feature const & feature);


private:

  Point const * point; /// Pointer to point

  // List of crossings
  unsigned  nCrossings;
  unsigned * crossings;

  // List of proximities
  unsigned  nProximits;
  unsigned * proximits;

  // Indices
  unsigned frstCrossIdx;
  unsigned lastCrossIdx;

  unsigned frstProxtIdx;
  unsigned lastProxtIdx;

  // Relationships with base segment
  bool baseSegmentCross;
  bool baseSegmentProxt;
  bool baseSegmentError;
};


inline Feature::Feature(Feature const & feature)
  : point(feature.point),

    nCrossings(feature.nCrossings), crossings(feature.crossings),
    nProximits(feature.nProximits), proximits(feature.proximits),
 
    frstCrossIdx(feature.frstCrossIdx), lastCrossIdx(feature.lastCrossIdx),
    frstProxtIdx(feature.frstProxtIdx), lastProxtIdx(feature.lastProxtIdx),

    baseSegmentCross(feature.baseSegmentCross),
    baseSegmentProxt(feature.baseSegmentProxt),
    baseSegmentError(feature.baseSegmentError)
{
  ++crossings[nCrossings];
}

inline Feature::~Feature()
{
  if (--crossings[nCrossings] == 0)
    {
      delete [] crossings;
      delete [] proximits;
    }
}


inline Point const & Feature::getPoint() const { return *point; }

inline unsigned Feature::getFrstCrossIdx() const { return frstCrossIdx; }
inline unsigned Feature::getLastCrossIdx() const { return lastCrossIdx; }

inline unsigned Feature::getFrstProxtIdx() const { return frstProxtIdx; }
inline unsigned Feature::getLastProxtIdx() const { return lastProxtIdx; }

inline bool Feature::getInconsistent() const
{
  return ((baseSegmentError) ||
	  (baseSegmentProxt && lastProxtIdx-frstProxtIdx == 0) ||
	  (baseSegmentCross+lastCrossIdx-frstCrossIdx)%2 == 1) ? true : false;
}


inline void Feature::setFrstCrossIdx(unsigned fcIdx) { frstCrossIdx = fcIdx; }
inline void Feature::setLastCrossIdx(unsigned lcIdx) { lastCrossIdx = lcIdx; }

inline void Feature::setFrstProxtIdx(unsigned fpIdx) { frstProxtIdx = fpIdx; }
inline void Feature::setLastProxtIdx(unsigned lpIdx) { lastProxtIdx = lpIdx; }

inline void Feature::setBaseSegmentCross(bool bsC) { baseSegmentCross = bsC; }
inline void Feature::setBaseSegmentProxt(bool bsP) { baseSegmentProxt = bsP; }
inline void Feature::setBaseSegmentError(bool bsE) { baseSegmentError = bsE; }

inline bool Feature::operator==(Feature const & feature) const
{ return *point == *feature.point; }

inline bool Feature::operator<(Feature const & feature) const
{
  return point->getY() == feature.point->getY()
    ?    point->getX() <  feature.point->getX()
    :    point->getY() >  feature.point->getY();
}




#endif // __FEATURE_HH_
