#include "map.hh"
#include "rips.hh"

#include <iomanip>
#include <algorithm>
#include <typeinfo>
#include <sys/time.h>




/**
 * @brief Simplifies polylines of this map using VIS algorithm for tolerance
 * 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void Map::simplifyMRTCVIS(double nearTolerance)
{
  // Setting width of polylines
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      polylines[polIdx].setWidth(nearTolerance);
      polylines[polIdx].simplifyMRVIS();
    }


  // Adding initial features
  initCRUs();


  // Simplifying for infty tolerance
  simplifyTCVIS(INFTY);


  // Correcting sidedness errors
  while (!CRU::cleanCRUs.empty())
    {
      double tolerance = CRU::cleanCRUs.front()->getMetric();

      CRU::cleanCRUs.front()->correct(tolerance);
      simplifyTCVIS(tolerance);
    }
}



/**
 *
 *****************************************************************************/
void Map::initCRUs()
{
  // List of rectangles
  list<Rectangle const *> rectangles;


  // Creating CRUs and pushing them in list of rectangles
  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
    if (polylines[polIdx].numberOfVertices() > 2)
      rectangles.push_back(new CRU(polylines[polIdx], 0,
				   polylines[polIdx].numberOfVertices()-1));
    else
      {
	rectangles.push_back(&polylines[polIdx].front());
	rectangles.push_back(&polylines[polIdx].back());
      }


  // Adding point features in list of rectangles
  for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
    rectangles.push_back(&points[pointIdx]);


  // Identifying neighbours
  RIPS rips;
  rips.solve(rectangles);


  // Adding points and setting neighbours
  RectPairList const & rectPairs = rips.getRectPairs();

  for (constRectPairIter_t rectPairIter = rectPairs.begin();
       rectPairIter != rectPairs.end(); ++rectPairIter)
    {
      Rectangle * frstRect = const_cast<Rectangle *>(rectPairIter->first);
      Rectangle * scndRect = const_cast<Rectangle *>(rectPairIter->second);

      if (typeid(*frstRect) == typeid(CRU) &&
	  typeid(*scndRect) == typeid(CRU))
	{
	  CRU & cru1 = *reinterpret_cast<CRU *>(frstRect);
	  CRU & cru2 = *reinterpret_cast<CRU *>(scndRect);

	  setNeighbouring(cru1, cru2);

	  cru1.pushFeature(cru2.getFrstPoint());
	  cru1.pushFeature(cru2.getLastPoint());

	  cru2.pushFeature(cru1.getFrstPoint());
	  cru2.pushFeature(cru1.getLastPoint());
	}

      else if (typeid(*frstRect) == typeid(CRU))
	reinterpret_cast<CRU *>(frstRect)->
	  pushFeature(*reinterpret_cast<Point *>(scndRect));

      else if (typeid(*scndRect) == typeid(CRU))
	reinterpret_cast<CRU *>(scndRect)->
	  pushFeature(*reinterpret_cast<Point *>(frstRect));
    }
}



/**
 * @brief Simplifies map 'maptcrdp' with Silva's version of VIS algorithm for
 * squared tolerance 'sqrTolerance', filling up the CRUs in 'crus'.
 *****************************************************************************/
void Map::simplifyTCVIS(double tolerance)
{
  // Correcting sidedness errors
  while (!CRU::dirtyCRUs.empty())
    CRU::dirtyCRUs.front()->correct(tolerance);
}



/**
 * @brieft Prints number of polylines, followed by polylines themselves.  Then,
 * prints number of points, followed by points themselves.
 *****************************************************************************/
void Map::print(ostream & ostr) const
{
  /* Setting needed flags */
  ostr.setf(ios::scientific, ios::floatfield);
  ostr.setf(ios::uppercase);
  ostr.precision(COORD_PRECISION);


  /* Printing polylines */
  ostr << numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      ostr << endl;
      polylines[polIdx].print(ostr);
    }


  /* Printing points */
  ostr << endl << endl << numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < numberOfPoints(); ++pointIdx)
    {
      points[pointIdx].print(ostr);
      ostr << endl;
    }
}



/**
 * @brief Prints number of polylines, followed by polylines themselves.  Then,
 * prints number of points, followed by points themselves.  Is a redefinition
 * of insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Map const & map)
{
  /* Setting needed flags */
  ostr.setf(ios::scientific, ios::floatfield);
  ostr.setf(ios::uppercase);
  ostr.precision(COORD_PRECISION);


  /* Printing polylines */
  ostr << map.numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    ostr << endl << map.polylines[polIdx];


  /* Printing points */
  ostr << endl << endl << map.numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    ostr << map.points[pointIdx] << endl;


  return ostr;
}



/**
 * @brief First reads number of polylines, followed by polylines themselves.
 * Then, reads number of points, followed by points themselves.  Is a
 * redefinition of the extraction operator '>>' in stream context. Accepts the
 * points in the given order by considering they are ordered correctly.
 *****************************************************************************/
istream & operator>>(istream & istr, Map & map)
{
  /* Reading polylines */
  unsigned nPolylines;

  istr >> nPolylines;
  map.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    istr >> map.polylines[polIdx];


  /* Reading points */
  unsigned nPoints;

  istr >> nPoints;
  map.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    istr >> map.points[pointIdx];


  return istr;
}
