#ifndef __POINT_HH_
#define __POINT_HH_


#include "rectangle.hh"

#include <list>
#include <cmath>
#include <iostream>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;



struct vertex_t {

  vertex_t() : vertexIdx(0), effectiveArea(0.0) {}
  vertex_t(unsigned vertexIdx, double effectiveArea)
    : vertexIdx(vertexIdx), effectiveArea(effectiveArea) {}

  bool operator<(vertex_t const & vertex) const
  {
    return effectiveArea != vertex.effectiveArea
      ? effectiveArea < vertex.effectiveArea
      : vertexIdx     < vertex.vertexIdx;;
  }

  unsigned vertexIdx;
  double effectiveArea;
};

typedef list<vertex_t>::iterator            vertexIter_t;
typedef list<vertex_t>::const_iterator constVertexIter_t;



/**
 * @brief This class contains cartesian coordinates of a point in 2D space. It
 * presents geometric transformations and util functions usually applied to
 * points.
 *****************************************************************************/
class Point
  : public Rectangle {

public:

  // Constructors
  Point();                    // Default
  Point(Point const & point); // Copy
  Point(double x, double y);  // Cartesian coordinates


  // Get rectangle functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;

  // Get coordinates functions
  double getX() const;
  double getY() const;

  // Indices access functions
  unsigned getPrevIdx() const;
  unsigned getNextIdx() const;

  // Functions 'get' for the right and left children' indices
  unsigned getLeftIdx() const;
  unsigned getRghtIdx() const;

  // Functions 'get' for level of detail
  unsigned getLevel() const;

  // Set vertex index function
  vertexIter_t getVertexIter() const;

  // Get metric function
  double getMetric() const;


  // Set functions
  void setX(double x);
  void setY(double y);

  // Indices modification functions
  void setPrevIdx(unsigned prevIdx);
  void setNextIdx(unsigned nextIdx);

  // Functions 'get' for the right and left sons' indices
  void setLeftIdx(unsigned leftIdx);
  void setRghtIdx(unsigned rghtIdx);

  // Functions 'set' for level of detail
  void setLevel (unsigned level);

  // Set vertex iterator function
  void setVertexIter(vertexIter_t vertexIter);

  // Set tolerance function
  void setMetric(double metric);
  void setMetric(Point const & p, Point const & q);


  // Squared distances beween points and between a point and a segment
  double sqrDistance(Point const & p) const;
  double sqrDistance(Point const & p, Point const & q) const;


  // Triangle are for three points
  double triangleArea(Point const & p, Point const & q) const;


  // The cross product between 'pq' and 'p(*this)'
  double crossProduct(Point const & p, Point const & q) const;

  // Test for sidedness
  bool atLeft (Point const & p, Point const & q) const;
  bool atRight(Point const & p, Point const & q) const;
  bool onLine (Point const & p, Point const & q) const;

  bool atLeftOrOnLine(Point const & p, Point const & q) const;
  bool atRghtOrOnLine(Point const & p, Point const & q) const;


  // Operators of equality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Friend operators for input and output handle
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);

  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Cartesian coordinates
  double x;
  double y;

  // Indices of previous and next vertices in polyline
  unsigned prevIdx;
  unsigned nextIdx;

  // Indices of children in RDP's tree
  unsigned leftIdx;
  unsigned rghtIdx;

  // Level of detail
  unsigned level;

  // Vertex iterator
  vertexIter_t vertexIter;

  // Metric
  double metric;
};



inline Point::Point()
  : x(0.0),
    y(0.0),

    prevIdx(0),
    nextIdx(0),

    leftIdx(0),
    rghtIdx(0),

    level (0),

    metric(0.0) {}

inline Point::Point(Point const & point)
  : x(point.x),
    y(point.y),

    prevIdx(point.prevIdx),
    nextIdx(point.nextIdx),

    leftIdx(point.leftIdx),
    rghtIdx(point.rghtIdx),

    level (point.level),

    metric(point.metric) {}

inline Point::Point(double x, double y)
  : x(x),
    y(y),

    prevIdx(0),
    nextIdx(0),

    leftIdx(0),
    rghtIdx(0),

    level (0),

    metric(0.0) {}


inline double Point::getXMin() const { return x; }
inline double Point::getXMax() const { return x; }
inline double Point::getYMin() const { return y; }
inline double Point::getYMax() const { return y; }

inline double Point::getX() const { return x; }
inline double Point::getY() const { return y; }

inline unsigned Point::getPrevIdx() const { return prevIdx; }
inline unsigned Point::getNextIdx() const { return nextIdx; }

inline unsigned Point::getLeftIdx() const { return leftIdx; }
inline unsigned Point::getRghtIdx() const { return rghtIdx; }

inline unsigned Point::getLevel() const  { return level; }

inline vertexIter_t Point::getVertexIter() const { return vertexIter; }

inline double Point::getMetric() const { return metric; }


inline void Point::setX(double x) { this->x = x; }
inline void Point::setY(double y) { this->y = y; }

inline void Point::setPrevIdx(unsigned prevIdx) { this->prevIdx = prevIdx; }
inline void Point::setNextIdx(unsigned nextIdx) { this->nextIdx = nextIdx; }

inline void Point::setLeftIdx(unsigned index) { this->leftIdx = index; }
inline void Point::setRghtIdx(unsigned index) { this->rghtIdx = index; }

inline void Point::setLevel (unsigned level)  { this->level  = level; }

inline void Point::setVertexIter(vertexIter_t vIt) { vertexIter = vIt; }

inline void Point::setMetric(double metric) { this->metric = metric; }
inline void Point::setMetric(Point const & p, Point const & q)
{ metric = triangleArea(p, q); }


inline bool Point::operator!=(Point const & p) const { return !operator==(p); }
inline bool Point::operator==(Point const & p) const { return ((x == p.x) &&
							       (y == p.y)); }

inline double Point::sqrDistance(Point const & p) const
{ return (p.x-x)*(p.x-x)+(p.y-y)*(p.y-y); }

inline double Point::triangleArea(Point const & p, Point const & q) const
{ return fabs(crossProduct(p, q)/2.0); }


inline double Point::crossProduct(Point const & p, Point const & q) const
{ return (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y); }

inline bool Point::atLeft(Point const & p, Point const & q) const
{ return crossProduct(p, q) > 0; }

inline bool Point::atRight(Point const & p, Point const & q) const
{ return crossProduct(p, q) < 0; }

inline bool Point::onLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) == 0; }

inline bool Point::atLeftOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) >= 0; }

inline bool Point::atRghtOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) <= 0; }



#endif // __POINT_HH_
