#include "polyline.hh"

#include <iomanip>
#include <values.h>




/**
 * @brief Applies Visvalingam's algorithm to this polyline.
 *
 * The simplification is carried out with a minimum effective area of
 * @tolerance.
 *
 * @param tolerance The minimum effective area.
 *****************************************************************************/
void Polyline::simplifyMRVIS()
{
  list<vertex_t> vertices;

  front().setLevel(0);
  front().setLevel(0);


  // First step
  for (unsigned vertexIdx = 1; vertexIdx < numberOfVertices()-1; ++vertexIdx)
    {
      (*this)[vertexIdx].setMetric((*this)[vertexIdx-1], (*this)[vertexIdx+1]);
      vertices.push_back(vertex_t(vertexIdx, (*this)[vertexIdx].getMetric()));
    }


  // Sorting list of vertices
  vertices.sort();

  for (vertexIter_t vertexIter = vertices.begin();
       vertexIter != vertices.end(); ++vertexIter)
    (*this)[vertexIter->vertexIdx].setVertexIter(vertexIter);


  // Second step: eliminate vertices with smallest effective area
  double previousEffectiveArea = -MAXDOUBLE;

  for (unsigned nVertices = numberOfVertices(); nVertices > 2; --nVertices)
    {
      // Getting vertex with smallest index and effective area
      double   smallestEffectiveArea = MAXDOUBLE;
      unsigned smallestIdx           = 0;

      if (!vertices.empty())
	{
	  smallestIdx           = vertices.front().vertexIdx;
	  smallestEffectiveArea = (*this)[smallestIdx].getMetric();
	  vertices.pop_front();
	}


      // Setting effective area
      if (previousEffectiveArea > smallestEffectiveArea)
	(*this)[smallestIdx].setMetric(previousEffectiveArea);


      // Updating previous effective area
      previousEffectiveArea = (*this)[smallestIdx].getMetric();


      // Initializing references to neighbours
      Point & prevPoint = (*this)[(*this)[smallestIdx].getPrevIdx()];
      Point & nextPoint = (*this)[(*this)[smallestIdx].getNextIdx()];


      // Updating previous and next vertices of its neighbours
      prevPoint.setNextIdx((*this)[smallestIdx].getNextIdx());
      nextPoint.setPrevIdx((*this)[smallestIdx].getPrevIdx());


      // Updating children of its neighbours
      prevPoint.setRghtIdx(smallestIdx);
      nextPoint.setLeftIdx(smallestIdx);


      // Updating levels of its neighbours
      prevPoint.setLevel(nVertices-2);
      nextPoint.setLevel(nVertices-2);


      // Updating effective area of its neighbours
      if (&prevPoint != &front())
	{
	  prevPoint.setMetric((*this)[prevPoint.getPrevIdx()],
			      (*this)[prevPoint.getNextIdx()]);

	  vertices.erase(prevPoint.getVertexIter());

	  vertex_t vertex((*this)[smallestIdx].getPrevIdx(),
			  prevPoint.getMetric());

	  vertexIter_t vertexIter = vertices.begin();
	  while (vertexIter != vertices.end() && *vertexIter < vertex)
	    ++vertexIter;

	  prevPoint.setVertexIter(vertices.insert(vertexIter, vertex));
	}

      if (&nextPoint != &back())
	{
	  nextPoint.setMetric((*this)[nextPoint.getPrevIdx()],
			      (*this)[nextPoint.getNextIdx()]);

	  vertices.erase(nextPoint.getVertexIter());

	  vertex_t vertex((*this)[smallestIdx].getNextIdx(),
			  nextPoint.getMetric());

	  vertexIter_t vertexIter = vertices.begin();
	  while (vertexIter != vertices.end() && *vertexIter < vertex)
	    ++vertexIter;

	  nextPoint.setVertexIter(vertices.insert(vertexIter, vertex));
	}
    }
}



/**
 * @brief Prints the list of points sequentially ordered, one per line.
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setfill('0') << setw(3) << vertexIdx << ") " << setfill(' ');
      (*this)[vertexIdx].print(ostr);
    }
}



/**
 * @brief Prints out the number of points, followed by the list of points
 * sequentially ordered, one per line. Is a redefinition of the insertion
 * operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx <
	 polyline.numberOfVertices(); ++vertexIdx)
    ostr << polyline[vertexIdx]
	 << setw(COORD_WIDTH) << polyline[vertexIdx].getMetric() << endl;


  return ostr;
}



/**
 * @brief First reads the number of points, and then the list of points. Is a
 * redefinition of the extraction operator '>>' in stream context. Accepts the
 * points in the given order by considering they are ordered correctely.
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);

  for (unsigned vertexIdx = 0; vertexIdx <
	 polyline.numberOfVertices(); ++vertexIdx)
    {
      istr >> polyline[vertexIdx];

      polyline[vertexIdx].setPrevIdx(vertexIdx-1);
      polyline[vertexIdx].setNextIdx(vertexIdx+1);
    }

  polyline.front().setMetric(INFTY);
  polyline.back(). setMetric(INFTY);

  return istr;
}
