#ifndef __CRU_HH_
#define __CRU_HH_


#include "rectangle.hh"
#include "polyline.hh"
#include "feature.hh"

#include <deque>
#include <list>


using namespace std;



/**
 * @brief This structure implements a neighbour CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class CRU;
struct neighbour_t
{
  neighbour_t(CRU * cru) : cru(cru) {}

  CRU * cru;
  list<neighbour_t>::iterator myIter;
};


/**
 * @brief This structure implements the state of a CRU.  It can be clean,
 * which means without inconsistencies, or dirty, which means with
 * inconsistencies.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
enum state_t { CLEAN = 0, DIRTY = 1 };


/**
 * Useful typedefs
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
typedef list<CRU *>::iterator            CRUIter_t;
typedef list<CRU *>::const_iterator constCRUIter_t;

typedef list<Feature>::iterator            featureIter_t;
typedef list<Feature>::const_iterator constFeatureIter_t;



/**
 * @brief This class implements a Consitency Recovery Unit.
 * 
 * It is used to recover the consistency between a line segment in a
 * simplified polyline with its correspondent subpolyline in the original
 * polyline.  It is a subclass of the class Rectangle, because CRUs have be
 * used by RIPS class transparently.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
class CRU
  : public Rectangle {

public:

  // Constructors
  CRU(Polyline & polyline, unsigned frstIdx, unsigned lastIdx);

  // Virtual destructor
  virtual ~CRU();


  // Determining bounding box of 'polyline' between 'frstIdx' and 'lastIdx'
  void set(unsigned frstIdx, unsigned lastIdx);


  // Spliting convex hull at 'splitIdx'
  void correct(double tolerance);


  // Determines if 'point' is inside its bounding box
  bool isInside(Point const & point);


  // Pushes a new feature into the features' list
  void pushFeature(Point   const & point);
  void pushFeature(Feature const & feature, bool firstSubPol,
		   unsigned findCrossIdx, unsigned findProxtIdx);


  // Get functions
  Polyline const & getPolyline() const;

  double getNearTolerance() const;

  unsigned getFrstIdx() const;
  unsigned getLastIdx() const;
  unsigned getSpltIdx() const;

  Vertex const & getFrstVertex() const;
  Vertex const & getLastVertex() const;
  Vertex const & getSpltVertex() const;

  unsigned numberOfFeatures() const;


  // Set function
  void setNearTolerance(double nearTolerance);


  // Test if cru overlaps this cru
  bool overlaps(CRU const & cru) const;


  // Bouding box get functions
  virtual double getXMin() const;
  virtual double getXMax() const;
  virtual double getYMin() const;
  virtual double getYMax() const;


  // Static list of CRUs
  static list<CRU *> cleanCRUs;
  static list<CRU *> dirtyCRUs;


  // Printing function
  void print(ostream & ostr) const;


  // Neighbouring function
  friend void setNeighbouring(CRU & cru1, CRU & cru2);


  // Friendship relation of insertion operator
  friend ostream & operator<<(ostream & ostr, CRU const & cru);


private:

  // Referenced polyline
  Polyline & polyline;

  // Limit indices of convex hull
  unsigned frstIdx;
  unsigned lastIdx;
  unsigned spltIdx;

  // Tolerance for minimum distance
  double    nearTolerance;
  double sqrNearTolerance;

  // List of features
  list<Feature> features;

  // Vector of neighbours
  list<neighbour_t> neighbours;

  // Sidedness test
  state_t state;

  // CRU iterator to list of CRUs (dirty or clean)
  CRUIter_t cruIter;

  // Bounding box coordinates
  double xMin, xMax;
  double yMin, yMax;

  // Bounding diamond coordinates
  double pMin, pMax;
  double qMin, qMax;
};



/**
 * @brief It constructs a new CRU to monitor the consistency between the
 * subpolyline of @a polyline in the range [@a frstIdx, @a lastIdx] and the
 * line segment from @a frstIdx to @a lastIdx..
 *
 * @param polyline The polyline to monitor the consistency.
 * @param frstIdx The index of the first vertex of subpolyline (line segment).
 * @param lastIdx The index of the last vertex of subpolyline (line segment).
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline CRU::CRU(Polyline & polyline, unsigned frstIdx, unsigned lastIdx)
  : polyline(polyline), state(CLEAN)
{
  nearTolerance    = polyline.getWidth();
  sqrNearTolerance = polyline.getWidth()*polyline.getWidth();

  set(frstIdx, lastIdx);
  cruIter = cleanCRUs.insert(cleanCRUs.end(), this);
}


/**
 * @brief The destructor.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline CRU::~CRU() {}


/**
 * @brief It returns the polyline this CRU monitors.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Polyline const & CRU::getPolyline() const { return polyline; }


/**
 * @brief It returns the proximity tolerance of this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double CRU::getNearTolerance() const { return nearTolerance; }


/**
 * @brief It returns the index of the first vertex of the subpolyline (or
 * line segment) this CRU monitors.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned CRU::getFrstIdx() const { return frstIdx; }


/**
 * @brief It returns the index of the last vertex of the subpolyline (or
 * line segment) this CRU monitors.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned CRU::getLastIdx() const { return lastIdx; }


/**
 * @brief It returns the index of the next vertex that is to be inserted if
 * this CRU presents inconsistency.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned CRU::getSpltIdx() const { return spltIdx; }


/**
 * @brief It returns the first vertex of the subpolyline (or line segment)
 * this CRU monitors.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex const & CRU::getFrstVertex() const { return polyline[frstIdx]; }


/**
 * @brief It returns the first vertex of the subpolyline (or line segment)
 * this CRU monitors.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex const & CRU::getLastVertex() const { return polyline[lastIdx]; }


/**
 * @brief It returns the next vertex that is to be inserted if this CRU
 * presents inconsistency.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex const & CRU::getSpltVertex() const { return polyline[spltIdx]; }


/**
 * @brief It returns the number of point/vertex features in this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned CRU::numberOfFeatures() const { return features.size(); }


/**
 * @brief It sets the proximity tolerance of this CRU from @a nearTolerance.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void CRU::setNearTolerance(double nearTolerance)
{
  this->   nearTolerance = nearTolerance;
  this->sqrNearTolerance = nearTolerance*nearTolerance;
}


/**
 * @brief It returns @c true if this CRU overlaps @a cru, and false otherwise.
 *
 * @param cru the CRU to be tested.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool CRU::overlaps(CRU const & cru) const
{
  // Testing bounding box
  if (cru.xMax < xMin || cru.xMin > xMax ||
      cru.yMax < yMin || cru.yMin > yMax)
    return false;

  // Testing bounding diamond
  if (cru.pMax < pMin || cru.xMin > xMax ||
      cru.pMax < pMin || cru.yMin > yMax)
    return false;

  return true;
}


/**
 * @brief It returns @c true if @a point is inside the bouding box of CRU.
 *
 * @param point The point to be tested.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool CRU::isInside(Point const & point)
{
  // Testing bounding box
  if (point.getX() < xMin || point.getX() > xMax ||
      point.getY() < yMin || point.getY() > yMax)
    return false;


  // Testing bounding diamond
  double p = point.getX()+point.getY(); if (p < pMin || p > pMax) return false;
  double q = point.getX()-point.getY(); if (q < qMin || q > qMax) return false;


  return true;
}


/**
 * @brief It returns the minimum x of this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double CRU::getXMin() const { return xMin; }


/**
 * @brief It returns the maximum x of this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double CRU::getXMax() const { return xMax; }


/**
 * @brief It returns the minimum y of this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double CRU::getYMin() const { return yMin; }


/**
 * @brief It returns the maximum y of this CRU.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double CRU::getYMax() const { return yMax; }



#endif // __CRU_HH_
