#include "map.hh"
#include "rips.hh"

#include <iomanip>
#include <typeinfo>




/**
 * @brief It simplifies the polylines of this map using the Consistent
 * version of Visvalingam's algorithm.
 *
 * Visvalingam's algorithm ensures that each simplified polyline will not
 * have vertices with effective area less the @a tolerance.  The consistency
 * of the simplified map will be ensured, because the polylines are
 * simplified together.  The objects will be spaced with at least the
 * distance @a nearTolerance.
 *
 * @param tolerance The minimum effective area.
 * @param nearTolerance The minimum distance between objects.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::simplifyCVIS(double tolerance, double nearTolerance)
{
  // RDP simplification, outputing convex hulls
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      polylines[polIdx].setWidth(nearTolerance);
      polylines[polIdx].simplifyVIS(tolerance);
    }


  // Adding initial features
  initCRUs();


  // Recovering consistency of simplified map
  recoverConsistency(tolerance*tolerance);


  // Deleting allocated memory for CRUs
  for (CRUIter_t cruIter = CRU::cleanCRUs.begin();
       cruIter != CRU::cleanCRUs.end(); ++cruIter)
    delete *cruIter;


  // Removing vertices not in simplified polylines
  removeVertices();
}



/**
 * @brief It creates the CRUs from the simplified polylines of this maps.
 *
 * The Consistent Recovery Units (CRUs) are later used to recover the
 * consistency of the simplified map that can be lost in the simplification
 * step.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::initCRUs()
{
  // List of rectangles
  list<Rectangle const *> rectangles;


  // Creating CRUs and pushing them in list of rectangles
  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
    {
      bool lastCRU = false;

      for (unsigned vertexIdx = 0;
	   vertexIdx < polylines[polIdx].numberOfVertices();
	   vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())

	if (vertexIdx == polylines[polIdx][vertexIdx].getNextIdx()-1)
	  {
	    // Pushing point in list of rectangles
	    if (!lastCRU) rectangles.push_back(&polylines[polIdx][vertexIdx]);

	    lastCRU = false;
	  }

	else
	  {
	    // Pushing CRU in list of CRUs and rectangles
	    CRU * cru = new CRU(polylines[polIdx],vertexIdx,
				polylines[polIdx][vertexIdx].getNextIdx());

	    rectangles.push_back(cru);

	    lastCRU = true;
	  }
    }


  // Adding point features in list of rectangles
  for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
    rectangles.push_back(&points[pointIdx]);


  // Identifying neighbours
  RIPS rips;
  rips.solve(rectangles);


  // Adding points and setting neighbours
  RectPairList const & rectPairs = rips.getRectPairs();

  for (constRectPairIter_t rectPairIter = rectPairs.begin();
       rectPairIter != rectPairs.end(); ++rectPairIter)
    {
      Rectangle * frstRect = const_cast<Rectangle *>(rectPairIter->first);
      Rectangle * scndRect = const_cast<Rectangle *>(rectPairIter->second);

      // Two CRUs intersect -- A edge is created in the interference graph
      if (typeid(*frstRect) == typeid(CRU) &&
	  typeid(*scndRect) == typeid(CRU))
	{
	  CRU & cru1 = *reinterpret_cast<CRU *>(frstRect);
	  CRU & cru2 = *reinterpret_cast<CRU *>(scndRect);

	  setNeighbouring(cru1, cru2);

	  cru1.pushFeature(cru2.getFrstVertex());
	  cru1.pushFeature(cru2.getLastVertex());

	  cru2.pushFeature(cru1.getFrstVertex());
	  cru2.pushFeature(cru1.getLastVertex());
	}

      // A point feature is inside a CRU -- first case
      else if (typeid(*frstRect) == typeid(CRU))
	reinterpret_cast<CRU *>(frstRect)->
	  pushFeature(*reinterpret_cast<Point *>(scndRect));

      // A point feature is inside a CRU -- second case
      else if (typeid(*scndRect) == typeid(CRU))
	reinterpret_cast<CRU *>(scndRect)->
	  pushFeature(*reinterpret_cast<Point *>(frstRect));
    }
}



/**
 * @brief Recover consistency of this map based on the CRUs created.
 *
 * @param sqrTolerance The maximum squared distance.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::recoverConsistency(double sqrTolerance)
{
  // Correcting sidedness errors
  while (!CRU::dirtyCRUs.empty()) // While there still are dirty CRUs
    CRU::dirtyCRUs.front()->correct(sqrTolerance);
}



/**
 * @brief Remove vertices not present in the simplification step neither in
 * the correction step.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::removeVertices()
{
  // Erasing vertices not in simplified polylines
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      unsigned currIdx = 1;

      for (unsigned vertexIdx = polylines[polIdx].front().getNextIdx();
	   vertexIdx < polylines[polIdx].numberOfVertices()-1;
	   vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())

	polylines[polIdx][currIdx++] = polylines[polIdx][vertexIdx];

      polylines[polIdx][currIdx] = polylines[polIdx].back();
      polylines[polIdx].resize(currIdx+1);
    }
}



/**
 * @brief Writes this map in @c ostr.
 *
 * First, writes the number of polylines, followed by the polylines.  Then,
 * writes the number of points, followed by the features points.  Uses the
 * Polyline::print and Point::print.
 *
 * @param ostr The output stream.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::print(ostream & ostr) const
{
  // Setting needed flags
  ostr.flags(ios::scientific|ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      ostr << endl;
      polylines[polIdx].print(ostr);
    }


  // Printing points
  ostr << endl << endl << numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < numberOfPoints(); ++pointIdx)
    {
      points[pointIdx].print(ostr);
      ostr << endl;
    }
}



/**
 * @brief Writes a map in an output stream.
 *
 * First, writes the number of polylines, followed by the polylines.  Then,
 * writes the number of points, followed by the points.  Is a redefinition of
 * the insertion operator in the context of streams.
 *
 * @param ostr The output stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Map const & map)
{
  // Setting needed flags
  ostr.setf(ios::scientific, ios::floatfield);
  ostr.setf(ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << map.numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    ostr << endl << map.polylines[polIdx];


  // Printing points
  ostr << endl << endl << map.numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    ostr << map.points[pointIdx] << endl;


  return ostr;
}



/**
 * @brief Reads a map from an input stream.
 *
 * First, reads the number of polylines, followed by the polylines.  Then,
 * reads the number of points, followed by the points.  Is a redefinition of
 * the extraction operator in the context of streams.
 *
 * @param istr The input stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Map & map)
{
  // Reading polylines
  unsigned nPolylines;

  istr >> nPolylines;
  map.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    istr >> map.polylines[polIdx];


  // Reading points
  unsigned nPoints;

  istr >> nPoints;
  map.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    istr >> map.points[pointIdx];


  return istr;
}
