#include "polyline.hh"

#include <iomanip>
#include <list>




/**
 * @brief Applies Visvalingam's algorithm to this polyline.
 *
 * The simplification is carried out with a minimum effective area of
 * @tolerance.
 *
 * @param tolerance The minimum effective area.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Polyline::simplifyVIS(double tolerance)
{
  // First step
  for (unsigned vertexIdx = 1; vertexIdx <
	 numberOfVertices()-1; ++vertexIdx)
    {
      // Compute effective area
      (*this)[vertexIdx].setMetric((*this)[vertexIdx-1],
				   (*this)[vertexIdx+1]);

      // Set neighbours
      (*this)[vertexIdx].setPrevIdx(vertexIdx-1);
      (*this)[vertexIdx].setNextIdx(vertexIdx+1);
    }


  // Second step - Eliminating vertices with smallest effective area
  for (unsigned level = numberOfVertices()-2; true; --level)
    {
      double   smallestEffectiveArea = tolerance;
      unsigned smallestIdx = 0;

      for (unsigned vertexIdx = (*this)[0].getNextIdx();
	   vertexIdx < (*this).numberOfVertices()-1;
	   vertexIdx = (*this)[vertexIdx].getNextIdx())
	{
	  double currentEffectiveArea = (*this)[vertexIdx].getMetric();

	  if (currentEffectiveArea < smallestEffectiveArea)
	    {
	      smallestEffectiveArea = currentEffectiveArea;
	      smallestIdx = vertexIdx;
	    }
	}

      if (smallestEffectiveArea < tolerance)
	{
	  // Updating previous and next vertices of its neighbours
	  unsigned prevIdx = (*this)[smallestIdx].getPrevIdx();
	  unsigned nextIdx = (*this)[smallestIdx].getNextIdx();

	  // Updating children of its neighbours
	  (*this)[prevIdx].setNextIdx(nextIdx);
	  (*this)[nextIdx].setPrevIdx(nextIdx);

	  // Updating children of its neighbours
	  (*this)[prevIdx].setRghtIdx(smallestIdx);
	  (*this)[nextIdx].setLeftIdx(smallestIdx);

	  // Updating levels of its neighbours
	  (*this)[prevIdx].setLevel(level);
	  (*this)[nextIdx].setLevel(level);

	  // Updating effective area of neighbours
	  if (prevIdx > 0)
	    (*this)[prevIdx].setMetric((*this)[prevIdx].getPrevIdx(),
				       (*this)[prevIdx].getNextIdx());

	  if (nextIdx < numberOfVertices()-1)
	    (*this)[nextIdx].setMetric((*this)[nextIdx].getPrevIdx(),
				       (*this)[nextIdx].getNextIdx());
	}

      // LOOP EXIT
      // No more vertices with effective area less than tolerance
      else break;
    }
}



/**
 * @brief Prints subpolyline of this polyline in @c ostr.
 *
 * First, prints the number of vertices of the subpolyline and the range.
 * Then, prints vertices of the subpolyline one per line, prefixed with its
 * index.
 *
 * @param ostr Output stream in where the polyline is printed.
 * @param frstIdx First index of subpolyline's range.
 * @param lastIdx Last  index of subpolyline's range.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setfill('0') << setw(3) << vertexIdx << ") " << setfill(' ');
      (*this)[vertexIdx].print(ostr);
    }
}



/**
 * @brief Writes @c polyline in @c ostr.
 *
 * First, writes the number of vertices.  Then, writes the vertices one per
 * line.  Is a redefinition of the insertion operator in the context of
 * streams.
 *
 * @param ostr Output stream in where the polyline is writed.
 * @param polyline Polyline to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;


  return ostr;
}



/**
 * @brief Reads @c polyline from @c ostr.
 *
 * First, reads number of vertices.  Then, reads vertices in the specified
 * order.  Is a redefinition of the extraction operator in the context of
 * streams.
 *
 * @param ostr Input stream from where the polyline is printed.
 * @param polyline Polyline to print.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    istr >> polyline[vertexIdx];


  return istr;
}
