#include "convexhull.hh"

#include <iomanip>

#define ERROR_DISTANCE 1E-20
#define CHULL 1


#ifdef CHULL
/**
 * @brief Determines the convex hull of the sub-polyline beginning at the
 * vertex 'frstIdx' and ending at the vertex 'lastIdx'.
 *****************************************************************************/
void ConvexHull::set(unsigned frstIdx, unsigned lastIdx)
{
  // Setting indices of first, last and farthest vertices
  this->frstIdx = frstIdx;
  this->lastIdx = lastIdx;
  this->farthestIdx =
    (polyline[frstIdx].getLevel() > polyline[lastIdx].getLevel())
    ? static_cast<unsigned>(polyline[frstIdx].getRghtIdx())
    : static_cast<unsigned>(polyline[lastIdx].getLeftIdx());


  // Inserting first point once and second point twice
  clear();
  push_front(frstIdx);
  push_front(frstIdx+1);
  push_back (frstIdx+1);


  // Handling rest of points
  for (unsigned pointIdx = frstIdx+2; pointIdx <= lastIdx; ++pointIdx)
    {
      if (polyline[pointIdx].atLeftOrOnLine(polyline[front(0)],
					    polyline[front(1)]) &&
	  polyline[pointIdx].atLeftOrOnLine(polyline[back(1)],
					    polyline[back(0)]))
	continue;

      while (polyline[pointIdx].atRight(polyline[front(0)],
					polyline[front(1)])) pop_front();
      push_front(pointIdx);

      while (polyline[pointIdx].atRight(polyline[back(1)],
					polyline[back(0)])) pop_back();
      push_back(pointIdx);
    }


  // Computing bouding box
  xMin = xMax = polyline[front()].getX();
  yMin = yMax = polyline[front()].getY();

  pMin = pMax = polyline[front()].getX()+polyline[front()].getY();
  qMin = qMax = polyline[front()].getX()-polyline[front()].getY();


  // Handling rest of points
  for (unsigned vertexIdx = 1; vertexIdx < size()-1; ++vertexIdx)
    {
      unsigned vIdx = front(vertexIdx);

      if      (polyline[vIdx].getX() < xMin) xMin = polyline[vIdx].getX();
      else if (polyline[vIdx].getX() > xMax) xMax = polyline[vIdx].getX();
      if      (polyline[vIdx].getY() < yMin) yMin = polyline[vIdx].getY();
      else if (polyline[vIdx].getY() > yMax) yMax = polyline[vIdx].getY();

      double pCurr = polyline[vIdx].getX()+polyline[vIdx].getY();
      double qCurr = polyline[vIdx].getX()-polyline[vIdx].getY();

      if      (pCurr < pMin) pMin = pCurr;
      else if (pCurr > pMax) pMax = pCurr;
      if      (qCurr < qMin) qMin = qCurr;
      else if (qCurr > qMax) qMax = qCurr;
    }


  xMin -= nearTolerance; xMax += nearTolerance;
  yMin -= nearTolerance; yMax += nearTolerance;

  pMin -= M_SQRT2*nearTolerance;  pMax += M_SQRT2*nearTolerance;
  qMin -= M_SQRT2*nearTolerance;  qMax += M_SQRT2*nearTolerance;
}
#else
/**
 * @brief Determines the convex hull of the sub-polyline beginning at the
 * vertex 'frstIdx' and ending at the vertex 'lastIdx'.
 *****************************************************************************/
void ConvexHull::set(unsigned frstIdx, unsigned lastIdx)
{
  // Setting indices of first, last and farthest vertices
  this->frstIdx = frstIdx;
  this->lastIdx = lastIdx;
  this->farthestIdx =
    (polyline[frstIdx].getLevel() > polyline[lastIdx].getLevel())
    ? static_cast<unsigned>(polyline[frstIdx].getRghtIdx())
    : static_cast<unsigned>(polyline[lastIdx].getLeftIdx());


  // Computing bouding box
  xMin = xMax = polyline[frstIdx].getX();
  yMin = yMax = polyline[frstIdx].getY();

  pMin = pMax = polyline[frstIdx].getX()+polyline[frstIdx].getY();
  qMin = qMax = polyline[frstIdx].getX()-polyline[frstIdx].getY();


  // Handling rest of points
  for (unsigned vIdx = frstIdx+1; vIdx <= lastIdx; ++vIdx)
    {
      if      (polyline[vIdx].getX() < xMin) xMin = polyline[vIdx].getX();
      else if (polyline[vIdx].getX() > xMax) xMax = polyline[vIdx].getX();
      if      (polyline[vIdx].getY() < yMin) yMin = polyline[vIdx].getY();
      else if (polyline[vIdx].getY() > yMax) yMax = polyline[vIdx].getY();

      double pCurr = polyline[vIdx].getX()+polyline[vIdx].getY();
      double qCurr = polyline[vIdx].getX()-polyline[vIdx].getY();

      if      (pCurr < pMin) pMin = pCurr;
      else if (pCurr > pMax) pMax = pCurr;
      if      (qCurr < qMin) qMin = qCurr;
      else if (qCurr > qMax) qMax = qCurr;
    }
}
#endif


/**
 * Split the convex hull into two convex hulls at the vertex 'splitIdx'.  The
 * two resulting convex hulls are placed in 'frstCHull' and 'lastCHull'.
 *****************************************************************************/
void ConvexHull::split(unsigned splitIdx,
		       ConvexHull & frstCHull, ConvexHull & scndCHull)
{
  /* Setting both convex hulls */
  frstCHull.set(frstIdx, splitIdx);
  scndCHull.set(splitIdx, lastIdx);


  /* Handling features for the new convex hulls */
  for (featureIter_t featureIter = features.begin();
       featureIter != features.end(); ++featureIter)
    {
      unsigned findCrossIdx = featureIter->findCrossIdx(splitIdx);
      unsigned findProxtIdx = featureIter->findProxtIdx(splitIdx);

      if (frstCHull.isInside(featureIter->getPoint()))
	frstCHull.pushFeature(*featureIter, true,  findCrossIdx, findProxtIdx);

      if (scndCHull.isInside(featureIter->getPoint()))
	scndCHull.pushFeature(*featureIter, false, findCrossIdx, findProxtIdx);
    }


  // Adding uncommon extremes
  if (frstCHull.isInside(polyline[lastIdx]))
    frstCHull.pushFeature(polyline[lastIdx]);

  if (scndCHull.isInside(polyline[frstIdx]))
    scndCHull.pushFeature(polyline[frstIdx]);
}



/**
 * Split the convex hull into two convex hulls at the vertex 'splitIdx'.  The
 * two resulting convex hulls are placed in 'frstCHull' and 'scndCHull'.
 *****************************************************************************/
void ConvexHull::splitSA(unsigned spltIdx,
			 ConvexHull & frstCHull, ConvexHull & scndCHull)
{
  // Setting both convex hulls
  frstCHull.set(frstIdx, spltIdx);
  scndCHull.set(spltIdx, lastIdx);


  /* Handling features for the new convex hulls */
  for (featureSAIter_t featureIter = featuresSA.begin();
       featureIter != featuresSA.end(); ++featureIter)
    {
      Point const & point = *featureIter->point;


      /* Setting side of feature */
      if (point.isInTriangle(polyline[frstIdx],
			     polyline[lastIdx],
			     polyline[spltIdx]))
	featureIter->wrongSide = !featureIter->wrongSide;


      /* Inserting features in new convex hulls */
      if (frstCHull.isInside(point)) frstCHull.pushFeatureSA(*featureIter);
      if (scndCHull.isInside(point)) scndCHull.pushFeatureSA(*featureIter);
    }


  /* Inserting vertices of the other subpolyline */
  for (unsigned vertexIdx = spltIdx+1; vertexIdx <= lastIdx; ++vertexIdx)
    if (frstCHull.isInside(polyline[vertexIdx]))
      frstCHull.pushFeatureSA(polyline[vertexIdx]);

  for (unsigned vertexIdx = frstIdx; vertexIdx < spltIdx; ++vertexIdx)
    if (scndCHull.isInside(polyline[vertexIdx]))
      scndCHull.pushFeatureSA(polyline[vertexIdx]);
}



#ifdef CHULL
/**
 * Returns 'true' if the point 'point' is inside the convex hull, and 'false'
 * otherwise.
 */
bool ConvexHull::isInside(Point const & point)
{
//   cerr << "isInside " << point << endl;
//   cerr << "Convex Hull fr: " << setw(2) << frstIdx << " ("
//        << polyline[frstIdx] << ')'<< endl
//        << "            to: " << setw(2) << lastIdx << " ("
//        << polyline[lastIdx] << ") ";


  if (point.getX() < xMin || point.getX() > xMax ||
      point.getY() < yMin || point.getY() > yMax)
    { // cerr << "FALSE BB" << endl;
    return false; }


  double p = point.getX()+point.getY();
  if (p < pMin || p > pMax) { // cerr << "FALSE BD1" << endl;
  return false; }
  double q = point.getX()-point.getY();
  if (q < qMin || q > qMax) { // cerr << "FALSE BD2" << endl;
  return false; }


  for (unsigned pointIdx = 0; pointIdx < size()-1; ++pointIdx)
    if (point.sqrDistance(polyline[front(pointIdx)],
			  polyline[front(pointIdx+1)]) <= sqrNearTolerance)
      { // cerr << "TRUE DIST" << endl;
      return true; }


  for (unsigned pointIdx = 0; pointIdx < size()-1; ++pointIdx)
    if (point.atRight(polyline[front(pointIdx)],
		      polyline[front(pointIdx+1)]))
      { // cerr << "FALSE CH" << endl;
      return false; }

//   if (inCH == true && (inBB == false || inBD == false))
//     {
//       cerr << endl << "pnt: ";
//       point.print(cerr);
//       cerr << "p < pMin: " << (p < pMin) << endl
// 	   << "p > pMax: " << (p > pMax) << endl
// 	   << "q < qMin: " << (q < qMin) << endl
// 	   << "q > qMax: " << (q > qMax) << endl;

//       cerr << inCH << " - " << inBB << " - " << inBD << endl  << "p = ("
// 	   << setw(14) << pMin << ", "
// 	   << setw(14) << pMax << ") :: q = ("
// 	   << setw(14) << qMin << ", "
// 	   << setw(14) << qMax << ")" << endl << "x = ("
// 	   << setw(14) << xMin << ", "
// 	   << setw(14) << xMax << ") :: y = ("
// 	   << setw(14) << yMin << ", "
// 	   << setw(14) << yMax << ")" << endl;

//       polyline.print(cerr, frstIdx, lastIdx);
//     }

//   cerr << "TRUE GEN" << endl;
  return true;
}
#else
/**
 * Returns 'true' if the point 'point' is inside the convex hull, and 'false'
 * otherwise.
 */
bool ConvexHull::isInside(Point const & point)
{
  if (point.getX() < xMin || point.getX() > xMax ||
      point.getY() < yMin || point.getY() > yMax)
    return false;

  double p = point.getX()+point.getY(); if (p < pMin || p > pMax) return false;
  double q = point.getX()-point.getY(); if (q < qMin || q > qMax) return false;

  return true;
}
#endif


/**
 * Pushes a new feature point 'point' into the features list.  It sets also the
 * sidedness of both feature point and convex hull.
 */
void ConvexHull::pushFeature(Point const & point)
{
  // Discarding limits
  if (point == polyline[frstIdx] ||
      point == polyline[lastIdx]) return;


  ++counter;

  // Storing crossings and proximities
  list<unsigned> crossingList;
  list<unsigned> proximitList;

  bool atLeft = polyline[frstIdx].getX() <= point.getX();

  for (unsigned prevIdx = frstIdx, currIdx = frstIdx+1;
       currIdx <= lastIdx; prevIdx = currIdx++)
    {
      // Crossings
      if (polyline[currIdx].getX() <= point.getX())
	{
	  if (!atLeft && point.atLeft(polyline[prevIdx], polyline[currIdx]))
	    crossingList.push_back(currIdx);

	  atLeft = true;
	}
      else if (polyline[currIdx].getX() > point.getX())
	{
	  if (atLeft && point.atRight(polyline[prevIdx], polyline[currIdx]))
	    crossingList.push_back(currIdx);

	  atLeft = false;
	}

      // Proximities
      if (point.sqrDistance(polyline[prevIdx],
			    polyline[currIdx]) <= sqrNearTolerance)
	proximitList.push_back(currIdx);
    }


  // Determining whether base segment cross vertical ray from feature
  bool lLeft = polyline[lastIdx].getX() <= point.getX();
  bool fLeft = polyline[frstIdx].getX() <= point.getX();
  bool baseSegmentCross =
    ( lLeft && !fLeft && point.atRight(polyline[lastIdx],polyline[frstIdx])) ||
    (!lLeft &&  fLeft && point.atLeft (polyline[lastIdx],polyline[frstIdx]));


  // Determining whether base segment is near to feature
  double baseSegmentSqrDistance = point.sqrDistance(polyline[frstIdx],
						    polyline[lastIdx]);

  bool baseSegmentProxt = baseSegmentSqrDistance <= sqrNearTolerance;
  bool baseSegmentError = baseSegmentSqrDistance <= ERROR_DISTANCE;


  /* Pushing feature in list */
  features.push_back(Feature(&point, crossingList, proximitList,
			     baseSegmentCross, baseSegmentProxt,
			     baseSegmentError));


  // Updating convexhull's consistency
  if (features.back().getInconsistent())
    inconsistent = true;
}



/**
 * @brief Pushes a new feature point 'feature' into the features list.  It sets
 * also the sidedness of the convex hull.
 */
void ConvexHull::pushFeature(Feature const & feature, bool firstCHull,
			     unsigned findCrossIdx,
			     unsigned findProxtIdx)
{
  // Pushing feature in list
  features.push_back(feature);
  Point const & point = features.back().getPoint();


  // Updating indices
  if (firstCHull)
    {
      features.back().setLastCrossIdx(findCrossIdx);
      features.back().setLastProxtIdx(findProxtIdx);
    }
  else
    {
      features.back().setFrstCrossIdx(findCrossIdx);
      features.back().setFrstProxtIdx(findProxtIdx);
    }


  // Determining whether base segment cross vertical upward ray
  bool lLeft = polyline[lastIdx].getX() <= point.getX();
  bool fLeft = polyline[frstIdx].getX() <= point.getX();
  bool baseSegmentCross =
    ( lLeft && !fLeft && point.atRight(polyline[lastIdx],polyline[frstIdx])) ||
    (!lLeft &&  fLeft && point.atLeft (polyline[lastIdx],polyline[frstIdx]));


  // Determining whether base segment is near to feature
  double baseSegmentSqrDistance = point.sqrDistance(polyline[frstIdx],
						    polyline[lastIdx]);

  bool baseSegmentProxt = baseSegmentSqrDistance <= sqrNearTolerance;
  bool baseSegmentError = baseSegmentSqrDistance <= ERROR_DISTANCE;


  // Setting relationships with base segment
  features.back().setBaseSegmentCross(baseSegmentCross);
  features.back().setBaseSegmentProxt(baseSegmentProxt);
  features.back().setBaseSegmentError(baseSegmentError);


  // Updating convexhull's consistency
  if (features.back().getInconsistent())
    inconsistent = true;
}



/**
 * Pushes a new feature point 'point' into the features list.  It sets also the
 * sidedness of both feature point and convex hull.
 *****************************************************************************/
void ConvexHull::pushFeatureSA(Point const & point)
{
  /* Pushing feature point */
  featuresSA.push_back(feature_t(&point));


  /* Testing for crossings */
  if (point != polyline[frstIdx] &&
      point != polyline[lastIdx])
    {
      bool atLeft = polyline[lastIdx].getX() <= point.getX();

      for (unsigned prevIdx = lastIdx, currIdx = frstIdx;
	   currIdx <= lastIdx; prevIdx = currIdx++)

	if (polyline[currIdx].getX() <= point.getX())
	  {
	    if (!atLeft && point.atLeft(polyline[prevIdx], polyline[currIdx]))
	      featuresSA.back().wrongSide = !featuresSA.back().wrongSide;

	    atLeft = true;
	  }
	else if (polyline[currIdx].getX() > point.getX())
	  {
	    if (atLeft && point.atRight(polyline[prevIdx], polyline[currIdx]))
	      featuresSA.back().wrongSide = !featuresSA.back().wrongSide;

	    atLeft = false;
	  }
    }


  if (featuresSA.back().wrongSide)
    inconsistent = true;
}



/**
 * Pushes a new feature point 'feature' into the features list.  It sets also
 * the sidedness of the convex hull.
 *****************************************************************************/
void ConvexHull::pushFeatureSA(feature_t const & feature)
{
  featuresSA.push_back(feature);

  if (feature.wrongSide)
    inconsistent = true;
}



/**
 * Prints the list of point indeces sequentially ordered, one per line.
 */
void ConvexHull::print(ostream & ostr) const
{
  ostr << size() << endl;

  for (unsigned pointIdx = 0; pointIdx < size(); ++pointIdx)
    ostr << setw(3) << (*this)[pointIdx] << endl;
}



/**
 * Prints out the number of indeces, followed by the list of point indeces
 * sequentially ordered, one per line.  It is a redefinition of the insertion
 * operator '<<' in stream context.
 */
ostream & operator<<(ostream & ostr, ConvexHull const & convexHull)
{
  ostr << "Convex hull fr: " << setw(2) << convexHull.frstIdx << " ("
       << convexHull.polyline[convexHull.frstIdx] << ')'<< endl
       << "            to: " << setw(2) << convexHull.lastIdx << " ("
       << convexHull.polyline[convexHull.lastIdx] << ')'<< endl;

//   convexHull.polyline.writeTree(ostr, convexHull.frstIdx, 0);

  ostr << "* xMin = " << setw(14) << convexHull.xMin
       << ", xMax = " << setw(14) << convexHull.xMax << endl
       << "  yMin = " << setw(14) << convexHull.yMin
       << ", yMax = " << setw(14) << convexHull.yMax << endl;

  ostr << "* pMin = " << setw(14) << convexHull.pMin
       << ", pMax = " << setw(14) << convexHull.pMax << endl
       << "  qMin = " << setw(14) << convexHull.qMin
       << ", qMax = " << setw(14) << convexHull.qMax << endl;

  ostr << convexHull.size() << endl;

  for (unsigned pointIdx = 0; pointIdx < convexHull.size(); ++pointIdx)
    ostr << setw(2) << convexHull[pointIdx] << " ("
	 << convexHull.polyline[convexHull[pointIdx]] << ')' << endl;


  ostr << " - features: " << convexHull.features.size() << endl;

  for (constFeatureIter_t featIter = convexHull.features.begin();
       featIter != convexHull.features.end(); ++featIter)
    ostr << "  " << *featIter << endl;


  return ostr;
}
