#ifndef __CONVEXHULL_HH_
#define __CONVEXHULL_HH_


#include "polyline.hh"
#include "feature.hh"

#include <deque>
#include <list>


using namespace std;




/******************************************************************************
 * External types
 *****************************************************************************/
typedef deque<unsigned> convexHull_t;

typedef deque<unsigned>::iterator            hullIter_t;
typedef deque<unsigned>::const_iterator constHullIter_t;

class ConvexHull;
typedef list<ConvexHull>::iterator            cHullIter_t;
typedef list<ConvexHull>::const_iterator constCHullIter_t;


/******************************************************************************
 * Internal types
 *****************************************************************************/
struct feature_t {

  feature_t(Point const * point)
    : point(point), wrongSide(false) {}

  feature_t(Point const * point, bool wrongSide)
    : point(point), wrongSide(wrongSide) {}


  Point const * point;
  bool wrongSide;


  bool operator==(feature_t const & f) const { return *point == *f.point; }
  bool operator< (feature_t const & f) const
  {
    return point->getX() == f.point->getX()
      ?    point->getY() <  f.point->getY()
      :    point->getX() <  f.point->getX();
  }
};

typedef list<feature_t>::iterator            featureSAIter_t;
typedef list<feature_t>::const_iterator constFeatureSAIter_t;

typedef list<Feature>::iterator            featureIter_t;
typedef list<Feature>::const_iterator constFeatureIter_t;


/******************************************************************************
 * 'ConvexHull' is a double-ended queue of indices.
 *****************************************************************************/
class ConvexHull
  : public deque<unsigned> {

public:

  /* Constructors */
  ConvexHull(Polyline const & polyline, double nearTolerance);
  ConvexHull(Polyline const & polyline, double nearTolerance,
	     unsigned frstIdx, unsigned lastIdx);


  /* Determining convex hull of 'polyline' between 'frstIdx' and 'lastIdx' */
  void set(unsigned frstIdx, unsigned lastIdx);

  /* Spliting convex hull at 'splitIdx' */
  void split  (unsigned splitIdx,ConvexHull& frstCHull, ConvexHull& scndCHull);
  void splitSA(unsigned splitIdx,ConvexHull& frstCHull, ConvexHull& scndCHull);

  /* Determines if 'point' is inside convex hulll */
  bool isInside(Point const & point);

  /* Pushes a new feature into the features' list */
  void pushFeature(Point const & point);
  void pushFeature(Feature const & feature, bool firstCHull,
		   unsigned findCrossIdx,
		   unsigned findProxtIdx);

  void pushFeatureSA(Point     const & point);
  void pushFeatureSA(feature_t const & feature);

  void uniqueFeatures();

  unsigned numberOfFeatures() const;


  /* Get functions */
  Polyline const & getPolyline() const;

  unsigned getFrstIdx() const;
  unsigned getLastIdx() const;
  unsigned getFarthestIdx() const;

  bool getInconsistent() const;
  bool getBiggerDistance(double tolerance) const;

  double getNearTolerance() const;


  // Set function
  void setNearTolerance(double nearTolerance);


  // Returns the front/back element with an offset of 'offset'
  unsigned       & front(unsigned offset = 0);
  unsigned const & front(unsigned offset = 0) const;

  unsigned       & back(unsigned offset = 0);
  unsigned const & back(unsigned offset = 0) const;


  // Test if polyline overlaps convex hull
  bool overlaps(Polyline const & polyline) const;


  /* Friendship relation of insertion operator */
  friend ostream & operator<<(ostream & ostr, ConvexHull const & convexHull);

  /* Printing function */
  void print(ostream & ostr) const;

//   void push_front(unsigned pointIdx);
//   void push_back (unsigned pointIdx);

//   void pop_front();
//   void pop_back ();

//   void clear();
//   unsigned size() const;
//   unsigned   operator[](unsigned i) const;
//   unsigned & operator[](unsigned i);

  unsigned long counter;

private:

  // Referenced polyline
  Polyline const & polyline;

  // Limit indices of convex hull
  unsigned frstIdx;
  unsigned lastIdx;
  unsigned farthestIdx;

  // Tolerance for minimum distance
  double    nearTolerance;
  double sqrNearTolerance;

  // List of features
  list<Feature>   features;
  list<feature_t> featuresSA;

  // Sidedness test
  bool inconsistent;

  // Bounding box coordinates
  double xMin, xMax;
  double yMin, yMax;

  // Bounding diamond coordinates
  double pMin, pMax;
  double qMin, qMax;

  // Vector of convex hull
//   vector<unsigned> v;

//   unsigned b;
//   unsigned e;
};

inline ConvexHull::ConvexHull(Polyline const & polyline, double nearTolerance)
  : counter(0), polyline(polyline), nearTolerance(nearTolerance),
    sqrNearTolerance(nearTolerance*nearTolerance), inconsistent(false) {}

inline ConvexHull::ConvexHull(Polyline const & polyline, double nearTolerance,
			      unsigned frstIdx, unsigned lastIdx)
  : counter(0), polyline(polyline), nearTolerance(nearTolerance),
    sqrNearTolerance(nearTolerance*nearTolerance), inconsistent(false)
{ set(frstIdx, lastIdx); }


inline void ConvexHull::uniqueFeatures()
{
  features.sort();
  features.unique();
}

inline Polyline const & ConvexHull::getPolyline() const { return polyline; }

inline unsigned ConvexHull::getFrstIdx() const     { return frstIdx; }
inline unsigned ConvexHull::getLastIdx() const     { return lastIdx; }
inline unsigned ConvexHull::getFarthestIdx() const { return farthestIdx; }

inline bool ConvexHull::getInconsistent() const { return inconsistent; }
inline bool ConvexHull::getBiggerDistance(double tolerance) const
{ return polyline[farthestIdx].getDistance() >= tolerance; }

inline double ConvexHull::getNearTolerance() const { return nearTolerance; }


inline void ConvexHull::setNearTolerance(double nearTolerance)
{
  this->nearTolerance    = nearTolerance;
  this->sqrNearTolerance = nearTolerance*nearTolerance;
}

inline unsigned & ConvexHull::front(unsigned d) { return (*this)[d]; }
inline unsigned & ConvexHull::back (unsigned d) { return (*this)[size()-1-d]; }

inline unsigned const& ConvexHull::front(unsigned d) const {return (*this)[d];}
inline unsigned const& ConvexHull::back (unsigned d) const
{ return (*this)[size()-1-d]; }


inline bool ConvexHull::overlaps(Polyline const & polyline) const
{
  if (xMax < polyline.getXMin()-nearTolerance ||
      xMin > polyline.getXMax()+nearTolerance ||
      yMax < polyline.getYMin()-nearTolerance ||
      yMin > polyline.getYMax()+nearTolerance) return false;

  return true;
}


inline unsigned ConvexHull::numberOfFeatures() const { return features.size();}


// inline unsigned & ConvexHull::front(unsigned d) { return v[b+d]; }
// inline unsigned & ConvexHull::back (unsigned d) { return v[e-d]; }

// inline unsigned const & ConvexHull::front(unsigned d) const { return v[b+d];}
// inline unsigned const & ConvexHull::back (unsigned d) const { return v[e-d]; }

// inline void ConvexHull::clear()
// {
//   v.resize(2*(lastIdx-frstIdx)+1);
//   b = e = lastIdx-frstIdx;
//   v[b] = frstIdx;
// }

// inline unsigned ConvexHull::size() const { return e-b+1; }

// inline unsigned   ConvexHull::operator[](unsigned i) const { return v[b+i]; }
// inline unsigned & ConvexHull::operator[](unsigned i)       { return v[b+i]; }

// inline void ConvexHull::push_front(unsigned i) { v[--b] = i; }
// inline void ConvexHull::push_back (unsigned i) { v[++e] = i; } 

// inline void ConvexHull::pop_front() { ++b; }
// inline void ConvexHull::pop_back () { --e; }


#endif // __CONVEXHULL_HH_
