#include <cstdlib>
#include <sys/time.h>

#include <iostream>
#include <fstream>
#include <iomanip>

#include "map.hh"



using namespace std;


// Type of algorithm to be applied
enum algorithm_t
  {
    RDP = 0, // The Ramer-Douglas-Peucker algorithm
    VIS = 1  // Visvalingam's algorithm
  };



int main(int argc, char * argv[])
{
  // Arguments
  string  inFileName; //  Input file name
  string outFileName; // Output file name

  string toleranceString = "0"; // String for tolerance
  string  nearnessString = "0"; // String for nearness

  double tolerance = 0.0; // Tolerance         (default: 0.0)
  double nearness  = 0.0; // Nearness distance (default: 0.0)

  bool saFlag = false; // Flag for saalfeld                  (default: no SA)
  bool tcFlag = false; // Flag for topologically consistency (default: no TC)
  bool mrFlag = false; // Flag for multi-resolutions         (default: no MR)

  algorithm_t algorithm =  RDP;  // Algorithm to be applied (default:  RDP)
  string algorithmName  = "RDP"; // Algorithm name          (default: "RDP")


  // strings
  string const helpString = "help";
  string const   saString = "saalfeld";
  string const   tcString = "consistent";
  string const   mrString = "scaleless";


  // Usage string
  string const usage =
    string("Basic usage: ")+argv[0]+
    " [-i <input file name>] [-o <ouput file name>]\n"
    "Other options:\n"
    "  -t <tolerance>\n"
    "  -n <nearness>\n"
    "  -a RDP|VIS\n"
    "  --saalfeld\n"
    "  --consistent\n"
    "  --scaleless";


  // Reading arguments
  for (int argIdx = 1; argIdx < argc; ++argIdx)

    if (argv[argIdx][0] == '-')
      switch (argv[argIdx][1])
	{
	case 'i': // input file name
	  if (argIdx+1 < argc) inFileName = argv[++argIdx];
	  else
	    {
	      cerr << argv[0] << ": could not find file name after -i" << endl
		   << usage << endl;
	      exit(-1);
	    }
	  break;

	case 'o': // output file name
	  if (argIdx+1 < argc) outFileName = argv[++argIdx];
	  else
	    {
	      cerr << argv[0] << ": could not find file name after -o" << endl
		   << usage << endl;
	      exit(-1);
	    }
	  break;

	case 't': // tolerance
	  if (argIdx+1 < argc)
	    {
	      toleranceString = argv[++argIdx];
	      tolerance = atof(toleranceString.c_str());
	    }
	  else
	    {
	      cerr << argv[0] << ": could not find tolerance after -t" << endl
		   << usage << endl;
	      exit(-1);
	    }
	  break;

	case 'n': // near tolerance
	  if (argIdx+1 < argc)
	    {
	      nearnessString = argv[++argIdx];
	      nearness = atof(nearnessString.c_str());
	    }
	  else
	    {
	      cerr << argv[0] << ": could not find nearness after -n" << endl
		   << usage << endl;
	      exit(-1);
	    }
	  break;

	case 'a': // output file name
	  if (argIdx+1 < argc)
	    {
	      algorithmName = argv[++argIdx];

	      if      (algorithmName == "RDP") algorithm = RDP;
	      else if (algorithmName == "VIS") algorithm = VIS;
	      else
		{
		  cerr << argv[0] << ": invalid algorithm after -a" << endl
		       << usage << endl;
		  exit(-1);
		}
	    }
	  else
	    {
	      cerr << argv[0] << ": could not find algorithm after -a" << endl
		   << usage << endl;
	      exit(-1);
	    }
	  break;

	case '-':
	  if (&argv[argIdx][2] == helpString)
	    {
	      cerr << usage << endl;
	      exit(1);
	    }
	  else if (&argv[argIdx][2] == saString) saFlag = true;
	  else if (&argv[argIdx][2] == tcString) tcFlag = true;
	  else if (&argv[argIdx][2] == mrString) mrFlag = true;
	  else
	    {
	      cerr << usage << endl;
	      exit(-1);
	    }
	  break;

	default:
	  cerr << argv[0] << ": invalid option " << argv[argIdx] << endl
	       << usage << endl;
	  exit(-1);
	}
    else
      {
	cerr << usage << endl;
	exit(-1);
      }


  // Setting default output file name
  if (outFileName == "default")
    {
      outFileName = inFileName;

      if (outFileName.substr(outFileName.size()-4) == ".inf")
	outFileName.erase(outFileName.size()-4);

      outFileName += "-";

      if (saFlag) outFileName += "SARDP-"+toleranceString;
      else
	{
	  if (mrFlag) outFileName += "MR";
	  if (tcFlag) outFileName += "TC"; 
 
	  outFileName += algorithmName;

	  if (!mrFlag) outFileName += "-"+toleranceString;
	  if ( tcFlag) outFileName += "-"+ nearnessString;
	}
 
      outFileName += ".inf";
    }


  // Setting default streams to standard input and output
  istream *  inStream = &cin;
  ostream * outStream = &cout;


  // File streams
  ifstream  inFileStream;
  ofstream outFileStream;


  // Testing for openning input file
  if (!inFileName.empty())
    {
      inFileStream.open(inFileName.c_str());
      if (!inFileStream.is_open())
	{
	  cerr << argv[0] << ": could not open in file "
	       << inFileName << endl;
	  exit(-1);
	}

      inStream = static_cast<istream *>(&inFileStream);
    }

  // Testing for openning output file
  if (!outFileName.empty())
    {
      outFileStream.open(outFileName.c_str());
      if (!outFileStream.is_open())
	{
	  cerr << argv[0] << ": could not open out file "
	       << outFileName << endl;
	  exit(-1);
	}

      outStream = static_cast<ostream *>(&outFileStream);
    }


  /////////////////////////////////////////////////////////////////////////////
  // Reading map
  Map map;
  *inStream >> map;


  // Setting needed flags
  cerr.setf(ios::scientific, ios::floatfield);
  cerr.setf(ios::uppercase);
  cerr.precision(COORD_PRECISION);


  // Simplification
  if (saFlag)
    {
      cerr << "Algorithm: SARDP" << endl;
      cerr << "Tolerance: " << setw(14) << tolerance << endl;
    }

  else
    {
      cerr << "Algorithm: " << (mrFlag ? "MR" : "") << (tcFlag ? "TC" : "")
	   << algorithmName << endl;

      if (!mrFlag) cerr << "Tolerance: " << setw(14) << tolerance << endl;
      if ( tcFlag) cerr << "Nearness:  " << setw(14) << nearness  << endl;
    }


  // Getting initial time
  timeval bgnTime;
  timeval endTime;
  gettimeofday(&bgnTime, NULL);
  

  // Applying chosed algorithm

  switch (algorithm)
    {
    case RDP:

      if      (saFlag)           map.simplifySARDP(tolerance);
      else if (mrFlag && tcFlag) map.simplifyMRTCRDP(nearness);
      else if (tcFlag)           map.simplifyTCRDP(tolerance, nearness);
      else if (mrFlag)           map.simplifyMRRDP();
      else                       map.simplifyRDP(tolerance);

      break;

    case VIS:

      if      (mrFlag && tcFlag) map.simplifyMRTCVIS(nearness);
      else if (tcFlag)           map.simplifyTCVIS(tolerance, nearness);
      else if (mrFlag)           map.simplifyMRVIS();
      else                       map.simplifyVIS(tolerance);

      break;
    }

  // Getting final time and computing difference
  gettimeofday(&endTime, NULL);
  double diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
  		     (endTime.tv_usec-bgnTime.tv_usec)/1E6);

  cerr << "Time:      " << setw(14) << diffTime << endl;


  // Writing simplified map
  *outStream << map;


  return 0;
}
