#ifndef __MAP_HH_
#define __MAP_HH_


#include "point.hh"
#include "polyline.hh"
#include "convexhull.hh"
#include "tolerance.hh"

#include <iostream>
#include <vector>
#include <list>


using namespace std;


/**
 * @brief this class contains a vector of polylines and a vector of points.
 *****************************************************************************/
class Map {

public:

  // Number of vertices functions
  unsigned numberOfPolylines() const;
  unsigned numberOfPoints() const;


  // Access and modification functions
  Polyline & getPolyline(unsigned polylineIdx);
  Point &    getPoint   (unsigned pointIdx);


  // RDP and Visvaligam's simplifications
  void simplifyRDP(double tolerance);
  void simplifyVIS(double tolerance);


  // Multi-Resolution RDP and Visvaligam's simplifications
  void simplifyMRRDP();
  void simplifyMRVIS();


  // Topologically Consistent RDP and Visvalingam's simplification
  void simplifySARDP(double tolerance);
  void simplifyTCRDP(double tolerance, double nearTolerance);
  void simplifyTCVIS(double tolerance, double nearTolerance);


  // Multi-Resolution TC RDP and TC Visvalingam's simplification
  void simplifyMRTCRDP(double nearTolerance);
  void simplifyMRTCVIS(double nearTolerance);


  // Friendship relation of insertion and extraction operators
  friend ostream & operator<<(ostream & ostr, Map const & map);
  friend istream & operator>>(istream & istr, Map       & map);

  // Printing function
  void print(ostream & ostr) const;


protected:

  // Internal variables
  vector<Polyline> polylines;
  vector<Point>    points;

  vector<Tolerance> tolerances;


  // Internal RDP and Visvaligam's simplifications
  void simplifyRDP(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
		   double sqrTolerance);
  void simplifyVIS(unsigned polIdx, double tolerance);


  // Internal Multi-Resolution RDP and Visvaligam's simplifications
  void simplifyMRRDP(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
		     unsigned level, double parentTolerance);
  void simplifyMRVIS(unsigned polIdx);


  // Internal RDP and Vivalingam's simplification for TC
  void simplifyRDP(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
		   double sqrTolerance,  list<ConvexHull> & cHulls);
  void simplifyVIS(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
		   double sqrTolerance,  list<ConvexHull> & cHulls);


  // Procedure for computing and adding initial features
  void initConvexHullsSA(list<ConvexHull> & cHulls);
  void initConvexHulls  (list<ConvexHull> * cHulls, double nearTolerance);

  // Internal Topologically Consistent  RDP and Visvalingam's simplifications
  void simplifySARDP(double sqrTolerance, ConvexHull & cHull);
  void simplifyTCRDP(double sqrTolerance, list<ConvexHull> * cHulls);
  void simplifyTCVIS(double sqrTolerance, list<ConvexHull> * cHulls);

  // Remove vertices not in the simplified polylines
  void removeVertices();


  // Correction procedures
  void correctSARDP(double sqrTolerance, ConvexHull & cHull);
  void correctTCRDP(unsigned polIdx, double sqrTolerance,
		    cHullIter_t cHullIter, list<ConvexHull> & cHulls,
		    list<Point *> & newVertices);
  void correctTCVIS(unsigned polIdx, double sqrTolerance,
		    cHullIter_t cHullIter, list<ConvexHull> & cHulls,
		    list<Point *> & newVertices);


  // Other procedures
  void fillTolerances();
  void addConvexHulls (unsigned tolIdx, list<ConvexHull> * cHulls);
  void simplifyMRTCRDP(unsigned tolIdx, list<ConvexHull> * cHulls);
  void correctMRTCRDP (unsigned tolIdx, cHullIter_t cHullIter,
		       list<ConvexHull> & cHulls, list<Point *> & newVertices);

  void simplifyMRTCVIS(unsigned tolIdx, list<ConvexHull> * cHulls);
  void correctMRTCVIS (unsigned tolIdx, cHullIter_t cHullIter,
		       list<ConvexHull> & cHulls, list<Point *> & newVertices);
};



inline unsigned Map::numberOfPolylines() const { return polylines.size(); }
inline unsigned Map::numberOfPoints() const    { return points.size(); }


inline Polyline & Map::getPolyline(unsigned idx) { return polylines[idx]; }
inline Point &    Map::getPoint   (unsigned idx) { return points[idx]; }



#endif // __MAP_HH_
