#include "map.hh"




/**
 * @brief Simplifies polylines of 'mapMRTCRDP' using Silva's algorithm for
 * tolerance 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void Map::simplifyMRTCRDP(double nearTolerance)
{
  // Allocating memory for convex hulls
  list<ConvexHull> * cHulls = new list<ConvexHull>[numberOfPolylines()];


  // Setting RDP distances
  simplifyMRRDP();


  // Filling vector of tolerances
  fillTolerances();


  // Set initial data and symplifying for tolerance infinity
  initConvexHulls(cHulls, nearTolerance);
  simplifyMRTCRDP(0, cHulls);


  // Computing rest of tolerances
  for (unsigned tolIdx = 1; tolerances[tolIdx].getValue() !=
	 tolerances.back().getValue(); ++tolIdx)
    {
      if (tolerances[tolIdx  ].getValid())  addConvexHulls(tolIdx, cHulls);
      if (tolerances[tolIdx  ].getValue() !=
	  tolerances[tolIdx+1].getValue()) simplifyMRTCRDP(tolIdx, cHulls);
    }


  // Deallocating memory for convex hulls
  delete [] cHulls;
}



/**
 *
 *****************************************************************************/
void Map::fillTolerances()
{
  // Calculating number of tolerances and allocating memory
  unsigned nTolerances = 1;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    nTolerances += polylines[polIdx].numberOfVertices()-2;

  tolerances.reserve(nTolerances);


  // Computing tolerances and storing in vector
  tolerances.push_back(Tolerance(MAXDOUBLE, 0, 0, 0));

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    for (unsigned vertexIdx = 1; vertexIdx <
	   polylines[polIdx].numberOfVertices()-1; ++vertexIdx)
      tolerances.
	push_back(Tolerance(polylines[polIdx][vertexIdx].getTolerance(),
			    polylines[polIdx][vertexIdx].getLevel(),
			    polIdx, vertexIdx));

  // Sorting vector of tolerances and setting indices of tolerances in vertices
  sort(tolerances.begin(), tolerances.end());

  for (unsigned tolIdx = 0; tolIdx < nTolerances; ++tolIdx)

    polylines[tolerances[tolIdx].getPolIdx()]
      [tolerances[tolIdx].getVertexIdx()].setTolIdx(tolIdx);
}



/**
 *
 *****************************************************************************/
void Map::addConvexHulls(unsigned tolIdx, list<ConvexHull> * cHulls)
{
  unsigned polIdx    = tolerances[tolIdx].getPolIdx();
  unsigned vertexIdx = tolerances[tolIdx].getVertexIdx();


  /* Setting vertex tolerance */
  polylines[polIdx][vertexIdx].setTolerance(tolerances[tolIdx].getValue());


  /* Searching for correct convex hull */
  cHullIter_t cHullIter = cHulls[polIdx].begin();
  for (; cHullIter != cHulls[polIdx].end(); ++cHullIter)
    if (vertexIdx > cHullIter->getFrstIdx() &&
	vertexIdx < cHullIter->getLastIdx()) break;


  /* Copying first and last indices of current convex hull */
  unsigned frstIdx = cHullIter->getFrstIdx();
  unsigned lastIdx = cHullIter->getLastIdx();


  /* Setting first and last convex hulls */
  ConvexHull frstCHull(polylines[polIdx], cHullIter->getNearTolerance());
  ConvexHull lastCHull(polylines[polIdx], cHullIter->getNearTolerance());
  cHullIter->split(vertexIdx, frstCHull, lastCHull);


  /* Erasing remaining convex hull */
  cHullIter = cHulls[polIdx].erase(cHullIter);


  /* Adding new convex hulls */
  if (vertexIdx > frstIdx+1) cHulls[polIdx].insert(cHullIter, frstCHull);
  if (vertexIdx < lastIdx-1) cHulls[polIdx].insert(cHullIter, lastCHull);


  /* Inserting new vertex in convex hulls */
  for (unsigned pol2Idx = 0; pol2Idx < numberOfPolylines(); ++pol2Idx)

    for (cHullIter_t cHullIter = cHulls[pol2Idx].begin();
	 cHullIter != cHulls[pol2Idx].end(); ++cHullIter)

      if (cHullIter->isInside(polylines[polIdx][vertexIdx]))
	cHullIter->pushFeature(polylines[polIdx][vertexIdx]);
}



/**
 * Simplifies map 'mapmrtcrdp' with Silva's version of RDP algorithm for
 * squared tolerance 'sqrTolerance', filling up the convex hulls in 'cHulls'.
 */
void Map::simplifyMRTCRDP(unsigned tolIdx, list<ConvexHull> * cHulls)
{
  // Correcting sidedness errors
  while (true)
    {
      // Applying Silva's correction
      list<Point *> newVertices;

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); )
	  correctMRTCRDP(tolIdx, cHullIter++, cHulls[polIdx], newVertices);


      // Test if there is no new vertex
      if (newVertices.empty()) break;


      // Test for new vertices
      for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); ++cHullIter)

	  for (list<Point *>::iterator vertexIter = newVertices.begin();
	       vertexIter != newVertices.end(); ++vertexIter)
	    if (cHullIter->isInside(**vertexIter))
	      cHullIter->pushFeature(**vertexIter);
    }
}



/**
 * Simplifies polyline 'polylines[polIdx]' of 'tcdpmap' using RDP algorithm for
 * squared tolerance 'sqrTolerance' with convex hull 'cHullIter' in convex
 * hulls list 'cHulls'.  New vertices are added to 'newVertices' list.
 *****************************************************************************/
void Map::correctMRTCRDP(unsigned tolIdx, cHullIter_t cHullIter,
			 list<ConvexHull> & cHulls, list<Point*> & newVertices)
{
  // Testing for inconsistency or vertex farther than tolerance
  if (cHullIter->getInconsistent() ||
      cHullIter->getBiggerDistance(tolerances[tolIdx].getValue()))
    {
      // Setting polyine, first, last and farthest indices from convex hull
      Polyline & polyline = const_cast<Polyline &>(cHullIter->getPolyline());

      unsigned frstIdx     = cHullIter->getFrstIdx();
      unsigned lastIdx     = cHullIter->getLastIdx();
      unsigned farthestIdx = cHullIter->getFarthestIdx();


      /* Setting tolerance of farthest vertex */
      polyline[farthestIdx].setTolerance(tolerances[tolIdx].getValue());
      tolerances[polyline[farthestIdx].getTolIdx()].setValid(false);


      // Adding new vertex
      newVertices.push_back(&polyline[farthestIdx]);


      // Setting first and last convex hulls
      ConvexHull frstCHull(polyline, cHullIter->getNearTolerance());
      ConvexHull lastCHull(polyline, cHullIter->getNearTolerance());
      cHullIter->split(farthestIdx, frstCHull, lastCHull);


      // Erasing remaining convex hull
      cHullIter = cHulls.erase(cHullIter);


      // Inserting new convex hulls and applying Saalfeld correction
      cHullIter_t frstCHullIter;
      cHullIter_t lastCHullIter;

      if (farthestIdx > frstIdx+1)
	{
	  frstCHullIter = cHulls.insert(cHullIter, frstCHull);
	  correctMRTCRDP(tolIdx, frstCHullIter, cHulls, newVertices);
	}

      if (farthestIdx < lastIdx-1)
	{
	  lastCHullIter = cHulls.insert(cHullIter, lastCHull);
	  correctMRTCRDP(tolIdx, lastCHullIter, cHulls, newVertices);
	}
    }
}
