#include "map.hh"




/**
 * @brief Simplifies polylines of 'mapMRTCVIS' using Silva's algorithm for
 * tolerance 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void Map::simplifyMRTCVIS(double nearTolerance)
{
  // Allocating memory for convex hulls
  list<ConvexHull> * cHulls = new list<ConvexHull>[numberOfPolylines()];


  // Setting Visvalingam's effective areas
  simplifyMRVIS();


  // Filling vector of tolerances
  fillTolerances();


  // Set initial data and symplifying for tolerance infinity
  initConvexHulls(cHulls, nearTolerance);
  simplifyMRTCVIS(0, cHulls);


  // Computing rest of tolerances
  for (unsigned tolIdx = 1; tolerances[tolIdx].getValue() !=
	 tolerances.back().getValue(); ++tolIdx)
    {
      if (tolerances[tolIdx  ].getValid())  addConvexHulls(tolIdx, cHulls);
      if (tolerances[tolIdx  ].getValue() !=
	  tolerances[tolIdx+1].getValue()) simplifyMRTCVIS(tolIdx, cHulls);
    }


  // Deallocating memory for convex hulls
  delete [] cHulls;
}



/**
 * Simplifies map 'mapmrtcrdp' with Silva's version of VIS algorithm for
 * squared tolerance 'sqrTolerance', filling up the convex hulls in 'cHulls'.
 */
void Map::simplifyMRTCVIS(unsigned tolIdx, list<ConvexHull> * cHulls)
{
  // Correcting sidedness errors
  while (true)
    {
      // Applying Silva's correction
      list<Point *> newVertices;

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); )
	  correctMRTCVIS(tolIdx, cHullIter++, cHulls[polIdx], newVertices);


      // Test if there is no new vertex
      if (newVertices.empty()) break;


      // Test for new vertices
      for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); ++cHullIter)

	  for (list<Point *>::iterator vertexIter = newVertices.begin();
	       vertexIter != newVertices.end(); ++vertexIter)
	    if (cHullIter->isInside(**vertexIter))
	      cHullIter->pushFeature(**vertexIter);
    }
}



/**
 * Simplifies polyline 'polylines[polIdx]' of 'tcdpmap' using VIS algorithm for
 * squared tolerance 'sqrTolerance' with convex hull 'cHullIter' in convex
 * hulls list 'cHulls'.  New vertices are added to 'newVertices' list.
 *****************************************************************************/
void Map::correctMRTCVIS(unsigned tolIdx, cHullIter_t cHullIter,
			 list<ConvexHull> & cHulls, list<Point*> & newVertices)
{
  Polyline & polyline = const_cast<Polyline &>(cHullIter->getPolyline());

  unsigned frstIdx = cHullIter->getFrstIdx();
  unsigned lastIdx = cHullIter->getLastIdx();

  unsigned farthestIdx =
    (polyline[frstIdx].getLevel() > polyline[lastIdx].getLevel())
    ? static_cast<unsigned>(polyline[frstIdx].getRghtIdx())
    : static_cast<unsigned>(polyline[lastIdx].getLeftIdx());


  // Computing smallest effective area
  double smallestEffectiveArea =
    polyline[farthestIdx].triangleArea(polyline[frstIdx], polyline[lastIdx]);


  // Testing for inconsistency or vertex farther than tolerance
  if (smallestEffectiveArea > tolerances[tolIdx].getValue() ||
      cHullIter->getInconsistent())
    {
      /* Setting tolerance of farthest vertex */
      polyline[farthestIdx].setTolerance(tolerances[tolIdx].getValue());
      tolerances[polyline[farthestIdx].getTolIdx()].setValid(false);


      // Adding new vertex
      newVertices.push_back(&polyline[farthestIdx]);


      // Setting first and last convex hulls
      ConvexHull frstCHull(polyline, cHullIter->getNearTolerance());
      ConvexHull lastCHull(polyline, cHullIter->getNearTolerance());
      cHullIter->split(farthestIdx, frstCHull, lastCHull);


      // Erasing remaining convex hull
      cHullIter = cHulls.erase(cHullIter);


      // Inserting new convex hulls and applying Saalfeld correction
      cHullIter_t frstCHullIter;
      cHullIter_t lastCHullIter;

      if (farthestIdx > frstIdx+1)
	{
	  frstCHullIter = cHulls.insert(cHullIter, frstCHull);
	  correctMRTCVIS(tolIdx, frstCHullIter, cHulls, newVertices);
	}

      if (farthestIdx < lastIdx-1)
	{
	  lastCHullIter = cHulls.insert(cHullIter, lastCHull);
	  correctMRTCVIS(tolIdx, lastCHullIter, cHulls, newVertices);
	}
    }
}
