#include "map.hh"

#include <values.h>




/**
 * @brief Simplifies polylines of this map using Visvalingam's algorithm.
 *
 * The RDP algorithms ensures that each simplified polyline will be at most at
 * the distance @a tolerance from its original version.
 *
 * @warning Topology of the map can be lost, because the polylines are
 * simplified in isolation.
 */
void Map::simplifyMRVIS()
{
  // Simplifying polylines of map
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      // Setting tolerances of end-vertices
      polylines[polIdx].front().setTolerance(MAXDOUBLE);
      polylines[polIdx]. back().setTolerance(MAXDOUBLE);


      // Setting level of end-vertices
      polylines[polIdx].front().setLevel(0);
      polylines[polIdx]. back().setLevel(0);


      // Computing effective area of vertices
      if (polylines[polIdx].numberOfVertices() > 2)
	simplifyMRVIS(polIdx);
    }
}



/**
 * @brief Simplifies this polyline using Visvalingam's algorithm.
 *****************************************************************************/
void Map::simplifyMRVIS(unsigned polIdx)
{
  // First step
  for (unsigned vertexIdx = 1; vertexIdx <
	 polylines[polIdx].numberOfVertices()-1; ++vertexIdx)

    polylines[polIdx][vertexIdx].setTolerance(polylines[polIdx][vertexIdx-1],
					      polylines[polIdx][vertexIdx+1]);


  // Second step: eliminate vertices with smallest effective area
  double previousEffectiveArea = -MAXDOUBLE;

  for (unsigned nVertices = polylines[polIdx].numberOfVertices();
       nVertices > 2; --nVertices)
    {
      double   smallestEffectiveArea = MAXDOUBLE;
      unsigned smallestIdx           = 0;

      for (unsigned vertexIdx = polylines[polIdx].front().getNextIdx();
	   vertexIdx < polylines[polIdx].numberOfVertices()-1;
	   vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())
	{
	  double currentEffectiveArea =
	    polylines[polIdx][vertexIdx].getTolerance();

	  if (currentEffectiveArea < smallestEffectiveArea)
	    {
	      smallestEffectiveArea = currentEffectiveArea;
	      smallestIdx           = vertexIdx;
	    }
	}


      // Setting level
      polylines[polIdx][smallestIdx].setLevel(nVertices-2);


      // Setting effective area
      if (previousEffectiveArea > smallestEffectiveArea)
	polylines[polIdx][smallestIdx].setTolerance(previousEffectiveArea);


      // Updating previous effective area
      previousEffectiveArea = polylines[polIdx][smallestIdx].getTolerance();


      // Initializing references to neighbours
      Point & prevPoint =
	polylines[polIdx][polylines[polIdx][smallestIdx].getPrevIdx()];

      Point & nextPoint =
	polylines[polIdx][polylines[polIdx][smallestIdx].getNextIdx()];


      // Updating previous and next vertices of its neighbours
      prevPoint.setNextIdx(polylines[polIdx][smallestIdx].getNextIdx());
      nextPoint.setPrevIdx(polylines[polIdx][smallestIdx].getPrevIdx());


      // Updating children of its neighbours
      prevPoint.setRghtIdx(smallestIdx);
      nextPoint.setLeftIdx(smallestIdx);


      // Updating effective area of its neighbours
      if (&prevPoint != &polylines[polIdx].front())
	prevPoint.setTolerance(polylines[polIdx][prevPoint.getPrevIdx()],
			       polylines[polIdx][prevPoint.getNextIdx()]);

      if (&nextPoint != &polylines[polIdx].back())
	nextPoint.setTolerance(polylines[polIdx][nextPoint.getPrevIdx()],
			       polylines[polIdx][nextPoint.getNextIdx()]);
    }
}
