#ifndef __POINT_HH_
#define __POINT_HH_


#include <cmath>
#include <iostream>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;



/**
 * @brief This class contains cartesian coordinates of a point in 2D space. It
 * presents geometric transformations and util functions usually applied to
 * points.
 *****************************************************************************/
class Point {

public:

  // Constructors
  Point();                    // Default
  Point(Point const & point); // Copy
  Point(double x, double y);  // Cartesian coordinates


  // Get functions
  double getX() const;
  double getY() const;

  // Functions 'get' for RDP's distance and tolerance
  double getDistance()  const;
  double getTolerance() const;

  // Functions 'get' for the right and left children' indices
  unsigned getLeftIdx() const;
  unsigned getRghtIdx() const;

  // Functions 'get' for level of detail and index of tolerance
  unsigned getLevel() const;
  unsigned getTolIdx() const;

  // Effective area access funtion
  double getEffectiveArea() const;

  // Indices access functions
  unsigned getPrevIdx() const;
  unsigned getNextIdx() const;

  // Flag access functions
  bool getErased() const;


  // Set functions
  void setX(double x);
  void setY(double y);

  // Functions 'set' for DP distance and Saalfeld's tolerance
  void setDistance (double distance);
  void setTolerance(double tolerance);
  void setTolerance(Point const & p, Point const & q);

  // Functions 'get' for the right and left sons' indices
  void setLeftIdx(unsigned leftIdx);
  void setRghtIdx(unsigned rghtIdx);

  // Functions 'set' for level of detail and index of tolerance
  void setLevel (unsigned level);
  void setTolIdx(unsigned tolIdx);

  // Indices modification functions
  void setPrevIdx(unsigned prevIdx);
  void setNextIdx(unsigned nextIdx);

  // Flag modification functions
  void setErased(bool erased);


  // Squared distances beween points and between a point and a segment
  double sqrDistance(Point const & p) const;
  double sqrDistance(Point const & p, Point const & q) const;

  // Triangle are for three points
  double triangleArea(Point const & p, Point const & q) const;


  // The cross product between 'pq' and 'p(*this)'
  double crossProduct(Point const & p, Point const & q) const;

  // Test for sidedness
  bool atLeft (Point const & p, Point const & q) const;
  bool atRight(Point const & p, Point const & q) const;
  bool onLine (Point const & p, Point const & q) const;

  bool atLeftOrOnLine(Point const & p, Point const & q) const;
  bool atRghtOrOnLine(Point const & p, Point const & q) const;

  // Test if point '(*this)' is inside triangle 'pqr'
  bool isInTriangle(Point const & p, Point const & q, Point const & r) const;


  // Operators of equality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Friend operators for input and output handle
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);

  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Cartesian coordinates
  double x;
  double y;

  // RDP's distance and tolerance
  double distance;
  double tolerance;

  // Indices of children in RDP's tree
  unsigned leftIdx;
  unsigned rghtIdx;

  // Level of detail and index of tolerance
  unsigned level;
  unsigned tolIdx;

  // Indices of previous and next vertices in polyline
  unsigned prevIdx;
  unsigned nextIdx;

  // Erased flag -- it indicates if this point has been erased from polyline
  bool erased;
};



inline Point::Point()
  : x(0.0),
    y(0.0),

    distance (0.0),
    tolerance(0.0),

    leftIdx(0),
    rghtIdx(0),

    level (0),
    tolIdx(0),

    prevIdx(0),
    nextIdx(0),

    erased(false) {}

inline Point::Point(Point const & point)
  : x(point.x),
    y(point.y),

    distance (point.distance),
    tolerance(point.tolerance),

    leftIdx(point.leftIdx),
    rghtIdx(point.rghtIdx),

    level (point.level),
    tolIdx(point.tolIdx),

    prevIdx(point.prevIdx),
    nextIdx(point.nextIdx),

    erased(point.erased) {}

inline Point::Point(double x, double y)
  : x(x),
    y(y),

    distance (0.0),
    tolerance(0.0),

    leftIdx(0),
    rghtIdx(0),

    level (0),
    tolIdx(0),

    prevIdx(0),
    nextIdx(0),

    erased(false) {}


inline double Point::getX() const { return x; }
inline double Point::getY() const { return y; }

inline double Point::getDistance() const  { return distance; }
inline double Point::getTolerance() const { return tolerance; }

inline unsigned Point::getLeftIdx() const { return leftIdx; }
inline unsigned Point::getRghtIdx() const { return rghtIdx; }

inline unsigned Point::getLevel() const  { return level; }
inline unsigned Point::getTolIdx() const { return tolIdx; }

inline unsigned Point::getPrevIdx() const { return prevIdx; }
inline unsigned Point::getNextIdx() const { return nextIdx; }

inline bool Point::getErased() const { return erased; }


inline void Point::setX(double x) { this->x = x; }
inline void Point::setY(double y) { this->y = y; }

inline void Point::setDistance (double dist)  { this->distance  = dist; }
inline void Point::setTolerance(double toler) { this->tolerance = toler; }
inline void Point::setTolerance(Point const & p, Point const & q)
{ this->tolerance = triangleArea(p, q); }

inline void Point::setLeftIdx(unsigned index) { this->leftIdx = index; }
inline void Point::setRghtIdx(unsigned index) { this->rghtIdx = index; }

inline void Point::setLevel (unsigned level)  { this->level  = level; }
inline void Point::setTolIdx(unsigned tolIdx) { this->tolIdx = tolIdx; }

inline void Point::setPrevIdx(unsigned prevIdx) { this->prevIdx = prevIdx; }
inline void Point::setNextIdx(unsigned nextIdx) { this->nextIdx = nextIdx; }

inline void Point::setErased(bool erased) { this->erased = erased; }


inline bool Point::operator!=(Point const & p) const { return !operator==(p); }
inline bool Point::operator==(Point const & p) const { return ((x == p.x) &&
							       (y == p.y)); }


inline double Point::sqrDistance(Point const & p) const
{ return (p.x-x)*(p.x-x)+(p.y-y)*(p.y-y); }

inline double Point::triangleArea(Point const & p, Point const & q) const
{ return fabs(crossProduct(p, q)/2.0); }


inline double Point::crossProduct(Point const & p, Point const & q) const
{ return (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y); }

inline bool Point::atLeft(Point const & p, Point const & q) const
{ return crossProduct(p, q) > 0; }

inline bool Point::atRight(Point const & p, Point const & q) const
{ return crossProduct(p, q) < 0; }

inline bool Point::onLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) == 0; }

inline bool Point::atLeftOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) >= 0; }

inline bool Point::atRghtOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) <= 0; }

inline bool Point::isInTriangle(Point const & p,
				Point const & q,
				Point const & r) const
{
  return r.atLeft(p, q)
    ? atLeft (p, q) && atLeft (q, r) && atLeft (r, p)
    : atRight(p, q) && atRight(q, r) && atRight(r, p);
}



#endif // __POINT_HH_
