#include "map.hh"




/**
 * @brief Simplifies polylines of this map using the RDP algorithm.
 *
 * The RDP algorithms ensures that each simplified polyline will be at most at
 * the distance @a tolerance from its original version.
 *
 * @param tolerance The tolerated distance.
 * @warning Topology of the map can be lost, because the polylines are
 * simplified in isolation.
 *****************************************************************************/
void Map::simplifyRDP(double tolerance)
{
  // Simplifying polyline
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    simplifyRDP(polIdx, 0, polylines[polIdx].numberOfVertices()-1,
		tolerance*tolerance);

  // Removing vertices
  removeVertices();
}



/**
 * @brief Applies the RDP algorithm to part this polyline.
 *
 * The simplification is carried out in the range [@a frstIdx, @a lastIdx] 'of
 * @polIdx th polyline of this map with a maximum distance of @a sqrTolerance,
 * which is given squared.
 *
 * @param polIdx The index of polyline to be simplified.
 * @param frstIdx The index of the first vertex in the range.
 * @param latsIdx The index of the last  vertex in the range.
 * @param sqrTolerance The squared maximum distance..
 *****************************************************************************/
void Map::simplifyRDP(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
		      double sqrTolerance)
{
  // Searching farthest vertex
  double   farthestSqrDistance = sqrTolerance;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	polylines[polIdx][vertexIdx].sqrDistance(polylines[polIdx][frstIdx],
						 polylines[polIdx][lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  // Testing whether distance is greater than tolerance
  if (farthestSqrDistance > sqrTolerance)
    {
      simplifyRDP(polIdx, frstIdx, farthestIdx, sqrTolerance);
      simplifyRDP(polIdx, farthestIdx, lastIdx, sqrTolerance);
    }

  else
    {
      polylines[polIdx][frstIdx].setNextIdx(lastIdx);
      polylines[polIdx][lastIdx].setPrevIdx(frstIdx);
    }
}
