#include "map.hh"

#include <iomanip>
#include <sys/time.h>




/**
 * @brief Simplifies polylines of this map using RDP algorithm for tolerance
 * 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void Map::simplifyTCRDP(double tolerance, double nearTolerance)
{
  // Allocating memory of list of convex hulls
  list<ConvexHull> * cHulls = new list<ConvexHull>[numberOfPolylines()];


  // Getting initial time
  timeval bgnTime;
  timeval endTime;
  double diffTime;

  gettimeofday(&bgnTime, NULL);
  

  // RDP simplification, outputing convex hulls
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    simplifyRDP(polIdx, 0, polylines[polIdx].numberOfVertices()-1,
		tolerance*tolerance);

  gettimeofday(&endTime, NULL);
  diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
	      (endTime.tv_usec-bgnTime.tv_usec)/1E6);

  cerr << "Time RDP:  " << setw(14) << diffTime << endl;


  // Adding initial features
  initConvexHulls(cHulls, nearTolerance);


  gettimeofday(&bgnTime, NULL);
  diffTime = ((bgnTime.tv_sec -endTime.tv_sec)+
	      (bgnTime.tv_usec-endTime.tv_usec)/1E6);

  cerr << "Time Init: " << setw(14) << diffTime << endl;


  // Writing cHulls
//   for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
//     for (cHullIter_t cHullIter = cHulls[polIdx].begin();
// 	 cHullIter != cHulls[polIdx].end(); ++cHullIter)
//       cerr << endl << *cHullIter << endl;


//   // Writing polyline contents
//   unsigned nPsFCalls = 0;
//   unsigned nFeatures = 0;

//   for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
//     {
//       list<unsigned> const & polIdces = polylines[polIdx].getPolylineIndices();

//       cerr << setw(4) << polIdx << " ("
// 	   << setw(5) << polylines[polIdx].numberOfVertices() << ") : ("
// 	   << setw(5) << polIdces.size() << ") : ("
// 	   << setw(5) << cHulls[polIdx].size() << ") :";

//       for (cHullIter_t cHullIter = cHulls[polIdx].begin();
// 	   cHullIter != cHulls[polIdx].end(); ++cHullIter)
// 	{
// 	  nPsFCalls += cHullIter->counter;
// 	  nFeatures += cHullIter->numberOfFeatures();
// 	  cerr << ' ' << cHullIter->numberOfFeatures() << ", "
// 	       << cHullIter->counter;
// 	}

//       cerr << endl;
//     }

//   cerr << "Number of features: " << nFeatures << endl
//        << "Number of psfcalls: " << nPsFCalls << endl;


  // Simplifying map
  simplifyTCRDP(tolerance*tolerance, cHulls);


  gettimeofday(&endTime, NULL);
  diffTime = ((endTime.tv_sec -bgnTime.tv_sec)+
	      (endTime.tv_usec-bgnTime.tv_usec)/1E6);

  cerr << "Time TC:   " << setw(14) << diffTime << endl;


  // Removing vertices not in simplified polylines
  removeVertices();


  gettimeofday(&bgnTime, NULL);
  diffTime = ((bgnTime.tv_sec -endTime.tv_sec)+
	      (bgnTime.tv_usec-endTime.tv_usec)/1E6);

  cerr << "Time Rem:  " << setw(14) << diffTime << endl
       << "-------------------------" << endl;


  // Deallocating memory previously allocated
  delete [] cHulls;
}



/**
 *
 *****************************************************************************/
void Map::initConvexHulls(list<ConvexHull> * cHulls, double nearTolerance)
{
  double diffTime0 = 0.0;
  double diffTime1 = 0.0;
  double diffTime2 = 0.0;

  timeval time1;
  timeval time2;


  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)

    for (unsigned vertexIdx = 0;
	 vertexIdx < polylines[polIdx].numberOfVertices()-1;
	 vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())
      {
 	gettimeofday(&time1, NULL);

	// Pushing convex hull in list of current polyline
	if (vertexIdx < polylines[polIdx][vertexIdx].getNextIdx()-1)
	  cHulls[polIdx].
	    push_back(ConvexHull(polylines[polIdx], nearTolerance, vertexIdx,
				 polylines[polIdx][vertexIdx].getNextIdx()));
	else continue;

	gettimeofday(&time2, NULL);
	diffTime0 += ((time2.tv_sec -time1.tv_sec)+
		      (time2.tv_usec-time1.tv_usec)/1E6);

	ConvexHull & cHull = cHulls[polIdx].back();


// 	// Setting near tolerance
// 	cHull.setNearTolerance(nearTolerance);


	// Adding external points
	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
	  if (cHull.isInside(points[pointIdx]))
	    cHull.pushFeature(points[pointIdx]);

	gettimeofday(&time1, NULL);
	diffTime1 += ((time1.tv_sec -time2.tv_sec)+
		      (time1.tv_usec-time2.tv_usec)/1E6);


	// Adding vertices
// 	list<unsigned> const & pIdces = polylines[polIdx].getPolylineIndices();

// 	for (list<unsigned>::const_iterator pol2Iter = pIdces.begin();
// 	     pol2Iter != pIdces.end(); ++pol2Iter)
// 	  {
// 	    if (cHull.overlaps(polylines[*pol2Iter]))
// 	      for (unsigned vertex2Idx = 0;
// 		   vertex2Idx < polylines[*pol2Iter].numberOfVertices();
// 		   vertex2Idx = polylines[*pol2Iter][vertex2Idx].getNextIdx())

// 		if (cHull.isInside(polylines[*pol2Iter][vertex2Idx]))
// 		  cHull.pushFeature(polylines[*pol2Iter][vertex2Idx]);
// 	  }

	for (unsigned vertex2Idx = 0;
	     vertex2Idx < polylines[polIdx].numberOfVertices();
	     vertex2Idx = polylines[polIdx][vertex2Idx].getNextIdx())

	  if (cHull.isInside(polylines[polIdx][vertex2Idx]))
	    cHull.pushFeature(polylines[polIdx][vertex2Idx]);


	for (unsigned pol2Idx = 0; pol2Idx < numberOfPolylines(); ++pol2Idx)

// 	  if (cHull.overlaps(polylines[pol2Idx]))
	    for (unsigned vertex2Idx = 0;
		 vertex2Idx < polylines[pol2Idx].numberOfVertices();
		 vertex2Idx = polylines[pol2Idx][vertex2Idx].getNextIdx())

	      if (cHull.isInside(polylines[pol2Idx][vertex2Idx]))
		cHull.pushFeature(polylines[pol2Idx][vertex2Idx]);

	gettimeofday(&time2, NULL);
	diffTime2 += ((time2.tv_sec -time1.tv_sec)+
		      (time2.tv_usec-time1.tv_usec)/1E6);
      }

  cerr << "Time Cre:  " << setw(14) << diffTime0 << endl;
  cerr << "Time Pnt:  " << setw(14) << diffTime1 << endl;
  cerr << "Time Pol:  " << setw(14) << diffTime2 << endl;
}



// /**
//  *
//  *****************************************************************************/
// void Map::initConvexHulls(list<ConvexHull> * cHulls, double nearTolerance)
// {
//   for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)

//     for (unsigned vertexIdx = 0;
// 	 vertexIdx < polylines[polIdx].numberOfVertices()-1;
// 	 vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())
//       {
// 	// Pushing convex hull in list of current polyline
// 	if (vertexIdx < polylines[polIdx][vertexIdx].getNextIdx()-1)
// 	  cHulls[polIdx].
// 	    push_back(ConvexHull(polylines[polIdx],vertexIdx,
// 				 polylines[polIdx][vertexIdx].getNextIdx()));
// 	else continue;

// 	ConvexHull & cHull = cHulls[polIdx].back();


// 	// Setting near tolerance
// 	cHull.setNearTolerance(nearTolerance);


// 	// Adding external points
// 	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
// 	  if (cHull.isInside(points[pointIdx]))
// 	    cHull.pushFeature(points[pointIdx]);
//       }


//   // Adding vertices
//   for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
//     for (unsigned pol2Idx = 0; pol2Idx < numberOfPolylines(); ++pol2Idx)
//       if (polylines[polIdx].overlaps(polylines[pol2Idx]))
// 	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
// 	     cHullIter != cHulls[polIdx].end(); ++cHullIter)
//  	  if (cHullIter->overlaps(polylines[pol2Idx]))
// 	    for (unsigned vertex2Idx = 0;
// 		 vertex2Idx < polylines[pol2Idx].numberOfVertices();
// 		 vertex2Idx = polylines[pol2Idx][vertex2Idx].getNextIdx())

// 	      if (cHullIter->isInside(polylines[pol2Idx][vertex2Idx]))
// 		cHullIter->pushFeature(polylines[pol2Idx][vertex2Idx]);
// }



/**
 * @brief Simplifies map 'maptcrdp' with Silva's version of RDP algorithm for
 * squared tolerance 'sqrTolerance', filling up the convex hulls in 'cHulls'.
 *****************************************************************************/
void Map::simplifyTCRDP(double sqrTolerance, list<ConvexHull> * cHulls)
{
  // Correcting sidedness errors
  while (true)
    {
      // Applying Silva's correction
      list<Point *> newVertices;

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); )
	  correctTCRDP(polIdx, sqrTolerance, cHullIter++,
		       cHulls[polIdx], newVertices);


      // Test if there is no new vertex
      if (newVertices.empty()) break;


      // Test for new vertices
      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); ++cHullIter)

	  for (list<Point *>::iterator vertexIter = newVertices.begin();
	       vertexIter != newVertices.end(); ++vertexIter)
	    if (cHullIter->isInside(**vertexIter))
	      cHullIter->pushFeature(**vertexIter);

//       // Writing polyline contents
//       unsigned nPsFCalls = 0;
//       unsigned nFeatures = 0;

//       for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
// 	{
// 	  list<unsigned> const & polIdces =
// 	    polylines[polIdx].getPolylineIndices();

// 	  cerr << setw(4) << polIdx << " ("
// 	       << setw(5) << polylines[polIdx].numberOfVertices() << ") : ("
// 	       << setw(5) << polIdces.size() << ") : ("
// 	       << setw(5) << cHulls[polIdx].size() << ") :";

// 	  for (cHullIter_t cHullIter = cHulls[polIdx].begin();
// 	       cHullIter != cHulls[polIdx].end(); ++cHullIter)
// 	    {
// 	      nPsFCalls += cHullIter->counter;
// 	      nFeatures += cHullIter->numberOfFeatures();
// 	      cerr << " (" << cHullIter->numberOfFeatures() << ", "
// 		   << cHullIter->counter << ')';
// 	    }

// 	  cerr << endl;
// 	}
    }
}



/**
 * Simplifies polyline 'polylines[polIdx]' of 'tcdpmap' using RDP algorithm for
 * squared tolerance 'sqrTolerance' with convex hull 'cHullIter' in convex
 * hulls list 'cHulls'.  New vertices are added to 'newVertices' list.
 *****************************************************************************/
void Map::correctTCRDP(unsigned polIdx, double sqrTolerance,
		       cHullIter_t cHullIter, list<ConvexHull> & cHulls,
		       list<Point *> & newVertices)
{
  unsigned frstIdx = cHullIter->getFrstIdx();
  unsigned lastIdx = cHullIter->getLastIdx();


  double   farthestSqrDistance = 0.0;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	polylines[polIdx][vertexIdx].sqrDistance(polylines[polIdx][frstIdx],
						 polylines[polIdx][lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
 	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  if (farthestSqrDistance > sqrTolerance || cHullIter->getInconsistent())
    {
      // Correcting links of neighbouring vertices
      polylines[polIdx][frstIdx].setNextIdx(farthestIdx);
      polylines[polIdx][lastIdx].setPrevIdx(farthestIdx);


      // Correcting its links
      polylines[polIdx][farthestIdx].setPrevIdx(frstIdx);
      polylines[polIdx][farthestIdx].setNextIdx(lastIdx);


      // Adding vertex in list of new vertices
      newVertices.push_back(&polylines[polIdx][farthestIdx]);


      // Setting first and last convex hulls
      ConvexHull frstCHull(polylines[polIdx], cHullIter->getNearTolerance());
      ConvexHull lastCHull(polylines[polIdx], cHullIter->getNearTolerance());
      cHullIter->split(farthestIdx, frstCHull, lastCHull);


      // Erasing remaining convex hull
      cHullIter = cHulls.erase(cHullIter);


      // Inserting new convex hulls and applying Saalfeld correction
      cHullIter_t frstCHullIter;
      cHullIter_t lastCHullIter;

      if (farthestIdx > frstIdx+1)
	{
	  frstCHullIter = cHulls.insert(cHullIter, frstCHull);
	  correctTCRDP(polIdx,sqrTolerance,frstCHullIter,cHulls,newVertices);
	}

      if (farthestIdx < lastIdx-1)
	{
	  lastCHullIter = cHulls.insert(cHullIter, lastCHull);
	  correctTCRDP(polIdx,sqrTolerance,lastCHullIter,cHulls,newVertices);
	}
    }
}
