#include "map.hh"




/**
 * @brief Simplifies polylines of this map using VIS algorithm for tolerance
 * 'tolerance', respecting relative position of internal points.
 *****************************************************************************/
void Map::simplifyTCVIS(double tolerance, double nearTolerance)
{
  // Allocating memory of list of convex hulls
  list<ConvexHull> * cHulls = new list<ConvexHull>[numberOfPolylines()];


  // VIS simplification, outputing convex hulls
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    simplifyVIS(polIdx, tolerance);

  // Adding initial features
  initConvexHulls(cHulls, nearTolerance);

//   for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
//     for (cHullIter_t cHullIter = cHulls[polIdx].begin();
// 	 cHullIter != cHulls[polIdx].end(); ++cHullIter)
//       cerr << cHullIter->getFrstIdx() << " -> "
// 	   << cHullIter->getLastIdx() << endl;


  // Simplifying map
  simplifyTCVIS(tolerance, cHulls);


  // Removing vertices not in simplified polylines
  removeVertices();


  // Deallocating memory previously allocated
  delete [] cHulls;
}



/**
 * @brief Simplifies map 'maptcrdp' with Silva's version of VIS algorithm for
 * squared tolerance 'sqrTolerance', filling up the convex hulls in 'cHulls'.
 *****************************************************************************/
void Map::simplifyTCVIS(double tolerance, list<ConvexHull> * cHulls)
{
  // Correcting sidedness errors
  while (true)
    {
      // Applying Silva's correction
      list<Point *> newVertices;

      for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); )
	  correctTCVIS(polIdx, tolerance, cHullIter++,
		       cHulls[polIdx], newVertices);


      // Test if there is no new vertex
      if (newVertices.empty()) break;


      // Test for new vertices
      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (cHullIter_t cHullIter = cHulls[polIdx].begin();
	     cHullIter != cHulls[polIdx].end(); ++cHullIter)
	  {
	    for (list<Point *>::iterator vertexIter = newVertices.begin();
		 vertexIter != newVertices.end(); ++vertexIter)
	      if (cHullIter->isInside(**vertexIter))
		cHullIter->pushFeature(**vertexIter);

	    // Sorting features and removing repeated ones
	    cHullIter->uniqueFeatures();
	  }
    }
}



/**
 * Simplifies polyline 'polylines[polIdx]' of 'tcdpmap' using VIS algorithm for
 * squared tolerance 'sqrTolerance' with convex hull 'cHullIter' in convex
 * hulls list 'cHulls'.  New vertices are added to 'newVertices' list.
 *****************************************************************************/
void Map::correctTCVIS(unsigned polIdx, double tolerance,
		       cHullIter_t cHullIter, list<ConvexHull> & cHulls,
		       list<Point *> & newVertices)
{
  unsigned frstIdx = cHullIter->getFrstIdx();
  unsigned lastIdx = cHullIter->getLastIdx();
  unsigned farthestIdx =
    (polylines[polIdx][frstIdx].getLevel() >
     polylines[polIdx][lastIdx].getLevel())
    ? static_cast<unsigned>(polylines[polIdx][frstIdx].getRghtIdx())
    : static_cast<unsigned>(polylines[polIdx][lastIdx].getLeftIdx());


//   cerr << *cHullIter << endl;
  

//   cerr << frstIdx << " -> " << lastIdx << ": " << farthestIdx << endl
//        << "frst level: " << polylines[polIdx][frstIdx].getLevel() << endl
//        << "last level: " << polylines[polIdx][lastIdx].getLevel() << endl
//        << "frst rght son: " << polylines[polIdx][frstIdx].getRghtIdx() << endl
//        << "last left son: " << polylines[polIdx][lastIdx].getLeftIdx() << endl;


  // Computing smallest effective area
  double smallestEffectiveArea =
    polylines[polIdx][farthestIdx].triangleArea(polylines[polIdx][frstIdx],
						polylines[polIdx][lastIdx]);


  if (smallestEffectiveArea > tolerance || cHullIter->getInconsistent())
    {
      // Correcting links of neighbouring vertices
      polylines[polIdx][frstIdx].setNextIdx(farthestIdx);
      polylines[polIdx][lastIdx].setPrevIdx(farthestIdx);


      // Adding new vertex
      newVertices.push_back(&polylines[polIdx][farthestIdx]);


      // Setting first and last convex hulls
      ConvexHull frstCHull(polylines[polIdx], cHullIter->getNearTolerance());
      ConvexHull lastCHull(polylines[polIdx], cHullIter->getNearTolerance());
      cHullIter->split(farthestIdx, frstCHull, lastCHull);


      // Erasing remaining convex hull
      cHullIter = cHulls.erase(cHullIter);


      // Inserting new convex hulls and applying Saalfeld correction
      cHullIter_t frstCHullIter;
      cHullIter_t lastCHullIter;

      if (farthestIdx > frstIdx+1)
	{
	  frstCHullIter = cHulls.insert(cHullIter, frstCHull);
	  correctTCVIS(polIdx, tolerance, frstCHullIter, cHulls, newVertices);
	}

      if (farthestIdx < lastIdx-1)
	{
	  lastCHullIter = cHulls.insert(cHullIter, lastCHull);
	  correctTCVIS(polIdx, tolerance, lastCHullIter, cHulls, newVertices);
	}
    }
}
