#include "map.hh"

#include <iomanip>



/**
 * @brief Simplifies polylines of this map using Visvalingam's algorithm.
 *****************************************************************************/
void Map::simplifyVIS(double tolerance)
{
  // Simplifying polylines of map
  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
    simplifyVIS(polIdx, tolerance);


  // Removing vertices
  removeVertices();
}



/**
 * @brief Simplifies a polyline of this map using Visvalingam's algorithm.
 *****************************************************************************/
void Map::simplifyVIS(unsigned polIdx, double tolerance)
{
  // First step
  for (unsigned vertexIdx = 1; vertexIdx <
	 polylines[polIdx].numberOfVertices()-1; ++vertexIdx)
    polylines[polIdx][vertexIdx].setTolerance(polylines[polIdx][vertexIdx-1],
					      polylines[polIdx][vertexIdx+1]);


  // Second step: eliminate vertices with smallest effective area
  for (unsigned level = polylines[polIdx].numberOfVertices()-3; true; --level)
    {
      // Searching vertex with smallest effective area
      double   smallestEffectiveArea = tolerance;
      unsigned smallestIdx           = 0;

      for (unsigned vertexIdx = polylines[polIdx].front().getNextIdx();
	   vertexIdx < polylines[polIdx].numberOfVertices()-1;
	   vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())
	{
	  double currentEffectiveArea =
	    polylines[polIdx][vertexIdx].getTolerance();

	  if (currentEffectiveArea < smallestEffectiveArea)
	    {
	      smallestEffectiveArea = currentEffectiveArea;
	      smallestIdx           = vertexIdx;
	    }
	}


      // Updating if under tolerance
      if (smallestEffectiveArea < tolerance)
	{
	  // Initializing references to neighbours
	  Point & prevPoint =
	    polylines[polIdx][polylines[polIdx][smallestIdx].getPrevIdx()];

	  Point & nextPoint =
	    polylines[polIdx][polylines[polIdx][smallestIdx].getNextIdx()];


	  // Updating previous and next vertices of its neighbours
	  prevPoint.setNextIdx(polylines[polIdx][smallestIdx].getNextIdx());
	  nextPoint.setPrevIdx(polylines[polIdx][smallestIdx].getPrevIdx());


	  // Updating children of its neighbours
	  prevPoint.setRghtIdx(smallestIdx);
	  nextPoint.setLeftIdx(smallestIdx);


	  // Updating levels of its neighbours
	  prevPoint.setLevel(level);
	  nextPoint.setLevel(level);


	  // Updating effective area of its neighbours
	  if (&prevPoint != &polylines[polIdx].front())
	    prevPoint.setTolerance(polylines[polIdx][prevPoint.getPrevIdx()],
				   polylines[polIdx][prevPoint.getNextIdx()]);

	  if (&nextPoint != &polylines[polIdx].back())
	    nextPoint.setTolerance(polylines[polIdx][nextPoint.getPrevIdx()],
				   polylines[polIdx][nextPoint.getNextIdx()]);
	}

      // No more vertices with effective area less than tolerance
      else break;
    }
}
