#include "map.hh"

#include <iomanip>




/**
 * @brief Simplifies polylines of this map using the RDP algorithm.
 *
 * The RDP algorithms ensures that each simplified polyline will be at most at
 * the distance @a tolerance from its original version.
 *
 * @param tolerance The maximum distance.
 * @warning Topology of the map can be lost, because the polylines are
 * simplified in isolation.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::simplifyRDP(double tolerance)
{
  // Simplifying polylines of map
  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
    {
      // Simplifying current polyline
      polylines[polIdx].simplifyRDP(tolerance);

      // Erasing vertices marked as erased from current polyline
      unsigned currIdx = 1;
      for (unsigned vertexIdx = 1; vertexIdx <
	     polylines[polIdx].size(); ++vertexIdx)
	{
	  while (polylines[polIdx][vertexIdx].getErased()) ++vertexIdx;
	  polylines[polIdx][currIdx++] = polylines[polIdx][vertexIdx];
	}
      polylines[polIdx].resize(currIdx);
    }
}



/**
 * @brief Writes this map in @c ostr.
 *
 * First, writes the number of polylines, followed by the polylines.  Then,
 * writes the number of points, followed by the features points.  Uses the
 * Polyline::print and Point::print.
 *
 * @param ostr The output stream.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::print(ostream & ostr) const
{
  // Setting needed flags
  ostr.flags(ios::scientific|ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      ostr << endl;
      polylines[polIdx].print(ostr);
    }


  // Printing points
  ostr << endl << endl << numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < numberOfPoints(); ++pointIdx)
    {
      points[pointIdx].print(ostr);
      ostr << endl;
    }
}



/**
 * @brief Writes a map in an output stream.
 *
 * First, writes the number of polylines, followed by the polylines.  Then,
 * writes the number of points, followed by the points.  Is a redefinition of
 * the insertion operator in the context of streams.
 *
 * @param ostr The output stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Map const & map)
{
  // Setting needed flags
  ostr.flags(ios::scientific|ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << map.numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    ostr << endl << map.polylines[polIdx];


  // Printing points
  ostr << endl << endl << map.numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    ostr << map.points[pointIdx] << endl;


  return ostr;
}



/**
 * @brief Reads a map from an input stream.
 *
 * First, reads the number of polylines, followed by the polylines.  Then,
 * reads the number of points, followed by the points.  Is a redefinition of
 * the extraction operator in the context of streams.
 *
 * @param istr The input stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Map & map)
{
  // Reading polylines
  unsigned nPolylines;

  istr >> nPolylines;
  map.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    istr >> map.polylines[polIdx];


  // Reading points
  unsigned nPoints;

  istr >> nPoints;
  map.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    istr >> map.points[pointIdx];


  return istr;
}
