#include "point.hh"

#include <iomanip>




/**
 * @brief It returns the shortest distance from this point to a straight
 * line segment.
 *
 * It it useful for the RDP algorithm.
 *
 * @param p First extreme of line segment.
 * @param q Second extreme of line segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
double Point::distance(Point const & p, Point const & q) const
{
  // 'p' or 'q' is the closest point to 'point'
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return distance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return distance(q);

  // 'point' is closest to a point between 'p' and 'q'
  double norm = p.distance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return fabs(determinant)/norm;
}



/**
 * @brief It returns the shortest distance from this point to a straight
 * line segment.  The distance is returned squared.
 *
 * Is is useful for the RDP algorithm.
 *
 * @param p First extreme of line segment.
 * @param q Second extreme of line segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
double Point::sqrDistance(Point const & p, Point const & q) const
{
  // 'p' or 'q' is the closest point to 'point'
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return sqrDistance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return sqrDistance(q);

  // 'point' is closest to a point between 'p' and 'q'
  double sqrNorm = p.sqrDistance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return determinant*determinant/sqrNorm;
}



/**
 * @brief It writes this point in the stream @a ostr.
 *
 * It writes the coordinates x and y separated by a comma in a single line.
 *
 * @param ostr Output stream.
 * @param width Width of the field coordinates will occupy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Point::print(ostream & ostr, int width) const
{
  ostr << setw(width) << x << ", "
       << setw(width) << y << endl;
}



/**
 * @brief It writes @a point in the stream @a ostr.
 *
 * It writes the coordinates x and y.  Each coordinate will occupy a field
 * of width COORD_WIDTH.
 *
 * @param ostr Output stream.
 * @param point The point to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Point const & point)
{
  ostr << setw(COORD_WIDTH) << point.x
       << setw(COORD_WIDTH) << point.y;


  return ostr;
}



/**
 * @brief It reads @a point from the stream @a istr.
 *
 * It reads coordinates x and y.
 *
 * @param istr Input stream.
 * @param point The point to store the coordinates.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Point & point)
{
  istr >> point.x >> point.y;


  return istr;
}
