#ifndef __POLYLINE_HH_
#define __POLYLINE_HH_


#include "vertex.hh"

#include <iostream>
#include <vector>


using namespace std;


/**
 * @brief This class implements a polyline.
 *
 * It is a subclass of the class vector<Vertex>, that is it stores a
 * vector of vertices.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Polyline
  : public vector<Vertex> {

public:

  // Returns the number of vertices of this polyline
  unsigned numberOfVertices() const;


  // RDP line simplification procedure
  void simplifyRDP(double tolerance);


  // Printing function
  void print(ostream & ostr) const;
  void print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const;

  // Friendship relation of insertion and extraction operators
  friend ostream & operator<<(ostream & ostr, Polyline const & polyline);
  friend istream & operator>>(istream & istr, Polyline       & polyline);


private:

  // Internal RDP line simplification procedure
  void simplifyRDP(double sqrTolerance, unsigned frstIdx, unsigned lastIdx);
};



/**
 * @brief Returns the number of vertices of this polyline.
 *
 * Uses vector<Point>::size().
 *
 * @return Number of vertices
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Polyline::numberOfVertices() const { return size(); }


/**
 * @brief Simplifies this polyline using the RDP algorithm.
 *
 * The RDP algorithms ensures that each simplified polyline will be at most at
 * the distance @a tolerance from its original version.
 *
 * @param tolerance The maximum distance.
 * @warning Topology of the map can be lost, because the polylines are
 * simplified in isolation.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Polyline::simplifyRDP(double tolerance)
{ simplifyRDP(tolerance*tolerance, 0, numberOfVertices()-1); }


/**
 * @brief Prints this polyline in @c ostr.
 *
 * Is equivalent to Polyline::print(@c ostr, 0, Polyline::numberOfVertices()-1)
 *
 * @param ostr Output stream in it prints this polyline.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Polyline::print(ostream & ostr) const { print(ostr, 0, size()-1); }



#endif // __POLYLINE_HH_
