#ifndef __VERTEX_HH_
#define __VERTEX_HH_


#include <iostream>

#include "point.hh"


using namespace std;


/**
 * @brief This class implements a 2D vertex of a polyline.
 * 
 * This class is a subclass of the class @c Point.  Besides its coordinates,
 * it stores a flag that indicates if this vertex has been erased from the
 * polyline it belongs to.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Vertex
  : public Point {

public:

  // Constructors
  Vertex(Vertex const & vertex);           // Copy
  Vertex(Point  const & point);            // Point
  Vertex(double x = 0.0, double y = 0.0);  // Cartesien coordinates


  // Get and set functions
  bool getErased() const;
  void setErased(bool erased);


  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Erased flag
  bool erased; /// Flag that indicates if this vertex has been erased from a
	       /// polyline it belongs to.
};



/**
 * @brief The copy constructor.
 *
 * It creates a new vertex from @a vertex.
 *
 * @param vertex The vertex to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Vertex const & vertex)
  : Point(vertex), erased(vertex.erased) {}


/**
 * @brief The point constructor.
 *
 * It creates a new vertex from @a point.
 *
 * @param point The point to copy coordinates.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Point const & point)
  : Point(point), erased(false) {}


/**
 * @brief The xy-coordinate constructor.
 *
 * It creates a new vertex with given coordinates.
 *
 * @param x The @a x coordinate.
 * @param y The @a y coordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(double x, double y) : Point(x, y), erased(false) {}


/**
 * @brief It returns the flag @c erased of this vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Vertex::getErased() const { return erased; }


/**
 * @brief It sets the flag @c erased of this vertex from @a erased.
 *
 * @param erased The new flag.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Vertex::setErased(bool erased) { this->erased = erased; }



#endif // __POINT_HH_
