#include "convexhull.hh"

#include <iomanip>




/******************************************************************************
 * convexHull.set(frstIdx, lastIdx)
 ******************************************************************************
 * Determines the convex hull of the sub-polyline beginning at the vertex
 * 'frstIdx' and ending at the vertex 'lastIdx'.
 *****************************************************************************/
void ConvexHull::set(unsigned frstIdx, unsigned lastIdx)
{
  /* Setting first and last indices */
  this->frstIdx = frstIdx;
  this->lastIdx = lastIdx;

  /* Inserting first point once */
  push_front(frstIdx);

  /* Inserting second point at bottom and top */
  push_front(frstIdx+1);
  push_back (frstIdx+1);


  /* Handling rest of points */
  for (unsigned pointIdx = frstIdx+2; pointIdx <= lastIdx; ++pointIdx)
    {
      if (polyline[pointIdx].atLeftOrOnLine(polyline[front(0)],
					    polyline[front(1)]) &&
	  polyline[pointIdx].atLeftOrOnLine(polyline[back(1)],
					    polyline[back(0)]))
	continue;

      while (polyline[pointIdx].atRight(polyline[front(0)],
					polyline[front(1)])) pop_front();
      push_front(pointIdx);

      while (polyline[pointIdx].atRight(polyline[back(1)],
					polyline[back(0)])) pop_back();
      push_back(pointIdx);
    }
}



/******************************************************************************
 * convexHull.splitConvexHull(splitIdx, frstCHull, lastCHull)
 ******************************************************************************
 * Split the convex hull into two convex hulls at the vertex 'splitIdx'.  The
 * two resulting convex hulls are placed in 'frstCHull' and 'lastCHull'.
 *****************************************************************************/
void ConvexHull::split(unsigned spltIdx,
		       ConvexHull & frstCHull, ConvexHull & scndCHull)
{
  // Setting both convex hulls
  frstCHull.set(frstIdx, spltIdx);
  scndCHull.set(spltIdx, lastIdx);


  /* Handling features for the new convex hulls */
  for (featureIter_t featureIter = features.begin();
       featureIter != features.end(); ++featureIter)
    {
      Point const & point = *featureIter->point;


      /* Setting side of feature */
      if (point.isInTriangle(polyline[frstIdx],
			     polyline[lastIdx],
			     polyline[spltIdx]))
	featureIter->wrongSide = !featureIter->wrongSide;


      /* Inserting features in new convex hulls */
      if (frstCHull.isInside(point)) frstCHull.pushFeature(*featureIter);
      if (scndCHull.isInside(point)) scndCHull.pushFeature(*featureIter);
    }


  /* Inserting vertices of the other subpolyline */
  for (unsigned vertexIdx = spltIdx+1; vertexIdx <= lastIdx; ++vertexIdx)
    if (frstCHull.isInside(polyline[vertexIdx]))
      frstCHull.pushFeature(polyline[vertexIdx]);

  for (unsigned vertexIdx = frstIdx; vertexIdx < spltIdx; ++vertexIdx)
    if (scndCHull.isInside(polyline[vertexIdx]))
      scndCHull.pushFeature(polyline[vertexIdx]);
}



/******************************************************************************
 * convexHull.isInside(point)
 ******************************************************************************
 * Returns 'true' if the point 'point' is inside the convex hull, and 'false'
 * otherwise.
 *****************************************************************************/
bool ConvexHull::isInside(Point const & point)
{
  for (unsigned pointIdx = 0; pointIdx < size()-1; ++pointIdx)
    if (point.atRight(polyline[front(pointIdx)],
		      polyline[front(pointIdx+1)]))
      return false;

  return true;
}



/******************************************************************************
 * polyline.pushFeature(point)
 ******************************************************************************
 * Pushes a new feature point 'point' into the features list.  It sets also the
 * sidedness of both feature point and convex hull.
 *****************************************************************************/
void ConvexHull::pushFeature(Point const & point)
{
  /* Pushing feature point */
  features.push_back(feature_t(&point));


  /* Testing for crossings */
  if (point != polyline[frstIdx] &&
      point != polyline[lastIdx])
    {
      bool atLeft = polyline[lastIdx].getX() <= point.getX();

      for (unsigned prevIdx = lastIdx, currIdx = frstIdx;
	   currIdx <= lastIdx; prevIdx = currIdx++)

	if (polyline[currIdx].getX() <= point.getX())
	  {
	    if (!atLeft && point.atLeft(polyline[prevIdx], polyline[currIdx]))
	      features.back().wrongSide = !features.back().wrongSide;

	    atLeft = true;
	  }
	else if (polyline[currIdx].getX() > point.getX())
	  {
	    if (atLeft && point.atRight(polyline[prevIdx], polyline[currIdx]))
	      features.back().wrongSide = !features.back().wrongSide;

	    atLeft = false;
	  }
    }


  if (features.back().wrongSide)
    wrongSide = true;
}



/******************************************************************************
 * polyline.pushFeature(feature)
 ******************************************************************************
 * Pushes a new feature point 'feature' into the features list.  It sets also
 * the sidedness of the convex hull.
 *****************************************************************************/
void ConvexHull::pushFeature(feature_t const & feature)
{
  features.push_back(feature);

  if (feature.wrongSide)
    wrongSide = true;
}



/******************************************************************************
 * convexHull.print(ostr)
 ******************************************************************************
 * Prints the list of point indeces sequentially ordered, one per line.
 *****************************************************************************/
void ConvexHull::print(ostream & ostr) const
{
  ostr << size() << endl;

  for (unsigned pointIdx = 0; pointIdx < size(); ++pointIdx)
    ostr << setw(3) << (*this)[pointIdx] << endl;
}



/******************************************************************************
 * ostr << convexhull
 ******************************************************************************
 * Prints out the number of indeces, followed by the list of point indeces
 * sequentially ordered, one per line.  It is a redefinition of the insertion
 * operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, ConvexHull const & convexHull)
{
  ostr << "  fr: " << setw(4) << convexHull.frstIdx << " ("
       << convexHull.polyline[convexHull.frstIdx] << ')'<< endl
       << "  to: " << setw(4) << convexHull.lastIdx << " ("
       << convexHull.polyline[convexHull.lastIdx] << ')'<< endl;

  ostr << convexHull.size() << endl;

  for (unsigned pointIdx = 0; pointIdx < convexHull.size(); ++pointIdx)
    ostr << setw(4) << convexHull[pointIdx] << " ("
	 << convexHull.polyline[convexHull[pointIdx]] << ')' << endl;


  ostr << " - features: " << convexHull.features.size() << endl;

  for (list<feature_t>::const_iterator featIter = convexHull.features.begin();
       featIter != convexHull.features.end(); ++featIter)
    ostr << " : " << *featIter->point
	 << " - " <<  (featIter->wrongSide ? "WRONG" : "CORRECT") << endl;


  return ostr;
}
