#ifndef __CONVEXHULL_HH_
#define __CONVEXHULL_HH_


#include "polyline.hh"

#include <deque>
#include <list>


using namespace std;




/******************************************************************************
 * External types
 *****************************************************************************/
typedef deque<unsigned> convexHull_t;

typedef deque<unsigned>::iterator            hullIter_t;
typedef deque<unsigned>::const_iterator constHullIter_t;

class ConvexHull;
typedef list<ConvexHull *>::iterator            cHullIter_t;
typedef list<ConvexHull *>::const_iterator constCHullIter_t;


/******************************************************************************
 * Internal types
 *****************************************************************************/
struct feature_t {

  feature_t(Point const * point)
    : point(point), wrongSide(false) {}

  feature_t(Point const * point, bool wrongSide)
    : point(point), wrongSide(wrongSide) {}


  Point const * point;
  bool wrongSide;


  bool operator==(feature_t const & f) const { return *point == *f.point; }
  bool operator< (feature_t const & f) const
  {
    return point->getX() == f.point->getX()
      ?    point->getY() <  f.point->getY()
      :    point->getX() <  f.point->getX();
  }
};

typedef list<feature_t>::iterator            featureIter_t;
typedef list<feature_t>::const_iterator constFeatureIter_t;



/******************************************************************************
 * 'ConvexHull' is a double-ended queue of indices.
 *****************************************************************************/
class ConvexHull
  : public deque<unsigned> {

public:

  /* Constructors */
  ConvexHull(Polyline const & polyline);
  ConvexHull(Polyline const & polyline, unsigned frstIdx, unsigned lastIdx);


  /* Determining convex hull of 'polyline' between 'frstIdx' and 'lastIdx' */
  void set(unsigned frstIdx, unsigned lastIdx);

  /* Spliting convex hull at 'splitIdx' */
  void split(unsigned splitIdx, ConvexHull& frstCHull, ConvexHull& scndCHull);

  /* Determines if 'point' is inside convex hulll */
  bool isInside(Point const & point);

  /* Pushes a new feature into the features' list */
  void pushFeature(Point const & point);
  void pushFeature(feature_t const & feature);

  void uniqueFeatures();


  /* Get functions */
  Polyline const & getPolyline() const;

  unsigned getFrstIdx() const;
  unsigned getLastIdx() const;

  bool getWrongSide() const;


  /* Returns the front/back element with an offset of 'offset' */
  unsigned       & front(unsigned offset = 0);
  unsigned const & front(unsigned offset = 0) const;

  unsigned       & back(unsigned offset = 0);
  unsigned const & back(unsigned offset = 0) const;


  /* Friendship relation of insertion operator */
  friend ostream & operator<<(ostream & ostr, ConvexHull const & convexHull);

  /* Printing function */
  void print(ostream & ostr) const;


private:

  /* Referenced polyline */
  Polyline const & polyline;

  /* Limit indices of convex hull */
  unsigned frstIdx;
  unsigned lastIdx;

  /* List of features */
  list<feature_t> features;

  /* Sidedness test */
  bool wrongSide;
};


inline ConvexHull::ConvexHull(Polyline const & polyline)
  : polyline(polyline), wrongSide(false) {}

inline ConvexHull::ConvexHull(Polyline const & polyline,
			      unsigned frstIdx, unsigned lastIdx)
  : polyline(polyline), wrongSide(false)
{
  set(frstIdx, lastIdx);
}


inline void ConvexHull::uniqueFeatures()
{
  features.sort();
//   features.unique();
}


inline Polyline const & ConvexHull::getPolyline() const { return polyline; }

inline unsigned ConvexHull::getFrstIdx() const { return frstIdx; }
inline unsigned ConvexHull::getLastIdx() const { return lastIdx; }

inline bool ConvexHull::getWrongSide() const { return wrongSide; }

inline unsigned & ConvexHull::front(unsigned offset)
{
  return (*this)[offset];
}

inline unsigned const & ConvexHull::front(unsigned offset) const
{
  return (*this)[offset];
}

inline unsigned & ConvexHull::back(unsigned offset)
{
  return (*this)[size()-1-offset];
}

inline unsigned const & ConvexHull::back(unsigned offset) const
{
  return (*this)[size()-1-offset];
}




#endif // __CONVEXHULL_HH_
