#include "maptcrdp.hh"

#include <iomanip>
#include <cstdlib>
#include <sys/time.h>




/**
 *
 *****************************************************************************/
void MapTCRDP::simplifyTCRDP(double tolerance)
{
  // RDP simplification
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    simplifyRDP(polIdx, 0, polylines[polIdx].numberOfVertices()-1,
		tolerance*tolerance);


  // Adding initial features
  list<ConvexHull *> cHulls;
  initConvexHulls(cHulls);


  // Testing for sidedness error
  for (cHullIter_t cHullIter = cHulls.begin();  cHullIter != cHulls.end(); )
    correct(cHulls, cHullIter++, tolerance*tolerance);


  // Removing erased vertices in lines of map
  removeVertices();
}



/**
 *
 *****************************************************************************/
void MapTCRDP::simplifyRDP(unsigned polIdx, unsigned frstIdx, unsigned lastIdx,
			   double sqrTolerance)
{
  // Searching fathest vertex
  double   farthestSqrDistance = sqrTolerance;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	polylines[polIdx][vertexIdx].sqrDistance(polylines[polIdx][frstIdx],
						 polylines[polIdx][lastIdx]);

      if (currentSqrDistance > farthestSqrDistance)
	{
	  farthestSqrDistance = currentSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  // Recursive calls
  if (farthestSqrDistance > sqrTolerance)
    {
      simplifyRDP(polIdx, frstIdx, farthestIdx, sqrTolerance);
      simplifyRDP(polIdx, farthestIdx, lastIdx, sqrTolerance);
    }

  else
    {
      polylines[polIdx][frstIdx].setNextIdx(lastIdx);
      polylines[polIdx][lastIdx].setPrevIdx(frstIdx);
    }
}



/**
 *
 *****************************************************************************/
void MapTCRDP::initConvexHulls(list<ConvexHull *> & cHulls)
{
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)

    for (unsigned vertexIdx = 0;
	 vertexIdx < polylines[polIdx].numberOfVertices()-1;
	 vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())
      {
	// Pushing convex hull in list of current polyline
	if (vertexIdx >= polylines[polIdx][vertexIdx].getNextIdx()-1)
	  continue;

	cHulls.
	  push_back(new ConvexHull(polylines[polIdx],vertexIdx,
				   polylines[polIdx][vertexIdx].getNextIdx()));
	ConvexHull & cHull = *cHulls.back();


	// Adding external points
	for (unsigned pointIdx = 0; pointIdx < numberOfPoints(); ++pointIdx)
	  if (cHull.isInside(points[pointIdx]))
	    cHull.pushFeature(points[pointIdx]);


	// Adding vertices of polylines
	for (unsigned pol2Idx = 0; pol2Idx < numberOfPolylines(); ++pol2Idx)
	  if (pol2Idx != polIdx)
	    {
	      for (unsigned vertex2Idx = 0; vertex2Idx <
		     polylines[pol2Idx].numberOfVertices(); ++vertex2Idx)
		if (cHull.isInside(polylines[pol2Idx][vertex2Idx]))
		  cHull.pushFeature(polylines[pol2Idx][vertex2Idx]);
	    }
	  else
	    {
	      for (unsigned vertex2Idx = 0; vertex2Idx <
		     cHull.getFrstIdx(); ++vertex2Idx)
		if (cHull.isInside(polylines[pol2Idx][vertex2Idx]))
		  cHull.pushFeature(polylines[pol2Idx][vertex2Idx]);

	      for (unsigned vertex2Idx= cHull.getLastIdx()+1; vertex2Idx <
		     polylines[pol2Idx].numberOfVertices(); ++vertex2Idx)
		if (cHull.isInside(polylines[pol2Idx][vertex2Idx]))
		  cHull.pushFeature(polylines[pol2Idx][vertex2Idx]);
	    }
      }
}



/**
 *
 *****************************************************************************/
void MapTCRDP::correct(list<ConvexHull *> & cHulls, cHullIter_t cHullIter,
		       double sqrTolerance)
{
  ConvexHull & cHull = **cHullIter;

  Polyline & polyline = const_cast<Polyline &>(cHull.getPolyline());

  unsigned frstIdx = cHull.getFrstIdx();
  unsigned lastIdx = cHull.getLastIdx();


  // Searching farthest vertex
  double   farthestSqrDistance = 0.0;
  unsigned farthestIdx         = frstIdx+1;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double vertexSqrDistance =
	polyline[vertexIdx].sqrDistance(polyline[frstIdx],
					polyline[lastIdx]);

      if (vertexSqrDistance > farthestSqrDistance)
	{
 	  farthestSqrDistance = vertexSqrDistance;
	  farthestIdx         = vertexIdx;
	}
    }


  if (farthestSqrDistance > sqrTolerance || cHull.getWrongSide())
    {
      // Correcting links of neighbouring vertices
      polyline[frstIdx].setNextIdx(farthestIdx);
      polyline[lastIdx].setPrevIdx(farthestIdx);


      // Correcting its links
      polyline[farthestIdx].setPrevIdx(frstIdx);
      polyline[farthestIdx].setNextIdx(lastIdx);


      // Setting first and last convex hulls
      ConvexHull * frstCHull = new ConvexHull(polyline);
      ConvexHull * scndCHull = new ConvexHull(polyline);
      cHull.split(farthestIdx, *frstCHull, *scndCHull);


      // Erasing current cru
      delete *cHullIter;
      cHulls.erase(cHullIter);


      // Recursive calls
      if (frstIdx < farthestIdx-1)
	{
	  cHulls.push_front(frstCHull);
	  correct(cHulls, cHulls.begin(), sqrTolerance);
	}
      else delete frstCHull;

      if (lastIdx > farthestIdx+1)
	{
	  cHulls.push_front(scndCHull);
	  correct(cHulls, cHulls.begin(), sqrTolerance);
	}
      else delete scndCHull;
    }

  else
    {
      delete *cHullIter;
      cHulls.erase(cHullIter);
    }
}



/**
 * @brieft Prints number of polylines, followed by polylines themselves.  Then,
 * prints number of points, followed by points themselves.
 *****************************************************************************/
void MapTCRDP::removeVertices()
{
  // Erasing vertices not in simplified polylines
  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      unsigned currIdx = 1;

      for (unsigned vertexIdx = polylines[polIdx].front().getNextIdx();
	   vertexIdx < polylines[polIdx].numberOfVertices()-1;
	   vertexIdx = polylines[polIdx][vertexIdx].getNextIdx())

	polylines[polIdx][currIdx++] = polylines[polIdx][vertexIdx];

      polylines[polIdx][currIdx] = polylines[polIdx].back();
      polylines[polIdx].resize(currIdx+1);
    }
}
