#ifndef __POINT_HH_
#define __POINT_HH_


#include <cmath>
#include <iostream>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;



/**
 * This class contains cartesian coordinates of a point in 2D space. It
 * presents geometric transformations and util functions usually applied to
 * points.
 *****************************************************************************/
class Point {

public:

  // Constructors
  Point();                    // Default
  Point(Point const & point); // Copy
  Point(double x, double y);  // Cartesien coordinates


  // Get coordinates functions
  double getX() const;
  double getY() const;

  // Indices access functions
  unsigned getPrevIdx() const;
  unsigned getNextIdx() const;


  // Set coordinates functions
  void setX(double x);
  void setY(double y);

  // Indices modification functions
  void setPrevIdx(unsigned prevIdx);
  void setNextIdx(unsigned nextIdx);


  /* Squared distances beween points and between a point and a segment */
  double sqrDistance(Point const & p) const;
  double sqrDistance(Point const & p, Point const & q) const;


  /* The cross product between 'pq' and 'p(*this)' */
  double crossProduct(Point const & p, Point const & q) const;

  /* Test for sidedness */
  bool atLeft (Point const & p, Point const & q) const;
  bool atRight(Point const & p, Point const & q) const;
  bool onLine (Point const & p, Point const & q) const;

  bool atLeftOrOnLine(Point const & p, Point const & q) const;
  bool atRghtOrOnLine(Point const & p, Point const & q) const;

  /* Test if point '(*this)' is inside triangle 'pqr' */
  bool isInTriangle(Point const & p, Point const & q, Point const & r) const;




  // Operators of equality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Friend operators for input and output handle
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);

  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;


private:

  // Cartesian coordinates
  double x;
  double y;

  // Indices of previous and next vertices in polyline
  unsigned prevIdx;
  unsigned nextIdx;
};


inline Point::Point() : x(0.0), y(0.0), prevIdx(0), nextIdx(0) {}
inline Point::Point(double x, double y) : x(x), y(y), prevIdx(0), nextIdx(0) {}
inline Point::Point(Point const & point)
  : x(point.x),
    y(point.y),

    prevIdx(point.prevIdx),
    nextIdx(point.nextIdx) {}

inline double Point::getX() const { return x; }
inline double Point::getY() const { return y; }

inline unsigned Point::getPrevIdx() const { return prevIdx; }
inline unsigned Point::getNextIdx() const { return nextIdx; }


inline void Point::setX(double x) { this->x = x; }
inline void Point::setY(double y) { this->y = y; }

inline void Point::setPrevIdx(unsigned prevIdx) { this->prevIdx = prevIdx; }
inline void Point::setNextIdx(unsigned nextIdx) { this->nextIdx = nextIdx; }


inline bool Point::operator!=(Point const & p) const { return !operator==(p); }
inline bool Point::operator==(Point const & p) const { return ((x == p.x) &&
							       (y == p.y)); }


inline double Point::sqrDistance(Point const & p) const
{ return (p.x-x)*(p.x-x)+(p.y-y)*(p.y-y); }


inline double Point::crossProduct(Point const & p, Point const & q) const
{ return (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y); }

inline bool Point::atLeft(Point const & p, Point const & q) const
{ return crossProduct(p, q) > 0; }

inline bool Point::atRight(Point const & p, Point const & q) const
{ return crossProduct(p, q) < 0; }

inline bool Point::onLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) == 0; }

inline bool Point::atLeftOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) >= 0; }

inline bool Point::atRghtOrOnLine(Point const & p, Point const & q) const
{ return crossProduct(p, q) <= 0; }

inline bool Point::isInTriangle(Point const & p,
				Point const & q,
				Point const & r) const
{
  return r.atLeft(p, q)
    ? atLeft (p, q) && atLeft (q, r) && atLeft (r, p)
    : atRight(p, q) && atRight(q, r) && atRight(r, p);
}



#endif // __POINT_HH_
