#include "point.hh"

#include <iomanip>




/**
 * @brief Returns the shortest distance from this point to a segment.
 *
 * Useful for the RDP algorithm.
 *
 * @param p First extreme of line segment.
 * @param q Second extreme of line segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
double Point::distance(Point const & p, Point const & q) const
{
  // 'p' or 'q' is the closest point to 'point'
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return distance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return distance(q);

  // 'point' is closest to a point between 'p' and 'q'
  double norm = p.distance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return fabs(determinant)/norm;
}



/**
 * @brief Returns the shortest distance from this point to a segment.
 *
 * The distance is returned squared.  Useful for the RDP algorithm.
 *
 * @param p First extreme of line segment.
 * @param q Second extreme of line segment.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
double Point::sqrDistance(Point const & p, Point const & q) const
{
  // 'p' or 'q' is the closest point to 'point'
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return sqrDistance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return sqrDistance(q);

  // 'point' is closest to a point between 'p' and 'q'
  double sqrNorm = p.sqrDistance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return determinant*determinant/sqrNorm;
}



/**
 * @brief Prints point in @a ostr.
 *
 * First, prints the coordinates x and y.
 *
 * @param ostr Output stream in where to printe.
 * @param width Width of the field coordinates occupy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Point::print(ostream & ostr, int width) const
{
  ostr << setw(width) << x << ", "
       << setw(width) << y << endl;
}



/**
 * @brief Writes @a point's coordinates in @a ostr.
 *
 * @param ostr Output stream in where to write.
 * @param point The point to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Point const & point)
{
  ostr << setw(COORD_WIDTH) << point.x
       << setw(COORD_WIDTH) << point.y;


  return ostr;
}



/**
 * @brief Reads @a point's coordinates from @a istr.
 *
 * @param istr Input stream from where to read.
 * @param point The point to store readed coordinates.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Point & point)
{
  istr >> point.x >> point.y;


  return istr;
}
