#ifndef __POINT_HH_
#define __POINT_HH_


#include <iostream>
#include <cmath>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;


/**
 * @brief This class implements a 2D point.
 * 
 * It stores its cartesian coordinates x (the absissa) and y (the ordinate).
 * This class also provides some geometric functions.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Point {

public:

  // Constructors
  Point(Point const & point);             // Copy
  Point(double x = 0.0, double y = 0.0);  // Cartesien coordinates


  // Get functions
  double getX() const;
  double getY() const;

  // Set functions
  void setX(double x);
  void setY(double y);


  // Operators of equality and inequality
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;


  // Distance and squared distance between points
  double    distance(Point const & point) const;
  double sqrDistance(Point const & point) const;

  // Distance and squared distance between point and segment
  double    distance(Point const & p, Point const & q) const;
  double sqrDistance(Point const & p, Point const & q) const;


  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;

  // Insertion and extractor operators for points
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);


protected:

  // Plane cartesian coordinates
  double x; /// x cartesian coordinate (the absissa)
  double y; /// y cartesian coordiante (the ordinate)
};



/**
 * @brief The copy constructor.
 *
 * It creates a new point from @a point.
 *
 * @param point The point to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point::Point(Point const & point) : x(point.x), y(point.y) {}


/**
 * @brief The xy-coordinates constructor.
 *
 * It creates a new point from the given coordinates @a x and @a y.
 *
 * @param x The absissa.
 * @param y The ordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point::Point(double x, double y) : x(x), y(y) {}


/**
 * @brief It returns the absissa @c x of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getX() const { return x; }


/**
 * @brief It returns the ordinate @c y of this point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::getY() const { return y; }


/**
 * @brief It sets the absissa @c x of this point from @a x.
 *
 * @param x The new absissa.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Point::setX(double x) { this->x = x; }


/**
 * @brief It sets the ordinate @c y of this point from @a y.
 *
 * @param y The new ordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Point::setY(double y) { this->y = y; }


/**
 * @brief It returns true if this point is different from @a point.
 *
 * It returns @c true if the at least one of the coordinates x and y of this
 * point is different from its correspondent coordinate of @a point, and @c
 * false, otherwise.
 *
 * @param point The point to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::operator!=(Point const & point) const
{ return !operator==(point); }


/**
 * @brief It returns true if this point is equal to @a point.
 *
 * It returns @c true if the the coordinates x and y of this point are equal
 * to their correspondent coordinates of @a point, and @c false, otherwise.
 *
 * @param point The point to be compared.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline bool Point::operator==(Point const & point) const
{ return ((x == point.x) && (y == point.y)); }


/**
 * @brief It computes the distance between this point and @a point.
 *
 * @param point The point to compute the distance.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::distance(Point const & point) const
{ return sqrt((point.x-x)*(point.x-x)+(point.y-y)*(point.y-y)); }


/**
 * @brief It computes the squared distance between this point and @a point.
 *
 * @param point The point to compute the distance.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Point::sqrDistance(Point const & point) const
{ return (point.x-x)*(point.x-x)+(point.y-y)*(point.y-y); }



#endif // __POINT_HH_
