#include "polyline.hh"

#include <iomanip>




/**
 * @brief Applies the RDP algorithm to part of this polyline in scaleless mode.
 *
 * The simplification is carried out in the range [@a frstIdx, @a lastIdx]
 * of this polyline.  The squared distance @a parentSqrDistance is used as a
 * maximum bound to the current node distance.
 *
 * @param parentSqrDistance The squared distance of parent vertex.
 * @param frstIdx The index of the first vertex in the range.
 * @param latsIdx The index of the last  vertex in the range.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Polyline::simplifySLRDP(double parentSqrDistance,
			     unsigned frstIdx, unsigned lastIdx)
{
  double   maximumSqrDistance = -DBL_MAX;
  unsigned farthestIdx = frstIdx;

  for (unsigned vertexIdx = frstIdx+1; vertexIdx < lastIdx; ++vertexIdx)
    {
      double currentSqrDistance =
	(*this)[vertexIdx].sqrDistance((*this)[frstIdx], (*this)[lastIdx]);

      if (currentSqrDistance > maximumSqrDistance)
	{
	  maximumSqrDistance = currentSqrDistance;
	  farthestIdx        = vertexIdx;
	}
    }


  // Computing and setting tolerance of farthest vertex
  double sqrDistance = (parentSqrDistance < maximumSqrDistance)
    ? parentSqrDistance : maximumSqrDistance;

  (*this)[farthestIdx].setDistance(sqrt(sqrDistance));


  // Recursive function calls
  if (farthestIdx > frstIdx+1)
    (*this).simplifySLRDP(sqrDistance, frstIdx, farthestIdx);

  if (farthestIdx < lastIdx-1)
    (*this).simplifySLRDP(sqrDistance, farthestIdx, lastIdx);
}



/**
 * @brief Prints subpolyline of this polyline in @c ostr.
 *
 * First, prints the number of vertices of the subpolyline and the range.
 * Then, prints vertices of the subpolyline one per line, prefixed with its
 * index.
 *
 * @param ostr Output stream in where the polyline is printed.
 * @param frstIdx First index of subpolyline's range.
 * @param lastIdx Last  index of subpolyline's range.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setfill('0') << setw(3) << vertexIdx << ") " << setfill(' ');
      (*this)[vertexIdx].print(ostr);
    }
}



/**
 * @brief Writes @c polyline in @c ostr.
 *
 * First, writes the number of vertices.  Then, writes the vertices one per
 * line.  Is a redefinition of the insertion operator in the context of
 * streams.
 *
 * @param ostr Output stream in where the polyline is writed.
 * @param polyline Polyline to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;


  return ostr;
}



/**
 * @brief Reads @c polyline from @c ostr.
 *
 * First, reads number of vertices.  Then, reads vertices in the specified
 * order.  Is a redefinition of the extraction operator in the context of
 * streams.
 *
 * @param ostr Input stream from where the polyline is printed.
 * @param polyline Polyline to print.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    istr >> polyline[vertexIdx];


  return istr;
}
