#ifndef __VERTEX_HH_
#define __VERTEX_HH_


#include <iostream>

#include "point.hh"


using namespace std;


/**
 * @brief This class implements a 2D vertex of a polyline.
 * 
 * This class is a subclass of the class @c Point.  Besides its coordinates,
 * it stores a flag that indicates if this vertex has been erased from the
 * polyline it belongs to.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007 
 *****************************************************************************/
class Vertex
  : public Point {

public:

  // Constructors
  Vertex(Vertex const & vertex);           // Copy
  Vertex(Point  const & point);            // Point
  Vertex(double x = 0.0, double y = 0.0);  // Cartesien coordinates


  // Get and set functions
  double getDistance() const;
  void   setDistance(double distance);


  // Print function
  void print(ostream & ostr, int width = COORD_WIDTH) const;

  // Extractor operator for vertices
  friend ostream & operator<<(ostream & ostr, Vertex const & vertex);


private:

  // Distance of this vertex
  double distance; /// Distance to label this vertex in scaleless mode
};



/**
 * @brief The copy constructor.
 *
 * It creates a new vertex from @a vertex.
 *
 * @param vertex The vertex to copy.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Vertex const & vertex)
  : Point(vertex), distance(vertex.distance) {}


/**
 * @brief The point constructor.
 *
 * It creates a new vertex from @a point.
 *
 * @param point The point to copy coordinates.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(Point const & point)
  : Point(point), distance(0.0) {}


/**
 * @brief The xy-coordinate constructor.
 *
 * It creates a new vertex with given coordinates.
 *
 * @param x The @a x coordinate.
 * @param y The @a y coordinate.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Vertex::Vertex(double x, double y) : Point(x, y), distance(0.0) {}


/**
 * @brief It returns the distance of this vertex.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline double Vertex::getDistance() const { return distance; }


/**
 * @brief It sets the distance of this vertex from @a distance.
 *
 * @param distance The new flag.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline void Vertex::setDistance(double distance) { this->distance = distance; }



#endif // __POINT_HH_
