#include "map.hh"

#include <iomanip>




/**
 * @brief It applies Visvalingam's algorithm in scaleless mode to the
 * polylines of this map.
 *
 * All the vertices of this polyline are labeled with its effective area
 * that can be used later to filter these vertices in a real-time
 * simplification.
 *
 * @warning Topology of the map can be lost, because the polylines are
 * simplified in isolation.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::simplifySLVIS()
{
  // Simplifying polylines of map
  for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
    polylines[polIdx].simplifySLVIS();
}



/**
 * @brief It writes this map in @c ostr.
 *
 * First, it writes the number of polylines, followed by the polylines.
 * Then, it writes the number of points, followed by the features points.
 * Uses the Polyline::print and Point::print.
 *
 * @param ostr The output stream.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
void Map::print(ostream & ostr) const
{
  // Setting needed flags
  ostr.flags(ios::scientific|ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < numberOfPolylines(); ++polIdx)
    {
      ostr << endl;
      polylines[polIdx].print(ostr);
    }


  // Printing points
  ostr << endl << endl << numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < numberOfPoints(); ++pointIdx)
    {
      points[pointIdx].print(ostr);
      ostr << endl;
    }
}



/**
 * @brief It writes a map in an output stream.
 *
 * First, it writes the number of polylines, followed by the polylines.
 * Then, it writes the number of points, followed by the points.  It is a
 * redefinition of the insertion operator in the context of streams.
 *
 * @param ostr The output stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Map const & map)
{
  // Setting needed flags
  ostr.flags(ios::scientific|ios::uppercase);
  ostr.precision(COORD_PRECISION);


  // Printing polylines
  ostr << map.numberOfPolylines() << endl;

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    ostr << endl << map.polylines[polIdx];


  // Printing points
  ostr << endl << endl << map.numberOfPoints() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    ostr << map.points[pointIdx] << endl;


  return ostr;
}



/**
 * @brief It reads a map from an input stream.
 *
 * First, it reads the number of polylines, followed by the polylines.
 * Then, it reads the number of points, followed by the points.  It is a
 * redefinition of the extraction operator in the context of streams.
 *
 * @param istr The input stream.
 * @param map The map to write.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
istream & operator>>(istream & istr, Map & map)
{
  // Reading polylines
  unsigned nPolylines;

  istr >> nPolylines;
  map.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < map.numberOfPolylines(); ++polIdx)
    istr >> map.polylines[polIdx];


  // Reading points
  unsigned nPoints;

  istr >> nPoints;
  map.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < map.numberOfPoints(); ++pointIdx)
    istr >> map.points[pointIdx];


  return istr;
}
