#ifndef __MAP_HH_
#define __MAP_HH_


#include "polyline.hh"
#include "point.hh"

#include <iostream>
#include <vector>


using namespace std;



/**
 * @brief This class implements a planar cartographic map.
 * 
 * It stores a vector of polylines and a vector of points.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
class Map {

public:

  // Polyline's number of vertices function
  unsigned numberOfPolylines() const;
  unsigned numberOfPoints() const;

  // Access and modification functions
  Polyline & getPolyline(unsigned polylineIdx);
  Point &    getPoint   (unsigned pointIdx);


  // Visvalingam's line simplification procedure in scaleless mode
  void simplifySLVIS();


  // Friendship relation of insertion and extraction operators
  friend ostream & operator<<(ostream & ostr, Map const & map);
  friend istream & operator>>(istream & istr, Map       & map);

  // Printing function
  void print(ostream & ostr) const;


protected:

  // Internal variables
  vector<Polyline> polylines;  /// Vector of polylines.
  vector<Point>    points;     /// Vecotr of points.
};



/**
 * @brief Returns the number of polylines of this map.
 *
 * Uses vector<Polyline>::size().
 *
 * @return Number of polylines.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Map::numberOfPolylines() const { return polylines.size(); }


/**
 * @brief Returns the number of points of this map.
 *
 * Uses vector<Point>::size().
 *
 * @return Number of points.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline unsigned Map::numberOfPoints() const { return points.size(); }


/**
 * @brief Returns the @a idx'th polyline of this map.
 *
 * @return The @a idx'th polyline
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Polyline & Map::getPolyline(unsigned idx) { return polylines[idx]; }


/**
 * @brief Returns the @a idx'th point of this map.
 *
 * @return The @a idx'th point.
 *
 * @author Adler da Silva
 * @date   Last update: 30/07/2007
 *****************************************************************************/
inline Point & Map::getPoint(unsigned idx) { return points[idx]; }



#endif // __MAP_HH_
