#include <iostream>
#include <fstream>
#include <iomanip>

#include <algorithm>
#include <string>
#include <vector>
#include <list>


using namespace std;



// Point structured type - 2D cartesian coordinates
struct point_t
{
  double x;
  double y;
};


// Polyline type - vector of points
typedef vector<point_t> polyline_t;


// Arc index type - index in polylines vector
struct arcIdx_t
{
  int arcIdx;
  bool reversed;
};


// Polygon structured type
typedef vector<arcIdx_t> polygon_t;


// ARC header structured type - connectivity data
struct arcHeader_t
{
  int coverageNumber;
  int coverageID;

  int frNode;
  int toNode;

  int leftPolygon;
  int rghtPolygon;
};


// ARC structured type - header + polyline
struct arc_t
{
  arcHeader_t header;
  polyline_t  polyline;

  arc_t(arcHeader_t const & header) : header(header) {}
};


// CNT (centroid) structured type - labels + centroid coordinates
struct cnt_t
{
  int nLabels;
  point_t centroid;

  vector<int> labels;

  cnt_t(int nLabels) : nLabels(nLabels) {}
};


// LAB (label) header structured type -  IDs + position
struct labHeader_t
{
  int labelID;
  int polygonID;

  point_t position;
};


// LAB (label) structured type - header + bounding box
struct lab_t 
{
  labHeader_t header;

  point_t upLeftPoint;
  point_t dnRghtPoint;

  lab_t(labHeader_t const & header) : header(header) {}
};


istream & operator>>(istream & istr, point_t     & point);
istream & operator>>(istream & istr, polyline_t  & polyline);
istream & operator>>(istream & istr, arcHeader_t & header);
istream & operator>>(istream & istr, arc_t       & arc);
istream & operator>>(istream & istr, labHeader_t & header);

ostream & operator<<(ostream & ostr, point_t     const & point);
ostream & operator<<(ostream & ostr, polyline_t  const & polyline);
ostream & operator<<(ostream & ostr, polygon_t   const & polygon);
ostream & operator<<(ostream & ostr, arcHeader_t const & header);
ostream & operator<<(ostream & ostr, arc_t       const & arc);


void eliminateRedudances(list<arc_t> & arcs);
void generatePolygons   (list<arc_t> const & arcs, list<polygon_t> & polygons);



int main(int argc, char * argv[])
{
  // Reading files' names
  string e00FileName;
  string infFileName;

  if (argc >= 2) e00FileName = argv[1];
  if (argc >= 3) infFileName = argv[2];


  // Streams
  istream * e00Stream = &cin;
  ostream * infStream = &cout;


  // File streams
  ifstream e00FileStream;
  ofstream infFileStream;


  // Testing for openning e00 file
  if (!e00FileName.empty())
    {
      e00FileStream.open(e00FileName.c_str());
      if (!e00FileStream.is_open())
	{
	  cerr << argv[0] << ": could not open e00 file "
	       << e00FileName << endl;
	  exit(-1);
	}

      e00Stream = static_cast<istream *>(&e00FileStream);
    }

  // Testing for openning inf file
  if (!infFileName.empty())
    {
      infFileStream.open(infFileName.c_str());
      if (!infFileStream.is_open())
	{
	  cerr << argv[0] << ": could not open inf file "
	       << infFileName << endl;
	  exit(-1);
	}

      infStream = static_cast<ostream *>(&infFileStream);
    }


  /////////////////////////////////////////////////////////////////////////////
  list<arc_t> arcs;
  list<cnt_t> centroids;
  list<lab_t> labels;
  list<polygon_t> polygons;


  string section;
  *e00Stream >> section;


  // Reading EXP section
  if (section == "EXP")
    {
      bool compressed;
      *e00Stream >> compressed;

      string fileName;
      *e00Stream >> fileName;

      *e00Stream >> section;
    }


  // Reading ARC section
  if (section == "ARC")
    {
      int precision;
      *e00Stream >> precision;

      arcHeader_t header;
      while (*e00Stream >> header, header.coverageNumber != -1)
	{
	  arcs.push_back(arc_t(header));
	  *e00Stream >> arcs.back().polyline;
	}

      *e00Stream >> section;
    }


  // Reading CNT section
  if (section == "CNT")
    {
      int precision;
      *e00Stream >> precision;

      int nLabels;
      while (*e00Stream >> nLabels, nLabels != -1)
	{
	  centroids.push_back(cnt_t(nLabels));
	  *e00Stream >> centroids.back().centroid;

	  for (; nLabels > 0; --nLabels)
	    {
	      centroids.back().labels.push_back(0);
	      *e00Stream >> centroids.back().labels.back();
	    }
	}

      *e00Stream >> section;
    }


  // Reading LAB section
  if (section == "LAB")
    {
      int precision;
      *e00Stream >> precision;

      labHeader_t header;
      while (*e00Stream >> header, header.labelID != -1)
	{
	  labels.push_back(lab_t(header));
	  *e00Stream >> labels.back().upLeftPoint
		     >> labels.back().dnRghtPoint;
	}
    }


  /////////////////////////////////////////////////////////////////////////////
  // Setting needed flags
  infStream->setf(ios::scientific, ios::floatfield);
  infStream->setf(ios::uppercase);
  infStream->precision(7);


  // Printing polylines
  *infStream << arcs.size() << endl;
  for (list<arc_t>::iterator arcIter = arcs.begin();
       arcIter != arcs.end(); ++arcIter)
    *infStream << endl << arcIter->polyline;


//    // Printing polygons
//   *infStream << endl << endl << polygons.size() << endl;
//   for (list<polygon_t>::iterator polygonIter = polygons.begin();
//        polygonIter != polygons.end(); ++polygonIter)
//     *infStream << endl << *polygonIter;


  // Printing feature points
  *infStream << endl << endl << labels.size() << endl;
  for (list<lab_t>::iterator labelIter = labels.begin();
       labelIter != labels.end(); ++labelIter)
    *infStream << endl << labelIter->header.position;
  *infStream << endl;


  return 0;
}



istream & operator>>(istream & istr, point_t & point)
{
  return istr >> point.x >> point.y;
}

istream & operator>>(istream & istr, polyline_t & polyline)
{
  unsigned nVertices;
  istr >> nVertices;

  polyline.resize(nVertices);
  for (unsigned vertexIdx = 0; vertexIdx < nVertices; ++vertexIdx)
    istr >> polyline[vertexIdx];

  return istr;
}

istream & operator>>(istream & istr, arcHeader_t & header)
{
  return istr >> header.coverageNumber
	      >> header.coverageID
	      >> header.frNode
	      >> header.toNode
	      >> header.leftPolygon
	      >> header.rghtPolygon;
}

istream & operator>>(istream & istr, arc_t & arc)
{
  return istr >> arc.header
	      >> arc.polyline;
}

istream & operator>>(istream & istr, labHeader_t & header)
{
  return istr >> header.labelID
	      >> header.polygonID
	      >> header.position;
}

ostream & operator<<(ostream & ostr, point_t const & point)
{
  return ostr << scientific << uppercase << setprecision(7)
	      << setw(14) << point.x
	      << setw(14) << point.y;
}

ostream & operator<<(ostream & ostr, polyline_t const & polyline)
{
  ostr << polyline.size() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;

  return ostr;
}

ostream & operator<<(ostream & ostr, polygon_t const & polygon)
{
  ostr << polygon.size() << endl;

  for (unsigned polylineIdx = 0; polylineIdx < polygon.size(); ++polylineIdx)
    ostr << polygon[polylineIdx].arcIdx << " "
	 << polygon[polylineIdx].reversed << endl;

  return ostr;
}

ostream & operator<<(ostream & ostr, arcHeader_t const & header)
{
  return ostr << "     " << header.coverageNumber
	      << "     " << header.coverageID
	      << "     " << header.frNode
	      << "     " << header.toNode
	      << "     " << header.leftPolygon
	      << "     " << header.rghtPolygon;
}

ostream & operator<<(ostream & ostr, arc_t const & arc)
{
  return ostr << arc.header << "     "
	      << arc.polyline;
}


struct node_t
{
  int nodeIdx;
  unsigned nOccurrences;

  node_t(int nodeIdx) : nodeIdx(nodeIdx), nOccurrences(0) {}
};


bool operator==(node_t const & node, int nodeIdx)
{
  return node.nodeIdx == nodeIdx;
}

bool operator!=(node_t const & node, int nodeIdx)
{
  return !operator==(node, nodeIdx);
}

bool operator==(int nodeIdx, node_t const & node) { return node == nodeIdx; }
bool operator!=(int nodeIdx, node_t const & node) { return node != nodeIdx; }



void eliminateRedudances(list<arc_t> & arcs)
{
  // Creating nodes list
  list<node_t> nodes;

  for (list<arc_t>::iterator arcIter = arcs.begin();
       arcIter != arcs.end(); ++arcIter)
    {
      // Finding from node
      list<node_t>::iterator frNodeIter = find(nodes.begin(), nodes.end(),
					       arcIter->header.frNode);
      if (frNodeIter == nodes.end())
	{
	  nodes.push_back(arcIter->header.frNode);
	  frNodeIter = --nodes.end();
	}

      // Finding to node
      list<node_t>::iterator toNodeIter = find(nodes.begin(), nodes.end(),
					       arcIter->header.toNode);
      if (toNodeIter == nodes.end())
	{
	  nodes.push_back(arcIter->header.toNode);
	  toNodeIter = --nodes.end();
	}

      // Updating nodes
      ++frNodeIter->nOccurrences;
      ++toNodeIter->nOccurrences;

      if (frNodeIter == toNodeIter)
	++frNodeIter->nOccurrences;
    }


  // Eliminating useless nodes
  for (list<node_t>::iterator nodeIter = nodes.begin();
       nodeIter != nodes.end(); ++nodeIter)

    if (nodeIter->nOccurrences == 2)
      {
	// Searching first arc
	list<arc_t>::iterator arc1Iter = arcs.begin();
	for (; arc1Iter != arcs.end(); ++arc1Iter)
	  if (arc1Iter->header.frNode == nodeIter->nodeIdx ||
	      arc1Iter->header.toNode == nodeIter->nodeIdx) break;
	if (arc1Iter == arcs.end()) break;

	// Searching second arc
	list<arc_t>::iterator arc2Iter = arc1Iter;
	for (++arc2Iter; arc2Iter != arcs.end(); ++arc2Iter)
	  if (arc2Iter->header.frNode == nodeIter->nodeIdx ||
	      arc2Iter->header.toNode == nodeIter->nodeIdx) break;
	if (arc2Iter == arcs.end()) break;


	// Inserting vertices
	if (arc1Iter->header.toNode == arc2Iter->header.frNode)
	  {
	    arc1Iter->polyline.insert(arc1Iter->polyline.end(),
				      arc2Iter->polyline.begin(),
				      arc2Iter->polyline.end());
	    arc1Iter->header.toNode = arc2Iter->header.toNode;
	  }

	else if (arc1Iter->header.toNode == arc2Iter->header.toNode)
	  {
	    arc1Iter->polyline.insert(arc1Iter->polyline.end(),
				      arc2Iter->polyline.rbegin(),
				      arc2Iter->polyline.rend());
	    arc1Iter->header.toNode = arc2Iter->header.frNode;
	  }

	else if (arc1Iter->header.frNode == arc2Iter->header.frNode)
	  {
	    arc1Iter->polyline.insert(arc1Iter->polyline.begin(),
				      arc2Iter->polyline.rbegin(),
				      arc2Iter->polyline.rend());
	    arc1Iter->header.frNode = arc2Iter->header.toNode;
	  }

	else if (arc1Iter->header.frNode == arc2Iter->header.toNode)
	  {
	    arc1Iter->polyline.insert(arc1Iter->polyline.begin(),
				      arc2Iter->polyline.begin(),
				      arc2Iter->polyline.end());
	    arc1Iter->header.frNode = arc2Iter->header.frNode;
	  }

	arcs.erase(arc2Iter);
      }
}


void generatePolygons(list<arc_t> const & arcs, list<polygon_t> & polygons)
{
}
