#include <iostream>
#include <fstream>
#include <iomanip>

#include <string>
#include <vector>
#include <list>

#include <cmath>
#include <cstdlib>
#include <values.h>


using namespace std;



// Structures and print functions
struct figHeader_t
{
  string version;          // Name and version
  string orientation;      // Landscape or Portrait
  string justification;    // Center or Flush Left
  string units;            // Metric or Inches
  string papersize;        // Letter, Legal, Ledger, Tabloid, A, B, C, D, E,
                           // A4, A3, A2, A1, A0, and B5
  float  magnification;    // Export and print magnification, %
  string multiplePage;     // Single or Multiple
  int    transparentColor; // Color number for transparent color for GIF export
			   // -3=background, -2=None, -1=Default,
			   // 0-31 for standard colors, 32- for user colors
  int    resolution;       // Fig units/inch
  int    coordSystem;      // Cordinate system:
			   //   1: origin at lower left corner (NOT USED)
			   //   2: upper left
};


istream & operator>>(istream & istr, figHeader_t & figHeader)
{
  getline(istr, figHeader.version);
  getline(istr, figHeader.orientation);
  getline(istr, figHeader.justification);
  getline(istr, figHeader.units);
  getline(istr, figHeader.papersize);

  return istr >> figHeader.magnification
	      >> figHeader.multiplePage
	      >> figHeader.transparentColor
	      >> figHeader.resolution
	      >> figHeader.coordSystem;
}


struct point_t
{
  point_t() : x(0), y(0) {}
  point_t(int x, int y) : x(x), y(y) {}

  int x;
  int y;
};


struct ellipse_t
{
  int   objectCode;       // Always 1
  int   subType;          // 1: radii, 2: diameters, 3: radius, 4: diameter
  int   lineStyle;        // Enumeration type
  int   thickness;        // 1/80 inch
  int   penColor;         // Enumeration type, pen color
  int   fillColor;        // Enumeration type, fill color
  int   depth;            // Enumeration type
  int   penStyle;         // Pen style, not used
  int   areaFill;         // Enumeration type, -1 = no fill
  float styleVal;         // 1/80 inch
  int   direction;        // Always 1
  float angle;            // Radians, the andle of the x-axis
  int   centerX, centerY; // Fig units
  int   radiusX, radiusY; // Fig units
  int    startX,  startY; // Fig units
  int      endX,    endY; // Fig units

  ellipse_t(int objectCode = 1) : objectCode(objectCode) {}
};


istream & operator>>(istream & istr, ellipse_t & ellipse)
{
  istr >> ellipse.subType
       >> ellipse.lineStyle
       >> ellipse.thickness
       >> ellipse.penColor
       >> ellipse.fillColor
       >> ellipse.depth
       >> ellipse.penStyle
       >> ellipse.areaFill
       >> ellipse.styleVal
       >> ellipse.direction
       >> ellipse.angle
       >> ellipse.centerX >> ellipse.centerY
       >> ellipse.radiusX >> ellipse.radiusY
       >> ellipse. startX >> ellipse. startY
       >> ellipse.   endX >> ellipse.   endY;

  return istr;
}


struct polyline_t
  : public vector<point_t>
{
  int   objectCode;    // Always 2
  int   subType;       // 1: line, 2: box, 3: pgon, 4: arcbox, 5: picture bb
  int   lineStyle;     // Enumeration type
  int   thickness;     // 1/80 inch
  int   penColor;      // Enumeration type, pen color
  int   fillColor;     // Enumeration type, fill color
  int   depth;         // Enumeration type
  int   penStyle;      // Pen style, not used
  int   areaFill;      // Enumeration type, -1 = no fill
  float styleVal;      // 1/80 inch
  int   joinStyle;     // Enumeration type
  int   capStyle;      // Enumeration type, only used for POLYLINE
  int   radius;        // 1/80 inch, radius of arc-boxes
  bool  forwardArrow;  // 0: off, 1: on
  bool  backwardArrow; // 0: off, 1: on
  int   nPoints;       // Number of points in line

  polyline_t(int objectCode = 2) : objectCode(objectCode) {}
};


istream & operator>>(istream & istr, polyline_t & polyline)
{
  istr >> polyline.subType
       >> polyline.lineStyle
       >> polyline.thickness
       >> polyline.penColor
       >> polyline.fillColor
       >> polyline.depth
       >> polyline.penStyle
       >> polyline.areaFill
       >> polyline.styleVal
       >> polyline.joinStyle
       >> polyline.capStyle
       >> polyline.radius
       >> polyline.forwardArrow
       >> polyline.backwardArrow
       >> polyline.nPoints;

  polyline.resize(polyline.nPoints);

  for (unsigned pointIdx = 0; pointIdx < polyline.size(); ++pointIdx)
    istr >> polyline[pointIdx].x
	 >> polyline[pointIdx].y;


  return istr;
}




int main(int argc, char *argv[])
{
  // Testing parameters for ratio
  if (argc < 2)
    {
      cerr << "Usage: fig2inf <ratio> [fig file] [inf file]" << endl;
      exit(-1);
    }


  // Getting ratio of scale
  double ratio = atof(argv[1]);


  // Reading files' names
  string figFileName;
  string infFileName;

  if (argc >= 3) figFileName = argv[2];
  if (argc >= 4) infFileName = argv[3];


  // Streams
  istream * figStream = &cin;
  ostream * infStream = &cout;


  // Input file stream
  ifstream figFileStream;
  ofstream infFileStream;


  // Testing for openning fig file
  if (!figFileName.empty())
    {
      figFileStream.open(figFileName.c_str());
      if (!figFileStream.is_open())
	{
	  cerr << argv[0] << ": could not open fig file "
	       << figFileName << endl;
	  exit(-1);
	}

      figStream = static_cast<istream *>(&figFileStream);
    }

  // Testing for openning inf file
  if (!infFileName.empty())
    {
      infFileStream.open(infFileName.c_str());
      if (!infFileStream.is_open())
	{
	  cerr << argv[0] << ": could not open inf file "
	       << infFileName << endl;
	  exit(-1);
	}

      infStream = static_cast<ostream *>(&infFileStream);
    }


  /////////////////////////////////////////////////////////////////////////////
  // Reading fig header
  figHeader_t figHeader;
  *figStream >> figHeader;


  // Reading polylines and points
  list<polyline_t> polylines;
  list<point_t>    points;
  int objectCode;

  while (*figStream >> objectCode)
    if (objectCode == 0)
      {
	int  colorNumber;
	char colorCode[8];

	*figStream >> colorNumber >> colorCode;
      }

    else if (objectCode == 1)
      {
	ellipse_t ellipse(objectCode);
	*figStream >> ellipse;

	points.push_back(point_t(ellipse.centerX,
				 ellipse.centerY));
      }

    else if (objectCode == 2)
      {
	polyline_t polyline(objectCode);
	*figStream >> polyline;

	switch (polyline.subType)
	  {
	  case 1:
	    polylines.push_back(polyline);
	    break;

	  case 2:
	    points.push_back(point_t((polyline[0].x+polyline[2].x)/2,
				     (polyline[0].y+polyline[2].y)/2));
	    break;
	  }
      }

    else 
      {
	cerr << polylines.size() << endl
	     << points.   size() << endl;
	cerr << "fig2inf: object code not 0/1/2" << endl;
	exit(-1);
      }


  /////////////////////////////////////////////////////////////////////////////
  // Setting needed flags
  infStream->setf(ios::scientific, ios::floatfield);
  infStream->setf(ios::uppercase);
  infStream->precision(7);


  // Writing polylines
  *infStream << polylines.size() << endl;
  for (list<polyline_t>::iterator polylineIter = polylines.begin();
       polylineIter != polylines.end(); ++polylineIter)
    {
      *infStream << endl << polylineIter->size() << endl;
      for (unsigned pointIdx= 0; pointIdx < polylineIter->size(); ++pointIdx)
	*infStream << setw(14) <<  (*polylineIter)[pointIdx].x/ratio
	     << setw(14) << -(*polylineIter)[pointIdx].y/ratio <<endl;
    }


  // Writing points
  *infStream << endl << endl << points.size() << endl << endl;
  for (list<point_t>::iterator pointIter = points.begin();
       pointIter != points.end(); ++pointIter)
    *infStream << setw(14) <<  pointIter->x/ratio
	       << setw(14) << -pointIter->y/ratio << endl;


  return 0;
}
