#ifndef __MMFFILE_HH_
#define __MMFFILE_HH_


#include <mmf/mmfpollayer.hh>
#include <mmf/mmfpoplayer.hh>
#include <mmf/mmfdrglayer.hh>
#include <mmf/mmfrodlayer.hh>
#include <mmf/mmfrarlayer.hh>
#include <mmf/mmfaptlayer.hh>
#include <mlf/mlffile.hh>

#include <iostream>
#include <string>
#include <vector>


using namespace std;



/******************************************************************************
 *
 *****************************************************************************/
class MMFFile {

public:

  /* Constructor */
  MMFFile();
  MMFFile(string const & name, unsigned nLayers);
  MMFFile(MLFFile const & mlfFile);

  /* Destructor */
  ~MMFFile();

  /* Get functions */
  string const & getName() const;
  unsigned       getNLayers() const;

  unsigned numberOfResolutions() const;
  double   getTolerance(unsigned resolutionIdx) const;
    

  MMFPOLLayer * getPOLLayer();
  MMFPOPLayer * getPOPLayer();
  MMFDRGLayer * getDRGLayer();
  MMFRODLayer * getRODLayer();
  MMFRARLayer * getRARLayer();
  MMFAPTLayer * getAPTLayer();

  MMFPOLLayer const * getPOLLayer() const;
  MMFPOPLayer const * getPOPLayer() const;
  MMFDRGLayer const * getDRGLayer() const;
  MMFRODLayer const * getRODLayer() const;
  MMFRARLayer const * getRARLayer() const;
  MMFAPTLayer const * getAPTLayer() const;

  /* Set functions */
  void setName(string const & name);
  void setNLayers(unsigned nLayers);

  void setNumberOfResolutions(unsigned nResolutions);
  void setTolerance(unsigned resolutionIdx, double tolerance);

  void setPOLLayer(MMFPOLLayer * polLayer);
  void setPOPLayer(MMFPOPLayer * popLayer);
  void setDRGLayer(MMFDRGLayer * drgLayer);
  void setRODLayer(MMFRODLayer * rodLayer);
  void setRARLayer(MMFRARLayer * rarLayer);
  void setAPTLayer(MMFAPTLayer * aptLayer);

  /* Insertion and extraction overloaded operators */
  friend istream & operator>>(istream & istr, MMFFile       & mmfFile);
  friend ostream & operator<<(ostream & ostr, MMFFile const & mmfFile);


private:

  string name;
  unsigned nLayers;
  vector<double> tolerances;

  MMFPOLLayer * polLayer;
  MMFPOPLayer * popLayer;
  MMFDRGLayer * drgLayer;
  MMFRODLayer * rodLayer;
  MMFRARLayer * rarLayer;
  MMFAPTLayer * aptLayer;
};


inline MMFFile::MMFFile()
  : polLayer(0), popLayer(0), drgLayer(0), rodLayer(0), rarLayer(0),
    aptLayer(0) {}

inline MMFFile::MMFFile(string const & name, unsigned nLayers)
  : name(name), nLayers(nLayers), polLayer(0), popLayer(0), drgLayer(0),
    rodLayer(0), rarLayer(0), aptLayer(0) {}

inline MMFFile::MMFFile(MLFFile const & mlfFile)
  : name(mlfFile.getName()), nLayers(mlfFile.getNLayers())
{
  polLayer = (mlfFile.getPOLLayer())
    ? new MMFPOLLayer(*mlfFile.getPOLLayer()) : 0;

  popLayer = (mlfFile.getPOPLayer())
    ? new MMFPOPLayer(*mlfFile.getPOPLayer()) : 0;

  drgLayer = (mlfFile.getDRGLayer())
    ? new MMFDRGLayer(*mlfFile.getDRGLayer()) : 0;

  rodLayer = (mlfFile.getRODLayer())
    ? new MMFRODLayer(*mlfFile.getRODLayer()) : 0;

  rarLayer = (mlfFile.getRARLayer())
    ? new MMFRARLayer(*mlfFile.getRARLayer()) : 0;

  aptLayer = (mlfFile.getAPTLayer())
    ? new MMFAPTLayer(*mlfFile.getAPTLayer()) : 0;
}

inline MMFFile::~MMFFile()
{
  if (polLayer) delete polLayer;
  if (popLayer) delete popLayer;
  if (drgLayer) delete drgLayer;
  if (rodLayer) delete rodLayer;
  if (rarLayer) delete rarLayer;
  if (aptLayer) delete aptLayer;
}

inline string const & MMFFile::getName() const    { return name; }
inline unsigned       MMFFile::getNLayers() const { return nLayers; }

inline unsigned MMFFile::numberOfResolutions() const
{
  return tolerances.size();
}
inline double MMFFile::getTolerance(unsigned resolutionIdx) const
{
  return tolerances[resolutionIdx];
}


inline MMFPOLLayer * MMFFile::getPOLLayer() { return polLayer; }
inline MMFPOPLayer * MMFFile::getPOPLayer() { return popLayer; }
inline MMFDRGLayer * MMFFile::getDRGLayer() { return drgLayer; }
inline MMFRODLayer * MMFFile::getRODLayer() { return rodLayer; }
inline MMFRARLayer * MMFFile::getRARLayer() { return rarLayer; }
inline MMFAPTLayer * MMFFile::getAPTLayer() { return aptLayer; }

inline MMFPOLLayer const * MMFFile::getPOLLayer() const { return polLayer; }
inline MMFPOPLayer const * MMFFile::getPOPLayer() const { return popLayer; }
inline MMFDRGLayer const * MMFFile::getDRGLayer() const { return drgLayer; }
inline MMFRODLayer const * MMFFile::getRODLayer() const { return rodLayer; }
inline MMFRARLayer const * MMFFile::getRARLayer() const { return rarLayer; }
inline MMFAPTLayer const * MMFFile::getAPTLayer() const { return aptLayer; }

inline void MMFFile::setName(string const & name) { this->name    = name; }
inline void MMFFile::setNLayers(unsigned nLayers) { this->nLayers = nLayers; }

inline void MMFFile::setTolerance(unsigned resolutionIdx, double tolerance)
{
  tolerances[resolutionIdx] = tolerance;
}

inline void MMFFile::setPOLLayer(MMFPOLLayer * layer) { polLayer = layer; }
inline void MMFFile::setPOPLayer(MMFPOPLayer * layer) { popLayer = layer; }
inline void MMFFile::setDRGLayer(MMFDRGLayer * layer) { drgLayer = layer; }
inline void MMFFile::setRODLayer(MMFRODLayer * layer) { rodLayer = layer; }
inline void MMFFile::setRARLayer(MMFRARLayer * layer) { rarLayer = layer; }
inline void MMFFile::setAPTLayer(MMFAPTLayer * layer) { aptLayer = layer; }



#endif // __MMFFILE_HH_
