#include <mlf/mlfdrglayer.hh>
#include <e00/arcsection.hh>
#include <e00/palsection.hh>
#include <e00/ifosection.hh>
#include <e00/tx7section.hh>
#include <e00/patfile.hh>

#include <iomanip>




/******************************************************************************
 * drgLayer.MLFDRGLayer(name, expfile)
 ******************************************************************************
 *
 *****************************************************************************/
MLFDRGLayer::MLFDRGLayer(string const & name, EXPFile const & expFile)
  : MLFLayer(name, "Drainage")
{
  /* Getting ARC section and setting number of polylines */
  ARCSection const & arcSection = *expFile.getARCSection();
  polylines.resize(arcSection.size());

  /* Setting polylines */
  ARCSection::const_iterator arcFieldIter = arcSection.begin();
  unsigned polylineIdx = 0;
  
  while (arcFieldIter != arcSection.end() &&
	 polylineIdx  <  polylines.size())
    {
      polylines[polylineIdx] = arcFieldIter->polyline;

      ++arcFieldIter;
      ++polylineIdx;
    }


  /* Getting PAL section and setting number of polygons */
  PALSection const & palSection = *expFile.getPALSection();
  polygons.resize(palSection.size()-1);

  /* Getting PAT file */
  PATFile const & patFile = *(*expFile.getIFOSection()).getPATFile();

  /* Setting polygons */
  PALSection::const_iterator palFieldIter = ++palSection.begin();
  unsigned patRecordIdx = 1;
  unsigned polygonIdx = 0;

  while (palFieldIter != palSection.end() &&
	 patRecordIdx <  patFile.getRecords().size() &&
	 polygonIdx   <  polygons.size())
    {
      /* Setting size and type */
      polygons[polygonIdx].resize(palFieldIter->arcInfos.size());
      polygons[polygonIdx].setType(patFile.getRecords()[patRecordIdx].type);

      /* Setting polylines' indeces */
      for (unsigned arcInfoIdx = 0; arcInfoIdx <
	     palFieldIter->arcInfos.size(); ++arcInfoIdx)
	if (palFieldIter->arcInfos[arcInfoIdx].arcID > 0)
	  {
	    polygons[polygonIdx][arcInfoIdx].reversed = false;
	    polygons[polygonIdx][arcInfoIdx].polylineIdx =
	      palFieldIter->arcInfos[arcInfoIdx].arcID-1;
	  }
	else 
	  {
	    polygons[polygonIdx][arcInfoIdx].reversed = true;
	    polygons[polygonIdx][arcInfoIdx].polylineIdx =
	      -palFieldIter->arcInfos[arcInfoIdx].arcID-1;
	  }

      ++palFieldIter;
      ++patRecordIdx;
      ++polygonIdx;
    }


  if (expFile.getTX7Section() != 0)
    {
      /* Getting TX7 section */
      TX7Section const & tx7Section = *expFile.getTX7Section();

      /* Getting number of labels */
      unsigned nLabels = 0;
      for (TX7Section::const_iterator tx7SubClassIter = tx7Section.begin();
	   tx7SubClassIter != tx7Section.end(); ++tx7SubClassIter)
	nLabels += tx7SubClassIter->tx7Fields.size();

      /* Setting number of labels */
      labels.resize(nLabels);

      /* Writing labels */
      unsigned labelIdx = 0;

      for (TX7Section::const_iterator tx7SubClassIter = tx7Section.begin();
	   tx7SubClassIter != tx7Section.end(); ++tx7SubClassIter)
	for (list<tx7Field_t>::const_iterator tx7FieldIter =
	       tx7SubClassIter->tx7Fields.begin(); tx7FieldIter !=
	       tx7SubClassIter->tx7Fields.end(); ++tx7FieldIter)
	  {
	    labels[labelIdx].setPosition(tx7FieldIter->vertices1[0]);
	    labels[labelIdx].setName    (tx7FieldIter->labelName);

	    ++labelIdx;
	  }

      if (tx7Section.getPrecision() > precision)
	precision = tx7Section.getPrecision();
    }

  if (arcSection.getPrecision() > precision)
    precision = arcSection.getPrecision();


  /* Finding bounding box */
  findBoundingBox();
}



/******************************************************************************
 * istr >> drgLayer
 ******************************************************************************
 *
 *****************************************************************************/
istream & operator>>(istream & istr, MLFDRGLayer & drgLayer)
{
  /* Reading precision, and title */
  istr >> static_cast<MLFLayer &>(drgLayer);

  /* Reading polylines */
  unsigned nPolylines;
  istr >> nPolylines;

  drgLayer.polylines.resize(nPolylines);
  for (unsigned polylineIdx = 0; polylineIdx < nPolylines; ++polylineIdx)
    istr >> drgLayer.polylines[polylineIdx];

  /* Reading polygons */
  unsigned nPolygons;
  istr >> nPolygons;

  drgLayer.polygons.resize(nPolygons);
  for (unsigned polygonIdx = 0; polygonIdx < nPolygons; ++polygonIdx)
    istr >> drgLayer.polygons[polygonIdx];

  /* Reading labels */
  unsigned nLabels;
  istr >> nLabels;

  drgLayer.labels.resize(nLabels);
  for (unsigned labelIdx = 0; labelIdx < nLabels; ++labelIdx)
    istr >> drgLayer.labels[labelIdx];

  if (nLabels == 0) istr.ignore(80, '\n');

  /* Finding bounding box */
  drgLayer.findBoundingBox();


  return istr;
}



/******************************************************************************
 * ostr >> drgLayer
 ******************************************************************************
 *
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MLFDRGLayer const & drgLayer)
{
  /* Writing name, precision, and title */
  ostr << static_cast<MLFLayer const &>(drgLayer)
       << scientific << uppercase << setprecision(7*(drgLayer.precision-1));

  /* Writing polylines */
  ostr << drgLayer.polylines.size() << endl;
  for (unsigned polylineIdx = 0; polylineIdx <
	 drgLayer.polylines.size(); ++polylineIdx)
    ostr << drgLayer.polylines[polylineIdx];

  /* Writing polygons */
  ostr << drgLayer.polygons.size() << endl;
  for (unsigned polygonIdx = 0; polygonIdx <
	 drgLayer.polygons.size(); ++polygonIdx)
    ostr << drgLayer.polygons[polygonIdx];

  /* Writing labels */
  ostr << drgLayer.labels.size() << endl;
  for (unsigned labelIdx = 0; labelIdx < drgLayer.labels.size(); ++labelIdx)
    ostr << drgLayer.labels[labelIdx];

  return ostr;
}



/******************************************************************************
 * drgLayer.findBoundingBox
 ******************************************************************************
 *
 *****************************************************************************/
void MLFDRGLayer::findBoundingBox()
{
  xMin = +1.7e+308;
  xMax = -1.7e+308;
  yMin = +1.7e+308;
  yMax = -1.7e+308;

  for (unsigned polylineIdx = 0; polylineIdx < polylines.size(); ++polylineIdx)
    for (unsigned vertexIdx = 0; vertexIdx <
	   polylines[polylineIdx].size(); ++vertexIdx)
      {
	MLFPoint const & point = polylines[polylineIdx][vertexIdx];

	if (point.getX() < xMin) xMin = point.getX();
	if (point.getX() > xMax) xMax = point.getX();
	if (point.getY() < yMin) yMin = point.getY();
	if (point.getY() > yMax) yMax = point.getY();
      }
}
