#include <mmf/mmfpoplayer.hh>
#include <e00/arcsection.hh>
#include <e00/labsection.hh>
#include <e00/palsection.hh>
#include <e00/ifosection.hh>
#include <e00/tx7section.hh>
#include <e00/patfile.hh>

#include <iomanip>




/******************************************************************************
 * popLayer.MMFPOPLayer(name, expFile1, expFile2)
 ******************************************************************************
 *
 *****************************************************************************/
MMFPOPLayer::MMFPOPLayer(string const & name,
		   EXPFile const & expFile1,
		   EXPFile const & expFile2)
  : MMFLayer(name, "Populated places")
{
  if (expFile1.getLABSection() != 0) readPPPTData(expFile1);
  if (expFile2.getARCSection() != 0) readPPPLData(expFile2);


  /* Finding bounding box */
  findBoundingBox();
}



/******************************************************************************
 * popLayer.readPPPTData(expFile)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFPOPLayer::readPPPTData(EXPFile const & expFile)
{
  /* Getting LAB section and setting number of points */
  LABSection const & labSection = *expFile.getLABSection();
  points.resize(labSection.size()-1);

  /* Setting points */
  LABSection::const_iterator labFieldIter = ++labSection.begin();
  unsigned pointIdx = 0;

  while (labFieldIter != labSection.end() &&
	 pointIdx     <  points.size())
    {
      points[pointIdx] = labFieldIter->labHeader.position;

      ++labFieldIter;
      ++pointIdx;
    }


  /* Getting TX7 section */
  TX7Section const & tx7Section = *expFile.getTX7Section();

  /* Getting number of labels */
  unsigned nLabels = 0;
  for (TX7Section::const_iterator tx7SubClassIter = tx7Section.begin();
       tx7SubClassIter != tx7Section.end(); ++tx7SubClassIter)
    nLabels += tx7SubClassIter->tx7Fields.size();

  /* Setting number of labels */
  labels.resize(nLabels);

  /* Writing labels */
  unsigned labelIdx = 0;

  for (TX7Section::const_iterator tx7SubClassIter = tx7Section.begin();
       tx7SubClassIter != tx7Section.end(); ++tx7SubClassIter)
    for (list<tx7Field_t>::const_iterator tx7FieldIter =
	   tx7SubClassIter->tx7Fields.begin(); tx7FieldIter !=
	   tx7SubClassIter->tx7Fields.end(); ++tx7FieldIter)
      {
	labels[labelIdx].setPosition(tx7FieldIter->vertices1[0]);
	labels[labelIdx].setName    (tx7FieldIter->labelName);

	++labelIdx;
      }

  if (tx7Section.getPrecision() > precision)
    precision = tx7Section.getPrecision();
}



/******************************************************************************
 * popLayer.readPPPLData(expFile)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFPOPLayer::readPPPLData(EXPFile const & expFile)
{
  /* Getting ARC section and setting number of polylines */
  ARCSection const & arcSection = *expFile.getARCSection();
  polylines.resize(arcSection.size());

  /* Setting polylines */
  ARCSection::const_iterator arcFieldIter = arcSection.begin();
  unsigned polylineIdx = 0;
  
  while (arcFieldIter != arcSection.end() &&
	 polylineIdx  <  polylines.size())
    {
      polylines[polylineIdx] = arcFieldIter->polyline;

      ++arcFieldIter;
      ++polylineIdx;
    }


  /* Getting PAL section and setting number of polygons */
  PALSection const & palSection = *expFile.getPALSection();
  polygons.resize(palSection.size()-1);

  /* Getting PAT file */
  PATFile const & patFile = *(*expFile.getIFOSection()).getPATFile();

  /* Setting polygons */
  PALSection::const_iterator palFieldIter = ++palSection.begin();
  unsigned patRecordIdx = 1;
  unsigned polygonIdx = 0;

  while (palFieldIter != palSection.end() &&
	 patRecordIdx <  patFile.getRecords().size() &&
	 polygonIdx   <  polygons.size())
    {
      /* Setting size and type */
      polygons[polygonIdx].resize(palFieldIter->arcInfos.size());
      polygons[polygonIdx].setType(patFile.getRecords()[patRecordIdx].type);

      /* Setting polylines' indeces */
      for (unsigned arcInfoIdx = 0; arcInfoIdx <
	     palFieldIter->arcInfos.size(); ++arcInfoIdx)
	if (palFieldIter->arcInfos[arcInfoIdx].arcID > 0)
	  {
	    polygons[polygonIdx][arcInfoIdx].reversed = false;
	    polygons[polygonIdx][arcInfoIdx].polylineIdx =
	      palFieldIter->arcInfos[arcInfoIdx].arcID-1;
	  }
	else 
	  {
	    polygons[polygonIdx][arcInfoIdx].reversed = true;
	    polygons[polygonIdx][arcInfoIdx].polylineIdx =
	      -palFieldIter->arcInfos[arcInfoIdx].arcID-1;
	  }

      ++palFieldIter;
      ++patRecordIdx;
      ++polygonIdx;
    }

  if (arcSection.getPrecision() > precision)
    precision = arcSection.getPrecision();
}



/******************************************************************************
 * istr >> popLayer
 ******************************************************************************
 *
 *****************************************************************************/
istream & operator>>(istream & istr, MMFPOPLayer & popLayer)
{
  /* Reading precision, and title */
  istr >> static_cast<MMFLayer &>(popLayer);

  /* Reading points */
  unsigned nPoints;
  istr >> nPoints;

  popLayer.points.resize(nPoints);
  for (unsigned pointIdx = 0; pointIdx < nPoints; ++pointIdx)
    istr >> popLayer.points[pointIdx];

  /* Reading labels */
  unsigned nLabels;
  istr >> nLabels;

  popLayer.labels.resize(nLabels);
  for (unsigned labelIdx = 0; labelIdx < nLabels; ++labelIdx)
    istr >> popLayer.labels[labelIdx];

  /* Reading polylines */
  unsigned nPolylines;
  istr >> nPolylines;

  popLayer.polylines.resize(nPolylines);
  for (unsigned polylineIdx = 0; polylineIdx < nPolylines; ++polylineIdx)
    {
      popLayer.polylines[polylineIdx].
	setNumberOfResolutions(popLayer.nResolutions);
      istr >> popLayer.polylines[polylineIdx];
    }

  /* Reading polygons */
  unsigned nPolygons;
  istr >> nPolygons;

  popLayer.polygons.resize(nPolygons);
  for (unsigned polygonIdx = 0; polygonIdx < nPolygons; ++polygonIdx)
    istr >> popLayer.polygons[polygonIdx];

  /* Finding bounding box */
  popLayer.findBoundingBox();


  return istr.ignore(80, '\n');
}



/******************************************************************************
 * ostr << popLayer
 ******************************************************************************
 *
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MMFPOPLayer const & popLayer)
{
  /* Writing name, precision, and title */
  ostr << static_cast<MMFLayer const &>(popLayer)
       << scientific << uppercase << setprecision(7*(popLayer.precision-1));

  /* Writing points */
  ostr << popLayer.points.size() << endl;
  for (unsigned pointIdx = 0; pointIdx < popLayer.points.size(); ++pointIdx)
    ostr << popLayer.points[pointIdx] << endl;

  /* Writing labels */
  ostr << popLayer.labels.size() << endl;
  for (unsigned labelIdx = 0; labelIdx < popLayer.labels.size(); ++labelIdx)
    ostr << popLayer.labels[labelIdx];

  /* Writing polylines */
  ostr << popLayer.polylines.size() << endl;
  for (unsigned polylineIdx = 0; polylineIdx <
	 popLayer.polylines.size(); ++polylineIdx)
    ostr << popLayer.polylines[polylineIdx];

  /* Writing polygons */
  ostr << popLayer.polygons.size() << endl;
  for (unsigned polygonIdx = 0; polygonIdx <
	 popLayer.polygons.size(); ++polygonIdx)
    ostr << popLayer.polygons[polygonIdx];

  return ostr;
}



/******************************************************************************
 * popLayer.findBoundingBox
 ******************************************************************************
 *
 *****************************************************************************/
inline void MMFPOPLayer::findBoundingBox()
{
  xMin = +1.7e+308;
  xMax = -1.7e+308;
  yMin = +1.7e+308;
  yMax = -1.7e+308;

  for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
    {
      if (points[pointIdx].getX() < xMin) xMin = points[pointIdx].getX();
      if (points[pointIdx].getX() > xMax) xMax = points[pointIdx].getX();
      if (points[pointIdx].getY() < yMin) yMin = points[pointIdx].getY();
      if (points[pointIdx].getY() > yMax) yMax = points[pointIdx].getY();
    }

  for (unsigned polylineIdx = 0; polylineIdx < polylines.size(); ++polylineIdx)
    for (unsigned resolIdx = 0; resolIdx <
	   polylines[polylineIdx].numberOfResolutions(); ++resolIdx)
      for (unsigned vertexIdx = 0; vertexIdx <
	     polylines[polylineIdx][resolIdx].numberOfVertices(); ++vertexIdx)
	{
	  MMFPoint const & point =
	    polylines[polylineIdx][resolIdx].getVertex(vertexIdx);

	  if (point.getX() < xMin) xMin = point.getX();
	  if (point.getX() > xMax) xMax = point.getX();
	  if (point.getY() < yMin) yMin = point.getY();
	  if (point.getY() > yMax) yMax = point.getY();
	}
}
