#include "point.hh"

#include <iomanip>




/******************************************************************************
 * point.distance(p, q)
 ******************************************************************************
 * Returns the distance from the point 'point' to the segment 'pq'.  This
 * function is used in Douglas-Peucker simplification algorithms.
 *****************************************************************************/
double Point::distance(Point const & p, Point const & q) const
{
  /* 'p' or 'q' is the closest point to 'point' */
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return distance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return distance(q);

  /* 'point' is closest to a point between 'p' and 'q' */
  double norm = p.distance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return fabs(determinant)/norm;
}



/******************************************************************************
 * point.sqrDistance(p, q)
 ******************************************************************************
 * Returns the distance from the point 'point' to the segment 'pq'.  This
 * function is used in Douglas-Peucker simplification algorithms.
 *****************************************************************************/
double Point::sqrDistance(Point const & p, Point const & q) const
{
  /* 'p' or 'q' is the closest point to 'point' */
  if ((x-p.x)*(q.x-p.x)+(y-p.y)*(q.y-p.y) <= 0) return sqrDistance(p);
  if ((x-q.x)*(p.x-q.x)+(y-q.y)*(p.y-q.y) <= 0) return sqrDistance(q);

  /* 'point' is closest to a point between 'p' and 'q' */
  double sqrNorm = p.sqrDistance(q);
  double determinant = (q.x-p.x)*(y-p.y)-(x-p.x)*(q.y-p.y);
  return determinant*determinant/sqrNorm;
}



/******************************************************************************
 * point.hasSameSharings(p)
 ******************************************************************************
 * Returns 'true', if 'point' has the same sharings thant 'p'.  Otherwise,
 * returns 'false'.
 *****************************************************************************/
bool Point::hasSameSharings(Point const & p) const
{
  if (sharings.size() != p.sharings.size())
    return false;

  for (unsigned sharingIdx = 0; sharingIdx < sharings.size(); ++sharingIdx)
    if (sharings[sharingIdx].polIdx != p.sharings[sharingIdx].polIdx ||
	abs(static_cast<int>(  sharings[sharingIdx].vertexIdx)-
	    static_cast<int>(p.sharings[sharingIdx].vertexIdx)) != 1)
      return false;

  return true;
}



/******************************************************************************
 * point.print(ostr, delim, width)
 ******************************************************************************
 * Prints the 'point''s members on 'ostr' as larg as 'width''delim' at the end.
 *****************************************************************************/
void Point::print(ostream & ostr, char const * delim, int width) const
{
  ostr << setw(width) << x << ' '
       << setw(width) << y << " [";

  for (unsigned sharingIdx = 0; sharingIdx < sharings.size(); ++sharingIdx)
    ostr << " (" << sharings[sharingIdx].polIdx
	 << ' '  << sharings[sharingIdx].vertexIdx << ')';

  ostr << " ]" << delim;
}



/******************************************************************************
 * ostr << point
 ******************************************************************************
 * Prints the cartesian coordinates of 'point' on 'ostr'.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Point const & point)
{
  ostr << setw(COORD_WIDTH) << point.x
       << setw(COORD_WIDTH) << point.y;


  return ostr;
}



/******************************************************************************
 * istr >> point
 ******************************************************************************
 * Reads the cartesian coordinates of 'point' from 'istr'.
 *****************************************************************************/
istream & operator>>(istream & istr, Point & point)
{
  istr >> point.x >> point.y;


  return istr;
}
