#ifndef __POINT_HH_
#define __POINT_HH_


#include <cmath>
#include <iostream>
#include <vector>

#define COORD_PRECISION  7
#define COORD_WIDTH     14


using namespace std;


struct sharing_t {

  sharing_t(unsigned polIdx, unsigned vertexIdx)
    : polIdx(polIdx), vertexIdx(vertexIdx) {}

  bool operator==(sharing_t const & s) const
  { return s.polIdx == polIdx && s.vertexIdx == vertexIdx; }

  unsigned polIdx;
  unsigned vertexIdx;
};


/******************************************************************************
 * 'Point' is the point class. It carries the cartesian coordinates of a
 * point in 2D space. It also presents geometric transformations and util
 * functions usually applied to points.
 *****************************************************************************/
class Point {

public:

  /* Constructors */
  Point(Point const & point);             // Copy
  Point(double x = 0.0, double y = 0.0);  // Cartesien coordinates


  /* Get functions */
  double getX() const;
  double getY() const;
  bool   getCanceled() const;
  bool   getCritical() const;

  unsigned numberOfSharings() const;
  sharing_t getSharing(unsigned sharingIdx);


  /* Set functions */
  void setX(double x);
  void setY(double y);
  void setCanceled(bool canceled);
  void setCritical(bool critical);

  void pushSharing(unsigned polIdx, unsigned vertexIdx);


  /* Distance from 'p' */
  double    distance(Point const & p) const;
  double sqrDistance(Point const & p) const;


  /* Squared distance from 'p' */
  double    distance(Point const & p, Point const & q) const;
  double sqrDistance(Point const & p, Point const & q) const;


  /* Operators of equality */
  bool operator==(Point const & point) const;
  bool operator!=(Point const & point) const;
  bool operator< (Point const & point) const;


  /* Operator of assignment */
  Point const & operator=(Point const & p);


  /* Sharing actions */
  bool hasSameSharings(Point const & point) const;


  /* Print function */
  void print(ostream & ostr, char const * delim = "\n",
	     int width = COORD_WIDTH) const;


  /* Friend operators for input and output handle */
  friend ostream & operator<<(ostream & ostr, Point const & point);
  friend istream & operator>>(istream & istr, Point       & point);


private:

  /* Cartesian coordinates */
  double x;
  double y;

  /* Critical flag */
  bool critical;

  /* Sharings */
  vector<sharing_t> sharings;
  bool canceled;
};



inline Point::Point(Point const & point)
  : x(point.x), y(point.y), critical(false), canceled(false) {}

inline Point::Point(double x, double y)
  : x(x), y(y), critical(false), canceled(false) {}


inline double Point::getX() const { return x; }
inline double Point::getY() const { return y; }

inline void Point::setX(double x) { this->x = x; }
inline void Point::setY(double y) { this->y = y; }

inline bool Point::operator!=(Point const & p) const { return !operator==(p); }
inline bool Point::operator==(Point const & p) const { return ((x == p.x) &&
							       (y == p.y)); }
inline bool Point::operator<(Point const & p) const
{ return (x < p.x ? true : (y < p.y ? true : false)); }

inline Point const & Point::operator=(Point const & p)
{
  if (this != &p)
    {
      x = p.x;
      y = p.y;
    }

  return *this;
}

inline bool      Point::getCritical() const      { return critical; }
inline bool      Point::getCanceled() const      { return canceled; }

inline unsigned  Point::numberOfSharings() const { return sharings.size(); }
inline sharing_t Point::getSharing(unsigned idx) { return sharings[idx]; }

inline void Point::setCritical(bool critical) { this->critical = critical; }
inline void Point::setCanceled(bool canceled) { this->canceled = canceled; }

inline void Point::pushSharing(unsigned polIdx, unsigned vertexIdx)
{ sharings.push_back(sharing_t(polIdx, vertexIdx)); }

inline double Point::distance(Point const & p) const
{ return sqrt((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y)); }

inline double Point::sqrDistance(Point const & p) const
{ return (p.x-x)*(p.x-x)+(p.y-y)*(p.y-y); }



#endif // __POINT_HH_
