#include "polyline.hh"

#include <iomanip>
#include <values.h>

#include <cstdlib>


/******************************************************************************
 * polyline.print(ostr, frstIdx, lastIdx)
 ******************************************************************************
 * Prints the list of points sequentially ordered, one per line.
 *****************************************************************************/
void Polyline::print(ostream & ostr, unsigned frstIdx, unsigned lastIdx) const
{
  ostr << numberOfVertices() << ": " << frstIdx << " - " << lastIdx << endl;

  for (unsigned vertexIdx = frstIdx; vertexIdx <= lastIdx; ++vertexIdx)
    {
      ostr << setw(3) << vertexIdx << ' ';
      (*this)[vertexIdx].print(ostr);
    }


  /* Printing intersections */
  ostr << "intr: " << intersections.size() << endl;
  for (list<Polyline::intersection_t>::const_iterator iter =
	 intersections.begin();
       iter != intersections.end(); ++iter)
    ostr << setw(2) << iter->first << " - " << iter->second << endl;


  /* Printing direction */
  ostr << "dirc: " << direction << endl;

  /* Printing prev/next polylines */
  ostr << "prev: " << nPrevPols << ' ' << prevPolIdx << endl;
  ostr << "next: " << nNextPols << ' ' << nextPolIdx << endl;
}



/******************************************************************************
 * ostr << pol
 ******************************************************************************
 * For each part of 'pol', prints the number of points, followed by the list of
 * points sequentially ordered, one per line. Is a redefinition of the
 * insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, Polyline const & polyline)
{
  /* Printing vertices */
  ostr << polyline.numberOfVertices() << endl;

  for (unsigned vertexIdx = 0; vertexIdx < polyline.size(); ++vertexIdx)
    ostr << polyline[vertexIdx] << endl;


  return ostr;
}



/******************************************************************************
 * istr >> pol
 ******************************************************************************
 * First reads the number of points, and then the list of points. Is a
 * redefinition of the extraction operator '>>' in stream context. Accepts the
 * points in the given order by considering they are ordered correctely.
 *****************************************************************************/
istream & operator>>(istream & istr, Polyline & pol)
{
  /* Reading number of vertices and allocating memory */
  unsigned nVertices;
  istr >> nVertices;

  pol.resize(nVertices);


  /* Inicializing bounding box */
  pol.xMin = MAXDOUBLE; pol.xMax = -MAXDOUBLE;
  pol.yMin = MAXDOUBLE; pol.yMax = -MAXDOUBLE;

  /* Reading vertices */
  for (unsigned vertexIdx = 0; vertexIdx < pol.size(); ++vertexIdx)
    {
      istr >> pol[vertexIdx];

      if (pol[vertexIdx].getX() < pol.xMin) pol.xMin = pol[vertexIdx].getX();
      if (pol[vertexIdx].getX() > pol.xMax) pol.xMax = pol[vertexIdx].getX();
      if (pol[vertexIdx].getY() < pol.yMin) pol.yMin = pol[vertexIdx].getY();
      if (pol[vertexIdx].getY() > pol.yMax) pol.yMax = pol[vertexIdx].getY();
    }


  return istr;
}



/******************************************************************************
 * polyline.pushBreakVertex(vertexIdx)
 ******************************************************************************
 * Inserts critical overlapping vertices in the breaking vertex index's list,
 * only if it is not already there.  The list is ordered.
 *****************************************************************************/
// void Polyline::pushBreakVertex(unsigned vertexIdx)
// {
//   /* Searching for position */
//   list<unsigned>::iterator brkVertexIter = brkVertices.begin();
//   while (brkVertexIter != brkVertices.end() && *brkVertexIter < vertexIdx)
//     ++brkVertexIter;

//   /* if vertex index is not already in the list */
//   if (brkVertexIter == brkVertices.end() || *brkVertexIter != vertexIdx)
//     brkVertices.insert(brkVertexIter, vertexIdx);
// }



/******************************************************************************
 * polyline.breakIntersections()
 ******************************************************************************
 *
 *****************************************************************************/
void Polyline::breakIntersections()
{
  /* Allocating memory */
  vector<Point> polyline;
  polyline.reserve(size()+intersections.size());

  /* Updating intersections */
  intersections.sort();
  intersections.unique();

  /* Adding intersection points */
  unsigned index = 0;

  for (list<intersection_t>::iterator iter = intersections.begin();
       iter != intersections.end(); ++iter)
    {
      polyline.insert(polyline.end(), &(*this)[index], &(*this)[iter->first]);
      polyline.push_back(iter->second);

      index = iter->first;
    }

  /* Adding last part */
  polyline.insert(polyline.end(), &(*this)[index], &back()+1);
  swap(polyline);

  /* Clearing intersections just inserted */
  intersections.clear();
}
