#include <bvfmap/bvfmap.hh>

#include <fstream>
#include <iomanip>
#include <list>

#include <GL/gl.h>
#include <GL/glu.h>




/******************************************************************************
 * bvfMap.read(ostr)
 ******************************************************************************
 * Reads 'bvfMap' from file given by 'fileName'.
 *****************************************************************************/
bool BVFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * bvfMap.read(ostr)
 ******************************************************************************
 * Writes 'bvfMap' in file given by 'fileName'.
 *****************************************************************************/
bool BVFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * bvfMap.draw()
 ******************************************************************************
 * Draws 'bvfMap' contents, including polylines and feature points.
 *****************************************************************************/
void BVFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting polylines */
  if (polylinesLayer)
    {
      glColor3f(0.0, 0.0, 0.0);
      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	{
	  glBegin(GL_LINE_STRIP);
	  {
	    for (unsigned vertexIdx = 0; vertexIdx <
		   polylines[polIdx].numberOfVertices(); ++vertexIdx)

	      if (polylines[polIdx][vertexIdx].getLevel() == 0 ||
		  polylines[polIdx][vertexIdx].
		  getCurrTolerance() >= tolerances[toleranceIdx])

		glVertex2d(polylines[polIdx][vertexIdx].getX(),
			   polylines[polIdx][vertexIdx].getY());
	  }
	  glEnd();
	}
    }


  /* Plotting points */
  if (pointsLayer)
    {
      glColor3f(1.0, 0.0, 0.0);
      glPointSize(5.0);
      glBegin(GL_POINTS);
      {
	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
	  glVertex2d(points[pointIdx].getX(),
		     points[pointIdx].getY());
      }
      glEnd();
    }


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> bvfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, BVFMap & bvfMap)
{
  /* Bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /* Reading polylines */
  unsigned nPolylines;
  istr.read(reinterpret_cast<char *>(&nPolylines), sizeof(unsigned));

  bvfMap.polylines.clear();
  bvfMap.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < bvfMap.polylines.size(); ++polIdx)
    {
      istr >> bvfMap.polylines[polIdx];

      if (bvfMap.polylines[polIdx].getXMin() < xMin)
	xMin = bvfMap.polylines[polIdx].getXMin();

      if (bvfMap.polylines[polIdx].getXMax() > xMax)
	xMax = bvfMap.polylines[polIdx].getXMax();

      if (bvfMap.polylines[polIdx].getYMin() < yMin)
	yMin = bvfMap.polylines[polIdx].getYMin();

      if (bvfMap.polylines[polIdx].getYMax() > yMax)
	yMax = bvfMap.polylines[polIdx].getYMax();
    }


  /* Reading feature points */
  unsigned nPoints;
  istr.read(reinterpret_cast<char*>(&nPoints), sizeof(unsigned));

  bvfMap.points.clear();
  bvfMap.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < bvfMap.points.size(); ++pointIdx)
    {
      bvfMap.points[pointIdx].readXY(istr);

      if (bvfMap.points[pointIdx].getX() < xMin)
	xMin = bvfMap.points[pointIdx].getX();

      if (bvfMap.points[pointIdx].getX() > xMax)
	xMax = bvfMap.points[pointIdx].getX();

      if (bvfMap.points[pointIdx].getY() < yMin)
	yMin = bvfMap.points[pointIdx].getY();

      if (bvfMap.points[pointIdx].getY() > yMax)
	yMax = bvfMap.points[pointIdx].getY();
    }


  /* Setting bounding box */
  bvfMap.setXMin(xMin);
  bvfMap.setXMax(xMax);
  bvfMap.setYMin(yMin);
  bvfMap.setYMax(yMax);


  /* Filling list of tolerances */
  list<double> tolList(1, bvfMap.polylines[0][0].getCurrTolerance());

  for (unsigned polIdx = 0; polIdx < bvfMap.polylines.size(); ++polIdx)
    for (unsigned vertexIdx = 1; vertexIdx <
	   bvfMap.polylines[polIdx].size()-1; ++vertexIdx)
      tolList.
	push_back(bvfMap.polylines[polIdx][vertexIdx].getCurrTolerance());


  /* Sorting and cleaning list of tolerances */
  tolList.sort();
  tolList.unique();
  tolList.reverse();


  /* Initiating index of tolerances and filling vector of tolerances */
  bvfMap.toleranceIdx = 0;
  bvfMap.tolerances.clear();
  bvfMap.tolerances.reserve(tolList.size());
  bvfMap.tolerances.insert(bvfMap.tolerances.begin(),
			   tolList.begin(), tolList.end());


  return istr;
}



/******************************************************************************
 * ostr << bvfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, BVFMap const & bvfMap)
{
  /* Writing polylines */
  unsigned nPolylines = bvfMap.polylines.size();
  ostr.write(reinterpret_cast<char const *>(&nPolylines), sizeof(unsigned));

  for (unsigned polIdx = 0; polIdx < bvfMap.polylines.size(); ++polIdx)
    ostr << bvfMap.polylines[polIdx];


  /* Writing points */
  unsigned nPoints = bvfMap.points.size();
  ostr.write(reinterpret_cast<char const *>(&nPoints), sizeof(unsigned));

  for (unsigned pointIdx = 0; pointIdx < bvfMap.points.size(); ++pointIdx)
    bvfMap.points[pointIdx].writeXY(ostr);


  return ostr;
}
