#ifndef __BVFMAP_HH_
#define __BVFMAP_HH_


#include <map.hh>
#include <bvfmap/bvfpolyline.hh>
#include <bvfmap/bvfpoint.hh>

#include <iostream>
#include <vector>


using namespace std;



/******************************************************************************
 * Class BVFMap
 *****************************************************************************/
class BVFMap
  : public Map {

public:

  /* Constructor */
  BVFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, BVFMap       & bvfMap);
  friend ostream & operator<<(ostream & ostr, BVFMap const & bvfMap);


private:

  enum layer_t { POLYLINES_LAYER = 0,
		    POINTS_LAYER = 1 };

  /* Flags for layers */
  bool polylinesLayer;
  bool    pointsLayer;

  /* Vectors of polylines and feature points */
  vector<BVFPolyline> polylines;
  vector<BVFPoint>    points;

  /* Vector of tolerances/resolutions */
  vector<double> tolerances;
  unsigned       toleranceIdx;
};


inline BVFMap::BVFMap() : polylinesLayer(true), pointsLayer(true) {}

inline char const * BVFMap::getName() const { return ""; }

inline unsigned     BVFMap::getNumberOfLayers() const { return 2; }
inline char const * BVFMap::getLayerName(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return "Polylines";
  if (layerIdx ==    POINTS_LAYER) return "Points";

  return "";
}

inline bool BVFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return polylinesLayer;
  if (layerIdx ==    POINTS_LAYER) return    pointsLayer;

  return false;
}

inline void BVFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx == POLYLINES_LAYER) polylinesLayer = activated;
  if (layerIdx ==    POINTS_LAYER)    pointsLayer = activated;
}

inline double BVFMap::getTolerance() const { return tolerances[toleranceIdx]; }
inline void   BVFMap::setTolerance(double tolerance)
{ for (toleranceIdx=0; tolerances[toleranceIdx]<tolerance; ++toleranceIdx) ; }

inline unsigned BVFMap::getNumberOfResolutions() const
{ return tolerances.size(); }

inline unsigned BVFMap::getResolution() const { return toleranceIdx; }

inline void BVFMap::setResolution(unsigned resIdx)
{ toleranceIdx = resIdx; }

inline void BVFMap::init() const {}

inline bool BVFMap::exportAsFig(char const * fileName, int colorMode,
				bool noFill) const
{
  fileName  = fileName;
  colorMode = colorMode;
  noFill    = noFill;

  return false;
}



#endif // __BVFMAP_HH_
