#include <imfmap/imfmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>




/******************************************************************************
 * imfMap.read(ostr)
 ******************************************************************************
 * Reads 'imfMap' from file given by 'fileName'.
 *****************************************************************************/
bool IMFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * imfMap.read(ostr)
 ******************************************************************************
 * Writes 'imfMap' in file given by 'fileName'.
 *****************************************************************************/
bool IMFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * imfMap.draw()
 ******************************************************************************
 * Draws 'imfMap' contents, including polylines and feature points.
 *****************************************************************************/
void IMFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting polylines */
  if (polylinesLayer)
    {
      glColor3f(0.0, 0.0, 0.0);
      for (unsigned polylineIdx = 0; polylineIdx <
	     polylines.size(); ++polylineIdx)
	{
	  glBegin(GL_LINE_STRIP);
	  {
	    for (unsigned resolutionIdx = 0; resolutionIdx <
		   tolerances.size(); ++resolutionIdx)
	      {
		resolution_t const & resolution =
		  polylines[polylineIdx].getResolution(resolutionIdx);

		for (unsigned vertexIdx = 0; vertexIdx <
		       resolution.size(); ++vertexIdx)

		  glVertex2d(resolution[vertexIdx].getX(),
			     resolution[vertexIdx].getY());
	      }
	  }
	  glEnd();
	}
    }


  /* Plotting points */
  if (pointsLayer)
    {
      glColor3f(1.0, 0.0, 0.0);
      glBegin(GL_POINTS);
      {
	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
	  glVertex2d(points[pointIdx].getX(),
		     points[pointIdx].getY());
      }
      glEnd();
    }


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> imfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, IMFMap & imfMap)
{
  /* Reading bounding box */
  double xMin, xMax;
  double yMin, yMax;

  istr >> xMin >> xMax
       >> yMin >> yMax;

  imfMap.setXMin(xMin);
  imfMap.setXMax(xMax);
  imfMap.setYMin(yMin);
  imfMap.setYMax(yMax);


  /* Reading tolerances for resolutions */
  unsigned nResolutions;
  istr >> nResolutions;

  imfMap.tolerances.clear();
  imfMap.tolerances.resize(nResolutions);

  for (unsigned toleranceIdx = 0; toleranceIdx < nResolutions; ++toleranceIdx)
    istr >> imfMap.tolerances[toleranceIdx];


  /* Reading polylines */
  unsigned nPolylines;
  istr >> nPolylines;

  imfMap.polylines.clear();
  imfMap.polylines.resize(nPolylines, IMFPolyline(imfMap.tolerances.size()));

  for (unsigned polIdx = 0; polIdx < imfMap.polylines.size(); ++polIdx)
    istr >> imfMap.polylines[polIdx];



  /* Reading feature points */
  unsigned nPoints;
  istr >> nPoints;

  imfMap.points.clear();
  imfMap.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < imfMap.points.size(); ++pointIdx)
    istr >> imfMap.points[pointIdx];


  return istr;
}



/******************************************************************************
 * ostr << imfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, IMFMap const & imfMap)
{
  /* Writing bounding box */
  ostr << scientific << uppercase << setprecision(7)
       << setw(14) << imfMap.xMin << setw(14) << imfMap.xMax << endl
       << setw(14) << imfMap.yMin << setw(14) << imfMap.yMax << endl
       << endl << endl;


  /* Writing tolerances for resolutions */
  ostr << endl << imfMap.tolerances.size() << endl;

  for (unsigned toleranceIdx = 0; toleranceIdx <
	 imfMap.tolerances.size(); ++toleranceIdx)
    ostr << setw(14) << imfMap.tolerances[toleranceIdx] << endl;


  /* Writing polylines */
  ostr << imfMap.polylines.size() << endl;

  for (unsigned polIdx = 0; polIdx < imfMap.polylines.size(); ++polIdx)
    ostr << endl << imfMap.polylines[polIdx];


  /* Writing feature points */
  ostr << imfMap.points.size() << endl << endl;

  for (unsigned pointIdx = 0; pointIdx < imfMap.points.size(); ++pointIdx)
    ostr << imfMap.points[pointIdx] << endl;


  return ostr;
}
