#ifndef __IMFMAP_HH_
#define __IMFMAP_HH_


#include <map.hh>
#include <imfmap/imfpolyline.hh>
#include <imfmap/imfpoint.hh>

#include <iostream>


using namespace std;



/******************************************************************************
 * Class IMFMap
 *****************************************************************************/
class IMFMap
  : public Map {

public:

  /* Constructor */
  IMFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, IMFMap       & imfMap);
  friend ostream & operator<<(ostream & ostr, IMFMap const & imfMap);


private:

  enum layer_t { POLYLINES_LAYER = 0,
		    POINTS_LAYER = 1 };

  /* Vector of tolerances */
  vector<double> tolerances;

  /* Flags for layers */
  bool polylinesLayer;
  bool    pointsLayer;

  /* Vectors of polylines and feature points */
  vector<IMFPolyline> polylines;
  vector<IMFPoint>    points;
};


inline IMFMap::IMFMap() : polylinesLayer(true), pointsLayer(true) {}

inline char const * IMFMap::getName() const { return ""; }

inline unsigned     IMFMap::getNumberOfLayers() const { return 2; }
inline char const * IMFMap::getLayerName(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return "Polylines";
  if (layerIdx ==    POINTS_LAYER) return "Points";

  return "";
}

inline bool IMFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return polylinesLayer;
  if (layerIdx ==    POINTS_LAYER) return    pointsLayer;

  return false;
}

inline void IMFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx == POLYLINES_LAYER) polylinesLayer = activated;
  if (layerIdx ==    POINTS_LAYER)    pointsLayer = activated;
}

inline double IMFMap::getTolerance() const { return 0.0; }
inline void   IMFMap::setTolerance(double tolerance) { tolerance = tolerance; }

inline unsigned IMFMap::getNumberOfResolutions() const { return 1; }
inline unsigned IMFMap::getResolution() const { return 0; }
inline void     IMFMap::setResolution(unsigned resIdx) { resIdx = resIdx; }

inline void IMFMap::init() const {}

inline bool IMFMap::exportAsFig(char const * fileName, int colorMode,
				bool noFill) const
{
  fileName  = fileName;
  colorMode = colorMode;
  noFill    = noFill;

  return false;
}



#endif // __IMFMAP_HH_
