#include <infmap/infmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>




/******************************************************************************
 * infMap.read(ostr)
 ******************************************************************************
 * Reads 'infMap' from file given by 'fileName'.
 *****************************************************************************/
bool INFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * infMap.read(ostr)
 ******************************************************************************
 * Writes 'infMap' in file given by 'fileName'.
 *****************************************************************************/
bool INFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * infMap.draw()
 ******************************************************************************
 * Draws 'infMap' contents, including polylines and feature points.
 *****************************************************************************/
void INFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting polylines */
  if (polylinesLayer)
    {
      glColor3f(0.0, 0.0, 0.0);
      for (unsigned polylineIdx = 0; polylineIdx <
	     polylines.size(); ++polylineIdx)
	{
	  glBegin(GL_LINE_STRIP);
	  {
	    for (unsigned vertexIdx = 0; vertexIdx <
		   polylines[polylineIdx].numberOfVertices(); ++vertexIdx)

	      glVertex2d(polylines[polylineIdx][vertexIdx].getX(),
			 polylines[polylineIdx][vertexIdx].getY());
	  }
	  glEnd();
	}
    }


  /* Plotting points */
  if (pointsLayer)
    {
      glColor3f(1.0, 0.0, 0.0);
      glPointSize(5.0);
      glBegin(GL_POINTS);
      {
	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
	  glVertex2d(points[pointIdx].getX(),
		     points[pointIdx].getY());
      }
      glEnd();
    }


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> infMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, INFMap & infMap)
{
  /* Bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /* Reading polylines */
  unsigned nPolylines;
  istr >> nPolylines;

  infMap.polylines.clear();
  infMap.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < infMap.polylines.size(); ++polIdx)
    {
      istr >> infMap.polylines[polIdx];

      double xMinAux = infMap.polylines[polIdx].getXMin();
      double xMaxAux = infMap.polylines[polIdx].getXMax();
      double yMinAux = infMap.polylines[polIdx].getYMin();
      double yMaxAux = infMap.polylines[polIdx].getYMax();

      if (xMinAux < xMin) xMin = xMinAux;
      if (xMaxAux > xMax) xMax = xMaxAux;
      if (yMinAux < yMin) yMin = yMinAux;
      if (yMaxAux > yMax) yMax = yMaxAux;
    }


  /* Reading feature points */
  unsigned nPoints;
  istr >> nPoints;

  infMap.points.clear();
  infMap.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < infMap.points.size(); ++pointIdx)
    {
      istr >> infMap.points[pointIdx];

      if (infMap.points[pointIdx].getX() < xMin)
	xMin = infMap.points[pointIdx].getX();

      if (infMap.points[pointIdx].getX() > xMax)
	xMax = infMap.points[pointIdx].getX();

      if (infMap.points[pointIdx].getY() < yMin)
	yMin = infMap.points[pointIdx].getY();

      if (infMap.points[pointIdx].getY() > yMax)
	yMax = infMap.points[pointIdx].getY();
    }


  /* Setting bounding box */
  infMap.setXMin(xMin);
  infMap.setXMax(xMax);
  infMap.setYMin(yMin);
  infMap.setYMax(yMax);


  return istr;
}



/******************************************************************************
 * ostr << infMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, INFMap const & infMap)
{
  /* Writing polylines */
  ostr << infMap.polylines.size() << endl
       << scientific << uppercase << setprecision(7);

  for (unsigned polIdx = 0; polIdx < infMap.polylines.size(); ++polIdx)
    ostr << endl << infMap.polylines[polIdx];


  /* Writing feature points */
  ostr << endl << infMap.points.size() << endl;

  for (unsigned pointIdx = 0; pointIdx < infMap.points.size(); ++pointIdx)
    ostr << endl << infMap.points[pointIdx];

  if (infMap.points.size()) ostr << endl;


  return ostr;
}
