#include <lvfmap/lvfmap.hh>

#include <fstream>
#include <iomanip>
#include <list>

#include <GL/gl.h>
#include <GL/glu.h>




/******************************************************************************
 * lvfMap.read(ostr)
 ******************************************************************************
 * Reads 'lvfMap' from file given by 'fileName'.
 *****************************************************************************/
bool LVFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * lvfMap.read(ostr)
 ******************************************************************************
 * Writes 'lvfMap' in file given by 'fileName'.
 *****************************************************************************/
bool LVFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * lvfMap.draw()
 ******************************************************************************
 * Draws 'lvfMap' contents, including polylines and feature points.
 *****************************************************************************/
void LVFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting polylines */
  if (polylinesLayer)
    {
      glColor3f(0.0, 0.0, 0.0);
      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	{
	  glBegin(GL_LINE_STRIP);
	  {
	    for (unsigned vertexIdx = 0; vertexIdx <
		   polylines[polIdx].numberOfVertices(); ++vertexIdx)

	      if (polylines[polIdx][vertexIdx].getLevel() == 0 ||
		  polylines[polIdx][vertexIdx].
		  getTolerance() >= tolerances[toleranceIdx])

		glVertex2d(polylines[polIdx][vertexIdx].getX(),
			   polylines[polIdx][vertexIdx].getY());
	  }
	  glEnd();
	}
    }


  /* Plotting points */
  if (pointsLayer)
    {
      glColor3f(1.0, 0.0, 0.0);
      glPointSize(5.0);
      glBegin(GL_POINTS);
      {
	for (unsigned pointIdx = 0; pointIdx < points.size(); ++pointIdx)
	  glVertex2d(points[pointIdx].getX(),
		     points[pointIdx].getY());
      }
      glEnd();
    }


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> lvfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, LVFMap & lvfMap)
{
  /* Bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /* Reading polylines */
  unsigned nPolylines;
  istr >> nPolylines;

  lvfMap.polylines.clear();
  lvfMap.polylines.resize(nPolylines);

  for (unsigned polIdx = 0; polIdx < lvfMap.polylines.size(); ++polIdx)
    {
      istr >> lvfMap.polylines[polIdx];

      double xMinAux = lvfMap.polylines[polIdx].getXMin();
      double xMaxAux = lvfMap.polylines[polIdx].getXMax();
      double yMinAux = lvfMap.polylines[polIdx].getYMin();
      double yMaxAux = lvfMap.polylines[polIdx].getYMax();

      if (xMinAux < xMin) xMin = xMinAux;
      if (xMaxAux > xMax) xMax = xMaxAux;
      if (yMinAux < yMin) yMin = yMinAux;
      if (yMaxAux > yMax) yMax = yMaxAux;
    }


  /* Reading feature points */
  unsigned nPoints;
  istr >> nPoints;

  lvfMap.points.clear();
  lvfMap.points.resize(nPoints);

  for (unsigned pointIdx = 0; pointIdx < lvfMap.points.size(); ++pointIdx)
    {
      istr >> lvfMap.points[pointIdx];

      if (lvfMap.points[pointIdx].getX() < xMin)
	xMin = lvfMap.points[pointIdx].getX();

      if (lvfMap.points[pointIdx].getX() > xMax)
	xMax = lvfMap.points[pointIdx].getX();

      if (lvfMap.points[pointIdx].getY() < yMin)
	yMin = lvfMap.points[pointIdx].getY();

      if (lvfMap.points[pointIdx].getY() > yMax)
	yMax = lvfMap.points[pointIdx].getY();
    }


  /* Setting bounding box */
  lvfMap.setXMin(xMin);
  lvfMap.setXMax(xMax);
  lvfMap.setYMin(yMin);
  lvfMap.setYMax(yMax);


  /* Filling list of tolerances */
  list<double> tolList(1, lvfMap.polylines[0][0].getTolerance());

  for (unsigned polIdx = 0; polIdx < lvfMap.polylines.size(); ++polIdx)
    for (unsigned vertexIdx = 1; vertexIdx <
	   lvfMap.polylines[polIdx].size()-1; ++vertexIdx)
      tolList.push_back(lvfMap.polylines[polIdx][vertexIdx].getTolerance());


  /* Sorting and cleaning list of tolerances */
  tolList.sort();
  tolList.unique();
  tolList.reverse();


  /* Initiating index of tolerances and filling vector of tolerances */
  lvfMap.toleranceIdx = 0;
  lvfMap.tolerances.clear();
  lvfMap.tolerances.reserve(tolList.size());
  lvfMap.tolerances.insert(lvfMap.tolerances.begin(),
			   tolList.begin(), tolList.end());


  return istr;
}



/******************************************************************************
 * ostr << lvfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, LVFMap const & lvfMap)
{
  /* Writing polylines */
  ostr << lvfMap.polylines.size() << endl
       << scientific << uppercase << setprecision(7);

  for (unsigned polIdx = 0; polIdx < lvfMap.polylines.size(); ++polIdx)
    ostr << endl << lvfMap.polylines[polIdx];


  /* Writing feature points */
  ostr << endl << lvfMap.points.size() << endl;

  for (unsigned pointIdx = 0; pointIdx < lvfMap.points.size(); ++pointIdx)
    ostr << endl << lvfMap.points[pointIdx];

  if (lvfMap.points.size()) ostr << endl;


  return ostr;
}
