#ifndef __LVFMAP_HH_
#define __LVFMAP_HH_


#include <map.hh>
#include <lvfmap/lvfpolyline.hh>
#include <lvfmap/lvfpoint.hh>

#include <iostream>
#include <vector>


using namespace std;



/******************************************************************************
 * Class LVFMap
 *****************************************************************************/
class LVFMap
  : public Map {

public:

  /* Constructor */
  LVFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, LVFMap       & lvfMap);
  friend ostream & operator<<(ostream & ostr, LVFMap const & lvfMap);


private:

  enum layer_t { POLYLINES_LAYER = 0,
		    POINTS_LAYER = 1 };

  /* Flags for layers */
  bool polylinesLayer;
  bool    pointsLayer;

  /* Vectors of polylines and feature points */
  vector<LVFPolyline> polylines;
  vector<LVFPoint>    points;

  /* Vector of tolerances/resolutions */
  vector<double> tolerances;
  unsigned       toleranceIdx;
};


inline LVFMap::LVFMap() : polylinesLayer(true), pointsLayer(true) {}

inline char const * LVFMap::getName() const { return ""; }

inline unsigned     LVFMap::getNumberOfLayers() const { return 2; }
inline char const * LVFMap::getLayerName(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return "Polylines";
  if (layerIdx ==    POINTS_LAYER) return "Points";

  return "";
}

inline bool LVFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx == POLYLINES_LAYER) return polylinesLayer;
  if (layerIdx ==    POINTS_LAYER) return    pointsLayer;

  return false;
}

inline void LVFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx == POLYLINES_LAYER) polylinesLayer = activated;
  if (layerIdx ==    POINTS_LAYER)    pointsLayer = activated;
}

inline double LVFMap::getTolerance() const { return tolerances[toleranceIdx]; }
inline void   LVFMap::setTolerance(double tolerance)
{ for (toleranceIdx=0; tolerances[toleranceIdx]<tolerance; ++toleranceIdx) ; }

inline unsigned LVFMap::getNumberOfResolutions() const
{ return tolerances.size(); }

inline unsigned LVFMap::getResolution() const { return toleranceIdx; }

inline void LVFMap::setResolution(unsigned resIdx)
{ toleranceIdx = resIdx; }

inline void LVFMap::init() const {}

inline bool LVFMap::exportAsFig(char const * fileName, int colorMode,
				bool noFill) const
{
  fileName  = fileName;
  colorMode = colorMode;
  noFill    = noFill;

  return false;
}



#endif // __LVFMAP_HH_
