#include <mlfmap/mlfmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>

typedef GLvoid (* cb_t)();




/* Point Radius */
#define POINT_RADIUS 2.5


/******************************************************************************
 * Terrain types
 *****************************************************************************/
#define MAP_BACKGROUND     0x0
#define POLITICAL_CONTENTS 0x1
#define POLITICAL_CONTOUR  0x2
#define BIG_CITY_CONTENTS  0x3
#define BIG_CITY_CONTOUR   0x4
#define CITY_LOCATION      0x5
#define PERENNIAL_WATER    0x6
#define	NONPERENNIAL_WATER 0x7
#define	WET_SAND           0x8
#define	SNOWFIELD          0x9
#define	RIVER_AND_CONTOUR  0xA
#define	ROAD               0xB
#define	RAILROAD           0xC
#define	AIRPORT            0xD 
#define NUMBER_OF_TYPES   (AIRPORT+1)


/******************************************************************************
 * Internal colors
 *****************************************************************************/
unsigned const colors[NUMBER_OF_TYPES][2][3] = {
  { { 255, 255, 255 }, { 255, 255, 255 } }, /* Map background */
  { { 255, 255, 180 }, { 240, 240, 240 } }, /* Political contents */
  { {   0,   0,   0 }, {   0,   0,   0 } }, /* Political contour */
  { { 230, 180, 180 }, { 200, 200, 200 } }, /* Big city contents */
  { { 255,   0,   0 }, { 130, 130, 130 } }, /* Big city contour */
  { { 255,   0,   0 }, {   0,   0,   0 } }, /* City location */
  { {  75, 155, 255 }, { 155, 155, 155 } }, /* Peren. inland water */
  { { 130, 205, 255 }, { 205, 205, 205 } }, /* Nonp. inland water */
  { { 205, 130, 255 }, { 205, 205, 205 } }, /* Wet sand */
  { { 230, 230, 255 }, { 255, 255, 255 } }, /* Snowfields */
  { {   0,   0, 255 }, {  75,  75,  75 } }, /* River, water contour*/
  { { 130, 130, 130 }, { 130, 130, 130 } }, /* Road */
  { { 205, 180,   0 }, { 180, 180, 180 } }, /* Railroad */
  { {   0, 180,   0 }, { 105, 105, 105 } }, /* Airport */
};


/******************************************************************************
 * Layers' makers definition
 *****************************************************************************/
#define END_FILE 0.0

#define BEGIN_POL_LAYER  1.0
#define   END_POL_LAYER -1.0

#define BEGIN_POP_LAYER  2.0
#define   END_POP_LAYER -2.0

#define BEGIN_DRG_LAYER  3.0
#define   END_DRG_LAYER -3.0

#define BEGIN_ROD_LAYER  4.0
#define   END_ROD_LAYER -4.0

#define BEGIN_RAR_LAYER  5.0
#define   END_RAR_LAYER -5.0

#define BEGIN_APT_LAYER  6.0
#define   END_APT_LAYER -6.0

#define BEGIN_POLYGON_SECTION  10.0
#define   END_POLYGON_SECTION -10.0

#define BEGIN_POINT_SECTION  11.0
#define   END_POINT_SECTION -11.0

#define BEGIN_LINE_SECTION  12.0
#define   END_LINE_SECTION -12.0

#define BEGIN_LINE_STRIP  13.0
#define   END_LINE_STRIP -13.0


/******************************************************************************
 * Internal local auxiliary variables
 *****************************************************************************/
namespace LocalVariablesMLF {

  double xLeft, xRight;
  double yBottom, yTop;

  double xScale, yScale;
  int pointRadius;

  int colorMode;
  bool noFill;
};


using namespace LocalVariablesMLF;



/******************************************************************************
 * mlfMap.read(ostr)
 ******************************************************************************
 * Reads 'mlfMap' from file given by 'fileName'.
 *****************************************************************************/
bool MLFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * mlfMap.read(ostr)
 ******************************************************************************
 * Writes 'mlfMap' in file given by 'fileName'.
 *****************************************************************************/
bool MLFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * mlfMap.init()
 ******************************************************************************
 * Creates all display list of 'mlfMap' and inialize all state variables.
 *****************************************************************************/
void MLFMap::init() const
{
  /* GL initial settings */
  glClearColor(1.0, 1.0, 1.0, 1.0);

  /* Creating display lists */
  int layerIdx = 1;

  if (mlfFile.getPOLLayer() != 0) createPOLList(layerIdx++);
  if (mlfFile.getPOPLayer() != 0) createPOPList(layerIdx++);
  if (mlfFile.getDRGLayer() != 0) createDRGList(layerIdx++);
  if (mlfFile.getRODLayer() != 0) createRODList(layerIdx++);
  if (mlfFile.getRARLayer() != 0) createRARList(layerIdx++);
  if (mlfFile.getAPTLayer() != 0) createAPTList(layerIdx);
}



/******************************************************************************
 * mlfMap.draw()
 ******************************************************************************
 * Draws 'mlfMap' contents, including polylines and feature points.
 *****************************************************************************/
void MLFMap::draw() const
{
  /* Drawing active layers */
  for (unsigned layerIdx = 0; layerIdx < nLayers; ++layerIdx)
    if (layerInfos[layerIdx].active)
      glCallList(layerInfos[layerIdx].ID);
  glPassThrough(END_FILE);
}



/******************************************************************************
 * mlfMap.exportAsFig(fileName, colorMode, noFill)
 ******************************************************************************
 * Export 'mlfMap' contents to a xFig file named 'fileName'.
 *****************************************************************************/
bool MLFMap::exportAsFig(char const * fileName, int colorMode,
			 bool noFill) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 

  /* Setting feedback buffer and entering feedback mode */
  GLfloat feedbackBuffer[2097152];
  glFeedbackBuffer(2097152, GL_2D, feedbackBuffer);
  glRenderMode(GL_FEEDBACK);

  /* Writing contents in feedback buffer and getting number of values writed */
  draw();
  if (glRenderMode(GL_RENDER) == -1) return false;


  /* Getting projection matrix and viewport */
  GLfloat projMatrix[16];
  GLint   viewport[4];

  glGetFloatv  (GL_PROJECTION_MATRIX, projMatrix);
  glGetIntegerv(GL_VIEWPORT,          viewport);

  /* Getting clip planes */
  GLdouble   leftPlane[4];
  GLdouble  rightPlane[4];
  GLdouble bottomPlane[4];
  GLdouble    topPlane[4];

  glGetClipPlane(GL_CLIP_PLANE0,   leftPlane);
  glGetClipPlane(GL_CLIP_PLANE1,  rightPlane);
  glGetClipPlane(GL_CLIP_PLANE2, bottomPlane);
  glGetClipPlane(GL_CLIP_PLANE3,    topPlane);

  /* Setting auxiliary variables */
  xLeft  = viewport[2]*(-projMatrix[0]* leftPlane[3]+projMatrix[12]+1)/2;
  xRight = viewport[2]*( projMatrix[0]*rightPlane[3]+projMatrix[12]+1)/2;

  yBottom = viewport[3]*(-projMatrix[5]*bottomPlane[3]+projMatrix[13]+1)/2;
  yTop    = viewport[3]*( projMatrix[5]*   topPlane[3]+projMatrix[13]+1)/2;

  if ((xRight-xLeft)/(yTop-yBottom) > 11.0/8.5)
    {
      xScale = 11.0*1200;
      yScale = 11.0*1200*(yTop-yBottom)/(xRight-xLeft);

      pointRadius = static_cast<int>(11.0*1200*POINT_RADIUS/(xRight-xLeft));
    }
  else
    {
      xScale = 8.5*1200*(xRight-xLeft)/(yTop-yBottom);
      yScale = 8.5*1200;

      pointRadius = static_cast<int>(8.5*1200*POINT_RADIUS/(yTop-yBottom));
    }

  /* Setting color/fill variables */
  LocalVariablesMLF::colorMode = colorMode;
  LocalVariablesMLF::noFill    = noFill;


  /* Writing buffer into map stream */
  writeBuffer(mapStream, feedbackBuffer);
  return true;
}



/******************************************************************************
 * istr >> mlfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, MLFMap & mlfMap)
{
  /* Reading MLF file */
  istr >> mlfMap.mlfFile;


  /* Initializing bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /***************************** Political layer *****************************/
  if (mlfMap.mlfFile.getPOLLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getPOLLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFPOLLayer const & polLayer = *mlfMap.mlfFile.getPOLLayer();

      if (polLayer.getXMin() < xMin) xMin = polLayer.getXMin();
      if (polLayer.getXMax() > xMax) xMax = polLayer.getXMax();
      if (polLayer.getYMin() < yMin) yMin = polLayer.getYMin();
      if (polLayer.getYMax() > yMax) yMax = polLayer.getYMax();
    }


  /************************* Populated places layer **************************/
  if (mlfMap.mlfFile.getPOPLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getPOPLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFPOPLayer const & popLayer = *mlfMap.mlfFile.getPOPLayer();

      if (popLayer.getXMin() < xMin) xMin = popLayer.getXMin();
      if (popLayer.getXMax() > xMax) xMax = popLayer.getXMax();
      if (popLayer.getYMin() < yMin) yMin = popLayer.getYMin();
      if (popLayer.getYMax() > yMax) yMax = popLayer.getYMax();
    }


  /***************************** Drainage layer ******************************/
  if (mlfMap.mlfFile.getDRGLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getDRGLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFDRGLayer const & drgLayer = *mlfMap.mlfFile.getDRGLayer();

      if (drgLayer.getXMin() < xMin) xMin = drgLayer.getXMin();
      if (drgLayer.getXMax() > xMax) xMax = drgLayer.getXMax();
      if (drgLayer.getYMin() < yMin) yMin = drgLayer.getYMin();
      if (drgLayer.getYMax() > yMax) yMax = drgLayer.getYMax();
    }


  /******************************* Roads layer *******************************/
  if (mlfMap.mlfFile.getRODLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getRODLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFRODLayer const & rodLayer = *mlfMap.mlfFile.getRODLayer();

      if (rodLayer.getXMin() < xMin) xMin = rodLayer.getXMin();
      if (rodLayer.getXMax() > xMax) xMax = rodLayer.getXMax();
      if (rodLayer.getYMin() < yMin) yMin = rodLayer.getYMin();
      if (rodLayer.getYMax() > yMax) yMax = rodLayer.getYMax();
    }


  /***************************** Railroads layer *****************************/
  if (mlfMap.mlfFile.getRARLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getRARLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFRARLayer const & rarLayer = *mlfMap.mlfFile.getRARLayer();

      if (rarLayer.getXMin() < xMin) xMin = rarLayer.getXMin();
      if (rarLayer.getXMax() > xMax) xMax = rarLayer.getXMax();
      if (rarLayer.getYMin() < yMin) yMin = rarLayer.getYMin();
      if (rarLayer.getYMax() > yMax) yMax = rarLayer.getYMax();
    }


  /***************************** Airports layer ******************************/
  if (mlfMap.mlfFile.getAPTLayer() != 0)
    {
      /* Updating layer info vector */
      mlfMap.layerInfos[mlfMap.nLayers].ID     = mlfMap.nLayers+1;
      mlfMap.layerInfos[mlfMap.nLayers].layer  = mlfMap.mlfFile.getAPTLayer();
      mlfMap.layerInfos[mlfMap.nLayers].active = true;
      ++mlfMap.nLayers;

      /* Searching for layer bounding box */
      MLFAPTLayer const & aptLayer = *mlfMap.mlfFile.getAPTLayer();

      if (aptLayer.getXMin() < xMin) xMin = aptLayer.getXMin();
      if (aptLayer.getXMax() > xMax) xMax = aptLayer.getXMax();
      if (aptLayer.getYMin() < yMin) yMin = aptLayer.getYMin();
      if (aptLayer.getYMax() > yMax) yMax = aptLayer.getYMax();
    }


  /* Setting bounding box */
  mlfMap.setXMin(xMin);
  mlfMap.setXMax(xMax);
  mlfMap.setYMin(yMin);
  mlfMap.setYMax(yMax);


  return istr;
}



/******************************************************************************
 * ostr << mlfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MLFMap const & mlfMap)
{
  /* Writing MLF file */
  ostr << mlfMap.mlfFile;


  return ostr;
}



/******************************************************************************
 * vertexMLF(vertex)
 ******************************************************************************
 * Static vertex callback function.
 *****************************************************************************/
void vertexMLF(GLvoid * vertex_data)
{
  MLFPoint const & point = *reinterpret_cast<MLFPoint const *>(vertex_data);
  glVertex2d(point.getX(), point.getY());
}



/******************************************************************************
 * combineMLF(coords, vertex_data, weight, dataOut)
 ******************************************************************************
 * Static combine callback function.
 *****************************************************************************/
void combineMLF(GLdouble coords[3], MLFPoint * vertex_data[4],
		GLfloat  weight[4], MLFPoint **dataOut)
{
  *dataOut = new MLFPoint(coords[0], coords[1]);
}



/******************************************************************************
 * mlfMap.createPOLList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s political layer display list.
 *****************************************************************************/
void MLFMap::createPOLList(int layerIdx) const
{
  MLFPOLLayer const & polLayer = *mlfFile.getPOLLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_POL_LAYER);
  {
    /********************** Political contents polygons **********************/
    /* Defining political contents color */
    glColor3f(1.0, 1.0, 0.7);

    /* Creating tesselator object */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMLF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMLF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing polical contents from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   polLayer.numberOfPolygons(); ++polygonIdx)
      {
	MLFPolygon const & polygon = polLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9) continue;

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    MLFPolyline const & polyline =
	      polLayer.getPolyline(polygon[polylineIdxIdx].polylineIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = polyline.numberOfVertices()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     polyline.numberOfVertices()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator */
    gluDeleteTess(tess);


    /************************ Political contour lines ************************/
    /* Defining political contour width and color */
    glLineWidth(1.0);
    glColor3f(0.0, 0.0, 0.0);

    /* Drawing contour from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   polLayer.numberOfPolylines(); ++polylineIdx)
      {
	MLFPolyline const & polyline = polLayer.getPolyline(polylineIdx);

	

// 	glPointSize(3*POINT_RADIUS);
// 	glColor3f(1.0, 0.0, 1.0);

// 	/* Drawing small cities from points */
// 	glPassThrough(BEGIN_LINE_STRIP);
// 	glBegin(GL_POINTS);
// 	{
// 	  glVertex2d(polyline.front().getX(),
// 		     polyline.front().getY());
// 	  glVertex2d(polyline.back().getX(),
// 		     polyline.back().getY());
// 	}
// 	glEnd();
// 	glPassThrough(END_LINE_STRIP);
	

	glColor3f(0.0, 0.0, 0.0);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx = 0; vertexIdx <
		 polyline.numberOfVertices(); ++vertexIdx)

	    glVertex2d(polyline[vertexIdx].getX(),
		       polyline[vertexIdx].getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_POL_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.createPOPList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s populated places layer display list.
 *****************************************************************************/
void MLFMap::createPOPList(int layerIdx) const
{
  MLFPOPLayer const & popLayer = *mlfFile.getPOPLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_POP_LAYER);
  {
    /********************* Big cities' contents polygons *********************/
    /* Defining big cities' contents color */
    glColor3f(0.9, 0.6, 0.6);

    /* Creating tesselator object and setting callback functions */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMLF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMLF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing big cities from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   popLayer.numberOfPolygons(); ++polygonIdx)
      {
	MLFPolygon const & polygon = popLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9) continue;

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    MLFPolyline const & polyline =
	      popLayer.getPolyline(polygon[polylineIdxIdx].polylineIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = polyline.numberOfVertices()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     polyline.numberOfVertices()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator object */
    gluDeleteTess(tess);


    /*********************** Big cities' contour lines ***********************/
    /* Defining big cities' contour width and color */
    glLineWidth(1.0);
    glColor3f(1.0, 0.0, 0.0);

    /* Drawing big cities from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   popLayer.numberOfPolylines(); ++polylineIdx)
      {
	MLFPolyline const & polyline = popLayer.getPolyline(polylineIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx = 0; vertexIdx <
		 polyline.numberOfVertices(); ++vertexIdx)

	    glVertex2d(polyline[vertexIdx].getX(),
		       polyline[vertexIdx].getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);


    /****************************** City points ******************************/
    /* Defining city points' size and color */
    glPointSize(2*POINT_RADIUS);
    glColor3f(1.0, 0.0, 0.0);

    /* Drawing small cities from points */
    glPassThrough(BEGIN_POINT_SECTION);
    glBegin(GL_POINTS);
    {
      for (unsigned pointIdx = 0; pointIdx <
	     popLayer.numberOfPoints(); ++pointIdx)

	glVertex2d(popLayer.getPoint(pointIdx).getX(),
		   popLayer.getPoint(pointIdx).getY());
    }
    glEnd();
    glPassThrough(END_POINT_SECTION);
  }
  glPassThrough(END_POP_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.createDRGList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s drainage layer display list.
 *****************************************************************************/
void MLFMap::createDRGList(int layerIdx) const
{
  MLFDRGLayer const & drgLayer = *mlfFile.getDRGLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_DRG_LAYER);
  {
    /*************************** Drainage contents ***************************/
    /* Creating tesselator object and setting callback functions */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMLF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMLF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing drainage contents from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   drgLayer.numberOfPolygons(); ++polygonIdx)
      {
	MLFPolygon const & polygon = drgLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9 ||
	    polygon.getType() == 0) continue;

	if      (polygon.getType() == 2) glColor3f(0.5, 0.8, 1.0);
	else if (polygon.getType() == 3) glColor3f(0.8, 0.5, 1.0);
	else if (polygon.getType() == 4) glColor3f(0.9, 0.9, 1.0);
	else                             glColor3f(0.3, 0.6, 1.0);

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    MLFPolyline const & polyline =
	      drgLayer.getPolyline(polygon[polylineIdxIdx].polylineIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = polyline.numberOfVertices()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     polyline.numberOfVertices()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { polyline[vertexIdx].getX(),
					 polyline[vertexIdx].getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MLFPoint *>
				 (&polyline[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator */
    gluDeleteTess(tess);


    /************************ Rivers and contour lines ***********************/
    /* Defining rivers and contours' width and color */
    glLineWidth(1.0);
    glColor3f(0.0, 0.0, 1.0);

    /* Drawing rivers and rivers' contours from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   drgLayer.numberOfPolylines(); ++polylineIdx)
      {
	MLFPolyline const & polyline = drgLayer.getPolyline(polylineIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx = 0; vertexIdx <
		 polyline.numberOfVertices(); ++vertexIdx)

	    glVertex2d(polyline[vertexIdx].getX(),
		       polyline[vertexIdx].getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_DRG_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.createRODList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s roads layer display list.
 *****************************************************************************/
void MLFMap::createRODList(int layerIdx) const
{
  MLFRODLayer const & rodLayer = *mlfFile.getRODLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_ROD_LAYER);
  {
    /******************************* Road lines ******************************/
    /* Defining roads width and color */
    glLineWidth(2.0);
    glColor3f(0.5, 0.5, 0.5);

    /* Drawing roads from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   rodLayer.numberOfPolylines(); ++polylineIdx)
      {
	MLFPolyline const & polyline = rodLayer.getPolyline(polylineIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx = 0; vertexIdx <
		 polyline.numberOfVertices(); ++vertexIdx)

	    glVertex2d(polyline[vertexIdx].getX(),
		       polyline[vertexIdx].getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_ROD_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.createRARList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s railroads layer display list.
 *****************************************************************************/
void MLFMap::createRARList(int layerIdx) const
{
  MLFRARLayer const & rarLayer = *mlfFile.getRARLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_RAR_LAYER);
  {
    /***************************** Railroad lines ****************************/
    /* Defining railroads width and color */
    glLineWidth(2.0);
    glColor3f(0.8, 0.7, 0.0);

    /* Drawing railroads from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   rarLayer.numberOfPolylines(); ++polylineIdx)
      {
	MLFPolyline const & polyline = rarLayer.getPolyline(polylineIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx = 0; vertexIdx <
		 polyline.numberOfVertices(); ++vertexIdx)

	    glVertex2d(polyline[vertexIdx].getX(),
		       polyline[vertexIdx].getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_RAR_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.createAPTList(layerIdx)
 ******************************************************************************
 * Create 'mlfMap''s airports layer display list.
 *****************************************************************************/
void MLFMap::createAPTList(int layerIdx) const
{
  MLFAPTLayer const & aptLayer = *mlfFile.getAPTLayer();

  /* Defining display list */
  glNewList(layerIdx, GL_COMPILE);
  glPassThrough(BEGIN_APT_LAYER);
  {
    /***************************** Airport points ****************************/
    /* Defining airports size and color */
    glPointSize(2*POINT_RADIUS);
    glColor3f(0.0, 0.7, 0.2);

    /* Drawing airports from points */
    glPassThrough(BEGIN_POINT_SECTION);
    glBegin(GL_POINTS);
    {
      for (unsigned pointIdx = 0; pointIdx <
	     aptLayer.numberOfPoints(); ++pointIdx)

	glVertex2d(aptLayer.getPoint(pointIdx).getX(),
		   aptLayer.getPoint(pointIdx).getY());
    }
    glEnd();
    glPassThrough(END_POINT_SECTION);
  }
  glPassThrough(END_APT_LAYER);
  glEndList();
}



/******************************************************************************
 * mlfMap.writeBuffer(stream, buffer)
 ******************************************************************************
 * 
 *****************************************************************************/
void MLFMap::writeBuffer(ostream & ostr, float * buffer) const
{
  /* Writing Xfig header */
  ostr << "#FIG 3.2"  << endl
       << "Landscape" << endl
       << "Center"    << endl
       << "Inches"    << endl
       << "Letter  "  << endl
       << "100.00"    << endl
       << "Single"    << endl
       << "-2"        << endl
       << "1200 2"    << endl;

  /* Writing colors */
  ostr << setfill('0') << dec;
  for (unsigned colorIdx = 1; colorIdx <= AIRPORT; ++colorIdx)
    ostr << "0 " << 31+colorIdx << " #" << hex
	 << setw(2) << colors[colorIdx][colorMode][0]
	 << setw(2) << colors[colorIdx][colorMode][1]
	 << setw(2) << colors[colorIdx][colorMode][2] << endl << dec;


  /* Writing layers */
  GLfloat token;

  while (token = *buffer++, token != END_FILE)
    if      (token == GL_PASS_THROUGH_TOKEN) ;
    else if (token == BEGIN_POL_LAYER) writePOLLayer(ostr, buffer);
    else if (token == BEGIN_POP_LAYER) writePOPLayer(ostr, buffer);
    else if (token == BEGIN_DRG_LAYER) writeDRGLayer(ostr, buffer);
    else if (token == BEGIN_ROD_LAYER) writeRODLayer(ostr, buffer);
    else if (token == BEGIN_RAR_LAYER) writeRARLayer(ostr, buffer);
    else if (token == BEGIN_APT_LAYER) writeAPTLayer(ostr, buffer);
    else break;
}



/******************************************************************************
 * mlfMap.writePOLLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writePOLLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POL_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, POLITICAL_CONTENTS);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, POLITICAL_CONTOUR);

    else break;
}



/******************************************************************************
 * mlfMap.writePOPLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writePOPLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POP_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, BIG_CITY_CONTENTS);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, BIG_CITY_CONTOUR);

    else if (token == BEGIN_POINT_SECTION)
      writePoints(ostr, buffer, CITY_LOCATION);

    else break;
}



/******************************************************************************
 * mlfMap.writeDRGLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeDRGLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_DRG_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, PERENNIAL_WATER);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, RIVER_AND_CONTOUR);

    else break;
}



/******************************************************************************
 * mlfMap.writeRODLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeRODLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_ROD_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, ROAD);

    else break;
}



/******************************************************************************
 * mlfMap.writeRARLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeRARLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_RAR_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, RAILROAD);

    else break;
}



/******************************************************************************
 * mlfMap.writeAPTLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeAPTLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_APT_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POINT_SECTION)
      writePoints(ostr, buffer, AIRPORT);

    else break;
}



/******************************************************************************
 * mlfMap.writePolygons(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writePolygons(ostream & ostr, float *& buffer, int type) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POLYGON_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_POLYGON_TOKEN)
      {
	unsigned nVertices = static_cast<unsigned>(*buffer++);

	if (noFill == false)
	  ostr << "2 1 0 0 0 " << 31+type
	       << " 50 -1 20 0.000 0 0 -1 0 0 " << nVertices;

	for (unsigned vertexIdx = 0; vertexIdx < nVertices; ++vertexIdx)
	  {
	    GLfloat x = *buffer++;
	    GLfloat y = *buffer++;

	    if (noFill) continue;

	    if (vertexIdx%6 == 0) ostr << endl << '\t';

	    ostr << ' ' << static_cast<int>(xScale*(x-xLeft)/(xRight-xLeft))
		 << ' ' << static_cast<int>(yScale*(yTop-y) /(yTop-yBottom));
	  }
	if (noFill == false) ostr << endl;
      }

    else break;
}



/******************************************************************************
 * mlfMap.writeLines(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeLines(ostream & ostr, float *& buffer, int type) const
{
  /* Counting number of vertices */
  GLfloat token;

  unsigned nLines = 0;
  float * bufferBU = 0;

  GLfloat prevX2 = 0.0;
  GLfloat prevY2 = 0.0;

  while (token = *buffer++, token != END_LINE_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_LINE_TOKEN ||
	     token == GL_LINE_RESET_TOKEN)
      {
	GLfloat x1 = *buffer++, y1 = *buffer++;
	GLfloat x2 = *buffer++, y2 = *buffer++;

	if (nLines > 0 && (x1 != prevX2 ||
			   y1 != prevY2))
	  {
	    writeLineStrip(ostr, bufferBU, nLines, type);

	    nLines = 1;
	    bufferBU = buffer-5;
	  }
	else
	  {
	    prevX2 = x2;
	    prevY2 = y2;

	    ++nLines;
	  }
      }

    else if (token == BEGIN_LINE_STRIP)
      {
	nLines = 0;
	bufferBU = buffer;
      }

    else if (token == END_LINE_STRIP)
      {
	if (nLines > 0)
	  writeLineStrip(ostr, bufferBU, nLines, type);
      }

    else break;
}



/******************************************************************************
 * mlfMap.writePoints(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writePoints(ostream & ostr, float *& buffer, int type) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POINT_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_POINT_TOKEN)
      { 
	ostr << "2 2 0 0 0 " << 31+type << " 50 -1 20 0.000 0 0 -1 0 0 5"
	     << endl << '\t';

	GLfloat x = *buffer++;
	GLfloat y = *buffer++;

	int xInt = static_cast<int>(xScale*(x-xLeft)/(xRight-xLeft));
	int yInt = static_cast<int>(yScale*(yTop-y) /(yTop-yBottom));

	ostr << ' ' << xInt-pointRadius << ' ' << yInt-pointRadius
	     << ' ' << xInt+pointRadius << ' ' << yInt-pointRadius
	     << ' ' << xInt+pointRadius << ' ' << yInt+pointRadius
	     << ' ' << xInt-pointRadius << ' ' << yInt+pointRadius
	     << ' ' << xInt-pointRadius << ' ' << yInt-pointRadius << endl;
      }

    else break;
}



/******************************************************************************
 * mlfMap.writeLineStrip(ostr, buffer, nLines, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MLFMap::writeLineStrip(ostream & ostr, float * buffer,
			    unsigned nLines, int type) const
{
  /* Writing header */
  if (type != ROAD && type != RAILROAD)
    ostr << "2 1 0 1 " << 31+type
	 << " 0 50 -1 -1 0.000 1 1 -1 0 0 " << nLines+1;

  else if (colorMode == FULL_COLOR_MODE)
    ostr << "2 1 0 2 " << 31+type
	 << " 0 50 -1 -1 0.000 1 1 -1 0 0 " << nLines+1;

  else
    ostr << "2 1 " << (type == ROAD ? 3 : 1) << " 2 " << 31+type
	 << " 0 50 -1 -1 "
	 << setprecision(3) << fixed << setw(4) << pointRadius/10.0
	 << " 1 1 -1 0 0 " << nLines+1;


  /* Writing vertices */
  GLfloat x1 = 0.0, y1 = 0.0;
  GLfloat x2 = 0.0, y2 = 0.0;

  for (unsigned vertexIdx = 0; vertexIdx < nLines; ++vertexIdx)
    {
      /* Skipping line token */
      ++buffer;

      if (vertexIdx%6 == 0) ostr << endl << '\t';

      x1 = *buffer++; y1 = *buffer++;
      x2 = *buffer++; y2 = *buffer++;

      ostr << ' ' << static_cast<int>(xScale*(x1-xLeft)/(xRight-xLeft))
	   << ' ' << static_cast<int>(yScale*(yTop-y1) /(yTop-yBottom));
    }


  /* Writing last vertex */
  if (nLines%6 == 0) ostr << endl << '\t';

  ostr << ' ' << static_cast<int>(xScale*(x2-xLeft)/(xRight-xLeft))
       << ' ' << static_cast<int>(yScale*(yTop-y2) /(yTop-yBottom)) << endl;
}
