#ifndef __MLFMAP_HH_
#define __MLFMAP_HH_


#include <map.hh>
#include <mlf/mlffile.hh>

#include <iostream>


using namespace std;



/******************************************************************************
 * Class MLFMap
 *****************************************************************************/
class MLFMap
  : public Map {

public:

  /* Constructor */
  MLFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill = false) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, MLFMap       & mlfMap);
  friend ostream & operator<<(ostream & ostr, MLFMap const & mlfMap);


private:

  /* Layer information structure */
  struct layerInfo_t
  {
    unsigned ID;
    MLFLayer const * layer;
    bool active;
  };

  /* Number of layers and array of layers informations */
  unsigned nLayers;
  layerInfo_t layerInfos[6];

  /* Vectors of polylines and feature points */
  MLFFile mlfFile;

  /* Internal display lists creation functions */
  void createPOLList(int layerIdx) const;
  void createPOPList(int layerIdx) const;
  void createDRGList(int layerIdx) const;
  void createRODList(int layerIdx) const;
  void createRARList(int layerIdx) const;
  void createAPTList(int layerIdx) const;

  /* Internal functions for writing layers in Xfig format */
  void writeBuffer(ostream & ostr, float * buffer) const;

  void writePOLLayer(ostream & ostr, float *& buffer) const;
  void writePOPLayer(ostream & ostr, float *& buffer) const;
  void writeDRGLayer(ostream & ostr, float *& buffer) const;
  void writeRODLayer(ostream & ostr, float *& buffer) const;
  void writeRARLayer(ostream & ostr, float *& buffer) const;
  void writeAPTLayer(ostream & ostr, float *& buffer) const;

  void writePolygons(ostream & ostr, float *& buffer, int type) const;
  void writeLines   (ostream & ostr, float *& buffer, int type) const;
  void writePoints  (ostream & ostr, float *& buffer, int type) const;

  void writeLineStrip(ostream & ostr, float * buffer,
		      unsigned nLines, int type) const;
};


inline MLFMap::MLFMap() : nLayers(0) {}

inline char const * MLFMap::getName() const
{
  return mlfFile.getName().c_str();
}

inline unsigned MLFMap::getNumberOfLayers() const
{
  return nLayers;
}

inline char const * MLFMap::getLayerName(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return "";

  return layerInfos[layerIdx].layer->getTitle().c_str();
}

inline bool MLFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return false;

  return layerInfos[layerIdx].active;
}

inline void MLFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return;

  layerInfos[layerIdx].active = activated;
}

inline double MLFMap::getTolerance() const { return 0.0; }
inline void   MLFMap::setTolerance(double tolerance) { tolerance = tolerance; }

inline unsigned MLFMap::getNumberOfResolutions() const { return 1; }
inline unsigned MLFMap::getResolution() const { return 0; }
inline void     MLFMap::setResolution(unsigned resIdx) { resIdx = resIdx; }



#endif // __MLFMAP_HH_
