#include <mmfmap/mmfmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>

typedef GLvoid (* cb_t)();




/* Point Radius */
#define POINT_RADIUS 2.5


/******************************************************************************
 * Terrain types
 *****************************************************************************/
#define MAP_BACKGROUND     0x0
#define POLITICAL_CONTENTS 0x1
#define POLITICAL_CONTOUR  0x2
#define BIG_CITY_CONTENTS  0x3
#define BIG_CITY_CONTOUR   0x4
#define CITY_LOCATION      0x5
#define PERENNIAL_WATER    0x6
#define	NONPERENNIAL_WATER 0x7
#define	WET_SAND           0x8
#define	SNOWFIELD          0x9
#define	RIVER_AND_CONTOUR  0xA
#define	ROAD               0xB
#define	RAILROAD           0xC
#define	AIRPORT            0xD 
#define NUMBER_OF_TYPES   (AIRPORT+1)


/******************************************************************************
 * Internal colors
 *****************************************************************************/
unsigned const colors[NUMBER_OF_TYPES][2][3] = {
  { { 255, 255, 255 }, { 255, 255, 255 } }, /* Map background */
  { { 255, 255, 180 }, { 240, 240, 240 } }, /* Political contents */
  { {   0,   0,   0 }, {   0,   0,   0 } }, /* Political contour */
  { { 230, 180, 180 }, { 200, 200, 200 } }, /* Big city contents */
  { { 255,   0,   0 }, { 130, 130, 130 } }, /* Big city contour */
  { { 255,   0,   0 }, {   0,   0,   0 } }, /* City location */
  { {  75, 155, 255 }, { 155, 155, 155 } }, /* Peren. inland water */
  { { 130, 205, 255 }, { 205, 205, 205 } }, /* Nonp. inland water */
  { { 205, 130, 255 }, { 205, 205, 205 } }, /* Wet sand */
  { { 230, 230, 255 }, { 255, 255, 255 } }, /* Snowfields */
  { {   0,   0, 255 }, {  75,  75,  75 } }, /* River, water contour*/
  { { 130, 130, 130 }, { 130, 130, 130 } }, /* Road */
  { { 205, 180,   0 }, { 180, 180, 180 } }, /* Railroad */
  { {   0, 180,   0 }, { 105, 105, 105 } }, /* Airport */
};


/******************************************************************************
 * Layers' makers definition
 *****************************************************************************/
#define END_FILE 0.0

#define BEGIN_POL_LAYER  1.0
#define   END_POL_LAYER -1.0

#define BEGIN_POP_LAYER  2.0
#define   END_POP_LAYER -2.0

#define BEGIN_DRG_LAYER  3.0
#define   END_DRG_LAYER -3.0

#define BEGIN_ROD_LAYER  4.0
#define   END_ROD_LAYER -4.0

#define BEGIN_RAR_LAYER  5.0
#define   END_RAR_LAYER -5.0

#define BEGIN_APT_LAYER  6.0
#define   END_APT_LAYER -6.0

#define BEGIN_POLYGON_SECTION  10.0
#define   END_POLYGON_SECTION -10.0

#define BEGIN_POINT_SECTION  11.0
#define   END_POINT_SECTION -11.0

#define BEGIN_LINE_SECTION  12.0
#define   END_LINE_SECTION -12.0

#define BEGIN_LINE_STRIP  13.0
#define   END_LINE_STRIP -13.0


/******************************************************************************
 * Internal auxiliary variables
 *****************************************************************************/
namespace LocalVariablesMMF {

  double xLeft, xRight;
  double yBottom, yTop;

  double xScale, yScale;
  int pointRadius;

  int colorMode;
  bool noFill;
};


using namespace LocalVariablesMMF;



/******************************************************************************
 * mmfMap.read(ostr)
 ******************************************************************************
 * Reads 'mmfMap' from file given by 'fileName'.
 *****************************************************************************/
bool MMFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * mmfMap.read(ostr)
 ******************************************************************************
 * Writes 'mmfMap' in file given by 'fileName'.
 *****************************************************************************/
bool MMFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * mmfMap.init()
 ******************************************************************************
 * Creates all display list of 'mmfMap' and inialize all state variables.
 *****************************************************************************/
void MMFMap::init() const
{
  /* GL initial settings */
  glClearColor(1.0, 1.0, 1.0, 1.0);

  /* Creating display lists */
  int listIdx = 1;

  for (unsigned resolIdx = 0; resolIdx <
	 mmfFile.numberOfResolutions(); ++resolIdx)
    {
      if (mmfFile.getPOLLayer() != 0) createPOLList(resolIdx, listIdx++);
      if (mmfFile.getPOPLayer() != 0) createPOPList(resolIdx, listIdx++);
      if (mmfFile.getDRGLayer() != 0) createDRGList(resolIdx, listIdx++);
      if (mmfFile.getRODLayer() != 0) createRODList(resolIdx, listIdx++);
      if (mmfFile.getRARLayer() != 0) createRARList(resolIdx, listIdx++);
      if (mmfFile.getAPTLayer() != 0) createAPTList(resolIdx, listIdx++);
    }
}



/******************************************************************************
 * mmfMap.draw()
 ******************************************************************************
 * Draws 'mmfMap' contents, including polylines and feature points.
 *****************************************************************************/
void MMFMap::draw() const
{
  /* Drawing active layers */
  for (unsigned layerIdx = 0; layerIdx < nLayers; ++layerIdx)
    if (layerInfos[layerIdx].active)
      glCallList(currResolIdx*nLayers+layerInfos[layerIdx].ID);

  glPassThrough(END_FILE);
}



/******************************************************************************
 * mmfMap.exportAsFig(fileName, colorMode, noFill)
 ******************************************************************************
 * Export 'mmfMap' contents to a xFig file named 'fileName'.
 *****************************************************************************/
bool MMFMap::exportAsFig(char const * fileName, int colorMode,
			 bool noFill) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 

  /* Setting feedback buffer and entering feedback mode */
  GLfloat feedbackBuffer[2097152];
  glFeedbackBuffer(2097152, GL_2D, feedbackBuffer);
  glRenderMode(GL_FEEDBACK);

  /* Writing contents in feedback buffer and getting number of values writed */
  draw();
  if (glRenderMode(GL_RENDER) == -1) return false;


  /* Getting projection matrix and viewport */
  GLfloat projMatrix[16];
  GLint   viewport[4];

  glGetFloatv  (GL_PROJECTION_MATRIX, projMatrix);
  glGetIntegerv(GL_VIEWPORT,          viewport);

  /* Getting clip planes */
  GLdouble   leftPlane[4];
  GLdouble  rightPlane[4];
  GLdouble bottomPlane[4];
  GLdouble    topPlane[4];

  glGetClipPlane(GL_CLIP_PLANE0,   leftPlane);
  glGetClipPlane(GL_CLIP_PLANE1,  rightPlane);
  glGetClipPlane(GL_CLIP_PLANE2, bottomPlane);
  glGetClipPlane(GL_CLIP_PLANE3,    topPlane);

  /* Setting auxiliary variables */
  xLeft  = viewport[2]*(-projMatrix[0]* leftPlane[3]+projMatrix[12]+1)/2;
  xRight = viewport[2]*( projMatrix[0]*rightPlane[3]+projMatrix[12]+1)/2;

  yBottom = viewport[3]*(-projMatrix[5]*bottomPlane[3]+projMatrix[13]+1)/2;
  yTop    = viewport[3]*( projMatrix[5]*   topPlane[3]+projMatrix[13]+1)/2;

  if ((xRight-xLeft)/(yTop-yBottom) > 11.0/8.5)
    {
      xScale = 11.0*1200;
      yScale = 11.0*1200*(yTop-yBottom)/(xRight-xLeft);

      pointRadius = static_cast<int>(11.0*1200*POINT_RADIUS/(xRight-xLeft));
    }
  else
    {
      xScale = 8.5*1200*(xRight-xLeft)/(yTop-yBottom);
      yScale = 8.5*1200;

      pointRadius = static_cast<int>(8.5*1200*POINT_RADIUS/(yTop-yBottom));
    }

  /* Setting color/fill variables */
  LocalVariablesMMF::colorMode = colorMode;
  LocalVariablesMMF::noFill    = noFill;


  /* Writing buffer into map stream */
  writeBuffer(mapStream, feedbackBuffer);
  return true;
}



/******************************************************************************
 * istr >> mmfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, MMFMap & mmfMap)
{
  /* Reading MMF file */
  istr >> mmfMap.mmfFile;


  /* Initializing bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /***************************** Political layer *****************************/
  if (mmfMap.mmfFile.getPOLLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getPOLLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFPOLLayer const & polLayer = *mmfMap.mmfFile.getPOLLayer();

      if (polLayer.getXMin() < xMin) xMin = polLayer.getXMin();
      if (polLayer.getXMax() > xMax) xMax = polLayer.getXMax();
      if (polLayer.getYMin() < yMin) yMin = polLayer.getYMin();
      if (polLayer.getYMax() > yMax) yMax = polLayer.getYMax();
    }


  /************************* Populated places layer **************************/
  if (mmfMap.mmfFile.getPOPLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getPOPLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFPOPLayer const & popLayer = *mmfMap.mmfFile.getPOPLayer();

      if (popLayer.getXMin() < xMin) xMin = popLayer.getXMin();
      if (popLayer.getXMax() > xMax) xMax = popLayer.getXMax();
      if (popLayer.getYMin() < yMin) yMin = popLayer.getYMin();
      if (popLayer.getYMax() > yMax) yMax = popLayer.getYMax();
    }


  /***************************** Drainage layer ******************************/
  if (mmfMap.mmfFile.getDRGLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getDRGLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFDRGLayer const & drgLayer = *mmfMap.mmfFile.getDRGLayer();

      if (drgLayer.getXMin() < xMin) xMin = drgLayer.getXMin();
      if (drgLayer.getXMax() > xMax) xMax = drgLayer.getXMax();
      if (drgLayer.getYMin() < yMin) yMin = drgLayer.getYMin();
      if (drgLayer.getYMax() > yMax) yMax = drgLayer.getYMax();
    }


  /******************************* Roads layer *******************************/
  if (mmfMap.mmfFile.getRODLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getRODLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFRODLayer const & rodLayer = *mmfMap.mmfFile.getRODLayer();

      if (rodLayer.getXMin() < xMin) xMin = rodLayer.getXMin();
      if (rodLayer.getXMax() > xMax) xMax = rodLayer.getXMax();
      if (rodLayer.getYMin() < yMin) yMin = rodLayer.getYMin();
      if (rodLayer.getYMax() > yMax) yMax = rodLayer.getYMax();
    }


  /***************************** Railroads layer *****************************/
  if (mmfMap.mmfFile.getRARLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getRARLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFRARLayer const & rarLayer = *mmfMap.mmfFile.getRARLayer();

      if (rarLayer.getXMin() < xMin) xMin = rarLayer.getXMin();
      if (rarLayer.getXMax() > xMax) xMax = rarLayer.getXMax();
      if (rarLayer.getYMin() < yMin) yMin = rarLayer.getYMin();
      if (rarLayer.getYMax() > yMax) yMax = rarLayer.getYMax();
    }


  /***************************** Airports layer ******************************/
  if (mmfMap.mmfFile.getAPTLayer() != 0)
    {
      /* Updating layer info vector */
      mmfMap.layerInfos[mmfMap.nLayers].ID     = mmfMap.nLayers+1;
      mmfMap.layerInfos[mmfMap.nLayers].layer  = mmfMap.mmfFile.getAPTLayer();
      mmfMap.layerInfos[mmfMap.nLayers].active = true;
      ++mmfMap.nLayers;

      /* Searching for layer bounding box */
      MMFAPTLayer const & aptLayer = *mmfMap.mmfFile.getAPTLayer();

      if (aptLayer.getXMin() < xMin) xMin = aptLayer.getXMin();
      if (aptLayer.getXMax() > xMax) xMax = aptLayer.getXMax();
      if (aptLayer.getYMin() < yMin) yMin = aptLayer.getYMin();
      if (aptLayer.getYMax() > yMax) yMax = aptLayer.getYMax();
    }


  /* Setting bounding box */
  mmfMap.setXMin(xMin);
  mmfMap.setXMax(xMax);
  mmfMap.setYMin(yMin);
  mmfMap.setYMax(yMax);


  return istr;
}



/******************************************************************************
 * ostr << mmfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MMFMap const & mmfMap)
{
  /* Writing MMF file */
  ostr << mmfMap.mmfFile;


  return ostr;
}



/******************************************************************************
 * vertexMMF(vertex)
 ******************************************************************************
 * Static vertex callback function.
 *****************************************************************************/
void vertexMMF(GLvoid * vertex_data)
{
  MMFPoint const & point = *reinterpret_cast<MMFPoint const *>(vertex_data);
  glVertex2d(point.getX(), point.getY());
}



/******************************************************************************
 * combineMMF(coords, vertex_data, weight, dataOut)
 ******************************************************************************
 * Static combine callback function.
 *****************************************************************************/
void combineMMF(GLdouble coords[3], MMFPoint * vertex_data[4],
		GLfloat  weight[4], MMFPoint **dataOut)
{
  *dataOut = new MMFPoint(coords[0], coords[1]);
}



/******************************************************************************
 * mmfMap.createPOLList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s political layer display list.
 *****************************************************************************/
void MMFMap::createPOLList(int resolutionIdx, int listIdx) const
{
  MMFPOLLayer const & polLayer = *mmfFile.getPOLLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_POL_LAYER);
  {
    /********************** Political contents polygons **********************/
    /* Defining political contents color */
    glColor3f(1.0, 1.0, 0.7);

    /* Creating tesselator object */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMMF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMMF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing polical contents from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   polLayer.numberOfPolygons(); ++polygonIdx)
      {
	MMFPolygon const & polygon = polLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9) continue;

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    vector<MMFPoint const *> const & vertices = polLayer.getPolyline
	      (polygon[polylineIdxIdx].polylineIdx).getVertices(resolutionIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = vertices.size()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     vertices.size()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator */
    gluDeleteTess(tess);


    /************************ Political contour lines ************************/
    /* Defining political contour width and color */
    glLineWidth(1.0);
    glColor3f(0.0, 0.0, 0.0);

    /* Drawing contour from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   polLayer.numberOfPolylines(); ++polylineIdx)
      {
	vector<MMFPoint const *> const & vertices =
	  polLayer.getPolyline(polylineIdx).getVertices(resolutionIdx);

	glColor3f(0.0, 0.0, 0.0);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx =0; vertexIdx < vertices.size(); ++vertexIdx)
	    glVertex2d(vertices[vertexIdx]->getX(),
		       vertices[vertexIdx]->getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_POL_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.createPOPList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s populated places layer display list.
 *****************************************************************************/
void MMFMap::createPOPList(int resolutionIdx, int listIdx) const
{
  MMFPOPLayer const & popLayer = *mmfFile.getPOPLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_POP_LAYER);
  {
    /********************* Big cities' contents polygons *********************/
    /* Defining big cities' contents color */
    glColor3f(0.9, 0.6, 0.6);

    /* Creating tesselator object and setting callback functions */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMMF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMMF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing big cities from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   popLayer.numberOfPolygons(); ++polygonIdx)
      {
	MMFPolygon const & polygon = popLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9) continue;

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    vector<MMFPoint const *> const & vertices = popLayer.getPolyline
	      (polygon[polylineIdxIdx].polylineIdx).getVertices(resolutionIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = vertices.size()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     vertices.size()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator object */
    gluDeleteTess(tess);


    /*********************** Big cities' contour lines ***********************/
    /* Defining big cities' contour width and color */
    glLineWidth(1.0);
    glColor3f(1.0, 0.0, 0.0);

    /* Drawing big cities from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   popLayer.numberOfPolylines(); ++polylineIdx)
      {
	vector<MMFPoint const *> const & vertices =
	  popLayer.getPolyline(polylineIdx).getVertices(resolutionIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx =0; vertexIdx < vertices.size(); ++vertexIdx)
	    glVertex2d(vertices[vertexIdx]->getX(),
		       vertices[vertexIdx]->getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);


    /****************************** City points ******************************/
    /* Defining city points' size and color */
    glPointSize(2*POINT_RADIUS);
    glColor3f(1.0, 0.0, 0.0);

    /* Drawing small cities from points */
    glPassThrough(BEGIN_POINT_SECTION);
    glBegin(GL_POINTS);
    {
      for (unsigned pointIdx = 0; pointIdx <
	     popLayer.numberOfPoints(); ++pointIdx)

	glVertex2d(popLayer.getPoint(pointIdx).getX(),
		   popLayer.getPoint(pointIdx).getY());
    }
    glEnd();
    glPassThrough(END_POINT_SECTION);
  }
  glPassThrough(END_POP_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.createDRGList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s drainage layer display list.
 *****************************************************************************/
void MMFMap::createDRGList(int resolutionIdx, int listIdx) const
{
  MMFDRGLayer const & drgLayer = *mmfFile.getDRGLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_DRG_LAYER);
  {
    /*************************** Drainage contents ***************************/
    /* Creating tesselator object and setting callback functions */
    GLUtesselator * tess = gluNewTess();

    /* Setting tesselator callback functions */
    gluTessCallback(tess, GLU_TESS_BEGIN,  reinterpret_cast<cb_t>(glBegin));
    gluTessCallback(tess, GLU_TESS_END,    reinterpret_cast<cb_t>(glEnd));
    gluTessCallback(tess, GLU_TESS_VERTEX, reinterpret_cast<cb_t>(vertexMMF));
    gluTessCallback(tess, GLU_TESS_COMBINE,reinterpret_cast<cb_t>(combineMMF));

    /* Setting tesselator properties */
    gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD);

    /* Drawing drainage contents from polygons */
    glPassThrough(BEGIN_POLYGON_SECTION);
    for (unsigned polygonIdx = 0; polygonIdx <
	   drgLayer.numberOfPolygons(); ++polygonIdx)
      {
	MMFPolygon const & polygon = drgLayer.getPolygon(polygonIdx);
	if (polygon.getType() == 9 ||
	    polygon.getType() == 0) continue;

	if      (polygon.getType() == 2) glColor3f(0.5, 0.8, 1.0);
	else if (polygon.getType() == 3) glColor3f(0.8, 0.5, 1.0);
	else if (polygon.getType() == 4) glColor3f(0.9, 0.9, 1.0);
	else                             glColor3f(0.3, 0.6, 1.0);

	gluTessBeginPolygon(tess, 0);
	gluTessBeginContour(tess);
	for (unsigned polylineIdxIdx = 0; polylineIdxIdx <
	       polygon.numberOfPolylines(); ++polylineIdxIdx)
	  {
	    if (polygon[polylineIdxIdx].polylineIdx == -1)
	      {
		gluTessEndContour(tess);
		gluTessBeginContour(tess);
		continue;
	      }

	    vector<MMFPoint const *> const & vertices = drgLayer.getPolyline
	      (polygon[polylineIdxIdx].polylineIdx).getVertices(resolutionIdx);

	    if (polygon[polylineIdxIdx].reversed)
	      for (unsigned vertexIdx = vertices.size()-1;
		   vertexIdx > 0; --vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}

	    else
	      for (unsigned vertexIdx = 0; vertexIdx <
		     vertices.size()-1; ++vertexIdx)
		{
		  GLdouble vertex[3] = { vertices[vertexIdx]->getX(),
					 vertices[vertexIdx]->getY(), 0.0 };
		  gluTessVertex(tess, vertex, static_cast<GLvoid *>
				(const_cast<MMFPoint *>(vertices[vertexIdx])));
		}
	  }
	gluTessEndContour(tess);
	gluTessEndPolygon(tess);
      }
    glPassThrough(END_POLYGON_SECTION);

    /* Deleting tesselator */
    gluDeleteTess(tess);


    /************************ Rivers and contour lines ***********************/
    /* Defining rivers and contours' width and color */
    glLineWidth(1.0);
    glColor3f(0.0, 0.0, 1.0);

    /* Drawing rivers and rivers' contours from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   drgLayer.numberOfPolylines(); ++polylineIdx)
      {
	vector<MMFPoint const *> const & vertices =
	  drgLayer.getPolyline(polylineIdx).getVertices(resolutionIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx =0; vertexIdx < vertices.size(); ++vertexIdx)
	    glVertex2d(vertices[vertexIdx]->getX(),
		       vertices[vertexIdx]->getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_DRG_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.createRODList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s roads layer display list.
 *****************************************************************************/
void MMFMap::createRODList(int resolutionIdx, int listIdx) const
{
  MMFRODLayer const & rodLayer = *mmfFile.getRODLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_ROD_LAYER);
  {
    /******************************* Road lines ******************************/
    /* Defining roads width and color */
    glLineWidth(2.0);
    glColor3f(0.5, 0.5, 0.5);

    /* Drawing roads from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   rodLayer.numberOfPolylines(); ++polylineIdx)
      {
	vector<MMFPoint const *> const & vertices =
	  rodLayer.getPolyline(polylineIdx).getVertices(resolutionIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx =0; vertexIdx < vertices.size(); ++vertexIdx)
	    glVertex2d(vertices[vertexIdx]->getX(),
		       vertices[vertexIdx]->getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_ROD_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.createRARList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s railroads layer display list.
 *****************************************************************************/
void MMFMap::createRARList(int resolutionIdx, int listIdx) const
{
  MMFRARLayer const & rarLayer = *mmfFile.getRARLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_RAR_LAYER);
  {
    /***************************** Railroad lines ****************************/
    /* Defining railroads width and color */
    glLineWidth(2.0);
    glColor3f(0.8, 0.7, 0.0);

    /* Drawing railroads from polylines */
    glPassThrough(BEGIN_LINE_SECTION);
    for (unsigned polylineIdx = 0; polylineIdx <
	   rarLayer.numberOfPolylines(); ++polylineIdx)
      {
	vector<MMFPoint const *> const & vertices =
	  rarLayer.getPolyline(polylineIdx).getVertices(resolutionIdx);

	glPassThrough(BEGIN_LINE_STRIP);
	glBegin(GL_LINE_STRIP);
	{
	  for (unsigned vertexIdx =0; vertexIdx < vertices.size(); ++vertexIdx)
	    glVertex2d(vertices[vertexIdx]->getX(),
		       vertices[vertexIdx]->getY());
	}
	glEnd();
	glPassThrough(END_LINE_STRIP);
      }
    glPassThrough(END_LINE_SECTION);
  }
  glPassThrough(END_RAR_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.createAPTList(resolutionIdx, listIdx)
 ******************************************************************************
 * Create 'mmfMap''s airports layer display list.
 *****************************************************************************/
void MMFMap::createAPTList(int resolutionIdx, int listIdx) const
{
  MMFAPTLayer const & aptLayer = *mmfFile.getAPTLayer();

  /* Defining display list */
  glNewList(listIdx, GL_COMPILE);
  glPassThrough(BEGIN_APT_LAYER);
  {
    /***************************** Airport points ****************************/
    /* Defining airports size and color */
    glPointSize(2*POINT_RADIUS);
    glColor3f(0.0, 0.7, 0.2);

    /* Drawing airports from points */
    glPassThrough(BEGIN_POINT_SECTION);
    glBegin(GL_POINTS);
    {
      for (unsigned pointIdx = 0; pointIdx <
	     aptLayer.numberOfPoints(); ++pointIdx)

	glVertex2d(aptLayer.getPoint(pointIdx).getX(),
		   aptLayer.getPoint(pointIdx).getY());
    }
    glEnd();
    glPassThrough(END_POINT_SECTION);
  }
  glPassThrough(END_APT_LAYER);
  glEndList();
}



/******************************************************************************
 * mmfMap.writeBuffer(stream, buffer)
 ******************************************************************************
 * 
 *****************************************************************************/
void MMFMap::writeBuffer(ostream & ostr, float * buffer) const
{
  /* Writing Xfig header */
  ostr << "#FIG 3.2"  << endl
       << "Landscape" << endl
       << "Center"    << endl
       << "Inches"    << endl
       << "Letter  "  << endl
       << "100.00"    << endl
       << "Single"    << endl
       << "-2"        << endl
       << "1200 2"    << endl;

  /* Writing colors */
  ostr << setfill('0') << dec;
  for (unsigned colorIdx = 1; colorIdx <= AIRPORT; ++colorIdx)
    ostr << "0 " << 31+colorIdx << " #" << hex
	 << setw(2) << colors[colorIdx][colorMode][0]
	 << setw(2) << colors[colorIdx][colorMode][1]
	 << setw(2) << colors[colorIdx][colorMode][2] << endl << dec;


  /* Writing layers */
  GLfloat token;

  while (token = *buffer++, token != END_FILE)
    if      (token == GL_PASS_THROUGH_TOKEN) ;
    else if (token == BEGIN_POL_LAYER) writePOLLayer(ostr, buffer);
    else if (token == BEGIN_POP_LAYER) writePOPLayer(ostr, buffer);
    else if (token == BEGIN_DRG_LAYER) writeDRGLayer(ostr, buffer);
    else if (token == BEGIN_ROD_LAYER) writeRODLayer(ostr, buffer);
    else if (token == BEGIN_RAR_LAYER) writeRARLayer(ostr, buffer);
    else if (token == BEGIN_APT_LAYER) writeAPTLayer(ostr, buffer);
    else break;
}



/******************************************************************************
 * mmfMap.writePOLLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writePOLLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POL_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, POLITICAL_CONTENTS);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, POLITICAL_CONTOUR);

    else break;
}



/******************************************************************************
 * mmfMap.writePOPLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writePOPLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POP_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, BIG_CITY_CONTENTS);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, BIG_CITY_CONTOUR);

    else if (token == BEGIN_POINT_SECTION)
      writePoints(ostr, buffer, CITY_LOCATION);

    else break;
}



/******************************************************************************
 * mmfMap.writeDRGLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeDRGLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_DRG_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POLYGON_SECTION)
      writePolygons(ostr, buffer, PERENNIAL_WATER);

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, RIVER_AND_CONTOUR);

    else break;
}



/******************************************************************************
 * mmfMap.writeRODLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeRODLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_ROD_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, ROAD);

    else break;
}



/******************************************************************************
 * mmfMap.writeRARLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeRARLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_RAR_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_LINE_SECTION)
      writeLines(ostr, buffer, RAILROAD);

    else break;
}



/******************************************************************************
 * mmfMap.writeAPTLayer(ostr, buffer)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeAPTLayer(ostream& ostr, float*& buffer) const
{
  GLfloat token;

  while (token = *buffer++, token != END_APT_LAYER)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == BEGIN_POINT_SECTION)
      writePoints(ostr, buffer, AIRPORT);

    else break;
}



/******************************************************************************
 * mmfMap.writePolygons(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writePolygons(ostream & ostr, float *& buffer, int type) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POLYGON_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_POLYGON_TOKEN)
      {
	unsigned nVertices = static_cast<unsigned>(*buffer++);

	if (noFill == false)
	  ostr << "2 1 0 0 0 " << 31+type
	       << " 50 -1 20 0.000 0 0 -1 0 0 " << nVertices;

	for (unsigned vertexIdx = 0; vertexIdx < nVertices; ++vertexIdx)
	  {
	    GLfloat x = *buffer++;
	    GLfloat y = *buffer++;

	    if (noFill) continue;

	    if (vertexIdx%6 == 0) ostr << endl << '\t';

	    ostr << ' ' << static_cast<int>(xScale*(x-xLeft)/(xRight-xLeft))
		 << ' ' << static_cast<int>(yScale*(yTop-y) /(yTop-yBottom));
	  }
	if (noFill == false) ostr << endl;
      }

    else break;
}



/******************************************************************************
 * mmfMap.writeLines(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeLines(ostream & ostr, float *& buffer, int type) const
{
  /* Counting number of vertices */
  GLfloat token;

  unsigned nLines = 0;
  float * bufferBU = 0;

  GLfloat prevX2 = 0.0;
  GLfloat prevY2 = 0.0;

  while (token = *buffer++, token != END_LINE_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_LINE_TOKEN ||
	     token == GL_LINE_RESET_TOKEN)
      {
	GLfloat x1 = *buffer++, y1 = *buffer++;
	GLfloat x2 = *buffer++, y2 = *buffer++;

	if (nLines > 0 && (x1 != prevX2 ||
			   y1 != prevY2))
	  {
	    writeLineStrip(ostr, bufferBU, nLines, type);

	    nLines = 1;
	    bufferBU = buffer-5;
	  }
	else
	  {
	    prevX2 = x2;
	    prevY2 = y2;

	    ++nLines;
	  }
      }

    else if (token == BEGIN_LINE_STRIP)
      {
	nLines = 0;
	bufferBU = buffer;
      }

    else if (token == END_LINE_STRIP)
      {
	if (nLines > 0)
	  writeLineStrip(ostr, bufferBU, nLines, type);
      }

    else break;
}



/******************************************************************************
 * mmfMap.writePoints(ostr, buffer, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writePoints(ostream & ostr, float *& buffer, int type) const
{
  GLfloat token;

  while (token = *buffer++, token != END_POINT_SECTION)

    if (token == GL_PASS_THROUGH_TOKEN) ;

    else if (token == GL_POINT_TOKEN)
      { 
	ostr << "2 2 0 0 0 " << 31+type << " 50 -1 20 0.000 0 0 -1 0 0 5"
	     << endl << '\t';

	GLfloat x = *buffer++;
	GLfloat y = *buffer++;

	int xInt = static_cast<int>(xScale*(x-xLeft)/(xRight-xLeft));
	int yInt = static_cast<int>(yScale*(yTop-y) /(yTop-yBottom));

	ostr << ' ' << xInt-pointRadius << ' ' << yInt-pointRadius
	     << ' ' << xInt+pointRadius << ' ' << yInt-pointRadius
	     << ' ' << xInt+pointRadius << ' ' << yInt+pointRadius
	     << ' ' << xInt-pointRadius << ' ' << yInt+pointRadius
	     << ' ' << xInt-pointRadius << ' ' << yInt-pointRadius << endl;
      }

    else break;
}



/******************************************************************************
 * mmfMap.writeLineStrip(ostr, buffer, nLines, type)
 ******************************************************************************
 *
 *****************************************************************************/
void MMFMap::writeLineStrip(ostream & ostr, float * buffer,
			    unsigned nLines, int type) const
{
  /* Writing header */
  if (type != ROAD && type != RAILROAD)
    ostr << "2 1 0 1 " << 31+type
	 << " 0 50 -1 -1 0.000 1 1 -1 0 0 " << nLines+1;

  else if (colorMode == FULL_COLOR_MODE)
    ostr << "2 1 0 2 " << 31+type
	 << " 0 50 -1 -1 0.000 1 1 -1 0 0 " << nLines+1;

  else
    ostr << "2 1 " << (type == ROAD ? 3 : 1) << " 2 " << 31+type
	 << " 0 50 -1 -1 "
	 << setprecision(3) << fixed << setw(4) << pointRadius/10.0
	 << " 1 1 -1 0 0 " << nLines+1;


  /* Writing vertices */
  GLfloat x1 = 0.0, y1 = 0.0;
  GLfloat x2 = 0.0, y2 = 0.0;

  for (unsigned vertexIdx = 0; vertexIdx < nLines; ++vertexIdx)
    {
      /* Skipping line token */
      ++buffer;

      if (vertexIdx%6 == 0) ostr << endl << '\t';

      x1 = *buffer++; y1 = *buffer++;
      x2 = *buffer++; y2 = *buffer++;

      ostr << ' ' << static_cast<int>(xScale*(x1-xLeft)/(xRight-xLeft))
	   << ' ' << static_cast<int>(yScale*(yTop-y1) /(yTop-yBottom));
    }


  /* Writing last vertex */
  if (nLines%6 == 0) ostr << endl << '\t';

  ostr << ' ' << static_cast<int>(xScale*(x2-xLeft)/(xRight-xLeft))
       << ' ' << static_cast<int>(yScale*(yTop-y2) /(yTop-yBottom)) << endl;
}
