#ifndef __MMFMAP_HH_
#define __MMFMAP_HH_


#include <map.hh>
#include <mmf/mmffile.hh>

#include <iostream>


using namespace std;



/******************************************************************************
 * Class MMFMap
 *****************************************************************************/
class MMFMap
  : public Map {

public:

  /* Constructor */
  MMFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill = false) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, MMFMap       & mmfMap);
  friend ostream & operator<<(ostream & ostr, MMFMap const & mmfMap);


private:

  /* Layer information structure */
  struct layerInfo_t
  {
    unsigned ID;
    MMFLayer const * layer;
    bool active;
  };

  /* Number of layers and array of layers informations */
  unsigned nLayers;
  layerInfo_t layerInfos[6];

  /* Multi-resolution data */
  double tolerance;
  unsigned currResolIdx;

  /* Vectors of polylines and feature points */
  MMFFile mmfFile;

  /* Internal display lists creation functions */
  void createPOLList(int resolutionIdx, int listIdx) const;
  void createPOPList(int resolutionIdx, int listIdx) const;
  void createDRGList(int resolutionIdx, int listIdx) const;
  void createRODList(int resolutionIdx, int listIdx) const;
  void createRARList(int resolutionIdx, int listIdx) const;
  void createAPTList(int resolutionIdx, int listIdx) const;

  /* Internal functions for writing layers in Xfig format */
  void writeBuffer(ostream & ostr, float * buffer) const;

  void writePOLLayer(ostream & ostr, float *& buffer) const;
  void writePOPLayer(ostream & ostr, float *& buffer) const;
  void writeDRGLayer(ostream & ostr, float *& buffer) const;
  void writeRODLayer(ostream & ostr, float *& buffer) const;
  void writeRARLayer(ostream & ostr, float *& buffer) const;
  void writeAPTLayer(ostream & ostr, float *& buffer) const;

  void writePolygons(ostream & ostr, float *& buffer, int type) const;
  void writeLines   (ostream & ostr, float *& buffer, int type) const;
  void writePoints  (ostream & ostr, float *& buffer, int type) const;

  void writeLineStrip(ostream & ostr, float * buffer,
		      unsigned nLines, int type) const;
};


inline MMFMap::MMFMap() : nLayers(0), tolerance(0.0), currResolIdx(0) {}

inline char const * MMFMap::getName() const
{
  return mmfFile.getName().c_str();
}

inline unsigned MMFMap::getNumberOfLayers() const
{
  return nLayers;
}

inline char const * MMFMap::getLayerName(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return "";

  return layerInfos[layerIdx].layer->getTitle().c_str();
}

inline bool MMFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return false;

  return layerInfos[layerIdx].active;
}

inline void MMFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return;

  layerInfos[layerIdx].active = activated;
}

inline double MMFMap::getTolerance() const { return tolerance; }
inline void   MMFMap::setTolerance(double tolerance)
{
  /* Setting tolerance */
  this->tolerance = tolerance;

  /* Searching for current resolution index */
  for (currResolIdx = 0; currResolIdx <
	 mmfFile.numberOfResolutions(); ++currResolIdx)
    if (mmfFile.getTolerance(currResolIdx) <= tolerance) break;
}

inline unsigned MMFMap::getNumberOfResolutions() const
{
  return mmfFile.numberOfResolutions();
}
inline unsigned MMFMap::getResolution() const { return currResolIdx; }
inline void     MMFMap::setResolution(unsigned idx) { currResolIdx = idx; }



#endif // __MMFMAP_HH_
