#include <mtfmap/mtfmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>




/* Point Radius */
#define POINT_RADIUS 2.5

#define POLLAYER 0
#define POPLAYER 1
#define DRGLAYER 2
#define RODLAYER 3
#define RARLAYER 4
#define APTLAYER 5


/******************************************************************************
 * mtfMap.read(ostr)
 ******************************************************************************
 * Reads 'mtfMap' from file given by 'fileName'.
 *****************************************************************************/
bool MTFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * mtfMap.read(ostr)
 ******************************************************************************
 * Writes 'mtfMap' in file given by 'fileName'.
 *****************************************************************************/
bool MTFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * mtfMap.addVertices(polyline, tolerance)
 ******************************************************************************
 *****************************************************************************/
void MTFMap::addVertices(MTFPolyline const & polyline, double tolerance)
{
  glVertex2d(polyline.front().getX(),
	     polyline.front().getY());

  addVertices(polyline, tolerance, polyline.front().getRghtSonIdx());

  glVertex2d(polyline.back().getX(),
	     polyline.back().getY());
}



/******************************************************************************
 * mtfMap.addVertices(polyline, tolerance, vertexIdx)
 ******************************************************************************
 *****************************************************************************/
void MTFMap::addVertices(MTFPolyline const & polyline, double tolerance,
			 unsigned vertexIdx)
{
  if (static_cast<int>(vertexIdx) == -1 ||
      polyline[vertexIdx].getTolerance() < tolerance) return ;

  addVertices(polyline, tolerance, polyline[vertexIdx].getLeftSonIdx());

  glVertex2d(polyline[vertexIdx].getX(),
	     polyline[vertexIdx].getY());

  addVertices(polyline, tolerance, polyline[vertexIdx].getRghtSonIdx());
}



/******************************************************************************
 * bvfMap.draw()
 ******************************************************************************
 * Draws 'bvfMap' contents, including polylines and feature points.
 *****************************************************************************/
void MTFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting POL Layer */
  for (unsigned layerIdx = 0; layerIdx < nLayers; ++layerIdx)
    if (layerInfos[layerIdx].active)

      if (layerInfos[layerIdx].ID == POLLAYER) /* Plotting POL Layer */
	{
	  MTFPOLLayer const & polLayer =
	    *reinterpret_cast<MTFPOLLayer const *>(layerInfos[layerIdx].layer);

	  /******************** Political contour lines **********************/
	  /* Defining political contour width and color */
	  glLineWidth(1.0);
	  glColor3f(0.0, 0.0, 0.0);

	  /* Drawing contour from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 polLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MTFPolyline const & polyline = polLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == POPLAYER) /* Plotting POP Layer */
	{
	  MTFPOPLayer const & popLayer =
	    *reinterpret_cast<MTFPOPLayer const *>(layerInfos[layerIdx].layer);

	  /******************** Big cities' contour lines ********************/
	  /* Defining big cities' contour width and color */
	  glLineWidth(1.0);
	  glColor3f(1.0, 0.0, 0.0);

	  /* Drawing big cities from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 popLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MTFPolyline const & polyline = popLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }

	  /*************************** City points ***************************/
	  /* Defining city points' size and color */
	  glPointSize(2*POINT_RADIUS);
	  glColor3f(1.0, 0.0, 0.0);

	  /* Drawing small cities from points */
	  glBegin(GL_POINTS);
	  {
	    for (unsigned pointIdx = 0; pointIdx <
		   popLayer.numberOfPoints(); ++pointIdx)

	      glVertex2d(popLayer.getPoint(pointIdx).getX(),
			 popLayer.getPoint(pointIdx).getY());
	  }
	  glEnd();
	}

      else if (layerInfos[layerIdx].ID == DRGLAYER) /* Plotting DRG Layer */
	{
	  MTFDRGLayer const & drgLayer =
	    *reinterpret_cast<MTFDRGLayer const *>(layerInfos[layerIdx].layer);

	  /********************* Rivers and contour lines ********************/
	  /* Defining rivers and contours' width and color */
	  glLineWidth(1.0);
	  glColor3f(0.0, 0.0, 1.0);

	  /* Drawing rivers and rivers' contours from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 drgLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MTFPolyline const & polyline = drgLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == RODLAYER) /* Plotting ROD Layer */
	{
	  MTFRODLayer const & rodLayer =
	    *reinterpret_cast<MTFRODLayer const *>(layerInfos[layerIdx].layer);

	  /**************************** Road lines ***************************/
	  /* Defining roads width and color */
	  glLineWidth(2.0);
	  glColor3f(0.5, 0.5, 0.5);

	  /* Drawing roads from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 rodLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MTFPolyline const & polyline = rodLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == RARLAYER) /* Plotting RAR Layer */
	{
	  MTFRARLayer const & rarLayer =
	    *reinterpret_cast<MTFRARLayer const *>(layerInfos[layerIdx].layer);

	  /************************** Railroad lines *************************/
	  /* Defining railroads width and color */
	  glLineWidth(2.0);
	  glColor3f(0.8, 0.7, 0.0);

	  /* Drawing railroads from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 rarLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MTFPolyline const & polyline = rarLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == APTLAYER) /* Plotting APT Layer */
	{
	  MTFAPTLayer const & aptLayer =
	    *reinterpret_cast<MTFAPTLayer const *>(layerInfos[layerIdx].layer);

	  /************************** Airport points *************************/
	  /* Defining airports size and color */
	  glPointSize(2*POINT_RADIUS);
	  glColor3f(0.0, 0.7, 0.2);

	  /* Drawing airports from points */
	  glBegin(GL_POINTS);
	  {
	    for (unsigned pointIdx = 0; pointIdx <
		   aptLayer.numberOfPoints(); ++pointIdx)

	      glVertex2d(aptLayer.getPoint(pointIdx).getX(),
			 aptLayer.getPoint(pointIdx).getY());
	  }
	  glEnd();
	}


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> mtfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, MTFMap & mtfMap)
{
  /* Reading MTF file */
  istr >> mtfMap.mtfFile;


  /* List of tolerances */
  list<double> tolList;

  /* Initializing bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /***************************** Political layer *****************************/
  if (mtfMap.mtfFile.getPOLLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = POLLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getPOLLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFPOLLayer const & polLayer = *mtfMap.mtfFile.getPOLLayer();

      if (polLayer.getXMin() < xMin) xMin = polLayer.getXMin();
      if (polLayer.getXMax() > xMax) xMax = polLayer.getXMax();
      if (polLayer.getYMin() < yMin) yMin = polLayer.getYMin();
      if (polLayer.getYMax() > yMax) yMax = polLayer.getYMax();

      vector<MTFPolyline> const & polylines = polLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /************************* Populated places layer **************************/
  if (mtfMap.mtfFile.getPOPLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = POPLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getPOPLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFPOPLayer const & popLayer = *mtfMap.mtfFile.getPOPLayer();

      if (popLayer.getXMin() < xMin) xMin = popLayer.getXMin();
      if (popLayer.getXMax() > xMax) xMax = popLayer.getXMax();
      if (popLayer.getYMin() < yMin) yMin = popLayer.getYMin();
      if (popLayer.getYMax() > yMax) yMax = popLayer.getYMax();

      vector<MTFPolyline> const & polylines = popLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Drainage layer ******************************/
  if (mtfMap.mtfFile.getDRGLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = DRGLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getDRGLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFDRGLayer const & drgLayer = *mtfMap.mtfFile.getDRGLayer();

      if (drgLayer.getXMin() < xMin) xMin = drgLayer.getXMin();
      if (drgLayer.getXMax() > xMax) xMax = drgLayer.getXMax();
      if (drgLayer.getYMin() < yMin) yMin = drgLayer.getYMin();
      if (drgLayer.getYMax() > yMax) yMax = drgLayer.getYMax();

      vector<MTFPolyline> const & polylines = drgLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /******************************* Roads layer *******************************/
  if (mtfMap.mtfFile.getRODLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = RODLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getRODLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFRODLayer const & rodLayer = *mtfMap.mtfFile.getRODLayer();

      if (rodLayer.getXMin() < xMin) xMin = rodLayer.getXMin();
      if (rodLayer.getXMax() > xMax) xMax = rodLayer.getXMax();
      if (rodLayer.getYMin() < yMin) yMin = rodLayer.getYMin();
      if (rodLayer.getYMax() > yMax) yMax = rodLayer.getYMax();

      vector<MTFPolyline> const & polylines = rodLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Railroads layer *****************************/
  if (mtfMap.mtfFile.getRARLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = RARLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getRARLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFRARLayer const & rarLayer = *mtfMap.mtfFile.getRARLayer();

      if (rarLayer.getXMin() < xMin) xMin = rarLayer.getXMin();
      if (rarLayer.getXMax() > xMax) xMax = rarLayer.getXMax();
      if (rarLayer.getYMin() < yMin) yMin = rarLayer.getYMin();
      if (rarLayer.getYMax() > yMax) yMax = rarLayer.getYMax();

      vector<MTFPolyline> const & polylines = rarLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Airports layer ******************************/
  if (mtfMap.mtfFile.getAPTLayer() != 0)
    {
      /* Updating layer info vector */
      mtfMap.layerInfos[mtfMap.nLayers].ID     = APTLAYER;//mtfMap.nLayers+1;
      mtfMap.layerInfos[mtfMap.nLayers].layer  = mtfMap.mtfFile.getAPTLayer();
      mtfMap.layerInfos[mtfMap.nLayers].active = true;
      ++mtfMap.nLayers;

      /* Searching for layer bounding box */
      MTFAPTLayer const & aptLayer = *mtfMap.mtfFile.getAPTLayer();

      if (aptLayer.getXMin() < xMin) xMin = aptLayer.getXMin();
      if (aptLayer.getXMax() > xMax) xMax = aptLayer.getXMax();
      if (aptLayer.getYMin() < yMin) yMin = aptLayer.getYMin();
      if (aptLayer.getYMax() > yMax) yMax = aptLayer.getYMax();
    }


  /* Setting bounding box */
  mtfMap.setXMin(xMin);
  mtfMap.setXMax(xMax);
  mtfMap.setYMin(yMin);
  mtfMap.setYMax(yMax);


  /* Sorting and cleaning list of tolerances */
  tolList.sort();
  tolList.unique();
  tolList.reverse();


  /* Initiating index of tolerances and filling vector of tolerances */
  mtfMap.toleranceIdx = 0;
  mtfMap.tolerances.clear();
  mtfMap.tolerances.reserve(tolList.size());
  mtfMap.tolerances.insert(mtfMap.tolerances.begin(),
			   tolList.begin(), tolList.end());


  return istr;
}



/******************************************************************************
 * ostr << mtfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MTFMap const & mtfMap)
{
  /* Writing MTF file */
  ostr << mtfMap.mtfFile;


  return ostr;
}
