#ifndef __MTFMAP_HH_
#define __MTFMAP_HH_


#include <map.hh>
#include <mtf/mtffile.hh>

#include <iostream>


using namespace std;



/******************************************************************************
 * Class MTFMap
 *****************************************************************************/
class MTFMap
  : public Map {

public:

  /* Constructor */
  MTFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill = false) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, MTFMap       & mtfMap);
  friend ostream & operator<<(ostream & ostr, MTFMap const & mtfMap);


private:

  void addVertices(MTFPolyline const & polyline, double tolerance);
  void addVertices(MTFPolyline const & polyline, double tolerance,
		   unsigned vertexIdx);

  /* Layer information structure */
  struct layerInfo_t
  {
    unsigned ID;
    MTFLayer const * layer;
    bool active;
  };

  /* Number of layers and array of layers informations */
  unsigned nLayers;
  layerInfo_t layerInfos[6];

  /* Vector of tolerances/resolutions */
  vector<double> tolerances;
  unsigned       toleranceIdx;

  /* Vectors of polylines and feature points */
  MTFFile mtfFile;
};


inline MTFMap::MTFMap() : nLayers(0) {}

inline char const * MTFMap::getName() const
{
  return mtfFile.getName().c_str();
}

inline unsigned MTFMap::getNumberOfLayers() const
{
  return nLayers;
}

inline char const * MTFMap::getLayerName(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return "";

  return layerInfos[layerIdx].layer->getTitle().c_str();
}

inline bool MTFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return false;

  return layerInfos[layerIdx].active;
}

inline void MTFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return;

  layerInfos[layerIdx].active = activated;
}

inline double MTFMap::getTolerance() const { return tolerances[toleranceIdx]; }
inline void   MTFMap::setTolerance(double tolerance)
{ for (toleranceIdx=0; tolerances[toleranceIdx]<tolerance; ++toleranceIdx) ; }

inline unsigned MTFMap::getNumberOfResolutions() const
{ return tolerances.size(); }

inline unsigned MTFMap::getResolution() const { return toleranceIdx; }

inline void MTFMap::setResolution(unsigned resIdx)
{ toleranceIdx = resIdx; }

inline void MTFMap::init() const {}

inline bool MTFMap::exportAsFig(char const * fileName, int colorMode,
				bool noFill) const
{
  fileName  = fileName;
  colorMode = colorMode;
  noFill    = noFill;

  return false;
}


#endif // __MTFMAP_HH_
