#include <mvfmap/mvfmap.hh>

#include <fstream>
#include <iomanip>

#include <GL/gl.h>
#include <GL/glu.h>




/* Point Radius */
#define POINT_RADIUS 2.5

#define POLLAYER 0
#define POPLAYER 1
#define DRGLAYER 2
#define RODLAYER 3
#define RARLAYER 4
#define APTLAYER 5


/******************************************************************************
 * mvfMap.read(ostr)
 ******************************************************************************
 * Reads 'mvfMap' from file given by 'fileName'.
 *****************************************************************************/
bool MVFMap::read(char const * fileName)
{
  /* Openning file in a ifstream */
  ifstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;

  /* Reading from map stream */
  mapStream >> *this;
  return true;
}



/******************************************************************************
 * mvfMap.read(ostr)
 ******************************************************************************
 * Writes 'mvfMap' in file given by 'fileName'.
 *****************************************************************************/
bool MVFMap::write(char const * fileName) const
{
  /* Openning file in a ifstream */
  ofstream mapStream(fileName);

  /* Testing if file was correctly open */
  if (!mapStream.is_open()) return false;
 
  /* Writing into map stream */
  mapStream << *this;
  return true;
}



/******************************************************************************
 * bvfMap.draw()
 ******************************************************************************
 * Draws 'bvfMap' contents, including polylines and feature points.
 *****************************************************************************/
void MVFMap::draw() const
{
  /* Clearing the window's background */
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);


  /* Plotting POL Layer */
  for (unsigned layerIdx = 0; layerIdx < nLayers; ++layerIdx)
    if (layerInfos[layerIdx].active)

      if (layerInfos[layerIdx].ID == POLLAYER) /* Plotting POL Layer */
	{
	  MVFPOLLayer const & polLayer =
	    *reinterpret_cast<MVFPOLLayer const *>(layerInfos[layerIdx].layer);

	  /******************** Political contour lines **********************/
	  /* Defining political contour width and color */
	  glLineWidth(1.0);
	  glColor3f(0.0, 0.0, 0.0);

	  /* Drawing contour from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 polLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MVFPolyline const & polyline = polLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == POPLAYER) /* Plotting POP Layer */
	{
	  MVFPOPLayer const & popLayer =
	    *reinterpret_cast<MVFPOPLayer const *>(layerInfos[layerIdx].layer);

	  /******************** Big cities' contour lines ********************/
	  /* Defining big cities' contour width and color */
	  glLineWidth(1.0);
	  glColor3f(1.0, 0.0, 0.0);

	  /* Drawing big cities from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 popLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MVFPolyline const & polyline = popLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }

	  /*************************** City points ***************************/
	  /* Defining city points' size and color */
	  glPointSize(2*POINT_RADIUS);
	  glColor3f(1.0, 0.0, 0.0);

	  /* Drawing small cities from points */
	  glBegin(GL_POINTS);
	  {
	    for (unsigned pointIdx = 0; pointIdx <
		   popLayer.numberOfPoints(); ++pointIdx)

	      glVertex2d(popLayer.getPoint(pointIdx).getX(),
			 popLayer.getPoint(pointIdx).getY());
	  }
	  glEnd();
	}

      else if (layerInfos[layerIdx].ID == DRGLAYER) /* Plotting DRG Layer */
	{
	  MVFDRGLayer const & drgLayer =
	    *reinterpret_cast<MVFDRGLayer const *>(layerInfos[layerIdx].layer);

	  /********************* Rivers and contour lines ********************/
	  /* Defining rivers and contours' width and color */
	  glLineWidth(1.0);
	  glColor3f(0.0, 0.0, 1.0);

	  /* Drawing rivers and rivers' contours from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 drgLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MVFPolyline const & polyline = drgLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == RODLAYER) /* Plotting ROD Layer */
	{
	  MVFRODLayer const & rodLayer =
	    *reinterpret_cast<MVFRODLayer const *>(layerInfos[layerIdx].layer);

	  /**************************** Road lines ***************************/
	  /* Defining roads width and color */
	  glLineWidth(2.0);
	  glColor3f(0.5, 0.5, 0.5);

	  /* Drawing roads from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 rodLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MVFPolyline const & polyline = rodLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == RARLAYER) /* Plotting RAR Layer */
	{
	  MVFRARLayer const & rarLayer =
	    *reinterpret_cast<MVFRARLayer const *>(layerInfos[layerIdx].layer);

	  /************************** Railroad lines *************************/
	  /* Defining railroads width and color */
	  glLineWidth(2.0);
	  glColor3f(0.8, 0.7, 0.0);

	  /* Drawing railroads from polylines */
	  for (unsigned polylineIdx = 0; polylineIdx <
		 rarLayer.numberOfPolylines(); ++polylineIdx)
	    {
	      MVFPolyline const & polyline = rarLayer.getPolyline(polylineIdx);

	      glBegin(GL_LINE_STRIP);
	      {
		for (unsigned vertexIdx = 0; vertexIdx <
		       polyline.numberOfVertices(); ++vertexIdx)

		  if (polyline[vertexIdx].
		      getTolerance() >= tolerances[toleranceIdx])
		    glVertex2d(polyline[vertexIdx].getX(),
			       polyline[vertexIdx].getY());
	      }
	      glEnd();
	    }
	}

      else if (layerInfos[layerIdx].ID == APTLAYER) /* Plotting APT Layer */
	{
	  MVFAPTLayer const & aptLayer =
	    *reinterpret_cast<MVFAPTLayer const *>(layerInfos[layerIdx].layer);

	  /************************** Airport points *************************/
	  /* Defining airports size and color */
	  glPointSize(2*POINT_RADIUS);
	  glColor3f(0.0, 0.7, 0.2);

	  /* Drawing airports from points */
	  glBegin(GL_POINTS);
	  {
	    for (unsigned pointIdx = 0; pointIdx <
		   aptLayer.numberOfPoints(); ++pointIdx)

	      glVertex2d(aptLayer.getPoint(pointIdx).getX(),
			 aptLayer.getPoint(pointIdx).getY());
	  }
	  glEnd();
	}


  /* Flushing the screen */
  glFlush();
}



/******************************************************************************
 * istr >> mvfMap
 ******************************************************************************
 * Redefinition of the extraction operator '>>' in stream context.
 *****************************************************************************/
istream & operator>>(istream & istr, MVFMap & mvfMap)
{
  /* Reading MVF file */
  istr >> mvfMap.mvfFile;


  /* List of tolerances */
  list<double> tolList;

  /* Initializing bounding box variables */
  double xMin = +1.7e+308;
  double xMax = -1.7e+308;
  double yMin = +1.7e+308;
  double yMax = -1.7e+308;


  /***************************** Political layer *****************************/
  if (mvfMap.mvfFile.getPOLLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = POLLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getPOLLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFPOLLayer const & polLayer = *mvfMap.mvfFile.getPOLLayer();

      if (polLayer.getXMin() < xMin) xMin = polLayer.getXMin();
      if (polLayer.getXMax() > xMax) xMax = polLayer.getXMax();
      if (polLayer.getYMin() < yMin) yMin = polLayer.getYMin();
      if (polLayer.getYMax() > yMax) yMax = polLayer.getYMax();

      vector<MVFPolyline> const & polylines = polLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /************************* Populated places layer **************************/
  if (mvfMap.mvfFile.getPOPLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = POPLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getPOPLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFPOPLayer const & popLayer = *mvfMap.mvfFile.getPOPLayer();

      if (popLayer.getXMin() < xMin) xMin = popLayer.getXMin();
      if (popLayer.getXMax() > xMax) xMax = popLayer.getXMax();
      if (popLayer.getYMin() < yMin) yMin = popLayer.getYMin();
      if (popLayer.getYMax() > yMax) yMax = popLayer.getYMax();

      vector<MVFPolyline> const & polylines = popLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Drainage layer ******************************/
  if (mvfMap.mvfFile.getDRGLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = DRGLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getDRGLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFDRGLayer const & drgLayer = *mvfMap.mvfFile.getDRGLayer();

      if (drgLayer.getXMin() < xMin) xMin = drgLayer.getXMin();
      if (drgLayer.getXMax() > xMax) xMax = drgLayer.getXMax();
      if (drgLayer.getYMin() < yMin) yMin = drgLayer.getYMin();
      if (drgLayer.getYMax() > yMax) yMax = drgLayer.getYMax();

      vector<MVFPolyline> const & polylines = drgLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /******************************* Roads layer *******************************/
  if (mvfMap.mvfFile.getRODLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = RODLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getRODLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFRODLayer const & rodLayer = *mvfMap.mvfFile.getRODLayer();

      if (rodLayer.getXMin() < xMin) xMin = rodLayer.getXMin();
      if (rodLayer.getXMax() > xMax) xMax = rodLayer.getXMax();
      if (rodLayer.getYMin() < yMin) yMin = rodLayer.getYMin();
      if (rodLayer.getYMax() > yMax) yMax = rodLayer.getYMax();

      vector<MVFPolyline> const & polylines = rodLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Railroads layer *****************************/
  if (mvfMap.mvfFile.getRARLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = RARLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getRARLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFRARLayer const & rarLayer = *mvfMap.mvfFile.getRARLayer();

      if (rarLayer.getXMin() < xMin) xMin = rarLayer.getXMin();
      if (rarLayer.getXMax() > xMax) xMax = rarLayer.getXMax();
      if (rarLayer.getYMin() < yMin) yMin = rarLayer.getYMin();
      if (rarLayer.getYMax() > yMax) yMax = rarLayer.getYMax();

      vector<MVFPolyline> const & polylines = rarLayer.getPolylines();

      tolList.push_back(polylines[0][0].getTolerance());

      for (unsigned polIdx = 0; polIdx < polylines.size(); ++polIdx)
	for (unsigned vertexIdx = 0; vertexIdx <
	       polylines[polIdx].size(); ++vertexIdx)
	  tolList.push_back(polylines[polIdx][vertexIdx].getTolerance());
    }


  /***************************** Airports layer ******************************/
  if (mvfMap.mvfFile.getAPTLayer() != 0)
    {
      /* Updating layer info vector */
      mvfMap.layerInfos[mvfMap.nLayers].ID     = APTLAYER;//mvfMap.nLayers+1;
      mvfMap.layerInfos[mvfMap.nLayers].layer  = mvfMap.mvfFile.getAPTLayer();
      mvfMap.layerInfos[mvfMap.nLayers].active = true;
      ++mvfMap.nLayers;

      /* Searching for layer bounding box */
      MVFAPTLayer const & aptLayer = *mvfMap.mvfFile.getAPTLayer();

      if (aptLayer.getXMin() < xMin) xMin = aptLayer.getXMin();
      if (aptLayer.getXMax() > xMax) xMax = aptLayer.getXMax();
      if (aptLayer.getYMin() < yMin) yMin = aptLayer.getYMin();
      if (aptLayer.getYMax() > yMax) yMax = aptLayer.getYMax();
    }


  /* Setting bounding box */
  mvfMap.setXMin(xMin);
  mvfMap.setXMax(xMax);
  mvfMap.setYMin(yMin);
  mvfMap.setYMax(yMax);


  /* Sorting and cleaning list of tolerances */
  tolList.sort();
  tolList.unique();
  tolList.reverse();


  /* Initiating index of tolerances and filling vector of tolerances */
  mvfMap.toleranceIdx = 0;
  mvfMap.tolerances.clear();
  mvfMap.tolerances.reserve(tolList.size());
  mvfMap.tolerances.insert(mvfMap.tolerances.begin(),
			   tolList.begin(), tolList.end());


  return istr;
}



/******************************************************************************
 * ostr << mvfMap
 ******************************************************************************
 * Redefinition of the insertion operator '<<' in stream context.
 *****************************************************************************/
ostream & operator<<(ostream & ostr, MVFMap const & mvfMap)
{
  /* Writing MVF file */
  ostr << mvfMap.mvfFile;


  return ostr;
}
