#ifndef __MVFMAP_HH_
#define __MVFMAP_HH_


#include <map.hh>
#include <mvf/mvffile.hh>

#include <iostream>


using namespace std;



/******************************************************************************
 * Class MVFMap
 *****************************************************************************/
class MVFMap
  : public Map {

public:

  /* Constructor */
  MVFMap();

  /* Get name function */
  virtual char const * getName() const;

  /* Get number of layers */
  virtual unsigned     getNumberOfLayers() const;
  virtual char const * getLayerName(int layerIdx) const;

  /* Activated layer functions */
  virtual bool  isLayerActivated(int layerIdx) const;
  virtual void setLayerActivated(int layerIdx, bool activated);

  /* Virtual read and write functions */
  virtual bool read (char const * fileName);
  virtual bool write(char const * fileName) const;

  /* Tolerance functions */
  virtual double getTolerance() const;
  virtual void   setTolerance(double tolerance);

  /* Resolution functioins */
  virtual unsigned getNumberOfResolutions() const;
  virtual unsigned getResolution() const;
  virtual void     setResolution(unsigned resolutionIdx);

  /* Init and draw functions */
  virtual void init() const;
  virtual void draw() const;

  /* Export as fig function */
  virtual bool exportAsFig(char const * fileName, int colorMode,
			   bool noFill = false) const;

  /* Friendship relation of insertion operator */
  friend istream & operator>>(istream & istr, MVFMap       & mvfMap);
  friend ostream & operator<<(ostream & ostr, MVFMap const & mvfMap);


private:

  /* Layer information structure */
  struct layerInfo_t
  {
    unsigned ID;
    MVFLayer const * layer;
    bool active;
  };

  /* Number of layers and array of layers informations */
  unsigned nLayers;
  layerInfo_t layerInfos[6];

  /* Vector of tolerances/resolutions */
  vector<double> tolerances;
  unsigned       toleranceIdx;

  /* Vectors of polylines and feature points */
  MVFFile mvfFile;
};


inline MVFMap::MVFMap() : nLayers(0) {}

inline char const * MVFMap::getName() const
{
  return mvfFile.getName().c_str();
}

inline unsigned MVFMap::getNumberOfLayers() const
{
  return nLayers;
}

inline char const * MVFMap::getLayerName(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return "";

  return layerInfos[layerIdx].layer->getTitle().c_str();
}

inline bool MVFMap::isLayerActivated(int layerIdx) const
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return false;

  return layerInfos[layerIdx].active;
}

inline void MVFMap::setLayerActivated(int layerIdx, bool activated)
{
  if (layerIdx <  0 ||
      layerIdx >= static_cast<int>(nLayers)) return;

  layerInfos[layerIdx].active = activated;
}

inline double MVFMap::getTolerance() const { return tolerances[toleranceIdx]; }
inline void   MVFMap::setTolerance(double tolerance)
{ for (toleranceIdx=0; tolerances[toleranceIdx]<tolerance; ++toleranceIdx) ; }

inline unsigned MVFMap::getNumberOfResolutions() const
{ return tolerances.size(); }

inline unsigned MVFMap::getResolution() const { return toleranceIdx; }

inline void MVFMap::setResolution(unsigned resIdx)
{ toleranceIdx = resIdx; }

inline void MVFMap::init() const {}

inline bool MVFMap::exportAsFig(char const * fileName, int colorMode,
				bool noFill) const
{
  fileName  = fileName;
  colorMode = colorMode;
  noFill    = noFill;

  return false;
}


#endif // __MVFMAP_HH_
