#ifndef __MVWIDGET_HH_
#define __MVWIDGET_HH_


#include <GL/gl.h>
#include <GL/glu.h>

#include <qgl.h>
#include <qsize.h>

#include <map.hh>
#include <iostream>
using namespace std;

#define MV_DEFAULT_LENGTH  400
#define MV_MINIMUM_LENGTH   40
#define MV_MAXIMUM_LENGTH 4000
#define MV_BORDER 0.05




class MVWidget
  : public QGLWidget {

public:

  /* Constructors */
  MVWidget(Map * map, QWidget * parent = 0, char const * name = 0);

  /* Get map function */
  Map *       getMap();
  Map const * getMap() const;

  /* Aspect ratio get function */
  double getMapAspectRatio() const;

  /* Aspect ratio correction functions */
  int widthToHeight(int width) const;
  int heightToWidth(int height) const;

  /* Clipping function */
  void  enableClipping();
  void disableClipping();

  void setClipRegion(double x, double y,
		     double w, double h);

  /* Export as fig function */
  bool exportAsFig(char const * fileName, int colorMode, bool noFill);

  /* Length functions */
  int getDefaultLength() const;
  int getMinimumLength() const;
  int getMaximumLength() const;

  void setDefaultLength(int defaultLength);
  void setMinimumLength(int minimumLength);
  void setMaximumLength(int maximumLength);

  /* Size functions */
  QSize const & getDefaultSize() const;
  QSize const & getMinimumSize() const;
  QSize const & getMaximumSize() const;

  int getDefaultWidth() const;
  int getMinimumWidth() const;
  int getMaximumWidth() const;

  int getDefaultHeight() const;
  int getMinimumHeight() const;
  int getMaximumHeight() const;


protected:

  /* Qt GL functions */
  virtual void paintGL();
  virtual void initializeGL();
  virtual void resizeGL(int width, int height);


private:

  /* Map pointer */
  Map * map;

  /* Map limits */
  double const mapLeft, mapRight;
  double const mapBottom, mapTop;

  /* Map aspect ratio */
  double mapAspectRatio;

  /* Lengths */
  int defaultLength;
  int minimumLength;
  int maximumLength;

  /* Sizes */
  QSize defaultSize;
  QSize minimumSize;
  QSize maximumSize;
};



inline MVWidget::MVWidget(Map * map, QWidget * parent, char const * name)
  : QGLWidget(parent, name), map(map),

    mapLeft  (map->getXMin()-map->getMaxDimension()*MV_BORDER),
    mapRight (map->getXMax()+map->getMaxDimension()*MV_BORDER),
    mapBottom(map->getYMin()-map->getMaxDimension()*MV_BORDER),
    mapTop   (map->getYMax()+map->getMaxDimension()*MV_BORDER),

    mapAspectRatio((mapRight-mapLeft)/(mapTop-mapBottom)),

    defaultLength(MV_DEFAULT_LENGTH),
    minimumLength(MV_MINIMUM_LENGTH),
    maximumLength(MV_MAXIMUM_LENGTH)
{
  if (mapAspectRatio > 1.0)
    {
      defaultSize.setWidth(defaultLength);
      defaultSize.setHeight(static_cast<int>(defaultLength/mapAspectRatio));

      minimumSize.setWidth(static_cast<int>(minimumLength*mapAspectRatio));
      minimumSize.setHeight(minimumLength);

      maximumSize.setWidth(maximumLength);
      maximumSize.setHeight(static_cast<int>(maximumLength/mapAspectRatio));
    }

  else
    {
      defaultSize.setWidth(static_cast<int>(defaultLength*mapAspectRatio));
      defaultSize.setHeight(defaultLength);

      minimumSize.setWidth(minimumLength);
      minimumSize.setHeight(static_cast<int>(minimumLength/mapAspectRatio));

      maximumSize.setWidth(static_cast<int>(maximumLength*mapAspectRatio));
      maximumSize.setHeight(maximumLength);
    }

  resize(defaultSize);
}


inline Map *       MVWidget::getMap()       { return map; }
inline Map const * MVWidget::getMap() const { return map; }

inline double MVWidget::getMapAspectRatio() const { return mapAspectRatio; }

inline int MVWidget::widthToHeight(int width) const
{
  return static_cast<int>(width/mapAspectRatio);
}

inline int MVWidget::heightToWidth(int height) const
{
  return static_cast<int>(height*mapAspectRatio);
}

inline void MVWidget::enableClipping()
{
  makeCurrent();
  glEnable(GL_CLIP_PLANE0);
  glEnable(GL_CLIP_PLANE1);
  glEnable(GL_CLIP_PLANE2);
  glEnable(GL_CLIP_PLANE3);
  doneCurrent();
}

inline void MVWidget::disableClipping()
{
  makeCurrent();
  glDisable(GL_CLIP_PLANE0);
  glDisable(GL_CLIP_PLANE1);
  glDisable(GL_CLIP_PLANE2);
  glDisable(GL_CLIP_PLANE3);
  doneCurrent();
}

inline bool MVWidget::exportAsFig(char const * fileName, int colorMode,
				  bool noFill)
{
  makeCurrent();
  bool fileExported = map->exportAsFig(fileName, colorMode, noFill);
  doneCurrent();

  return fileExported;
}

inline int MVWidget::getDefaultLength() const { return defaultLength; }
inline int MVWidget::getMinimumLength() const { return minimumLength; }
inline int MVWidget::getMaximumLength() const { return maximumLength; }

inline QSize const & MVWidget::getDefaultSize() const { return defaultSize; }
inline QSize const & MVWidget::getMinimumSize() const { return minimumSize; }
inline QSize const & MVWidget::getMaximumSize() const { return maximumSize; }

inline int MVWidget::getDefaultWidth() const { return defaultSize.width(); }
inline int MVWidget::getMinimumWidth() const { return minimumSize.width(); }
inline int MVWidget::getMaximumWidth() const { return maximumSize.width(); }

inline int MVWidget::getDefaultHeight() const { return defaultSize.height(); }
inline int MVWidget::getMinimumHeight() const { return minimumSize.height(); }
inline int MVWidget::getMaximumHeight() const { return maximumSize.height(); }


#endif // __MVWIDGET_HH_
